/*
 * This file is in the public domain.  It is mostly assembled and modified by isaac porat
 * from sources in the public domain 
 */

package core.utilities;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintStream;
import java.io.RandomAccessFile;
import java.io.Reader;
import java.net.MalformedURLException;
import java.net.URL;
import java.nio.ByteBuffer;
import java.nio.channels.FileChannel;
import java.nio.charset.Charset;

/**
 * Directory and File general utilities
 */
public class DirFileUtilities
{

	/**
	 * Reads a file to a string
	 * 
	 * @param filePath - file name including directory path
	 * @return - string containing the complete file
	 */
	public static String fileToString(String filePath) throws IOException
	{
		BufferedReader bufferedReader = null;
		try
		{
			bufferedReader = new BufferedReader(new FileReader(filePath));
			StringBuilder stringBuilder = new StringBuilder();
			String line;
			while ((line = bufferedReader.readLine()) != null)
			{
				stringBuilder.append(line + "\n");
			}
			return (stringBuilder.toString());
		} finally
		{
			try
			{
				if (bufferedReader != null)
					bufferedReader.close();
			} catch (IOException ex)
			{
			}
		}
	}

	/**
	 * Reads a file to a string
	 * 
	 * @param filePath - file name including directory path
	 *  @param format - encoding format 
	*	 @return - string containing the complete file
	 */
	public static String fileToString(String filePath, String format) throws IOException
	{
		BufferedReader bufferedReader = null;
		try
		{
			Reader reader = new InputStreamReader(new FileInputStream(filePath), format);
			bufferedReader = new BufferedReader(reader);
			StringBuilder stringBuilder = new StringBuilder();
			String line;
			while ((line = bufferedReader.readLine()) != null)
			{
				stringBuilder.append(line + "\n");
			}
			return (stringBuilder.toString());
		} finally
		{
			try
			{
				if (bufferedReader != null)
					bufferedReader.close();
			} catch (IOException ex)
			{
			}
		}
	}

	/**
	 * Reads a file to a string
	 * 
	 * @param fileURL - URL of file
	 * @return - string containing the complete file
	 */
	public static String fileToString(URL fileURL) throws IOException, MalformedURLException
	{
		BufferedReader bufferedReader = null;
		try
		{
			bufferedReader = new BufferedReader(new InputStreamReader(fileURL.openStream()));
			StringBuilder stringBuilder = new StringBuilder();
			String line;
			while ((line = bufferedReader.readLine()) != null)
			{
				stringBuilder.append(line + "\n");
			}
			return (stringBuilder.toString());
		} finally
		{
			if (bufferedReader != null)
				bufferedReader.close();
		}
	}

	/**
	 * Write string to file 
	 * @param filePath - file name including directory path.
	 * @param string - string to write.
	 */
	public static void stringToFile(String filePath, String string) throws IOException
	{
		PrintStream printStream = null;
		try
		{
			FileOutputStream fileOutputStream = new FileOutputStream(filePath);
			printStream = new PrintStream(fileOutputStream);
			printStream.println(string);
		} finally
		{
			if (printStream != null)
				printStream.close();
		}
	}

	/**
	 * Locks and reads a file to a string
	 * 
	 * @param filePath - file name including directory path
	 * @return - string containing the complete file
	 */
	public static String fileToStringLocked(String filePath) throws IOException
	{
		FileChannel fileChannel = null;
		try
		{
			// read write to allow lock
			RandomAccessFile randomAccessFile = new RandomAccessFile(filePath, "rw");
			fileChannel = randomAccessFile.getChannel();
			fileChannel.lock();
			/*
			 * MappedByteArray has a bug which prevents file deletion.
			 * Therefore in memory file was not used
			 */

			/*
			 * The first method is simple and less flexible for other use
			 *			Note buffer allocation larger than file size to allow loop termination
			*It is not clear if the while loop is necessary as single reading might not 
			*read the complete file
			*/

			ByteBuffer byteBuffer = ByteBuffer.allocate((int) fileChannel.size() + 10);
			while ((fileChannel.read(byteBuffer) != -1))
				;
			byteBuffer.flip();
			return Charset.defaultCharset().decode(byteBuffer).toString();

			/*
			 * The second method  is more general (and less elegant) solution which
			 * could lend itself to other variations of this method,
			 * An expandable ByteArray ByteArrayOutputStream is used.
			 */

			/* Disabled
			// Note ByteBuffer can be set to any size
			ByteBuffer byteBuffer = ByteBuffer.allocate(1024);
			byte[] bytes = new byte[0];
			ByteArrayOutputStream expandableByteArray = new ByteArrayOutputStream();
			while ((fileChannel.read(byteBuffer) != -1))
			{
				bytes = new byte[byteBuffer.position()];
				byteBuffer.flip();
				byteBuffer.get(bytes, 0, bytes.length);
				expandableByteArray.write(bytes);
				byteBuffer.clear();
			}
			return (expandableByteArray.toString(Charset.defaultCharset().name()));
			*/
		} finally
		{
			// Closing the channel unlocks it
			if (fileChannel != null)
				fileChannel.close();
		}
	}

	/**
	 * Locks and write string to file 
	 * @param filePath - file name including directory path.
	 * @param string - string to write.
	 */
	public static void stringToFileLocked(String filePath, String string) throws IOException
	{
		FileChannel fileChannel = null;
		try
		{
			FileOutputStream outputStream = new FileOutputStream(new File(filePath));
			fileChannel = outputStream.getChannel();
			fileChannel.lock();
			byte[] bytes = string.getBytes();
			ByteBuffer byteBuffer = ByteBuffer.allocate(bytes.length);
			byteBuffer.put(bytes);
			byteBuffer.flip();
			while (byteBuffer.hasRemaining())
				fileChannel.write(byteBuffer);
		} finally
		{
			// Closing the channel unlocks it
			if (fileChannel != null)
				fileChannel.close();
		}
	}

	public static URL locationToURL(String urlContextPath, String parentDirectory, String location)
			throws MalformedURLException
	{
		URL url, urlContext;
		// First try to create URL context
		// If urlContext is defined, it must complement url to work properly
		try
		{
			urlContext = new URL(urlContextPath);
		} catch (Exception ex)
		{
			urlContext = null;
		}
		//  First try to construct a URL
		try
		{
			// use the urlContext if defined
			if (urlContext != null)
				url = new URL(urlContext, location);
			else
				url = new URL(location);
		} catch (MalformedURLException e)
		{ // Construct a URL from a path
			File file = new File(location);
			// If File relative construct and absolute path
			if ((parentDirectory != null) && (!file.isAbsolute()))
				file = new File(parentDirectory, location);
			url = (file.toURI()).toURL();
		}
		return (url);
	}

	public static boolean mkdirsParent(File file)
	{
		String parentDirectory = file.getParent();
		File parentDirectoryFile = new File(parentDirectory);
		if (parentDirectoryFile.exists() == false)
			return (parentDirectoryFile.mkdirs());
		return (true);
	}

	public static void copyFile(File sourceFile, File targetFile) throws IOException, FileNotFoundException
	{
		if (sourceFile.equals(targetFile))
			return; // Can't copy to itself
		FileInputStream fis = null;
		FileOutputStream fos = null;
		try
		{
			fis = new FileInputStream(sourceFile);
			fos = new FileOutputStream(targetFile);
			byte[] buf = new byte[1024];
			int i = 0;
			while ((i = fis.read(buf)) != -1)
			{
				fos.write(buf, 0, i);
			}
		} finally
		{
			try
			{
				if (fis != null)
					fis.close();
			} catch (IOException ex)
			{
			}
			try
			{
				if (fos != null)
					fos.close();
			} catch (IOException ex)
			{
			}
		}
	}

	public static void copyFile2(File sourceFile, File destinationFile) throws IOException
	{
		/*
		 * Note that this method does a channel I/O copy which is significantly more
		 * efficient than the "normal" java method of copying byte-by-byte
		 * through an InputStream. The first method is attempted and if fails (there were a number of issues reported with this method) a traditional method is attempted.   
		*/
		String copyErrorMessage = null;
		boolean copied = false;
		FileChannel in = null;
		FileChannel out = null;
		try
		{
			in = new FileInputStream(sourceFile).getChannel();
			out = new FileOutputStream(destinationFile, false).getChannel();
			long destinationSize = out.transferFrom(in, 0, in.size());
			if (destinationSize == in.size())
			{
				copied = true;
			} else
			{
				copyErrorMessage = "An error occurred during channel file copy." + " Expected to copy " + in.size()
						+ " bytes but only copied " + destinationSize + " bytes.";
				copied = false;
			}
		} catch (Exception ex)
		{
			copyErrorMessage = ex.getMessage();
			copied = false;
		} finally
		{
			if (in != null)
				in.close();
			if (out != null)
				out.close();
		}
		/*
		* If for some reason the channel file copy failed then execute a 
		* different (slower) copy method.
		*/
		if (copied == false)
		{
			copyFile(sourceFile, destinationFile);
		}
		if (copyErrorMessage != null)
			throw new IOException(copyErrorMessage);
	}

	public static void copyDirectory(File sourceFile, File targetFile) throws FileNotFoundException, IOException
	{
		if (sourceFile.isDirectory())
		{
			if (targetFile.exists() == false)
				targetFile.mkdirs();
			String files[] = sourceFile.list();
			for (String file : files)
			{
				File sourceChildFile = new File(sourceFile, file);
				File targetChildFile = new File(targetFile, file);
				copyDirectory(sourceChildFile, targetChildFile);
			}
		} else
			copyFile(sourceFile, targetFile);
	}

	public static void copyDirectory(File sourceFile, File destinationFile, FileListFilter fileListFilter)
			throws IOException
	{
		if (sourceFile.isDirectory())
		{
			if (!destinationFile.exists())
			{
				if (destinationFile.mkdir() == false)
					return;
			}
			File[] children = fileListFilter.getSortedFilteredChildren(sourceFile, false);
			for (int i = 0; i < children.length; i++)
			{
				copyDirectory(new File(sourceFile, children[i].getName()),
						new File(destinationFile, children[i].getName()), fileListFilter);
			}
		} else
			DirFileUtilities.copyFile(sourceFile, destinationFile);
	}

	public static long getDirectorySize(File directoryFile)
	{
		long directorySize = 0;
		File[] files = directoryFile.listFiles();
		for (int i = 0; i < files.length; i++)
		{
			if (files[i].isDirectory())
				directorySize += getDirectorySize(files[i]);
			else
				directorySize += files[i].length();
		}
		return (directorySize);
	}

	public static boolean startsWithProtocol(String string)
	{
		return (string.matches("^[a-z][a-z0-9]+:.*$+")) ? true : false;
	}

	/**
	 * Empty directory by deleting all files and subdirectories under dir.
	 * Returns true if all deletions were successful. If a deletion fails, the
	 * method continues to delete and returns false.
	 * 
	 * @param directory - the directory to delete.
	 * @return - True if successfull False if not.
	 */
	public static boolean emptyDirectory(File directory)
	{
		boolean returnFlag = true;
		if (directory.isDirectory())
		{
			String[] children = directory.list();
			for (int i = 0; i < children.length; i++)
			{
				boolean success = deleteDirectory(new File(directory, children[i]));
				if (!success)
					returnFlag = false;
			}
		}
		return (returnFlag);
	}

	/**
	 * Delete directory by deleting all files and subdirectories under dir.
	 * Returns true if all deletions were successful. If a deletion fails, the
	 * method stops attempting to delete and returns false.
	 * 
	 * @param directory - the directory to delete.
	 * @return - True if successfull False if not.
	 */
	public static boolean deleteDirectory(File directory)
	{
		if (directory.isDirectory())
		{
			String[] children = directory.list();
			for (int i = 0; i < children.length; i++)
			{
				deleteDirectory(new File(directory, children[i]));
			}
		}
		// The directory should now be empty so attempt to delete it
		return directory.delete();
	}

	/**
	 * Combine absolute and relative paths including ../ in the relative path.
	 * <br/>For internal use only - no error checking <br/>
	 * 
	 * @param parentPath - Absolute path portion.
	 * @param relativePath - relative path portion.
	 * @return - combined path.
	 */
	private static String internalCombinePaths(String parentPath, String relativePath)
	{
		File directoryPathFile = new File(parentPath);
		String parts[] = relativePath.split("[/\\\\]");
		int n;
		for (n = 0; n < parts.length; n++)
		{
			if (parts[n].equals(".."))
			{
				if (directoryPathFile != null)
				{
					parentPath = directoryPathFile.getParent();
					if (parentPath != null)
						directoryPathFile = new File(parentPath);
				}
			} else
				break;
		}
		relativePath = "";
		StringBuilder relativePathBuffer = new StringBuilder();
		for (; n < parts.length; n++)
		{
			relativePathBuffer.append(File.separator + parts[n]);
		}
		relativePath = relativePathBuffer.toString();
		parentPath = directoryPathFile.getPath();
		parentPath = parentPath.replaceAll("[\\\\/]$", "");
		String path = parentPath + relativePath;
		return (path);
	}

	/**
	 * Gets absolute normilized path from absolute or relative path including '../' and './'
	  * @param path either relative or absolute path 
	 * @return Normalized absolute path
	 */
	public static String getNormalizedAbsolutePath(String path)
	{
		// This method was not tested with Linux
		if (path == null)
			return (null);
		File file = new File(path);
		if (file.isAbsolute() == false)
			path = file.getAbsolutePath();
		path = path.replaceAll("\\\\", "/");
		boolean endPathSlash = path.endsWith("/");
		String[] directoryArray = path.split("/");
		int removeCounter = 0;
		for (int n = directoryArray.length - 1; n >= 0; n--)
		{
			if (directoryArray[n].equals("."))
				directoryArray[n] = null;
			else if (directoryArray[n].equals(".."))
			{
				removeCounter++;
				directoryArray[n] = null;
			} else if (removeCounter > 0)
			{
				directoryArray[n] = null;
				removeCounter--;
			}
		}
		path = "";
		StringBuilder pathBuffer = new StringBuilder();
		for (int n = 0; n < directoryArray.length; n++)
		{
			if (directoryArray[n] != null)
				pathBuffer.append(directoryArray[n] + "/");
		}
		path = pathBuffer.toString();
		if (endPathSlash == false)
			path = path.replaceAll("/$", "");
		return (path);
	}

	/**
	 * Combine absolute and relative paths including ../ in the relative path.
	 * <br/>Support file: and http: <br/>
	 * 
	 * @param parent - Absolute path portion, can be URL format.
	 * @param relative - relative path portion, can be URL format.
	 * @return - combined path.
	 */
	public static String combinePaths(String parent, String relative) throws CombinePathsException
	{
		// Check for non supported protocols by URL and assume absolute path
		// Microsoft streamming
		if (relative.toLowerCase().startsWith("mms:"))
			return (relative);
		// RTSP: Real-Time Streaming Protocol
		else if (relative.toLowerCase().startsWith("rtsp:"))
			return (relative);
		String urlParentString = null;
		String parentPath = null;
		String relativePath = null;
		// Process parent
		if (DirFileUtilities.startsWithProtocol(parent))
		{
			URL urlParent = null;
			try
			{
				urlParent = new URL(parent);
			} catch (MalformedURLException ex)
			{
				throw new CombinePathsException(ex.toString());
			}
			String protocol = urlParent.getProtocol();
			parentPath = urlParent.getPath();
			if (parentPath.equals("")) // Path does not exist
			{
				throw new CombinePathsException("No path found in parent URL");
			}
			// Act on supported parent protocols
			if (parent.startsWith("file:"))
			{
				urlParentString = protocol + ":";
			} else if (parent.startsWith("http:"))
			{
				String host = urlParent.getHost();
				urlParentString = protocol + "://" + host;
			} else
			{
				throw new CombinePathsException("Protocol not supported in parent URL");
			}
		}
		// Process relative
		if (DirFileUtilities.startsWithProtocol(relative))
		{
			URL urlRelative = null;
			try
			{
				urlRelative = new URL(relative);
			} catch (MalformedURLException ex)
			{
				throw new CombinePathsException(ex.toString());
			}
			relativePath = urlRelative.getPath();
			if (relativePath.equals("")) // Path does not exist
			{
				throw new CombinePathsException("No path found in relative URL");
			}
			//  Act on supported relative protocols
			if (relative.startsWith("file:"))
			{
				// Do nothing
			} else if (relative.startsWith("http:"))
			{
				throw new CombinePathsException("Relative URL can't start with 'http:' ");
			} else
			{
				throw new CombinePathsException("Protocol not supported in relative URL");
			}
		}
		// Check for URL parent and relative paths conformity
		if (parentPath != null) // URL
		{
			if (!parentPath.startsWith("/"))
			{
				throw new CombinePathsException("Parent path must be absolute");
			}
			if (parentPath.matches(".+\\.[a-zA-Z0-9]+$"))
			{
				throw new CombinePathsException("Print path must be a directory");
			}
		}
		if (relativePath != null)
		{
			if (relativePath.startsWith("/"))
			{
				throw new CombinePathsException("Relative path must be relative");
			}
		}
		//Check for non URL parent and relative paths conformity
		if (parentPath == null) // Non URL
		{
			File parentFile = new File(parent);
			if (!parentFile.isAbsolute())
			{
				throw new CombinePathsException("Parent path must be absolute");
			}
			if (parent.matches(".+\\.[a-zA-Z0-9]+$"))
			{
				throw new CombinePathsException("Parent path must be a directory");
			}
		}
		if (relativePath == null) // non URL
		{
			File relativeFile = new File(relative);
			if (relativeFile.isAbsolute())
			{
				throw new CombinePathsException("Relative path must be relative");
			}
		}
		// Transfer values as appropriate
		if (parentPath == null)
			parentPath = parent;
		if (relativePath == null)
			relativePath = relative;
		String path = DirFileUtilities.internalCombinePaths(parentPath, relativePath);
		if (urlParentString != null)
		{
			path = urlParentString + path;
			path = path.replaceAll("\\\\", "/");
		}
		return (path);
	}

	/**
	 * Compare two paths regardless of '/' or '\'
	 * 
	 * @param path1 First path
	 * @param path2 Second path
	 * @return true if the same, false if not
	 */
	public static boolean isPathEqual(String path1, String path2)
	{
		if (path1 == null)
			return (false);
		if (path2 == null)
			return (false);
		path1 = path1.replaceAll("\\\\", "/");
		path2 = path2.replaceAll("\\\\", "/");
		return (path1.equals(path2));
	}

}
