/*
* Copyright (c) Isaac Porat 2011 - 2013 all rights reserved.
* This file is released under the GNU General Public License, version 2, or (at your option) any later version.
* See the file SpeechHub_License.txt for more details.
*/

package core.utilities;

import java.io.File;
import java.io.FileFilter;
import java.text.Collator;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Comparator;
import java.util.List;

/*
 * Filter for sorting and rejecting in a file list
 */
public class FileListFilter
{
	private FileComparator fFileComparator = new FileComparator();
	private FileExtensionFilter fFileFilter = new FileExtensionFilter();
	private List<String> fAcceptExtensionList = new ArrayList<String>();
	private List<String> fRejectExtensionList = new ArrayList<String>();
	private List<String> fRejectDirectoryNameList = new ArrayList<String>();
	private List<String> fRejectFileNameList = new ArrayList<String>();
	private boolean fDirectoryOnly = false;

	public void setDirectoryOnly(boolean directoryOnly)
	{
		fDirectoryOnly = directoryOnly;
	}

	public void addAcceptExtension(String extension)
	{
		// Remove '.'.' if necessary
		extension = extension.replaceAll("[\\.]+", "");
		fAcceptExtensionList.add(extension);
	}

	public void addRejectExtension(String extension)
	{
		// Remove '.'.' if necessary
		extension = extension.replaceAll("[\\.]+", "");
		fRejectExtensionList.add(extension);
	}

	public void addRejectDirectoryName(String directoryName)
	{
		fRejectDirectoryNameList.add(directoryName);
	}

	public void addRejectFileName(String fileName)
	{
		fRejectFileNameList.add(fileName);
	}

	public void clearAcceptExtensionList()
	{
		fAcceptExtensionList.clear();
	}

	public void clearRejectExtensionList()
	{
		fRejectExtensionList.clear();
	}

	public void clearRejectDirectoryNameList()
	{
		fRejectDirectoryNameList.clear();
	}

	public void clearRejectFileNameList()
	{
		fRejectFileNameList.clear();
	}

	public File[] getSortedFilteredChildren(File file, boolean sort)
	{
		if (file == null)
			return (null);
		if (file.isFile())
			return (null);
		File[] childrenFiles = file.listFiles(fFileFilter);
		if (childrenFiles == null)
			return (null);
		if (sort)
			Arrays.sort(childrenFiles, fFileComparator);
		return (childrenFiles);
	}

	@edu.umd.cs.findbugs.annotations.SuppressWarnings(value = "SE_COMPARATOR_SHOULD_BE_SERIALIZABLE", justification = "serializable Not required")
	private static class FileComparator implements Comparator<File>
	{
/*
  Note there is behaviour incompatibility between Java 7 and previous versions
 See:
http://www.oracle.com/technetwork/java/javase/compatibility-417013.html#source

Area: API: Utilities
Synopsis: Updated sort behavior for Arrays and Collections may throw an IllegalArgumentException
Description: The sorting algorithm used by java.util.Arrays.sort and (indirectly) by java.util.Collections.sort has been replaced. The new sort implementation
may throw an IllegalArgumentException if it detects a Comparable that violates the Comparable contract. The previous implementation silently ignored such
a situation. If the previous behavior is desired, you can use the new system property, java.util.Arrays.useLegacyMergeSort, to restore previous mergesort
behavior.
*/
		
		
		private Collator fCollator = Collator.getInstance();

		@Override
		public int compare(File f1, File f2)
		{
if (f1.isDirectory() && f2.isFile())
				return -1;
			if (f1.isFile() && f2.isDirectory())
				return 1;
			return fCollator.compare(f1.getName(), f2.getName());
		}

		}

	private class FileExtensionFilter implements FileFilter
	{

		@Override
		public boolean accept(File file)
		{
			try
			{
				String fileName = file.getName();
				if (file.isDirectory())
				{
					if (fRejectDirectoryNameList.size() > 0)
					{
						if (fRejectDirectoryNameList.contains(fileName))
							return (false);
						return (true);
					}
					return (true);
				}
				//      Files processed from here
				if (fDirectoryOnly)
					return (false);
				if (fRejectFileNameList.size() > 0)
				{
					if (fRejectFileNameList.contains(fileName))
						return (false);
				}
				// Extract the file extension if exists 
				if (fileName.contains("."))
				{
					String[] parts = fileName.split("\\.");
					String extension = parts[parts.length - 1].toLowerCase();
					if (fAcceptExtensionList.size() > 0)
					{
						if (fAcceptExtensionList.contains(extension))
							return (true);
						return (false);
					}
					if (fRejectExtensionList.contains(extension))
						return (false);
				}
				return (true);
			} catch (Exception ex)
			{
				return (false);
			}
		}

	}

}
