/*
* Copyright (c) Isaac Porat 2011 - 2013 all rights reserved.
* This file is released under the GNU General Public License, version 2, or (at your option) any later version.
* See the file SpeechHub_License.txt for more details.
*/

package hrpc.handler;

import hrpc.common.HRHandlerInterface;
import hrpc.common.HRMessageHandlerInterface;
import hrpc.common.HRMessageHandlerJSONInterface;
import hrpc.common.HRPCConstants;
import hrpc.common.HRReceiverHandlerJSONInterface;
import hrpc.common.HRSenderHandlerJSONInterface;
import hrpc.common.HRTransportInterface;

import java.io.IOException;

import core.utilities.StackTrace;

import log.common.CascadingLoggerInterface;

public class HRMessageHandlerServerMultiProtocolBidirectional extends HRMessageHandlerBase implements
		HRMessageHandlerInterface, HRMessageHandlerJSONInterface
{
	private final static String RESPONSE_REGEXP = "^" + HRPCConstants.HPR_REQUEST;
	private final static String JSON_REGEXP = "^" + HRPCConstants.HPR_JSON;
	private HRSenderHandlerJSONInterface fSenderHandlerJSON;
	private HRReceiverHandlerJSONInterface fReceiverHandlerJSON;
	private HRHandlerInterface fHandlerCustom;

	public HRMessageHandlerServerMultiProtocolBidirectional(HRTransportInterface server,
			HRSenderHandlerJSONInterface senderJSON, HRReceiverHandlerJSONInterface receiverJSON,
			HRHandlerInterface handlerCustom, CascadingLoggerInterface logger) throws IOException
	{
		log = logger.clone(HRMessageHandlerServerMultiProtocolBidirectional.class);
		fTransport = server;
		fSenderHandlerJSON = senderJSON;
		if (fSenderHandlerJSON != null)
			fSenderHandlerJSON.setMessageHandler(this);
		fReceiverHandlerJSON = receiverJSON;
		if (fReceiverHandlerJSON != null)
			fReceiverHandlerJSON.setMessageHandler(this);
		fHandlerCustom = handlerCustom;
		if (fHandlerCustom != null)
			fHandlerCustom.setMessageHandler(this);
	}

	@Override
	public void close() throws IOException
	{
		closeTransport();
	}

	@Override
	public void dispatchJSONRequestMessage(String requestMessage)
	{
		// Insert request marker
		dispatchMessage(HRPCConstants.HPR_REQUEST + HRPCConstants.HPR_JSON + requestMessage);
	}

	@Override
	public void dispatchJSONResponseMessage(String responseMessage)
	{
		// No Response marker in server
		dispatchMessage(HRPCConstants.HPR_JSON + responseMessage);
	}

	@Override
	public boolean processMessage(String message)
	{
		try
		{
			// Custom message does not support bi directional comm
			if (message.startsWith(HRPCConstants.HPR_RESPONSE))
			{
				message.replaceAll(RESPONSE_REGEXP, "");
				if (message.startsWith(HRPCConstants.HPR_JSON))
					message = message.replaceAll(JSON_REGEXP, "");
				return (fSenderHandlerJSON.processResponseMessage(message));
			} else
			{
				// Route to JSON or custom
				if (message.startsWith(HRPCConstants.HPR_JSON))
				{
					message = message.replaceAll(JSON_REGEXP, "");
					return (fReceiverHandlerJSON.processRequestMessage(message));
				} else
				{
					return (fHandlerCustom.processMessage(message));
				}
			}
		} catch (Exception ex)
		{
			log.error("Error in process method:\n" + StackTrace.stackToString(ex));
			return (true); // Do not close connection
		}
	}

}
