/*
* Copyright (c) Isaac Porat 2011 - 2013 all rights reserved.
* This file is released under the GNU General Public License, version 2, or (at your option) any later version.
* See the file SpeechHub_License.txt for more details.
*/

package hrpc.sh.server;

import hrpc.common.HRReceiverHandlerJSONInterface;
import hrpc.handler.HRReceiverMethodHandlerJSON;

import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;

import log.common.CascadingLoggerInterface;

import sh.connections.SHEngineConnectorInterface;

import com.thetransactioncompany.jsonrpc2.JSONRPC2Error;
import com.thetransactioncompany.jsonrpc2.JSONRPC2ParseException;
import com.thetransactioncompany.jsonrpc2.JSONRPC2Request;
import com.thetransactioncompany.jsonrpc2.util.NamedParamsRetriever;

public class SHPrimaryServerHandler implements Runnable
{
	private CascadingLoggerInterface log;
	private BlockingQueue<String> fJSONRequestStringList = new LinkedBlockingQueue<String>();
	private HRReceiverMethodHandlerJSON fRMH;
	private SHEngineConnectorInterface fEngine;

	public SHPrimaryServerHandler(HRReceiverHandlerJSONInterface receiverHandler, SHEngineConnectorInterface engine,
			CascadingLoggerInterface logger)
	{
		fEngine = engine;
		log = logger.clone(SHPrimaryServerHandler.class);
		fRMH = new HRReceiverMethodHandlerJSON(receiverHandler, log);
		Thread methodsThread = new Thread(this);
		methodsThread.setName(SHPrimaryServerHandler.class.getName());
		methodsThread.start();
	}

	public void handleMethods(String jsonRequestString)
	{
		try
		{
			fJSONRequestStringList.put(jsonRequestString);
		} catch (InterruptedException ex)
		{
			log.error("Error in handleMethods method:\n" + ex);
		}
	}

	private void handleMethodsFromQueue()
	{
		String jsonRequestString = null;
		try
		{
			jsonRequestString = fJSONRequestStringList.take();
		} catch (InterruptedException ex)
		{
			log.error("Error in handleMethodsFromQueue method:\n" + ex);
		}
		JSONRPC2Request request = null;
		try
		{
			request = JSONRPC2Request.parse(jsonRequestString);
			String methodName = request.getMethod();
			// For efficiency methods order  is bbased according to their expected relative frequency
			if (methodName.equals("speak"))
				handleSpeak(request);
			else if (methodName.equals("selectLanguage"))
				handleSelectLanguage(request);
			else if (methodName.equals("getEngineIdentifiers"))
				handleGetEngineIdentifiers(request);
			else if (methodName.equals("selectEngine"))
				handleSelectEngine(request);
			else if (methodName.equals("getVoiceIdentifiers"))
				handleGetVoiceIdentifiers(request);
			else if (methodName.equals("selectVoice"))
				handleSelectVoice(request);
			else if (methodName.equals("getVariantIdentifiers"))
				handleGetVariantIdentifiers(request);
			else if (methodName.equals("selectVariant"))
				handleSelectVariant(request);
			else if (methodName.equals("setVolume"))
				handleSetVolume(request);
			else if (methodName.equals("setRate"))
				handleSetRate(request);
			else if (methodName.equals("setPitch"))
				handleSetPitch(request);
			else if (methodName.equals("getProperty"))
				handleGetProperty(request);
			else if (methodName.equals("setProperty"))
				handleSetProperty(request);
			else if (methodName.equals("initialize"))
				handleInitialize(request);
			else if (methodName.equals("close"))
				handleClose(request);
			else
			{
				log.error("Error in routeMethod method:\n Method " + methodName + " not found");
				fRMH.dispatchResponseError(JSONRPC2Error.METHOD_NOT_FOUND, (String) request.getID());
				return;
			}
		} catch (JSONRPC2ParseException ex)
		{
			log.error("Error in routeMethod method:\n" + ex);
			fRMH.dispatchResponseError(JSONRPC2Error.PARSE_ERROR, "0");
		} catch (JSONRPC2Error ex)
		{
			log.error("Error in routeMethod method:\n" + ex);
			fRMH.dispatchResponseError(ex, (String) request.getID());
		}
	}

	private void handleInitialize(JSONRPC2Request request) throws JSONRPC2Error
	{
		boolean status = fEngine.initialize();
		fRMH.dispatchResponseBoolean(status, (String) request.getID());
	}

	private void handleClose(JSONRPC2Request request) throws JSONRPC2Error
	{
		boolean status = fEngine.close();
		fRMH.dispatchResponseBoolean(status, (String) request.getID());
	}

	private void handleSelectLanguage(JSONRPC2Request request) throws JSONRPC2Error
	{
		NamedParamsRetriever npr = fRMH.getNamedParamsRetriever(request);
		String languageCode = npr.getString("languageCode");
		boolean status = fEngine.selectLanguage(languageCode);
		fRMH.dispatchResponseBoolean(status, (String) request.getID());
	}

	private void handleGetEngineIdentifiers(JSONRPC2Request request) throws JSONRPC2Error
	{
		String[] engineIdentifiers = fEngine.getEngineIdentifiers();
		fRMH.dispatchResponseStringArray(engineIdentifiers, (String) request.getID());
	}

	private void handleSelectEngine(JSONRPC2Request request) throws JSONRPC2Error
	{
		NamedParamsRetriever npr = fRMH.getNamedParamsRetriever(request);
		String engineIdentifier = npr.getString("engineIdentifier");
		boolean status = fEngine.selectEngine(engineIdentifier);
		fRMH.dispatchResponseBoolean(status, (String) request.getID());
	}

	private void handleGetVoiceIdentifiers(JSONRPC2Request request) throws JSONRPC2Error
	{
		String[] voiceIdentifiers = fEngine.getVoiceIdentifiers();
		fRMH.dispatchResponseStringArray(voiceIdentifiers, (String) request.getID());
	}

	private void handleSelectVoice(JSONRPC2Request request) throws JSONRPC2Error
	{
		NamedParamsRetriever npr = fRMH.getNamedParamsRetriever(request);
		String voiceIdentifier = npr.getString("voiceIdentifier");
		boolean status = fEngine.selectVoice(voiceIdentifier);
		fRMH.dispatchResponseBoolean(status, (String) request.getID());
	}

	private void handleGetVariantIdentifiers(JSONRPC2Request request) throws JSONRPC2Error
	{
		String[] variantIdentifiers = fEngine.getVariantIdentifiers();
		fRMH.dispatchResponseStringArray(variantIdentifiers, (String) request.getID());
	}

	private void handleSelectVariant(JSONRPC2Request request) throws JSONRPC2Error
	{
		NamedParamsRetriever npr = fRMH.getNamedParamsRetriever(request);
		String variantIdentifier = npr.getString("variantIdentifier");
		boolean status = fEngine.selectVariant(variantIdentifier);
		fRMH.dispatchResponseBoolean(status, (String) request.getID());
	}

	private void handleSpeak(JSONRPC2Request request) throws JSONRPC2Error
	{
		NamedParamsRetriever npr = fRMH.getNamedParamsRetriever(request);
		int messageIdentifier = npr.getInt("messageIdentifier");
		String message = npr.getString("message");
		boolean speakSync = npr.getBoolean("speakSync");
		boolean status = fEngine.speak(messageIdentifier, message, speakSync);
		fRMH.dispatchResponseBoolean(status, (String) request.getID());
	}

	public void handleSetVolume(JSONRPC2Request request) throws JSONRPC2Error
	{
		NamedParamsRetriever npr = fRMH.getNamedParamsRetriever(request);
		float volume = npr.getFloat("volume");
		boolean status = fEngine.setVolume(volume);
		fRMH.dispatchResponseBoolean(status, (String) request.getID());
	}

	public void handleSetRate(JSONRPC2Request request) throws JSONRPC2Error
	{
		NamedParamsRetriever npr = fRMH.getNamedParamsRetriever(request);
		float rate = npr.getFloat("rate");
		boolean status = fEngine.setRate(rate);
		fRMH.dispatchResponseBoolean(status, (String) request.getID());
	}

	public void handleSetPitch(JSONRPC2Request request) throws JSONRPC2Error
	{
		NamedParamsRetriever npr = fRMH.getNamedParamsRetriever(request);
		float pitch = npr.getFloat("pitch");
		boolean status = fEngine.setPitch(pitch);
		fRMH.dispatchResponseBoolean(status, (String) request.getID());
	}

	public void handleGetProperty(JSONRPC2Request request) throws JSONRPC2Error
	{
		NamedParamsRetriever npr = fRMH.getNamedParamsRetriever(request);
		String key = npr.getString("key");
		String value = fEngine.getProperty(key);
		fRMH.dispatchResponseString(value, (String) request.getID());
	}

	public void handleSetProperty(JSONRPC2Request request) throws JSONRPC2Error
	{
		NamedParamsRetriever npr = fRMH.getNamedParamsRetriever(request);
		String key = npr.getString("key");
		String value = npr.getString("value");
		boolean status = fEngine.setProperty(key, value);
		String data = "key: " + key + ", value: " + value;
		fRMH.dispatchResponseBoolean(status, data, (String) request.getID());
	}

	@Override
	public void run()
	{
		for (;;)
		{
			handleMethodsFromQueue();
		}
	}

}
