/*
* Copyright (c) Isaac Porat 2011 - 2013 all rights reserved.
* This file is released under the GNU General Public License, version 2, or (at your option) any later version.
* See the file SpeechHub_License.txt for more details.
*/

package hrpc.transport;

import hrpc.common.HRTransportInterface;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.PrintStream;

import core.utilities.StackTrace;

import log.common.BasicLogger;
import log.common.CascadingLoggerInterface;

public class HRTransportStdioClient implements HRTransportInterface
{
	private CascadingLoggerInterface log;
	private BufferedReader fIn;
	private InputStream fInRaw;
	private PrintStream fOut;
	private Process fProcess;
	private String[] fCommandLineArray;
	private String fCurrentDirectory;
	private String fFormat;
	private String fProcessIdentifierMessage;
	private StderrStreamConsumer fErrorConsumer;

	public HRTransportStdioClient(String[] commandLineArray, String currentDirectory, String processIdentifierMessage,
			String format, CascadingLoggerInterface logger)
	{
		log = logger.clone(HRTransportStdioClient.class);
		fCommandLineArray = commandLineArray;
		fCurrentDirectory = currentDirectory;
		fFormat = format;
		fProcessIdentifierMessage = processIdentifierMessage;
	}

	@Override
	public BufferedReader getStdin()
	{
		return (fIn);
	}

	@Override
	public InputStream getStdinRaw()
	{
		return (fInRaw);
	}

	@Override
	public PrintStream getStdout()
	{
		return (fOut);
	}

	public void open() throws IOException
	{
		if (fCurrentDirectory != null)
			fProcess = Runtime.getRuntime().exec(fCommandLineArray, null, new File(fCurrentDirectory));
		else
			fProcess = Runtime.getRuntime().exec(fCommandLineArray);
		// Start the process stderr consumer which transfers process error to SH stderr  
		fErrorConsumer = new StderrStreamConsumer(fProcessIdentifierMessage, fProcess.getErrorStream());
		fErrorConsumer.start();
		fInRaw = fProcess.getInputStream();
		InputStreamReader reader = new InputStreamReader(fInRaw, fFormat);
		fIn = new BufferedReader(reader);
		fOut = new PrintStream(fProcess.getOutputStream(), true, fFormat);
	}

	public void close() throws IOException
	{
		try
		{
			if (fProcess != null)
				fProcess.destroy();
		} finally
		{
			try
			{
				fProcess = null;
				if (fIn != null)
					fIn.close();
				fIn = null;
				fInRaw = null;
				if (fOut != null)
					fOut.close();
				fOut = null;
				if (fErrorConsumer != null)
					fErrorConsumer.closeStream();
				fErrorConsumer = null;
			} catch (Exception ex)
			{
			}
		}
	}

	/**
	 * Consumer thread for capturing stderr from external process.
	 * 
	 */
	private class StderrStreamConsumer extends Thread
	{
		private InputStream fInputStream;
		private String fProcessIdentifierMessage;

		/**
		 * @param processIdentifier identifier to display with error message 
		 * @param stderrStream - stderr Input Stream to capture
		 */
		public StderrStreamConsumer(String processIdentifierMessage, InputStream stderrStream)
		{
			fProcessIdentifierMessage = processIdentifierMessage;
			fInputStream = stderrStream;
		}

		public void closeStream() throws IOException
		{
			fInputStream.close();
		}

		@Override
		public void run()
		{
			try
			{
				InputStreamReader inputStreamReader = new InputStreamReader(fInputStream);
				BufferedReader bufferedReader = new BufferedReader(inputStreamReader);
				String line = null;
				while ((line = bufferedReader.readLine()) != null)
				{
					if (fProcessIdentifierMessage != null)
					{
						BasicLogger.notifyPrintln("From " + fProcessIdentifierMessage + " process (prefixed with p::");
						fProcessIdentifierMessage = null;
					}
					BasicLogger.notifyPrintln("p:: " + line);
				}
			} catch (Exception ex)
			{
				log.error("Error in StderrStreamConsumer run method: in " + fProcessIdentifierMessage + " process:\n"
						+ StackTrace.stackToString(ex));
			}
		}
	}

}
