/*
* Copyright (c) Bill Cox 2011 - 2013 all rights reserved.
* This file is released under the GNU General Public License, version 2, or (at your option) any later version.
* See the file SpeechHub_License.txt for more details.
*/

package sh.audio;

import java.util.LinkedHashMap;

import log.common.CascadingLoggerInterface;

/**
 * Caches short speech audio messages to reduce load on an engine
 * @author Bill Cox
 *
 */
public class SHAudioSpeechCache
{
	@SuppressWarnings("unused")
	private CascadingLoggerInterface log;
	private final static int MAX_CACHE_SIZE = (1 << 24); // 4MB cache size
	private LinkedHashMap<String, byte[]> fSpeechCache = null;
	private byte fSpeechBuf[] = null;
	private int fSpeechBufSize = 0;
	private int fSpeechBufPos = 0;
	private int fSpeechCacheSize = 0;
	private float fPitch = 0;

	public SHAudioSpeechCache(CascadingLoggerInterface logger)
	{
		log = logger.clone(SHAudioSpeechCache.class);
		fSpeechCache = new LinkedHashMap<String, byte[]>(128, .75f, true);
		fSpeechBufSize = 4096;
		fSpeechBufPos = 0;
		fSpeechBuf = new byte[fSpeechBufSize];
	}

	public void addSamplesToSpeechBuf(byte[] samples)
	{
		if (fSpeechBufPos + samples.length >= fSpeechBufSize)
		{
			fSpeechBufSize *= 2;
			byte[] newArray = new byte[fSpeechBufSize];
			System.arraycopy(fSpeechBuf, 0, newArray, 0, fSpeechBufPos);
			fSpeechBuf = newArray;
		}
		System.arraycopy(samples, 0, fSpeechBuf, fSpeechBufPos, samples.length);
		fSpeechBufPos += samples.length;
	}

	public void resetSpeechBufPos()
	{
		fSpeechBufPos = 0;
	}

	public void addMessageToCache(String message)
	{
		if (fSpeechBufPos == 0)
			return;
		byte[] speechSamples = new byte[fSpeechBufPos];
		System.arraycopy(fSpeechBuf, 0, speechSamples, 0, fSpeechBufPos);
		fSpeechCache.put(getKey(message), speechSamples);
		fSpeechCacheSize += fSpeechBufPos;
		ensureSize();
	}

	private void ensureSize()
	{
		while (fSpeechCacheSize > MAX_CACHE_SIZE)
		{
			String key = fSpeechCache.keySet().iterator().next();
			byte[] data = fSpeechCache.get(key);
			fSpeechCache.remove(key);
			fSpeechCacheSize -= data.length;
		}
	}

	public void setMessageToCache(String message, byte[] speechSamples)
	{
		if (containsAudioMessage(message))
		{
			byte[] existingSpeechSamples = getAudioMessage(message);
			fSpeechBufSize -= existingSpeechSamples.length;
		}
		fSpeechCache.put(getKey(message), speechSamples);
		fSpeechBufSize += speechSamples.length;
		ensureSize();
	}

	public boolean containsAudioMessage(String message)
	{
		return (fSpeechCache.containsKey(getKey(message)));
	}

	public byte[] getAudioMessage(String message)
	{
		return (fSpeechCache.get(getKey(message)));
	}

	public void clearCache()
	{
		fSpeechCache.clear();
		fSpeechCacheSize = 0;
	}

	public void setPitch(float pitch)
	{
		fPitch = pitch;
	}

	private String getKey(String message)
	{
		if (fPitch == 0)
			return (message);
		return (Float.toString(fPitch) + message);
	}

}
