/*
* Copyright (c) Bill Cox 2011 - 2013 all rights reserved.
* This file is released under the GNU General Public License, version 2, or (at your option) any later version.
* See the file SpeechHub_License.txt for more details.
*/

package sh.audio;

import java.util.Set;

import log.common.BasicLogger;
import log.common.CascadingLoggerInterface;
import sh.audio.javasound.SHAudioDeliveryCombinedJavasound;
import sh.connections.SHAudioSpeechPlayerInterface;
import sh.connections.SHConstants;
import sh.connections.SHConstants.MessageType;
import sh.connections.SHMessageEnvelopeInterface;
import sh.connections.SHMessageQueuePeekHandlerInterface;
import sh.connections.SHModuleInterface;
import sh.connections.SHAudioDataHandlerInterface;
import sh.connections.SHUpdateParametersHandlerInterface;
import sh.settings.SHEngineParameters;
import core.utilities.StackTrace;

/**
 * Plays and controls speech audio generated by an engine
 *  
 * @author Bill Cox
 */
public class SHAudioSpeechPlayerCombined implements SHAudioSpeechPlayerInterface, SHAudioDataHandlerInterface
{
	private CascadingLoggerInterface log;
	private final static float MIN_PITCH = 0.75f;
	private final static float MAX_PITCH = 1.33f;
	private final static float MIN_RATE = 0.2f;
	private final static float MAX_RATE = 6.0f;
	private final static int MAX_CACHED_MESSAGE_LENGTH = 15;
	private SHModuleInterface fModule;
	private SHUpdateParametersHandlerInterface fUpdateParametersHandler;
	private SHSonic fSonic = null;
	private boolean fUseSonic;
	private boolean fUseSonicPitch;
	private int fSampleRate = -1;
	private boolean fCancelPending = false;
	private boolean fPaused = false;
	private SHMessageQueuePeekHandlerInterface fMessagePeekHandler;
	private Set<String> fConnectorEngineSet;
	private boolean fDebugMode = false;
	private SHAudioDataHandlerInterface fAudioHandler = null;
	private boolean fSpeakingChar = false;
	private float fVolume = 1.0f;
	private float fCharVolume = 1.0f;
	private boolean fScaleCharVolume = false;
	private SHAudioSpeechCache fAudioSpeechCache;

	private SHEngineParameters fEngineParameters;
	private boolean fUserUseSonic = false;
	private boolean fClientUseSonic = false;
	private String fEngineIdentifier;
	private SHAudioDeliveryCombinedInterface fAudioDelivery;
	private float fRate = 0;

	public SHAudioSpeechPlayerCombined(SHUpdateParametersHandlerInterface updateParametersHandler,
			CascadingLoggerInterface logger)
	{
		log = logger.clone(SHAudioSpeechPlayerCombined.class);
		fUpdateParametersHandler = updateParametersHandler;
		fEngineParameters = SHEngineParameters.getInstance(logger);
		loadEngineUserProperties();
	}

	@Override
	public void setConnectorEngineSet(Set<String> connectorEngineSet)
	{
		fConnectorEngineSet = connectorEngineSet;
	}

	@Override
	public void setMessagePeekHandler(SHMessageQueuePeekHandlerInterface messagePeekHandler)
	{
		fMessagePeekHandler = messagePeekHandler;
	}

	@Override
	public boolean initialize(SHModuleInterface module, SHAudioSpeechCache audioSpeechCache)
	{
		fModule = module;
		fAudioSpeechCache = audioSpeechCache;
		return (true);
	}

	@Override
	public boolean close()
	{
		try
		{
			if (fAudioDelivery != null)
				fAudioDelivery.close();
			fAudioDelivery = null;
			return true;
		} catch (Exception ex)
		{
			log.error("Error in method close:\n" + ex);
			return (false);
		}
	}

	// Scale the volume.
	private byte[] scaleSamples(byte[] samples)
	{
		float volume = fVolume;
		if (fSpeakingChar && fScaleCharVolume)
			volume *= fCharVolume * 32767.0 / findMaxValue(samples);
		byte[] newSamples = new byte[samples.length];
		for (int i = 0; i < samples.length; i += 2)
		{
			short sample = (short) ((samples[i] & 0xff) | (samples[i + 1] << 8));
			sample *= volume;
			newSamples[i] = (byte) sample;
			newSamples[i + 1] = (byte) (sample >> 8);
		}
		return newSamples;
	}

	// Find the max of the absolute value of any sample in the array.
	private short findMaxValue(byte[] samples)
	{
		short max = 0;

		for (int i = 0; i < samples.length; i += 2)
		{
			short sample = (short) ((samples[i] & 0xff) | (samples[i + 1] << 8));
			if (sample < 0)
				sample = (short) -sample;
			if (sample > max)
				max = sample;
		}
		return max;
	}

	// Strip leading zeros from the audio sample.
	private byte[] stripLeadingSilence(byte[] samples)
	{
		int i;
		for (i = 0; i < samples.length && samples[i] == 0 && samples[i + 1] == 0; i += 2)
			;
		byte[] newSamples = new byte[samples.length - i];
		System.arraycopy(samples, i, newSamples, 0, samples.length - i);
		return newSamples;
	}

	private boolean isSilence(byte[] samples)
	{
		for (int i = 0; i < samples.length; i++)
			if (samples[i] != 0)
				return false;
		return true;
	}

	private void writeSamplesToAudio(byte samples[])
	{
		try
		{
			if (fAudioHandler == null)
			{
				if (!fCancelPending)
				{
					if (fDebugMode)
					{
						log.notify("Sending " + samples.length / 2 + " samples to audio system");
					}
					if (fScaleCharVolume && !isSilence(samples))
						samples = stripLeadingSilence(samples);
					if (fVolume != 1.0f || (fSpeakingChar && fScaleCharVolume))
					{
						byte[] scaledSamples = scaleSamples(samples);
						fAudioDelivery.write(scaledSamples);
					} else
					{
						fAudioDelivery.write(samples);
					}
				}
			} else
			{
				fAudioHandler.processAudio(samples);
			}
			fAudioDelivery.setTotalWritten(fAudioDelivery.getTotalWritten() + samples.length);
		} catch (Exception ex)
		{
			log.error("Error in writeSamplesToAudio method:\n" + ex);
		}
	}

	private void writeSonicOutput()
	{
		int available = fSonic.bytesAvailable();
		if (available > 0)
		{
			byte modifiedSamples[] = new byte[available];
			fSonic.readBytesFromStream(modifiedSamples, available);
			if (fDebugMode)
			{
				log.notify("Sending " + available + " samples to audio system");
			}
			if (!fCancelPending)
			{
				writeSamplesToAudio(modifiedSamples);
			} else
			{
				fSonic.clear();
			}
		}
	}

	// If the user has canceled playback, return false so the synthesizer can stop.
	@Override
	public boolean processAudio(byte[] samples)
	{
		fAudioSpeechCache.addSamplesToSpeechBuf(samples);
		synchronized (this)
		{
			while (fPaused)
			{
				try
				{
					fAudioDelivery.stop();
					wait();
					fAudioDelivery.start();
				} catch (Exception ex)
				{
					log.error("Error in processAudio method:\n" + ex);
				}
			}
		}
		if (!fSpeakingChar)
		{
			writeSamples(samples);
		}
		return !fCancelPending;
	}

	@Override
	public boolean processAudio(SHMessageEnvelopeInterface messageEnvelope)
	{
		// Not implemented
		return (false);
	}

	// Write samples, using either sonic or direct to audio, with volume scaling.
	private void writeSamples(byte[] samples)
	{
		if (fUseSonic)
		{
			if (!fCancelPending)
			{
				fSonic.writeBytesToStream(samples, samples.length);
			}
			writeSonicOutput();
		} else
		{
			if (!fCancelPending)
			{
				writeSamplesToAudio(samples);
			}
		}
	}

	// Cancel speech.  This does not cancel the engine we're wrapping, but instead we allow
	// processAudio cancel speech by returning false.
	@Override
	public boolean cancel()
	{
		try
		{
			if (fAudioDelivery == null)
			{
				// Not initialize, so just return
				return (true);
			}
			fCancelPending = true;
			fAudioDelivery.indicateCancel();
			fModule.cancel();
			// With small buffer audio line the next line needs to be commented out
			//			fAudioLine.stop();
			if (fUseSonic)
				fSonic.clear();
			fAudioDelivery.flush();
			return true;
		} catch (Exception ex)
		{
			log.error("Error in cancel method:\n" + ex);
			return (false);
		}
	}

	@Override
	public boolean notifyEngineChanged(String engineIdentifier)
	{
		fEngineIdentifier = engineIdentifier;
		return (updateAudioLine());
	}

	private boolean updateAudioLine()
	{
		try
		{
			// Load engine properties
			int oldSampleRate = fSampleRate;
			String sampleRate = fModule.getProperty(SHConstants.PROP_ENGINE_SAMPLE_RATE);
			if (sampleRate == null)
				return false;
			fSampleRate = Integer.parseInt(sampleRate);
			String property = fEngineParameters.getEngineFactoryProperty(fEngineIdentifier,
					SHConstants.PROP_ENGINE_USE_SONIC);
			boolean useSonic = property != null && property.equals("true");
			property = fEngineParameters.getEngineFactoryProperty(fEngineIdentifier,
					SHConstants.PROP_ENGINE_USE_SONIC_PITCH);
			fUseSonicPitch = property != null && property.equals("true");
			property = fEngineParameters.getEngineFactoryProperty(fEngineIdentifier,
					SHConstants.PROP_ENGINE_CHAR_VOLUME);
			if (property != null)
			{
				fScaleCharVolume = true;
				fCharVolume = Float.parseFloat(property);
			}
			fUseSonicPitch = property != null && property.equals("true");
			if (fUserUseSonic || fClientUseSonic)
				useSonic = true;
			if (useSonic)
			{
				if (fSonic == null)
				{
					fSonic = new SHSonic(fSampleRate, 1);
				} else
				{
					if (fSampleRate != oldSampleRate)
						fSonic.setSampleRate(fSampleRate);
				}
			}
			fUseSonic = useSonic;
			if (fAudioDelivery == null || fSampleRate != oldSampleRate)
			{
				if (fAudioDelivery != null)
					fAudioDelivery.close();
				fAudioDelivery = new SHAudioDeliveryCombinedJavasound(fEngineIdentifier, fSampleRate, log);
				// To use the old direct audio line connector
				// Do not forget to uncomment the fAudioLine.start() and fAudioLine.stop() in the speak and cancel methods
				// Do not forget to comment these again when using the small buffer version
				//fAudioLine = new SHOriginalLargeBufferAudioLineJavasound(fSampleRate, log);

				fAudioDelivery.open();
			}
			return (true);
		} catch (Exception ex)
		{
			log.error("Error in updateAudioLine method\n" + StackTrace.stackToString(ex));
			return (false);
		}
	}

	@Override
	public boolean pause()
	{
		fPaused = true;
		return (true);
	}

	@Override
	public boolean resume()
	{
		fPaused = false;
		synchronized (this)
		{
			notifyAll();
		}
		return (true);
	}

	@Override
	public boolean setVolume(float volume)
	{
		fVolume = volume / 100.0f;
		return true;
	}

	@Override
	public boolean isPlayerHandlesRate()
	{
		return (fUseSonic);
	}

	@Override
	public boolean setRate(float rate)
	{
		try
		{
			fRate = rate;
			if (isPlayerHandlesRate() == false)
			{
				fAudioSpeechCache.clearCache();
				return (true);
			}
			if (rate >= 0.0f)
			{
				rate = 1.0f + rate * (MAX_RATE - 1.0f) / 100.0f;
			} else
			{
				rate = 1.0f + rate * (1.0f - MIN_RATE) / 100.0f;
			}
			fSonic.setSpeed(rate);
			return true;
		} catch (Exception ex)
		{
			log.error("Error in setRate method:\n" + StackTrace.stackToString(ex));
			return (false);
		}
	}

	@Override
	public boolean isPlayerHandlesPitch()
	{
		return (fUseSonic || fUseSonicPitch);
	}

	@Override
	public boolean setPitch(float pitch)
	{
		try
		{
			if (isPlayerHandlesPitch() == false)
				return (true);
			if (pitch >= 0.0f)
			{
				pitch = 1.0f + pitch * (MAX_PITCH - 1.0f) / 100.0f;
			} else
			{
				pitch = 1.0f + pitch * (1.0f - MIN_PITCH) / 100.0f;
			}
			fSonic.setPitch(pitch);
			return true;
		} catch (Exception ex)
		{
			log.error("Error in setPitch method:\n" + StackTrace.stackToString(ex));
			return (false);
		}
	}

	@Override
	public boolean setClientUseSonic(boolean mode)
	{
		if (mode == fClientUseSonic)
			return (true);
		fClientUseSonic = mode;
		if (updateAudioLine())
		{
			setRate(fRate);
			return (true);
		}
		return (false);
	}

	private boolean speakMessageWaiting()
	{
		if (fMessagePeekHandler == null)
		{
			return false;
		}
		SHMessageEnvelopeInterface messageEnvelope = fMessagePeekHandler.peekNextMessage(fConnectorEngineSet);
		if (messageEnvelope == null)
		{
			return false;
		}
		return true;
	}

	@Override
	public boolean speak(SHMessageEnvelopeInterface messageEnvelope)
	{
		try
		{
			String message = messageEnvelope.getMessage();
			if (fAudioDelivery == null)
				return (true); // This can happen while resetting an engine
			if (fUpdateParametersHandler.updateParameters(messageEnvelope) == false)
			{
				log.error("Error in speak method: updateParameters method failed");
				return (false);
			}
			fAudioDelivery.indicateSpeak(message);
			// With small buffer audio line the next line needs to be commented out
			//fAudioLine.start();
			debug("Start speaking " + message);
			fAudioSpeechCache.resetSpeechBufPos();
			fCancelPending = false;
			fAudioDelivery.setTotalWritten(0);
			// We force single characters to be synthesized in full so we can scale them to 100% volume.
			fSpeakingChar = messageEnvelope.getMessageType() == MessageType.CHAR;
			if (fAudioSpeechCache.containsAudioMessage(message))
			{
				byte[] cachedSpeechSamples = fAudioSpeechCache.getAudioMessage(message);
				writeSamples(cachedSpeechSamples);
			} else if (!fModule.speak(messageEnvelope))
			{
				return false;
			} else if (!fCancelPending && message.length() < MAX_CACHED_MESSAGE_LENGTH)
			{
				fAudioSpeechCache.addMessageToCache(message);
				if (fSpeakingChar)
				{
					byte[] cachedSpeechSamples = fAudioSpeechCache.getAudioMessage(message);
					writeSamples(cachedSpeechSamples);
				}
			}
			// If a message is waiting, let's return early to move things along.
			if (!fCancelPending && !speakMessageWaiting())
				fAudioDelivery.drain();
			fCancelPending = false;
			debug("End speaking " + message);
			return true;
		} catch (Exception ex)
		{
			log.error("Error in speak method:\n" + StackTrace.stackToString(ex));
			return (false);
		}
	}

	@Override
	public boolean setAudioDataHandler(SHAudioDataHandlerInterface audioDataHandler)
	{
		fAudioHandler = audioDataHandler;
		return (true);
	}

	private void debug(String message)
	{
		if (fDebugMode)
		{
			BasicLogger.notifyPrintln(message);
		}
	}

	private void loadEngineUserProperties()
	{
		try
		{
			fDebugMode = false;
			String debugModeConnectorIdentifier = fEngineParameters
					.getEngineCommonProperty(SHConstants.PROP_ENGINE_DEBUG_MODE_CONNECTOR_IDENTIFIER);
			if ((debugModeConnectorIdentifier == null) || debugModeConnectorIdentifier.equals("none"))
				fDebugMode = false;
			else if ((debugModeConnectorIdentifier.equals("all"))
					|| (debugModeConnectorIdentifier.equals(SHConstants.ENGINE_CONNECTOR_IDENTIFIER_MULTI_NATIVE)))
				fDebugMode = true;
			String property = fEngineParameters.getEngineCommonProperty(SHConstants.PROP_ENGINE_USE_SONIC);
			fUserUseSonic = ((property != null) && (property.equals("true")));
		} catch (Exception ex)
		{
			log.error("Error in loadEngineUserProperties method:\n" + ex);
		}
	}

}
