/*
* Copyright (c) Bill Cox and Isaac Porat 2011 - 2013 all rights reserved.
* This file is released under the GNU General Public License, version 2, or (at your option) any later version.
* See the file SpeechHub_License.txt for more details.
*/

package sh.audio;

import java.util.Arrays;
import java.util.Set;

import log.common.BasicLogger;
import log.common.CascadingLoggerInterface;
import sh.audio.javasound.SHWindowsTrimThreadAudioDeliveryParallelJavasound;
import sh.connections.SHAudioDataHandlerInterface;
import sh.connections.SHAudioSpeechPlayerInterface;
import sh.connections.SHConstants;
import sh.connections.SHConstants.MessageTextToAudioConversionState;
import sh.connections.SHConstants.MessageType;
import sh.connections.SHMessageEnvelopeInterface;
import sh.connections.SHMessageQueuePeekHandlerInterface;
import sh.connections.SHModuleInterface;
import sh.connections.SHUpdateParametersHandlerInterface;
import sh.settings.SHEngineParameters;
import core.utilities.OS;
import core.utilities.StackTrace;
import core.utilities.Stopwatch;

/**
 * Controls in parallel speech text to audio data generated by an engine
 *  and passes the audio data for conversion to sound to an audio delivery class
 *  Audio data and related information in stored in each message envelope
 *  Short messages are stored in a cache to increase speech response
 *  Sonic is used automatically for rate and pitch based on user preference and engine requirements
 *   
 * @author Bill Cox
 * @author Isaac Porat
 */
public class SHAudioSpeechPlayerParallel implements SHAudioSpeechPlayerInterface, SHAudioDataHandlerInterface, Runnable
{
	private CascadingLoggerInterface log;
	// Enable debug timing mode below to use Stopwatch to measure audio available and speak method return 
	private final static boolean DEBUG_TIMING_MODE = false;
	private static final float MIN_PITCH = 0.75f;
	private final static float MAX_PITCH = 1.33f;
	private final static float MIN_RATE = 0.2f;
	private final static float MAX_RATE = 6.0f;
	private final static int MAX_CACHED_MESSAGE_LENGTH = 15;
	private SHModuleInterface fModule;
	private SHUpdateParametersHandlerInterface fUpdateParametersHandler;
	private SHSonic fSonic = null;
	private boolean fUseSonic;
	private boolean fUseSonicPitch;
	private int fSonicMaxSampleSize = 2048; // Calculated dynamically based on sample rate
	private int fAudioSampleRate = -1;
	private float fTrailingSilenceTrimRatio = 1f;
	private boolean fCancelPending = false;
	private boolean fPaused = false;
	private SHMessageQueuePeekHandlerInterface fMessagePeekHandler;
	private Set<String> fConnectorEngineSet;
	private boolean fDebugMode = false;
	private SHAudioDataHandlerInterface fAudioDataHandler;
	private boolean fSpeakingChar = false;
	private float fVolume = 1.0f;
	private float fCharVolume = 1.0f;
	private boolean fScaleCharVolume = false;
	private boolean fSilenceDetectionAllowed = true;
	private SHAudioSpeechCache fAudioSpeechCache;
	private String fEngineIdentifier;
	private SHEngineParameters fEngineParameters;
	private boolean fUserUseSonic = false;
	private boolean fClientUseSonic = false;
	private SHAudioDeliveryParallelInterface fAudioDelivery;
	private SHMessageEnvelopeInterface fMessageEnvelopeAudioDataToSound;
	private SHMessageEnvelopeInterface fMessageEnvelopeTextToAudioData;
	private Thread fTextToAudioDataThread;
	private boolean fMessageAvailable = false;
	@SuppressWarnings("unused")
	private boolean fMessageWaitingInQueue = false;
	private boolean fCatchupModeMessageDelivered = false;
	private float fRate = 0;
	private boolean fMessagePeekEnabled = true;
	// Debug Stopwatch 
	private Stopwatch fStopwatch;
	private long fStopwatchAudioAvailable = -1;

	public SHAudioSpeechPlayerParallel(SHUpdateParametersHandlerInterface updateParametersHandler,
			CascadingLoggerInterface logger)
	{
		log = logger.clone(SHAudioSpeechPlayerParallel.class);
		if (DEBUG_TIMING_MODE)
			fStopwatch = new Stopwatch();
		fUpdateParametersHandler = updateParametersHandler;
		loadAudioDelivery();
		fEngineParameters = SHEngineParameters.getInstance(logger);
		loadEngineUserProperties();
	}

	private void loadAudioDelivery()
	{
		try
		{
			if (fAudioDelivery != null)
				fAudioDelivery.close();
			if (fAudioDataHandler == null)
			{
				if (OS.isWindows())
					fAudioDelivery = new SHWindowsTrimThreadAudioDeliveryParallelJavasound(log);
				else
				{
					/*
					* Assume Linux
					* Replace the Windows class with your own Linux class
					* A good start would be to copy the existing Windows class and modify as necessary
					* Note old Linux specific code left in the Windows class for future Linuxx implementation
					*/
					fAudioDelivery = new SHWindowsTrimThreadAudioDeliveryParallelJavasound(log);
				}
			} else
				fAudioDelivery = new SHAudioDeliveryParallelAudioHandlerPlugin(fAudioDataHandler, log);
		} catch (Exception ex)
		{
			log.error("Error in loadAudioLine method:\n" + StackTrace.stackToString(ex));
		}
	}

	@Override
	public void setConnectorEngineSet(Set<String> connectorEngineSet)
	{
		fConnectorEngineSet = connectorEngineSet;
	}

	@Override
	public void setMessagePeekHandler(SHMessageQueuePeekHandlerInterface messagePeekHandler)
	{
		fMessagePeekHandler = messagePeekHandler;
	}

	@Override
	public boolean initialize(SHModuleInterface module, SHAudioSpeechCache audioSpeechCache)
	{
		fModule = module;
		fAudioSpeechCache = audioSpeechCache;
		return (true);
	}

	private void ConditionallyStartThread()
	{
		if (fTextToAudioDataThread != null)
			return;
		fTextToAudioDataThread = new Thread(this);
		fTextToAudioDataThread.setName(SHAudioSpeechPlayerParallel.class.getName());
		fTextToAudioDataThread.start();
	}

	@Override
	public boolean close()
	{
		try
		{
			if (fTextToAudioDataThread != null)
			{
				fTextToAudioDataThread.interrupt();
				fTextToAudioDataThread = null;
			}
			if (fAudioDelivery != null)
				fAudioDelivery.close();
			return true;
		} catch (Exception ex)
		{
			log.error("Error in method close:\n" + ex);
			return (false);
		}
	}

	// Scale the volume.
	private byte[] scaleSamplesVolume(byte[] samples, boolean completeMessage)
	{
		float volume = fVolume;
		if (fSpeakingChar && fScaleCharVolume && completeMessage)
		{
			volume *= fCharVolume * 32767.0 / findMaxValue(samples);
		}
		byte[] newSamples = new byte[samples.length];
		for (int i = 0; i < samples.length; i += 2)
		{
			short sample = (short) ((samples[i] & 0xff) | (samples[i + 1] << 8));
			sample *= volume;
			newSamples[i] = (byte) sample;
			newSamples[i + 1] = (byte) (sample >> 8);
		}
		return newSamples;
	}

	// Find the max of the absolute value of any sample in the array.
	private short findMaxValue(byte[] samples)
	{
		short max = 0;
		for (int i = 0; i < samples.length; i += 2)
		{
			short sample = (short) ((samples[i] & 0xff) | (samples[i + 1] << 8));
			if (sample < 0)
				sample = (short) -sample;
			if (sample > max)
				max = sample;
		}
		return max;
	}

	// Strip leading zeros from the audio sample.
	private byte[] stripLeadingSilence(byte[] samples)
	{
		int i;
		for (i = 0; i < samples.length && samples[i] == 0 && samples[i + 1] == 0; i += 2)
			;
		byte[] newSamples = new byte[samples.length - i];
		System.arraycopy(samples, i, newSamples, 0, samples.length - i);
		return newSamples;
	}

	// Trim trailing bytes above a threshold
	private byte[] stripTrailingSilence(byte[] samples, short threshold)
	{
		if (samples == null)
			return (new byte[0]);
		boolean foundSample = false;
		int n;
		for (n = samples.length - 2; n >= 0; n -= 2)
		{
			short sample = (short) ((samples[n] & 0xff) | (samples[n + 1] << 8));
			if (sample < 0)
				sample = (short) -sample;
			if (sample > threshold)
			{
				foundSample = true;
				break;
			}
		}
		if (foundSample == false)
			return (new byte[0]);
		samples = Arrays.copyOfRange(samples, 0, n);
		return (samples);
	}

	private boolean isSilence(byte[] samples)
	{
		for (int i = 0; i < samples.length; i++)
			if (samples[i] != 0)
				return false;
		return true;
	}

	private void writeMessageEnvelopeBuffer(byte[] bytes)
	{
		try
		{
			if (fCancelPending && (bytes.length > 0)) // Not an end of message marker
				return;
			/*
						// Test simulation of slow computer
						int miliSleep = 0;
						// Typical values
						// 500 multiplier for PicoTTS 20000 multiplier for MaryTTS
						int nanoSleep = bytes.length * 500;
						if (nanoSleep > 999999)
						{
							miliSleep = nanoSleep / 1000000;
							nanoSleep = nanoSleep % 1000000;
							//nanoSleep = 999999;
						}
						if ((miliSleep > 0) || (nanoSleep > 0))
						{
							Thread.sleep(miliSleep, nanoSleep);
						}
			*/
			fMessageEnvelopeTextToAudioData.putAudio(bytes);
		} catch (Exception ex)
		{
			log.error("Error in writeBuffer method:\n" + ex);
		}
	}

	private void writeSamplesToAudio(byte samples[], boolean completeMessage)
	{
		try
		{
			if (!completeMessage)
				fAudioSpeechCache.addSamplesToSpeechBuf(samples);
			if (fVolume != 1.0f)
				samples = scaleSamplesVolume(samples, completeMessage);
			if (!fCancelPending)
			{
				if (fDebugMode)
				{
					log.notify("Sending " + samples.length / 2 + " samples to audio system");
				}
				writeMessageEnvelopeBuffer(samples);
			}
			fMessageEnvelopeTextToAudioData.setTotalAudioWritten(fMessageEnvelopeTextToAudioData.getTotalAudioWritten()
					+ samples.length);
		} catch (Exception ex)
		{
			log.error("Error in writeSamplesToAudio method:\n" + ex);
		}
	}

	private void writeSonicOutput()
	{
		int available = fSonic.bytesAvailable();
		if (available > 0)
		{
			byte modifiedSamples[] = new byte[available];
			fSonic.readBytesFromStream(modifiedSamples, available);
			if (fDebugMode)
			{
				log.notify("Sending " + available + " samples to audio system");
			}
			if (!fCancelPending)
			{
				writeSamplesToAudio(modifiedSamples, false);
			} else
			{
				fSonic.clear();
			}
		}
	}

	// If the user has canceled playback, return false so the synthesizer can stop.
	@Override
	public boolean processAudio(byte[] samples)
	{
		synchronized (this)
		{
			while (fPaused)
			{
				try
				{
					fAudioDelivery.stop();
					wait();
					fAudioDelivery.start();
				} catch (Exception ex)
				{
					log.error("Error in processAudio method:\n" + ex);
				}
			}
		}
		if (fSpeakingChar && (samples.length > 0)) // Single char and not terminating sample
		{
			// Remove silence from single characters
			if (fSilenceDetectionAllowed) // Silence removal is only allowed from start message until a sample with no silence is detected
			{
				if (isSilence(samples)) // No need to process
					return !fCancelPending;
				samples = stripLeadingSilence(samples);
				fSilenceDetectionAllowed = false;
			}
		}
		routeSamples(samples, false);
		return !fCancelPending;
	}

	@Override
	public boolean processAudio(SHMessageEnvelopeInterface messageEnvelope)
	{
		// Not implemented
		return (false);
	}

	// Route samples either to sonic or direct to audio.
	private void routeSamples(byte[] samples, boolean completeMessage)
	{
		if (fUseSonic && !completeMessage)
		{
			/*
						 * The following splits the samples to MAX_SONIC_SAMPLE_bytes length
						 * If the sample chunk is too small it can cause delay in Javasound (2 bytes size observed
						 * to cause delay of 10ms).  Small size can lead to audio starvation which can cause popping sounds and must avoided as it can cause drain problems in Javasound
						 */
			int sampleLength = samples.length;
			int sampleFrom = 0;
			int sampleTo = 0;
			while (!fCancelPending && sampleTo < sampleLength)
			{
				sampleTo = sampleFrom + fSonicMaxSampleSize;
				if (sampleTo > sampleLength)
					sampleTo = sampleLength;
				byte[] subSample = Arrays.copyOfRange(samples, sampleFrom, sampleTo);
				fSonic.writeBytesToStream(subSample, subSample.length);
				writeSonicOutput();
				sampleFrom += fSonicMaxSampleSize;
				if (fCancelPending)
					return;
			}
		} else
		// No Sonic direct to audio
		{
			if (!fCancelPending)
			{
				writeSamplesToAudio(samples, completeMessage);
			}
		}
	}

	// Cancel speech.  This does not cancel the engine we're wrapping, but instead we allow
	// processAudio cancel speech by returning false.
	@Override
	public boolean cancel()
	{
		try
		{
			if (fAudioDelivery == null)
			{
				// Not initialize, so just return
				return (true);
			}
			fCancelPending = true;
			fAudioDelivery.cancel();
			fModule.cancel();
			if (fUseSonic)
				fSonic.clear();
			return (true);
		} catch (Exception ex)
		{
			log.error("Error in cancel method:\n" + ex);
			return (false);
		}
	}

	@Override
	public boolean notifyEngineChanged(String engineIdentifier)
	{
		fEngineIdentifier = engineIdentifier;
		return (updateAudioLine());
	}

	private boolean updateAudioLine()
	{
		try
		{
			if (fEngineIdentifier == null)
				return (false);
			// Load engine properties
			int oldSampleRate = fAudioSampleRate;
			String sampleRate = fModule.getProperty(SHConstants.PROP_ENGINE_SAMPLE_RATE);
			if (sampleRate == null)
				return false;
			fAudioSampleRate = Integer.parseInt(sampleRate);
			String property = fEngineParameters.getEngineFactoryProperty(fEngineIdentifier,
					SHConstants.PROP_ENGINE_TRAILING_SILENCE_TRIM_RATIO);
			if (property == null)
			{
				fTrailingSilenceTrimRatio = 1f;
			} else
			{
				try
				{
					fTrailingSilenceTrimRatio = Float.parseFloat(property);
				} catch (Exception ex)
				{
					log.error("Error in updateAudioLine method: error converting trailingSilenceTrimRatio to float");
					fTrailingSilenceTrimRatio = 1f;
				}
			}
			// Sonic max sample size must not be too large as this can cause delay as conversion  is done in real time
			// If it is too small especially at high speed the sample size produced can be too small and cause line starvation
			fSonicMaxSampleSize = fAudioSampleRate / 5;
			if (fSonicMaxSampleSize % 2 != 0)
				fSonicMaxSampleSize++;
			property = fEngineParameters.getEngineFactoryProperty(fEngineIdentifier, SHConstants.PROP_ENGINE_USE_SONIC);
			boolean useSonic = property != null && property.equals("true");
			property = fEngineParameters.getEngineFactoryProperty(fEngineIdentifier,
					SHConstants.PROP_ENGINE_USE_SONIC_PITCH);
			fUseSonicPitch = property != null && property.equals("true");
			property = fEngineParameters.getEngineFactoryProperty(fEngineIdentifier,
					SHConstants.PROP_ENGINE_CHAR_VOLUME);
			if (property != null)
			{
				fScaleCharVolume = true;
				fCharVolume = Float.parseFloat(property);
			}
			fUseSonicPitch = property != null && property.equals("true");
			if (fUserUseSonic || fClientUseSonic)
				useSonic = true;
			if (useSonic)
			{
				if (fSonic == null)
				{
					fSonic = new SHSonic(fAudioSampleRate, 1);
				} else
				{
					if (fAudioSampleRate != oldSampleRate)
						fSonic.setSampleRate(fAudioSampleRate);
				}
			}
			fUseSonic = useSonic;
			fAudioDelivery.disableCatchupMode(); // Reset to  default conversion mode
			return (true);
		} catch (Exception ex)
		{
			log.error("Error in updateAudioLine method\n" + StackTrace.stackToString(ex));
			return (false);
		}
	}

	@Override
	public boolean pause()
	{
		fPaused = true;
		return (true);
	}

	@Override
	public boolean resume()
	{
		fPaused = false;
		synchronized (this)
		{
			notifyAll();
		}
		return (true);
	}

	@Override
	public boolean setVolume(float volume)
	{
		fVolume = volume / 100.0f;
		return true;
	}

	@Override
	public boolean isPlayerHandlesRate()
	{
		return (fUseSonic);
	}

	@Override
	public boolean setRate(float rate)
	{
		try
		{
			fAudioDelivery.disableCatchupMode(); // Reset to default conversion mode
			if (isPlayerHandlesRate() == false)
			{
				if (fRate != rate)
					fAudioSpeechCache.clearCache();
				fRate = rate;
				return (true);
			}
			fRate = rate;
			if (rate >= 0.0f)
			{
				rate = 1.0f + rate * (MAX_RATE - 1.0f) / 100.0f;
			} else
			{
				rate = 1.0f + rate * (1.0f - MIN_RATE) / 100.0f;
			}
			fSonic.setSpeed(rate);
			return true;
		} catch (Exception ex)
		{
			log.error("Error in setRate method:\n" + StackTrace.stackToString(ex));
			return (false);
		}
	}

	@Override
	public boolean isPlayerHandlesPitch()
	{
		return (fUseSonic || fUseSonicPitch);
	}

	@Override
	public boolean setPitch(float pitch)
	{
		try
		{
			if (isPlayerHandlesPitch() == false)
				return (true);
			if (pitch >= 0.0f)
			{
				pitch = 1.0f + pitch * (MAX_PITCH - 1.0f) / 100.0f;
			} else
			{
				pitch = 1.0f + pitch * (1.0f - MIN_PITCH) / 100.0f;
			}
			fSonic.setPitch(pitch);
			return true;
		} catch (Exception ex)
		{
			log.error("Error in setPitch method:\n" + StackTrace.stackToString(ex));
			return (false);
		}
	}

	@Override
	public boolean setClientUseSonic(boolean mode)
	{
		if (mode == fClientUseSonic)
			return (true);
		fClientUseSonic = mode;
		fAudioSpeechCache.clearCache();
		if (updateAudioLine())
		{
			setRate(fRate);
			return (true);
		}
		return (false);
	}

	private SHMessageEnvelopeInterface getMessageWaiting()
	{
		try
		{
			if (fMessagePeekEnabled == false)
				return (null);
			if (fMessagePeekHandler == null)
				return null;
			return (fMessagePeekHandler.peekNextMessage(fConnectorEngineSet));
		} catch (Exception ex)
		{
			log.error("Error in getMessageWaiting method:\n" + ex);
			return (null);
		}
	}

	private void conditionallyNotifyCatchMode(SHMessageEnvelopeInterface messageEnvelope)
	{
		if (fAudioDelivery == null)
			return;
		if (fAudioDelivery.getCatchupMode())
		{
			if (messageEnvelope.isEmptyAudio() == false) // Audio exists already, postponed message
				return;
			if (messageEnvelope.getMessageType() == MessageType.CHAR) // Postponed  message
				return;
			if (fCatchupModeMessageDelivered == false)
			{
				String message = "SpeechHub message: Slow synthesizer response detected.  ";
				// Replace current message for maximum effect
				messageEnvelope.setMessage(message);
				fCatchupModeMessageDelivered = true;
			}
		} else
			fCatchupModeMessageDelivered = false;
	}

	@Override
	public boolean speak(SHMessageEnvelopeInterface messageEnvelope)
	{
		try
		{
			if (DEBUG_TIMING_MODE)
			{
				if (fStopwatch.isRunning() == false)
					fStopwatch.start();
			}
			conditionallyNotifyCatchMode(messageEnvelope);
			ConditionallyStartThread();
			fCancelPending = false;
			fMessageEnvelopeAudioDataToSound = messageEnvelope;
			fMessageAvailable = true;
			synchronized (this)
			{
				notifyAll();
			}
			// Wait for first audio data to be generated by the textToAudioData thread
			if (messageEnvelope.waitForMessageAudio() == false)
				log.error("Timeout in speak method on waiting for audio conversion data");
			if (DEBUG_TIMING_MODE)
			{
				if (fStopwatch.isRunning())
				{
					fStopwatch.stop();
					fStopwatchAudioAvailable = fStopwatch.toValue();
				}
				fStopwatch.start();
			}
			if (audioDataToSound() == false)
				return (false); // Trigger engine restart as audio system failed
			// Text to audio conversion might have completed only partially - check state
			MessageTextToAudioConversionState state = MessageTextToAudioConversionState.UNKNOWN;
			boolean timeout = true;
			for (int n = 0; n < 10; n++)
			{
				state = fMessageEnvelopeAudioDataToSound.getMessageTextToAudioConversionState();
				if (state.equals(MessageTextToAudioConversionState.UNKNOWN) == false) // Text to audio data conversion finished 
				{
					timeout = false;
					break;
				}
				Thread.sleep(1000);
			}
			if (timeout)
				return (false); // Trigger engine restart as 1 minute passed 
			if (DEBUG_TIMING_MODE)
			{
				if (fStopwatch.isRunning())
				{
					fStopwatch.stop();
					log.notify(messageEnvelope.getMessage() + "\nSpeak audio available: " + fStopwatchAudioAvailable
							+ " ms\nSpeak return: " + (fStopwatchAudioAvailable + fStopwatch.toValue()) + " ms");
				}
				fStopwatchAudioAvailable = -1;
			}
			return (state.equals(MessageTextToAudioConversionState.SUCCESS));
		} catch (Exception ex)
		{
			log.error("Error in speak method:\n" + ex);
			return (false);
		}
	}

	private boolean audioDataToSound()
	{
		if (fAudioDelivery == null)
			return (true); // This can happen while resetting an engine
		return (fAudioDelivery.playAudioData(fMessageEnvelopeAudioDataToSound));
	}

	private boolean textToAudioData()
	{
		try
		{
			if (fUpdateParametersHandler.updateParameters(fMessageEnvelopeTextToAudioData) == false)
			{
				log.error("Error in textToAudioData method: updateParameters method failed");
				return (false);
			}
			fMessageEnvelopeTextToAudioData.setAudioSampleRate(fAudioSampleRate);
			fMessageEnvelopeTextToAudioData.setTrailingSilenceTrimRatio(fTrailingSilenceTrimRatio);
			// Disable peek message if in delay  mode to allow more uniform interval between  messages
			fMessagePeekEnabled = true;
			if (fMessageEnvelopeTextToAudioData.getSilenceBeforeMessage() > 0)
				fMessagePeekEnabled = false;
			// Buffer not bound but sentences are limited and not more than two message envelopes containing audio exists at the same time
			fMessageEnvelopeTextToAudioData.initializeAudio();
			String message = fMessageEnvelopeTextToAudioData.getMessage();
			debug("Start converting speech to audio data" + message);
			fAudioSpeechCache.resetSpeechBufPos();
			fMessageEnvelopeTextToAudioData.setTotalAudioWritten(0);
			fSpeakingChar = fMessageEnvelopeTextToAudioData.getMessageType() == MessageType.CHAR;
			fSilenceDetectionAllowed = true;
			if (fAudioSpeechCache.containsAudioMessage(message))
			{
				byte[] cachedSpeechSamples = fAudioSpeechCache.getAudioMessage(message);
				routeSamples(cachedSpeechSamples, true);
				fMessageEnvelopeTextToAudioData.setAudioCacheUsed(true);
			} else if (!fModule.speak(fMessageEnvelopeTextToAudioData))
			{
				return false;
			} else if (!fCancelPending && message.length() < MAX_CACHED_MESSAGE_LENGTH)
			{
				fAudioSpeechCache.addMessageToCache(message);
				if (fSpeakingChar)
				{
					byte[] samples = fAudioSpeechCache.getAudioMessage(message);
					samples = stripTrailingSilence(samples, (short) 0);
					fAudioSpeechCache.setMessageToCache(message, samples);
				}
			}
			debug("End speak conversion to audio data " + message);
			return true;
		} catch (Exception ex)
		{
			log.error("Error in speak method:\n" + StackTrace.stackToString(ex));
			return (false);
		}
	}

	@Override
	public boolean setAudioDataHandler(SHAudioDataHandlerInterface audioDataHandler)
	{
		/*
		 * The following arrangement ensures that if a handler is inserted repeatedly it does not result in unnecessary delivery object creation
		 * To change handlers a null object needs to be inserted first 
		 */
		if ((fAudioDataHandler != null) && (audioDataHandler != null))
			return (true);
		if ((fAudioDataHandler == null) && (audioDataHandler == null))
			return (true);
		fAudioDataHandler = audioDataHandler;
		loadAudioDelivery();
		return (true);
	}

	private void debug(String message)
	{
		if (fDebugMode)
		{
			BasicLogger.notifyPrintln(message);
		}
	}

	private void loadEngineUserProperties()
	{
		try
		{
			fDebugMode = false;
			String debugModeConnectorIdentifier = fEngineParameters
					.getEngineCommonProperty(SHConstants.PROP_ENGINE_DEBUG_MODE_CONNECTOR_IDENTIFIER);
			if ((debugModeConnectorIdentifier == null) || debugModeConnectorIdentifier.equals("none"))
				fDebugMode = false;
			else if ((debugModeConnectorIdentifier.equals("all"))
					|| (debugModeConnectorIdentifier.equals(SHConstants.ENGINE_CONNECTOR_IDENTIFIER_MULTI_NATIVE)))
				fDebugMode = true;
			String property = fEngineParameters.getEngineCommonProperty(SHConstants.PROP_ENGINE_USE_SONIC);
			fUserUseSonic = ((property != null) && (property.equals("true")));
		} catch (Exception ex)
		{
			log.error("Error in loadEngineUserProperties method:\n" + ex);
		}
	}

	@Override
	public void run()
	{
		for (;;)
		{
			try
			{
				synchronized (this)
				{
					while (fMessageAvailable == false)
						wait();
				}
				fMessageAvailable = false;
				fMessageWaitingInQueue = false;
				fMessageEnvelopeTextToAudioData = fMessageEnvelopeAudioDataToSound;
				if (fMessageEnvelopeTextToAudioData.isEmptyAudio()) // No audio data
				{
					boolean status = textToAudioData();
					if (status)
						fMessageEnvelopeTextToAudioData
								.setMessageTextToAudioConversionState(MessageTextToAudioConversionState.SUCCESS);
					else
					// Conversion failed
					{
						log.error("Engine crash in current conversion , last message sent to engine:\n"
								+ fMessageEnvelopeTextToAudioData.getMessage());
						fAudioSpeechCache.clearCache();
						fMessageEnvelopeTextToAudioData
								.setMessageTextToAudioConversionState(MessageTextToAudioConversionState.FAILURE);
					}
					// Mark the end of message with an empty byte array
					writeMessageEnvelopeBuffer(new byte[0]);
					if (status == false)
					{
						fTextToAudioDataThread = null;
						return; // Terminate thread
					}
				}
				if (fCancelPending)
				{
					continue;
				}
				// Convert any messages waiting in the queue
				fMessageEnvelopeTextToAudioData = getMessageWaiting();
				if (fMessageEnvelopeTextToAudioData != null)
				{
					fMessageWaitingInQueue = true;
					if (fMessageEnvelopeTextToAudioData.isEmptyAudio()) // No audio data
					{
						boolean status = textToAudioData();
						if (status)
							fMessageEnvelopeTextToAudioData
									.setMessageTextToAudioConversionState(MessageTextToAudioConversionState.SUCCESS);
						else
						// Conversion failed
						{
							log.error("Engine crash in look ahead conversion, last message sent to engine:\n"
									+ fMessageEnvelopeTextToAudioData.getMessage());
							fAudioSpeechCache.clearCache();
							fMessageEnvelopeTextToAudioData
									.setMessageTextToAudioConversionState(MessageTextToAudioConversionState.FAILURE);
						}
						// Mark the end of message with an empty byte array
						writeMessageEnvelopeBuffer(new byte[0]);
						if (status == false)
						{
							fTextToAudioDataThread = null;
							return; // Terminate thread
						}
					}
				}
			} catch (InterruptedException ex)
			{
				// Do nothing exception generated on close
			} catch (Exception ex)
			{
				log.error("Error in run method:\n" + ex);
				// Mark the end of message with an empty byte array
				writeMessageEnvelopeBuffer(new byte[0]);
				fTextToAudioDataThread = null;
				return; // Terminate thread
			}
		}
	}

}
