// This file is in the public domain.  It was published on stackoverflow.com
// by user663321.  Thanks, user663321!

package sh.audio;

import java.io.*;
import javax.sound.sampled.*;

import log.common.CascadingLoggerInterface;

/**
 * This class handles the reading, writing, and playing of wav files. It is
 * also capable of converting the file to its raw byte [] form.
 *
 * based on code by Evan Merz modified by Dan Vargo
 * @author dvargo
 */
public class SHWave
{
	private CascadingLoggerInterface log;
	/*
	WAV File Specification
	FROM http://ccrma.stanford.edu/courses/422/projects/WaveFormat/
	The canonical WAVE format starts with the RIFF header:
	0         4   ChunkID          Contains the letters "RIFF" in ASCII form
	(0x52494646 big-endian form).
	4         4   ChunkSize        36 + SubChunk2Size, or more precisely:
	4 + (8 + SubChunk1Size) + (8 + SubChunk2Size)
	This is the size of the rest of the chunk
	following this number.  This is the size of the
	entire file in bytes minus 8 bytes for the
	two fields not included in this count:
	ChunkID and ChunkSize.
	8         4   Format           Contains the letters "WAVE"
	(0x57415645 big-endian form).

	The "WAVE" format consists of two subchunks: "fmt " and "data":
	The "fmt " subchunk describes the sound data's format:
	12        4   Subchunk1ID      Contains the letters "fmt "
	(0x666d7420 big-endian form).
	16        4   Subchunk1Size    16 for PCM.  This is the size of the
	rest of the Subchunk which follows this number.
	20        2   AudioFormat      PCM = 1 (i.e. Linear quantization)
	Values other than 1 indicate some
	form of compression.
	22        2   NumChannels      Mono = 1, Stereo = 2, etc.
	24        4   SampleRate       8000, 44100, etc.
	28        4   ByteRate         == SampleRate * NumChannels * BitsPerSample/8
	32        2   BlockAlign       == NumChannels * BitsPerSample/8
	The number of bytes for one sample including
	all channels. I wonder what happens when
	this number isn't an integer?
	34        2   BitsPerSample    8 bits = 8, 16 bits = 16, etc.

	The "data" subchunk contains the size of the data and the actual sound:
	36        4   Subchunk2ID      Contains the letters "data"
	(0x64617461 big-endian form).
	40        4   Subchunk2Size    == NumSamples * NumChannels * BitsPerSample/8
	This is the number of bytes in the data.
	You can also think of this as the size
	of the read of the subchunk following this
	number.
	44        *   Data             The actual sound data.


	The thing that makes reading wav files tricky is that java has no unsigned types.  This means that the
	binary data can't just be read and cast appropriately.  Also, we have to use larger types
	than are normally necessary.

	In many languages including java, an integer is represented by 4 bytes.  The issue here is
	that in most languages, integers can be signed or unsigned, and in wav files the  integers
	are unsigned.  So, to make sure that we can store the proper values, we have to use longs
	to hold integers, and integers to hold shorts.

	Then, we have to convert back when we want to save our wav data.

	It's complicated, but ultimately, it just results in a few extra functions at the bottom of
	this file.  Once you understand the issue, there is no reason to pay any more attention
	to it.

	ALSO:

	This code won't read ALL wav files.  This does not use to full specification.  It just uses
	a trimmed down version that most wav files adhere to.

	 */

	ByteArrayOutputStream byteArrayOutputStream;
	AudioFormat audioFormat;
	TargetDataLine targetDataLine;
	AudioInputStream audioInputStream;
	SourceDataLine sourceDataLine;
	float frequency = 8000.0F; //8000,11025,16000,22050,44100
	int samplesize = 16;
	private String myPath;

	// I made this public so that you can toss whatever you want in here
	// maybe a recorded buffer, maybe just whatever you want
	private byte[] myData = new byte[2048];
	private int myDataPosition = 0;
	private int fSampleRate;

	// constructor takes a wav path
	public SHWave(String tmpPath, int sampleRate, CascadingLoggerInterface logger)
	{
		log = logger.clone(SHWave.class);
		myPath = tmpPath;
		fSampleRate = sampleRate;
	}

	public void addData(byte[] samples)
	{
		if (myDataPosition + samples.length >= myData.length)
		{
			int newLength = (myDataPosition + samples.length) << 1;
			byte newArray[] = new byte[newLength];
			if(myDataPosition > 0)
			{
			    System.arraycopy(myData, 0, newArray, 0, myDataPosition);
			}
			myData = newArray;
		}
		System.arraycopy(samples, 0, myData, myDataPosition, samples.length);
		myDataPosition += samples.length;
	}

	// write out the wav file
	public boolean save()
	{
		try
		{
			DataOutputStream outFile = new DataOutputStream(new FileOutputStream(myPath));

			// write the wav file per the wav file format
			outFile.writeBytes("RIFF"); // 00 - RIFF
			outFile.write(intToByteArray(36 + myDataPosition), 0, 4); // 04 - how big is the rest of this file?
			outFile.writeBytes("WAVE"); // 08 - WAVE
			outFile.writeBytes("fmt "); // 12 - fmt
			outFile.write(intToByteArray(16), 0, 4); // 16 - size of this chunk
			outFile.write(shortToByteArray((short) 1), 0, 2); // 20 - what is the audio format? 1 for PCM = Pulse Code Modulation
			outFile.write(shortToByteArray((short) 1), 0, 2); // 22 - mono or stereo? 1 or 2?  (or 5 or ???)
			outFile.write(intToByteArray(fSampleRate), 0, 4); // 24 - samples per second (numbers per second)
			outFile.write(intToByteArray(fSampleRate * 2), 0, 4); // 28 - bytes per second
			outFile.write(shortToByteArray((short) 2), 0, 2); // 32 - # of bytes in one sample, for all channels
			outFile.write(shortToByteArray((short) 16), 0, 2); // 34 - how many bits in a sample(number)?  usually 16 or 24
			outFile.writeBytes("data"); // 36 - data
			outFile.write(intToByteArray((int) myDataPosition), 0, 4); // 40 - how big is this data chunk
			outFile.write(myData, 0, myDataPosition); // 44 - the actual data itself - just a long string of numbers
		} catch (Exception ex)
		{
			log.error("error in save method:\n" + ex);
			return false;
		}

		return true;
	}

	// ===========================
	// CONVERT BYTES TO JAVA TYPES
	// ===========================
	// convert a short to a byte array
	public static byte[] intToByteArray(int data)
	{
		return new byte[]
		{ (byte) (data & 0xff), (byte) ((data >>> 8) & 0xff), (byte) ((data >>> 16) & 0xff),
				(byte) ((data >>> 24) & 0xff) };
	}

	// convert a short to a byte array
	public static byte[] shortToByteArray(short data)
	{
		return new byte[]
		{ (byte) (data & 0xff), (byte) ((data >>> 8) & 0xff) };
	}

}
