/*
* Copyright (c) Bill Cox and Isaac Porat 2011 - 2013 all rights reserved.
* This file is released under the GNU General Public License, version 2, or (at your option) any later version.
* See the file SpeechHub_License.txt for more details.
*/

package sh.audio.javasound;

import java.util.concurrent.ArrayBlockingQueue;

import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.BooleanControl;
import javax.sound.sampled.Control;
import javax.sound.sampled.DataLine;
import javax.sound.sampled.FloatControl;
import javax.sound.sampled.SourceDataLine;

import core.utilities.Stopwatch;

import sh.audio.SHAudioDeliveryException;
import sh.audio.SHAudioDeliveryCombinedInterface;

import log.common.CascadingLoggerInterface;

/**
* Delivers audio data to Javasound from combined speech to audio conversion and delivery player  
*  
 * @author Isaac Porat
 * @author Bill Cox
 *
 */
public class SHAudioDeliveryCombinedJavasound implements SHAudioDeliveryCombinedInterface, Runnable
{
	private CascadingLoggerInterface log;
	// Enable debug mode below to print additional audio parameters and use Stopwatch to measure response between cancel and writing to line
	private final static boolean DEBUG_MODE = false;
	private final static float MIN_PLAY_LENGTH = 0.6f; // In seconds - if we are smaller than this we add silence.
	private ArrayBlockingQueue<byte[]> fAudioBuffer = new ArrayBlockingQueue<byte[]>(1024, false);
	private SourceDataLine fLine;
	private BooleanControl fMuteControl;
	private FloatControl fGainControl;
	private int fAudioSampleRate;
	private int fTotalWritten = 0;
	private int fSilenceSamples = 0;
	private boolean fOSWindows = false;
	private boolean fFlushPending = false;
	private boolean fControlsErrorFlag = false;
	/*	
	From an answer to question in stack overflow website:
	buffer size in the open(AudioFormat,int) method. 
	A delay of 10ms-100ms will be acceptable for realtime audio. 
	Very low latencies
	will not work on all computer systems, and 100ms or more will probably be annoying for users. 
	A good tradeoff is, e.g. 50ms. 
	For audio format,
	8-bit, mono at 44100Hz, a good buffer size is 2200 bytes, which is almost 50ms.

	Note that different OS's have different audio capabilities in Java. On Windows and Linux you can work with quite small buffer sizes, but OS X uses
	an old implementation with significantly larger delay.

	Writing data byte by byte to the SourceDataLine is very inefficient (the buffer size is set in the open() method, not in write() )
	as a rule of thumb always write one full buffer size to the SourceDataLine.
	
	IP - actually a quarter of the buffer size found to give good performance in Windows
	50ms  buffer size seems to be the minimum where Javasound works reliably without hunting in Windows
	*/
	private int fLineBufSize = 1024; // Approx 25ms at 22050 bps 16 bits  	
	private int fMaxWriteSize = 256; // Approx 6ms
	private byte[] fCancelSilenceBuf; // Calculated to provide approx 6ms silence after cancel
	private boolean fWriting = false;
	// Debug stopwatch
	private Stopwatch fStopwatch = new Stopwatch();

	public SHAudioDeliveryCombinedJavasound(String engineIdentifier, int sampleRate, CascadingLoggerInterface logger)
			throws SHAudioDeliveryException
	{
		try
		{
			log = logger.clone(SHAudioDeliveryCombinedJavasound.class);
			fAudioSampleRate = sampleRate;
			log.notify("Engine identifier: " + engineIdentifier + " Sample rate: " + fAudioSampleRate);
			fLineBufSize = fAudioSampleRate * 2 / 20; // 50ms
			if (fLineBufSize % 2 != 0)
				fLineBufSize++;
			fMaxWriteSize = fLineBufSize / 4; // 12.5 ms
			if (fMaxWriteSize % 2 != 0)
				fMaxWriteSize++;
			int cancelSilenceBufSize = fAudioSampleRate * 2 / 160; // 6ms
			if (cancelSilenceBufSize % 2 != 0)
				cancelSilenceBufSize++;
			fCancelSilenceBuf = new byte[cancelSilenceBufSize];
			if (DEBUG_MODE)
			{
				log.notify("fLineMaxSize: " + fLineBufSize);
				log.notify("fMaxWriteSize: " + fMaxWriteSize);
				log.notify("cancelSilenceBufSize: " + cancelSilenceBufSize);
			}
			AudioFormat audioFormat = new AudioFormat(fAudioSampleRate, 16, 1, true, false);
			DataLine.Info info = new DataLine.Info(SourceDataLine.class, audioFormat);
			fLine = (SourceDataLine) AudioSystem.getLine(info);
			fOSWindows = core.utilities.OS.isWindows();
			Thread writeThread = new Thread(this);
			//writeThread.setPriority(Thread.MAX_PRIORITY);
			writeThread.setName(SHAudioDeliveryCombinedJavasound.class.getName());
			writeThread.start();
		} catch (Exception ex)
		{
			throw new SHAudioDeliveryException(ex.toString());
		}
	}

	@Override
	public void open() throws SHAudioDeliveryException
	{
		try
		{
			AudioFormat audioFormat = new AudioFormat(fAudioSampleRate, 16, 1, true, false);
			fLine.open(audioFormat, fLineBufSize);
			try
			{
				fControlsErrorFlag = false;
				fGainControl = (FloatControl) fLine.getControl(FloatControl.Type.MASTER_GAIN);
				fMuteControl = (BooleanControl) fLine.getControl(BooleanControl.Type.MUTE);
			} catch (Exception ex)
			{
				log.error("Error in obtaining controls in open method:\n" + ex);
				fControlsErrorFlag = true;
				log.notify("The controls available for the audio line are:");
				try
				{
					Control[] controls = fLine.getControls();
					for (int n = 0; n < controls.length; n++)
					{
						log.notify(controls[n].toString());
					}
				} catch (Exception e)
				{
				}
			}
			fLine.start();
		} catch (Exception ex)
		{
			throw new SHAudioDeliveryException(ex.toString());
		}
	}

	@Override
	public void close() throws SHAudioDeliveryException
	{
		try
		{
			fLine.close();
		} catch (Exception ex)
		{
			throw new SHAudioDeliveryException(ex.toString());
		}
	}

	@Override
	public void start() throws SHAudioDeliveryException
	{
		fLine.start();
	}

	@Override
	public void stop() throws SHAudioDeliveryException
	{
		try
		{
			fLine.stop();
		} catch (Exception ex)
		{
			throw new SHAudioDeliveryException(ex.toString());
		}
	}

	@Override
	public void write(byte[] bytes) throws SHAudioDeliveryException
	{
		try
		{
			fAudioBuffer.put(bytes);
		} catch (Exception ex)
		{
			log.error("Error in write method:\n" + ex);
		}
	}

	@Override
	public void flush() throws SHAudioDeliveryException
	{
		fFlushPending = true;
		flushBuffer();
		flushLine();
	}

	private synchronized void flushBuffer() throws SHAudioDeliveryException
	{
		// Synchronized so that flushLine and flushBuffer are not entered at the same time
		try
		{
			if (fAudioBuffer.isEmpty() == false)
			{
				fAudioBuffer.clear();
			}
		} catch (Exception ex)
		{
			throw new SHAudioDeliveryException(ex.toString());
		}
	}

	private synchronized void flushLine() throws SHAudioDeliveryException
	{
		// Synchronized so that flushLine and flushBuffer are not entered at the same time
		try
		{
			if (fLine.available() < fLineBufSize)
			{
				fLine.flush();
				// It seems that a small silence sequence eliminate crackling
				fLine.write(fCancelSilenceBuf, 0, fCancelSilenceBuf.length);
			}
		} catch (Exception ex)
		{
			throw new SHAudioDeliveryException(ex.toString());
		}
	}

	@Override
	public int available()
	{
		return (fLine.available());
	}

	@Override
	public void indicateSpeak(String message)
	{
		fFlushPending = false;
		if (DEBUG_MODE)
			log.notify("m: " + message);
	}

	@Override
	public void indicateCancel()
	{
		fSilenceSamples = 0;
		fFlushPending = true;
		if (DEBUG_MODE)
		{
			log.notify("c:");
			if (fStopwatch.isRunning())
				fStopwatch.stop();
			fStopwatch.start();
		}
	}

	@Override
	public void setTotalWritten(int totalWritten)
	{
		fTotalWritten = totalWritten;
	}

	@Override
	public int getTotalWritten()
	{
		return (fTotalWritten);
	}

	// The drain method is broken in both Linux and Windows.  Further, we can't even get short speech
	// samples to play unless we either call the broken drain, or add silence.  Since we are forced
	// to add silence, we want to detect the case where a new speak command has arrived, so we can
	// abort the playing of silence.  In this method, we determine if we need to add any silence to
	// start playback, and if so, we record how much we've added to the play queue.
	private synchronized void playSilenceIfNeeded() throws SHAudioDeliveryException
	{
		if (fTotalWritten > fLineBufSize)
		{
			return;
		}
		int neededSamples = (int) (MIN_PLAY_LENGTH * fAudioSampleRate);
		if (neededSamples * 2 <= fTotalWritten)
			return;
		fSilenceSamples = 2 * neededSamples - fTotalWritten;
		int available = fLine.available();
		if (fSilenceSamples > available)
		{
			// Avoid blocking
			fSilenceSamples = available;
		}
		byte[] samples = new byte[fSilenceSamples];
		fLine.write(samples, 0, samples.length);
	}

	@Override
	public void drain() throws SHAudioDeliveryException
	{
		try
		{
			while (fAudioBuffer.isEmpty() == false)
			{
				Thread.sleep(2);
				if (fFlushPending)
					return;
			}
			while (fWriting)
			{
				Thread.sleep(2);
				if (fFlushPending)
					return;
			}
			// SourceDataLine.drain seems to be broken on Windows 7.  This hack waits for samples to be played.
			if (fOSWindows)
			{
				while (fLine.available() < fLineBufSize)
				{
					Thread.sleep(2);
					if (fFlushPending)
						return;
				}
				fLine.drain();
			} else
			{
				playSilenceIfNeeded();
				while (fLine.isActive() && fLine.isRunning())
				{
					Thread.sleep(2);
					if (fFlushPending)
						return;
				}
			}
		} catch (Exception ex)
		{
			log.error("Error in drain method:\n" + ex);
		}
	}

	public void setMute(boolean status)
	{
		// If controls do not work we don't want to generate exceptions frequently
		if (fControlsErrorFlag)
			return;
		fMuteControl.setValue(status);
	}

	public void setGain(float gain)
	{
		// Gain control in Windows 7 pro 32 bits might be different with different OS
		// Master Gain with default value: 0.0 dB (range: -80.0 - 6.0206)
		// If controls do not work we don't want to generate exceptions frequently
		if (fControlsErrorFlag)
			return;
		fGainControl.setValue(gain);
	}

	private void playAudio(byte[] bytes)
	{
		try
		{
			fWriting = true;
			int length = bytes.length;
			int offset = 0;
			int bufSize = fMaxWriteSize;

			// Write the bytes in no more than MAX_WRITE_SIZE
			for (;;)
			{
				if (bufSize > length)
					bufSize = length;
				else if (length - offset > fLineBufSize)
					bufSize = fLineBufSize;
				else
					bufSize = length - offset;
				if (bufSize > fMaxWriteSize)
					bufSize = fMaxWriteSize;
				fLine.write(bytes, offset, bufSize);
				offset += bufSize;

				if (fFlushPending)
				{
					flushLine();
					fWriting = false;
					return;
				}
				if (offset >= length)
				{
					fWriting = false;
					return;
				}
			}
		} catch (Exception ex)
		{
			log.error("Error in playAudio method:\n" + ex);
			fWriting = false;
		}
	}

	@Override
	public void run()
	{
		try
		{
			for (;;)
			{
				byte[] bytes = fAudioBuffer.take();
				if (DEBUG_MODE)
				{
					if (fStopwatch.isRunning())
					{
						fStopwatch.stop();
						log.notify(fStopwatch.toString());
					}
				}
				playAudio(bytes);
			}
		} catch (Exception ex)
		{
			log.error("Error in run method:\n" + ex);
		}
	}

}
