/*
* Copyright (c) Isaac Porat 2011 - 2013 all rights reserved.
* This file is released under the GNU General Public License, version 2, or (at your option) any later version.
* See the file SpeechHub_License.txt for more details.
*/

package sh.audio.javasound;

import java.util.ArrayList;
import java.util.List;

import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.Line;
import javax.sound.sampled.Mixer;
import javax.sound.sampled.Port;
import javax.sound.sampled.SourceDataLine;

import core.utilities.StackTrace;

import log.common.CascadingLoggerInterface;

/**
 * Enumerate and select audio output device / driver using mixers 
 * @author Isaac Porat
 * 
 * The virtual driver in Windows "Microsoft sound mapper" is not displayed on the test machine
 * but is reported to be displayed on other machines based on various forums
 * It is not clear if the above driver is used by Javasound by default or not
 * As this virtual device is important for switching automatically between drivers set by the user in control panel 
 * this issues needs to be investigated perhaps with a USB driver
 * Default mixer is obtained by calling the getMixer method on AudioSystem with a null parameter
 * This might not be the exact equivalent as it returns the 'Primary Sound Driver'
 * This is the direct sound driver which by default uses the Primary Sound Device which is selected by the user in Control Panel 
 * So probably achieve the same thing as the 'Microsoft Sound Mapper'
 * 
 * Tester example in Windows as displayed by NVDA Output Device in the synthesizer dialog configuration
* Output device: Headset Earphone (Microsoft Lif  
* Output device: Digital Audio (HDMI) (2- High D
* Output device: Digital Audio (S/PDIF) (High De,
  * Output device: Speakers (High Definition Audio
 *     
 */
public class SHAudioMixersJavasound
{
	private CascadingLoggerInterface log;
	public static String SYSTEM_DEFAULT_MIXER = "default";

	public SHAudioMixersJavasound(CascadingLoggerInterface logger)
	{
		log = logger.clone(SHAudioMixersJavasound.class);
	}

	public Mixer getMixer(String mixerIdentifier) throws Exception
	{
		if (mixerIdentifier.equals(SYSTEM_DEFAULT_MIXER))
			return (getSystemDefaultMixer());
		Mixer mixer = getMixerSourceDataLine(mixerIdentifier);
		if (mixer != null)
		{
			if (openMixerPorts(mixerIdentifier))
				return (mixer);
		}
		log.error("Mixer " + mixerIdentifier + " not found, reverting to default mixer");
		return (getSystemDefaultMixer());
	}

	private Mixer getSystemDefaultMixer()
	{
		try
		{
			// In Windows 7 it is "Primary Sound driver"
			return (AudioSystem.getMixer(null));
		} catch (Exception ex)
		{
			log.error("Error in getSystemDefaultMixer method:\n" + StackTrace.stackToString(ex));
			return (null);
		}
	}

	public String[] getMixerWithPortsIdentifiers() throws Exception
	{
		try
		{
			List<String> mixerIdentifierList = new ArrayList<String>();
			for (Mixer.Info mixerInfo : AudioSystem.getMixerInfo())
			{
				Mixer mixer = AudioSystem.getMixer(mixerInfo);
				if ((mixer.isLineSupported(Port.Info.SPEAKER)) || (mixer.isLineSupported(Port.Info.HEADPHONE)))
				// Output line test disabled in case users loose speech				
				//(mixer.isLineSupported(Port.Info.LINE_OUT)))
				{
					// found a Mixer with the desired port
					String mixerName = mixerInfo.getName();
					// Some mixers are port mixers which means they do not contain source line (presumably connection in hardware) 
					// Other mixers have only source lines without port and a matching name without the port prefix hence 'port' removal
					mixerName = mixerName.replaceAll("^Port", "").trim();
					//log.notify(mixerName);
					//log.notify(mixerInfo);
					//log.notify(mixerInfo.getDescription());
					mixerIdentifierList.add(mixerName);
				}
			}
			return (mixerIdentifierList.toArray(new String[0]));
		} catch (Exception ex)
		{
			log.error("Error in getMixersWithPortsIdentifiers method:\n" + StackTrace.stackToString(ex));
			return (null);
		}
	}

	private boolean openMixerPorts(String mixerIdentifier) throws Exception
	{
		try
		{
			boolean status = false;
			for (Mixer.Info mixerInfo : AudioSystem.getMixerInfo())
			{
				Mixer mixer = AudioSystem.getMixer(mixerInfo);
				Line.Info portInfo = new Line.Info(Port.class);
				if (mixer.isLineSupported(portInfo))
				{
					String mixerName = mixerInfo.getName();
					if (mixerName.toLowerCase().contains(mixerIdentifier.toLowerCase()) == false)
						continue;
					if (mixer.isLineSupported(Port.Info.SPEAKER))
					{
						mixer.getLine(Port.Info.SPEAKER).open();
						status = true;
					}
					if (mixer.isLineSupported(Port.Info.HEADPHONE))
					{
						mixer.getLine(Port.Info.HEADPHONE).open();
						status = true;
					}
					// Output line test disabled in case users loose speech				
					/*
					if(mixer.isLineSupported(Port.Info.LINE_OUT))
					{
						mixer.getLine(Port.Info.LINE_OUT).open();
								status = true;
					}
					*/
				}
			}
			return (status);
		} catch (Exception ex)
		{
			log.error("Error in openMixersWithPorts method:\n" + StackTrace.stackToString(ex));
			return (false);
		}
	}

	private Mixer getMixerSourceDataLine(String mixerIdentifier) throws Exception
	{
		try
		{
			for (Mixer.Info mixerInfo : AudioSystem.getMixerInfo())
			{
				Mixer mixer = AudioSystem.getMixer(mixerInfo);
				Line.Info sourceDataLineInfo = new Line.Info(SourceDataLine.class);
				if (mixer.isLineSupported(sourceDataLineInfo))
				{
					String mixerName = mixerInfo.getName();
					if (mixerName.toLowerCase().contains(mixerIdentifier.toLowerCase()))
						return (mixer);
				}
			}
			return (null);
		} catch (Exception ex)
		{
			log.error("Error in getMixerSourceDataLine method:\n" + StackTrace.stackToString(ex));
			return (null);
		}
	}

	public void allMixersProbe()
	{
		try
		{
			Line.Info showAll = new Line.Info(Object.class);
			//iterate mixers
			for (Mixer.Info info : AudioSystem.getMixerInfo())
			{
				Mixer mixer = AudioSystem.getMixer(info);
				//print mixer details
				log.notify(info.toString() + "(" + mixer.toString() + ")\n");
				//iterate SourceDataLines for the mixer, printing info
				for (Line.Info lineinfo : mixer.getSourceLineInfo(showAll))
				{
					log.notify("...source line: " + lineinfo.toString() + " (" + lineinfo.getClass().getName() + ")\n");
				}
				//iterate TargetDataLines for the mixer, printing info
				for (Line.Info lineinfo : mixer.getTargetLineInfo(showAll))
				{
					log.notify("...target line: " + lineinfo.toString() + " (" + lineinfo.getClass().getName() + ")\n");
				}
			}
		} catch (Exception ex)
		{
			log.error("Error in allMixersProb method:\n" + StackTrace.stackToString(ex));
		}
	}

}
