/*
* Copyright (c) Isaac Porat and Bill Cox 2011 - 2013 all rights reserved.
* This file is released under the GNU General Public License, version 2, or (at your option) any later version.
* See the file SpeechHub_License.txt for more details.
*/

package sh.audio.javasound;

import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.BooleanControl;
import javax.sound.sampled.Control;
import javax.sound.sampled.DataLine;
import javax.sound.sampled.FloatControl;
import javax.sound.sampled.Mixer;
import javax.sound.sampled.SourceDataLine;

import log.common.CascadingLoggerInterface;
import sh.audio.SHAudioDeliveryException;
import sh.audio.SHAudioDeliveryParallelInterface;
import sh.connections.SHMessageEnvelopeInterface;
import sh.connections.SHUpdateParametersInterface;
import sh.connections.SHConstants.MessageTextToAudioConversionState;
import sh.connections.SHConstants.MessageType;
import core.utilities.StackTrace;
import core.utilities.Stopwatch;

/**
 * This implementation uses an additional  drain thread
 * It is important to use the switches to enable concurrent garbage collector with automatic sweap mode to reduce pauses caused by the GC kicking in
 * This implementation is not currently used as both the single thread and even more the trim silence approaches work better 
* In this implementation for single characters and upon cancel a different drain thread is used and the playAudioData method is returned immediatly after writing the audio data to the line  
 * The advantage of the double threads method is potentially no delay in case where characters are following each other without cancel for example during spelling or with key echo on when navigating in the desktop
 * This is achieved with a modified playSilenceIfNeeded method where the trailing silence is terminated upon arrival of a new message
 * In practice some silence had to be introduced between messages to reduce popping sound
 * Nevertheless spelling sounds a bit harsh and not absolutely reliable 
 * Typing is reliable but rarely there are missing typed echo characters
 * Since the single thread implementation is reliable this class was not debugged fully
 * Applies to Windows XP, Vista, Windows 7 and Windows 8
 * At present contains some Linux related code which will be removed once the Linux version is developed 
* Delivers audio data to Javasound from parallel speech to audio conversion and delivery player  
* Process audio data generated by the synthesizer
* optimize data to reduce crackling when cancellin speech by optimizing
* flow to audio line and inserting a minimal period of silence
* Optimize data flow to line for slow computer by detecting audio data gaps 
* and inserting catchup delay before a message
* to allow slow computer to cope with the slow engine
*     
 * @author Isaac Porat
 * @author Bill Cox
 *
 */
public class SHWindowsDrainThreadAudioDeliveryParallelJavasound implements SHAudioDeliveryParallelInterface
{
	private CascadingLoggerInterface log;
	// Enable debug timing mode below to print additional audio parameters and use Stopwatch to measure response between cancel and next message first writing audio to line
	private final static boolean DEBUG_TIMING_MODE = false;
	// Debug audio starvation mode provides information on audio starvation when occurs
	private final static boolean DEBUG_AUDIO_STARVATION_MODE = false;
	// Debug play silence mode gives information on the operation of the playSilenceIfNeeded method 
	private final static boolean DEBUG_PLAY_SILENCE_MODE = false;
	private final static float MIN_PLAY_LENGTH = 0.6f; // In seconds - if the actual time is smaller than this then silence is added.
	private int fSilenceNeededSamples = 0;
	private SHAudioMixersJavasound fAudioMixers;
	private SourceDataLine fLine;
	private BooleanControl fMuteControl;
	private FloatControl fGainControl;
	private int fAudioSampleRate;
	private int fOldAudioSampleRate;
	private int fSilenceSamples = 0;
	private boolean fFlushPending = false;
	private boolean fControlsErrorFlag = false;
	/*	
	From an answer to question in stack overflow website:
	buffer size in the open(AudioFormat,int) method. 
	A delay of 10ms-100ms will be acceptable for realtime audio. 
	Very low latencies
	will not work on all computer systems, and 100ms or more will probably be annoying for users. 
	A good tradeoff is, e.g. 50ms. 
	For audio format,
	8-bit, mono at 44100Hz, a good buffer size is 2200 bytes, which is almost 50ms.

	Note that different OS's have different audio capabilities in Java. On Windows and Linux you can work with quite small buffer sizes, but OS X uses
	an old implementation with significantly larger delay.

	Writing data byte by byte to the SourceDataLine is very inefficient (the buffer size is set in the open() method, not in write() )
	as a rule of thumb always write one full buffer size to the SourceDataLine.
	IP - actually after experimentation in ppractice a different approach was used here:
	It turned out that buffer size can be alot larger we use here 1 second.
	To prevent scratching noise during cancel it is important to make sure the audio line
	does not starve and not to use the start and stop methods of the line
	It is important to feed the line in small chunks (25ms were found to be adequate so that cancel
	in case of line blocking can be rapid after a flush.
	It was found that feeding the line with a small period of silence reduce cancel noise.
	*/
	private int fLineBufSize = 1024; // Approx 25msat 22050 bps 16 bits  	
	private int fMaxWriteSize = 256; // Approx 6 ms at above conditions
	private byte[] fCancelSilenceBuf; // Calculated to provide approx 6ms silence after cancel
	@SuppressWarnings("unused")
	private boolean fWriting = false;
	private SHMessageEnvelopeInterface fMessageEnvelope;
	private boolean fAudioMessageStarted = false;
	private int fAudioDataStarvationCounter = 0;
	private boolean fPreviousMessageAudioDataStarved = false;
	private boolean fCatchupMode = false;
	private float fCatchupWaitPerChar = 0.0f;
	private boolean fRequiredLineReset = false;
	private Thread fDrainThread;
	private boolean fDrainRequired = false;
	private String fMessage = "";
	private boolean fSpeakingChar = false;
	private int fSamplesChunkCounter = 0;
	private int fPreviousSamplesChunkLength = 0;
	private boolean fNewMessageArrived = false;
	// Debug Stopwatch 
	private Stopwatch fStopwatch;

	public SHWindowsDrainThreadAudioDeliveryParallelJavasound(CascadingLoggerInterface logger)
	{
		log = logger.clone(SHWindowsDrainThreadAudioDeliveryParallelJavasound.class);
		if (DEBUG_TIMING_MODE)
			fStopwatch = new Stopwatch();
		fAudioMixers = new SHAudioMixersJavasound(logger);
		fDrainThread = new Thread(new DrainThread());
		fDrainThread.setName(SHWindowsDrainThreadAudioDeliveryParallelJavasound.class.getName());
		fDrainThread.start();
	}

	private void open() throws SHAudioDeliveryException
	{
		try
		{
			calculateAudioParameters();
			openAudioLine();
		} catch (Exception ex)
		{
			throw new SHAudioDeliveryException("Error in open method:\n" + ex);
		}
	}

	private void calculateAudioParameters() throws SHAudioDeliveryException
	{
		try
		{
			// configuration 1 (as applied in the combined version together with configuration 1 below) 
			//fLineBufSize = fAudioSampleRate * 2 / 20; // 50ms
			// later configuration 2 (together with configuration 2 below) 			
			fLineBufSize = fAudioSampleRate * 2; // 1 sec (16 bits resolution) 
			if (fLineBufSize % 2 != 0)
				fLineBufSize++;
			// Configuration 1 (together with configuration 1 above)  
			//			fMaxWriteSize = fLineBufSize / 4; // 12.5 ms
			// Configuration 2 (together with configuration 2 above)
			fMaxWriteSize = fLineBufSize / 40; // 25 ms
			if (fMaxWriteSize % 2 != 0)
				fMaxWriteSize++;
			fSilenceNeededSamples = (int) (MIN_PLAY_LENGTH * fAudioSampleRate * 2);
			if (fSilenceNeededSamples % 2 != 0)
				fSilenceNeededSamples++;
			int cancelSilenceBufSize = fAudioSampleRate * 2 / 160; // Approax 6 ms
			if (cancelSilenceBufSize % 2 != 0)
				cancelSilenceBufSize++;
			fCancelSilenceBuf = new byte[cancelSilenceBufSize];
			if (DEBUG_TIMING_MODE)
			{
				log.notify("fLineMaxSize: " + fLineBufSize);
				log.notify("fMaxWriteSize: " + fMaxWriteSize);
				log.notify("cancelSilenceBufSize: " + cancelSilenceBufSize);
			}
		} catch (Exception ex)
		{
			throw new SHAudioDeliveryException(ex.toString());
		}
	}

	private void openAudioLine() throws SHAudioDeliveryException
	{
		try
		{
			AudioFormat audioFormat = new AudioFormat(fAudioSampleRate, 16, 1, true, false);
			DataLine.Info info = new DataLine.Info(SourceDataLine.class, audioFormat);
			Mixer mixer = fAudioMixers.getMixer(SHAudioMixersJavasound.SYSTEM_DEFAULT_MIXER);
			log.notify("Mixer used: " + mixer.getMixerInfo().getName());
			fLine = (SourceDataLine) mixer.getLine(info);
			// Previous method of obtaining the default line directly
			//fLine = (SourceDataLine) AudioSystem.getLine(info);
			fLine.open(audioFormat, fLineBufSize);
			try
			{
				fControlsErrorFlag = false;
				fGainControl = (FloatControl) fLine.getControl(FloatControl.Type.MASTER_GAIN);
				fMuteControl = (BooleanControl) fLine.getControl(BooleanControl.Type.MUTE);
			} catch (Exception ex)
			{
				log.error("Error in obtaining controls in open method:\n" + ex);
				fControlsErrorFlag = true;
				log.notify("The controls available for the audio line are:");
				try
				{
					Control[] controls = fLine.getControls();
					for (int n = 0; n < controls.length; n++)
					{
						log.notify(controls[n].toString());
					}
				} catch (Exception e)
				{
				}
			}
			fLine.start();
		} catch (Exception ex)
		{
			throw new SHAudioDeliveryException("Error in openAudioLine method:\n" + ex);
		}
	}

	@Override
	public void close() throws SHAudioDeliveryException
	{
		try
		{
			if (fLine != null)
				fLine.close();
		} catch (Exception ex)
		{
			throw new SHAudioDeliveryException("Error in close method:\n" + ex);
		} finally
		{
			fLine = null;
		}
	}

	@Override
	public void start() throws SHAudioDeliveryException
	{
		try
		{
			fLine.start();
		} catch (Exception ex)
		{
			throw new SHAudioDeliveryException("Error in start method:\n" + ex);
		}
	}

	@Override
	public void stop() throws SHAudioDeliveryException
	{
		try
		{
			fLine.stop();
		} catch (Exception ex)
		{
			throw new SHAudioDeliveryException("Error in stop method:\n" + ex);
		}
	}

	@Override
	public void flush()
	{
		try
		{
			fFlushPending = true;
			flushBuffer();
			flushLine();
		} catch (Exception ex)
		{
			log.error("Error in flush method:\n" + ex);
		}
	}

	private synchronized void flushBuffer() throws Exception
	{
		// Synchronized so that flushLine and flushBuffer are not entered at the same time
		if (fMessageEnvelope == null)
			return;
		if (fMessageEnvelope.isEmptyAudio() == false)
			fMessageEnvelope.clearAudio();
		// Markf with end of message indicator
		fMessageEnvelope.putAudio(new byte[0]);
	}

	private synchronized void flushLine() throws Exception
	{
		// Synchronized so that flushLine and flushBuffer are not entered at the same time
		if (fLine == null)
			return;
		if (fLine.available() < fLineBufSize)
		{
			fLine.flush();
			// It seems that a small silence sequence eliminate crackling
			fLine.write(fCancelSilenceBuf, 0, fCancelSilenceBuf.length);
		}
	}

	@Override
	public void disableCatchupMode()
	{
		fCatchupWaitPerChar = 0.0f;
		fCatchupMode = false;
	}

	@Override
	public boolean getCatchupMode()
	{
		return (fCatchupMode);
	}

	@Override
	public boolean playAudioData(SHMessageEnvelopeInterface messageEnvelope)
	{
		try
		{
			fNewMessageArrived = true;
			// Make sure previous message is drained or flushed
			synchronized (this)
			{
				while (fDrainRequired)
					wait(5000); // Time not required just as added assurance needs to be bigger than fLineBufSize in time 
			}
			fSamplesChunkCounter = 0;
			fFlushPending = false;
			fMessageEnvelope = messageEnvelope;
			fMessage = fMessageEnvelope.getMessage();
			fSpeakingChar = fMessageEnvelope.getMessageType() == MessageType.CHAR;
			int silenceBeforeMessage = fMessageEnvelope.getSilenceBeforeMessage();
			if (silenceBeforeMessage > 0)
				Thread.sleep(silenceBeforeMessage);
			fAudioSampleRate = fMessageEnvelope.getAudioSampleRate();
			if ((fLine == null) || (fAudioSampleRate != fOldAudioSampleRate))
			{
				close();
				open();
				fOldAudioSampleRate = fAudioSampleRate;
				SHUpdateParametersInterface updateParameters = fMessageEnvelope.getUpdateParameters();
				// UpdateParameters can be null if the player is used in the TTS stack 
				if (updateParameters != null)
				{
					log.notify("Engine identifier: " + updateParameters.getEngineIdentifier() + " Voice Identifier: "
							+ updateParameters.getVoiceIdentifier() + " Sample rate: " + fAudioSampleRate);
				}
			}
			if (waitForAudioData())
				playAudioBuffer();
			if (fRequiredLineReset)
			{
				close();
				open();
				fRequiredLineReset = false;
			}
			if (fAudioDataStarvationCounter > 1)
				fPreviousMessageAudioDataStarved = true;
			else
				fPreviousMessageAudioDataStarved = false;
			/*
						 * It is important to drain all messages either directly or on separate thread
						 * to speed return of this method and hence the player's speak method
						 * Drain is important for two reasons, first to make sure messages are delivered completely and not truncated
						 * The second reason is that due to a bug in Javasound if drain is not called sometimes 
										 * it results in a repeated echo of a fraction of the last audio fragment
										 */
			if (fSpeakingChar || (fFlushPending))
			{
				// Drain on a different thread and return early			
				fDrainRequired = true;
				synchronized (this)
				{
					notifyAll();
				}
			} else
			{
				// Messages of more than one character block until draine or cancelled
				drain();
				fDrainRequired = false;
			}
			return (true);
		} catch (Exception ex)
		{
			log.error("Error in playAudioData method:\n" + StackTrace.stackToString(ex));
			return (false);
		}
	}

	private boolean waitForAudioData() throws InterruptedException
	{
		if (fCatchupMode == false)
			return (true);
		return (waitForAudioDataPerChar());
		//return(waitForAudioDataComplete());	
	}

	/**
	 * Wait for a period based on nominal value multiplied by the number of chars in a message 
	 * @throws InterruptedException
	 * @return false if interrupted by cancel 
	 */
	private boolean waitForAudioDataPerChar() throws InterruptedException
	{
		if (fMessageEnvelope.getAudioCacheUsed()) // Audio stored in cache, no need for delay 
			return (true);
		int calculationLength = fMessage.length();
		if (calculationLength > 50)
			calculationLength = 50;
		int delay = (int) ((fCatchupWaitPerChar) * calculationLength);
		for (;;)
		{
			if (fFlushPending)
				return (false);
			if (delay < 25)
			{
				Thread.sleep(delay);
				break;
			} else
			{
				delay -= 25;
				Thread.sleep(25);
			}
		}
		return (true);
	}

	/**
	 * Waits until the message is completely converted to audio data
	 * @throws InterruptedException
	 */
	@SuppressWarnings("unused")
	private boolean waitForAudioDataComplete() throws InterruptedException
	{
		if (fMessageEnvelope.getAudioCacheUsed()) // Audio stored in cache, no need for delay 
			return (true);
		for (;;)
		{
			if (fMessageEnvelope.getMessageTextToAudioConversionState().equals(
					MessageTextToAudioConversionState.UNKNOWN) == false) // Text to audio data conversion finished
				return (true);
			Thread.sleep(50);
		}
	}

	@Override
	public void cancel()
	{
		fSilenceSamples = 0;
		if (DEBUG_TIMING_MODE)
		{
			if (fStopwatch.isRunning())
				fStopwatch.stop();
			fStopwatch.start();
		}
		if (fFlushPending == false)
			flush();
	}

	// The drain method is broken in both Linux and Windows.  Further, we can't even get short speech
	// samples to play unless we either call the broken drain, or add silence.  Since we are forced
	// to add silence, we want to detect the case where a new speak command has arrived, so we can
	// abort the playing of silence.  In this method, we determine if we need to add any silence to
	// start playback, and if so, we record how much we've added to the play queue.
	private void playSilenceIfNeeded() throws SHAudioDeliveryException
	{
		if (fMessageEnvelope == null)
			return;
		fNewMessageArrived = false;
		int totalWritten = fMessageEnvelope.getTotalAudioWritten();
		if (totalWritten > fLineBufSize)
			return;
		// Note here samples refer to individual bytes
		if (fSilenceNeededSamples <= totalWritten)
			return;
		fSilenceSamples = fSilenceNeededSamples - totalWritten;
		int available = fLine.available();
		if (fSilenceSamples > available)
		{
			// Avoid blocking
			fSilenceSamples = available;
		}
		/*
		* Originally lines below delivered delay in one chunk which created delay
		* when characters delivered one after the other without cancel
		* This is still done when audio starvation detected
		*/
		if (DEBUG_PLAY_SILENCE_MODE)
			log.notify("Play silence: " + fMessage);
		if (fAudioDataStarvationCounter > 0) // starvation detected
		{
			byte[] samples = new byte[fSilenceSamples];
			fLine.write(samples, 0, samples.length);
			if (DEBUG_PLAY_SILENCE_MODE)
				log.notify("Complete silence delivered in response to audio starvation");
			return;
		}
		/*
		 * If on another thread (single character or cancel) write silence in chunks and watch for another audio message arriving
		 * If the same thread (long message) this method is not likely to get here 
		 */
		int totalDelivered = 0;
		byte[] samples;
		int chunkCounter = 0;
		for (;;)
		{
			totalDelivered += fMaxWriteSize;
			chunkCounter++;
			if (totalDelivered > fSilenceSamples)
			{
				samples = new byte[totalDelivered - fSilenceSamples];
				fLine.write(samples, 0, samples.length);
				if (DEBUG_PLAY_SILENCE_MODE)
					log.notify("complete: " + chunkCounter);
				return;
			}
			samples = new byte[fMaxWriteSize];
			fLine.write(samples, 0, samples.length);
			for (int n = 0; n < 10; n++) // 20ms, total delay needs to be close but smaller than fMaxWriteSize
			{
				try
				{
					Thread.sleep(2);
				} catch (InterruptedException e)
				{
				}
				if (fFlushPending)
				{
					if (DEBUG_PLAY_SILENCE_MODE)
						log.notify("partial cancel: " + chunkCounter);
					return;
				}
				if (fNewMessageArrived)
				{
					// Increase gap between sequences of messages without cancel for clearer spelling and reduced popping sound
					// Spelling operation is not absolutely reliable, sometimes a letter is missed
					if (chunkCounter < 9)
						break; // Add another silence chunk if available 
					if (DEBUG_PLAY_SILENCE_MODE)
						log.notify("partial new message: " + chunkCounter);
					return;
				}
			}
		}
	}

	@Override
	public void drain()
	{
		try
		{
			/*
			* SourceDataLine.drain seems to be broken on Windows  (and Linux) This hack waits for samples to be played.
			* Extensive testing revealed that characters are correctly delivered to the audio buffer
			* So any missing characters are due to a bug in the line drain method 
			*/
			playSilenceIfNeeded();
			while (fLine.available() < fLineBufSize)
			{
				Thread.sleep(2);
				if (fFlushPending)
				{
					flushLine();
					return;
				}
			}
			fLine.drain();
		} catch (Exception ex)
		{
			log.error("Error in drain method:\n" + ex);
		}
	}

	@Override
	public void setMute(boolean status)
	{
		// If controls do not work we don't want to generate exceptions frequently
		if (fControlsErrorFlag)
			return;
		fMuteControl.setValue(status);
	}

	@Override
	public void setGain(float gain)
	{
		// Gain control in Windows 7 pro 32 bits might be different with different OS
		// Master Gain with default value: 0.0 dB (range: -80.0 - 6.0206)
		// If controls do not work we don't want to generate exceptions frequently
		if (fControlsErrorFlag)
			return;
		fGainControl.setValue(gain);
	}

	private void playAudioLine(byte[] sample)
	{
		try
		{
			fWriting = true;
			int length = sample.length;
			int offset = 0;
			int bufSize = fMaxWriteSize;
			// Write the bytes in no more than fMaxWriteSize
			for (;;)
			{
				if (bufSize > length)
					bufSize = length;
				else if (length - offset > fLineBufSize)
					bufSize = fLineBufSize;
				else
					bufSize = length - offset;
				if (bufSize > fMaxWriteSize)
					bufSize = fMaxWriteSize;
				fSamplesChunkCounter++;
				if ((fLineBufSize == fLine.available()) && fAudioMessageStarted) // Audio data starvation detected
				{
					processAudioStarvation();
					if (DEBUG_AUDIO_STARVATION_MODE)
					{
						log.notify(fMessage);
						log.notify("Samples chunk: " + fSamplesChunkCounter);
						log.notify(" Single char: " + fSpeakingChar + " In cache: "
								+ fMessageEnvelope.getAudioCacheUsed());
						log.notify("Previous chunk length: " + fPreviousSamplesChunkLength + " This chunk length: "
								+ sample.length);
					}
				}
				fPreviousSamplesChunkLength = sample.length;
				fLine.write(sample, offset, bufSize);
				offset += bufSize;
				fAudioMessageStarted = true;
				if (fFlushPending && (fSpeakingChar == false))
				{
					flushLine();
					fWriting = false;
					return;
				}
				if (offset >= length)
				{
					fWriting = false;
					return;
				}
			}
		} catch (Exception ex)
		{
			log.error("Error in playAudio method:\n" + ex);
			fWriting = false;
		}
	}

	private void processAudioStarvation()
	{
		try
		{
			fAudioDataStarvationCounter++;
			if (fAudioDataStarvationCounter == 1) // First starvation detected
			{
				String printedMessage = fMessage;
				if (printedMessage.length() > 20) // Print only the start of the message
					printedMessage = printedMessage.substring(0, 19) + "...";
				log.notify("FMAS: " + printedMessage);
				if (fPreviousMessageAudioDataStarved)
				{
					fAudioDataStarvationCounter = 3; // Trigger reset and added delay below  
					log.error("Previous message starvation detected, audio line reset triggered");
				}
			}
			if (fAudioDataStarvationCounter > 2) // React on more than 2 failures per message
			{
				// Indicate line reset to prevent periodic scratching cause by internal javazoom bug following sometimes audio starvation
				fRequiredLineReset = true;
				if (fCatchupWaitPerChar > 0.0f)
					log.notify("Adding delay to catchup mode");
				else
					log.notify("Detected slow engine response , switching to catchup mode");
				fAudioDataStarvationCounter = 0;
				// Add a nominal delay per character based on sample rate
				fCatchupWaitPerChar += fAudioSampleRate / 1000;
				fCatchupMode = true;
				waitForAudioData();
			}
		} catch (Exception ex)
		{
			log.error("Error in processAudioStarvation method:\n" + ex);
		}
	}

	private boolean playAudioBuffer()
	{
		try
		{
			fAudioMessageStarted = false;
			fAudioDataStarvationCounter = 0;
			int chunkAudioCounter = 0;
			for (;;)
			{
				byte[] sample = fMessageEnvelope.takeAudio();
				chunkAudioCounter++;
				if (sample.length == 0)
				{
					synchronized (this)
					{
						fMessageEnvelope.clearAudio();
					}
					if (chunkAudioCounter > 1)
						return (true);
					if (DEBUG_TIMING_MODE)
					{
						if (fStopwatch.isRunning())
							fStopwatch.stop();
						log.notify(fMessage);
						log.notify("Message with no audio detected, audio not delivered to line");
					}
					return (false); // Abort audio delivery to audio line
				}
				if (DEBUG_TIMING_MODE)
				{
					if (fStopwatch.isRunning())
					{
						fStopwatch.stop();
						log.notify("After cancel:\n" + fMessageEnvelope.getMessage() + "\nAudio delivered: "
								+ fStopwatch.toString());
					}
				}
				playAudioLine(sample);
			}
		} catch (Exception ex)
		{
			log.error("Error in playAudioBuffer method:\n" + ex);
			return (false);
		}
	}

	private void drainAudio()
	{
		try
		{
			synchronized (this)
			{
				while (fDrainRequired == false)
					wait();
			}
			drain();
		} catch (Exception ex)
		{
			log.error("Error in DrainAudio method:\n" + ex);
		} finally
		{
			fDrainRequired = false;
		}
		synchronized (this)
		{
			notifyAll();
		}
	}

	private class DrainThread implements Runnable
	{

		@Override
		public void run()
		{
			for (;;)
			{
				drainAudio();
			}
		}

	}

}
