/*
* Copyright (c) Isaac Porat 2011 - 2013 all rights reserved.
* This file is released under the GNU General Public License, version 2, or (at your option) any later version.
* See the file SpeechHub_License.txt for more details.
*/

package sh.comm;

import core.utilities.StackTrace;
import log.common.CascadingLoggerInterface;
import sh.common.SHMessageEnvelope;
import sh.connections.SHConstants;
import sh.connections.SHConstants.MessageState;
import sh.connections.SHMessageEnvelopeInterface;
import sh.connections.SHAudioDataHandlerInterface;
import sh.connections.SHEventsListenerInterface;
import sun.misc.BASE64Decoder;

/**
 * Client message data communication 
 * Receives message data including metadata and audio data from the  server 
 * @author Isaac Porat
 *
 */
public class SHClientCommMessageData implements SHEventsListenerInterface, Runnable
{
	private CascadingLoggerInterface log;
	private final static boolean DEBUG_MODE = false;
	private SHMessageEnvelopeInterface fClientMessageEnvelope;
	private SHCommMessageDataHandlerInterface fAudioDataMessageHandler;
	private SHAudioDataHandlerInterface fClientAudioDataHandler;
	private int fReceivedMessageIdentifier = -1;
	@SuppressWarnings("unused")
	private String fReceivedMessage = "unknown";
	private int fReceivedAudioSampleRate = 0;
	private MessageState fReceivedMessageState = MessageState.SINGLE;
	private BASE64Decoder fBASE64Decoder = new BASE64Decoder();
	private int fCurrentMessageIdentifier = 0;
	private int fCancelledMessageIdentifier = 0;
	private boolean fCancelPending = false;
	private boolean fClientMessageEnvelopeReady = false;
	private boolean fPlayerReady = true;
	private boolean fAudioDataAvailable = false;
	@SuppressWarnings("unused")
	private MessageEnvelopeProcessing fMessageEnvelopeProcessing;

	public SHClientCommMessageData(SHCommMessageDataHandlerInterface audioDataMessageHandler,
			SHAudioDataHandlerInterface clientAudioDataHandler, CascadingLoggerInterface logger)
	{
		log = logger.clone(SHClientCommMessageData.class);
		fAudioDataMessageHandler = audioDataMessageHandler;
		fClientAudioDataHandler = clientAudioDataHandler;
		Thread dataProcessingThread = new Thread(this);
		dataProcessingThread.setName(SHClientCommMessageData.class.getName());
		dataProcessingThread.start();
		fMessageEnvelopeProcessing = new MessageEnvelopeProcessing();
	}

	public void indicateSpeak(int messageIdentifier, String message, boolean speakSync)
	{
		fCurrentMessageIdentifier = messageIdentifier;
	}

	public boolean cancel()
	{
		fCancelledMessageIdentifier = fCurrentMessageIdentifier;
		fCancelPending = true;
		return (true);
	}

	private boolean processMetadataGroupHeader(String metadata)
	{
		try
		{
			String[] parts = metadata.split(SHConstants.CMDP_METADATA_SEPARATOR);
			String metadataGroupIdentifier = parts[1];
			if (metadataGroupIdentifier.equals(SHConstants.CMDP_GROUP_HEADER))
			{
				fReceivedMessageIdentifier = Integer.parseInt(parts[2]);
				fReceivedAudioSampleRate = Integer.parseInt(parts[3]);
				if (processMetadataMessageState(parts[4]) == false)
					return (false);
				return (true);
			}
			return (false);
		} catch (Exception ex)
		{
			log.error("Error in processMetadataHeader method:\n" + ex);
			return (false);
		}
	}

	private boolean processMetadataMessageState(String metadataMessageState)
	{
		try
		{
			fReceivedMessageState = MessageState.MIDDLE;
			if (metadataMessageState.equals(SHConstants.CMDP_GROUP_HEADER_MESSAGESTATE_SINGLE))
				fReceivedMessageState = MessageState.SINGLE;
			else if (metadataMessageState.equals(SHConstants.CMDP_GROUP_HEADER_MESSAGESTATE_BEGIN))
				fReceivedMessageState = MessageState.BEGIN;
			else if (metadataMessageState.equals(SHConstants.CMDP_GROUP_HEADER_MESSAGESTATE_END))
				fReceivedMessageState = MessageState.END;
			return (true);
		} catch (Exception ex)
		{
			log.error("Error in processMetadataMessageState method:\n" + ex);
			return (false);
		}
	}

	@SuppressWarnings("unused")
	private boolean processMetadataGroupMessage(String metadata)
	{
		try
		{
			// Note split method limits array to 3 to ensure that message can contain also the separator characters
			String[] parts = metadata.split(SHConstants.CMDP_METADATA_SEPARATOR, 3);
			String metadataGroupIdentifier = parts[1];
			if (metadataGroupIdentifier.equals(SHConstants.CMDP_GROUP_MESSAGE))
			{
				fReceivedMessage = parts[2];
				return (true);
			}
			return (false);
		} catch (Exception ex)
		{
			log.error("Error in processMetadataMessage method:\n" + ex);
			return (false);
		}
	}

	private String getMessageData(String type, String identifier)
	{
		return (fAudioDataMessageHandler.getMessageData(type, identifier));
	}

	private String getMessageAudioData()
	{
		try
		{
			return (getMessageData(SHConstants.CMDP_MESSAGE_AUDIODATA, SHConstants.STATUS_NONE));
		} catch (Exception ex)
		{
			log.error("Error in getMessageAudioData method:\n" + ex);
			return (null);
		}
	}

	private String getMessageMetadata(String metadataGroupIdentifier)
	{
		try
		{
			return (getMessageData(SHConstants.CMDP_MESSAGE_METADATA, metadataGroupIdentifier));
		} catch (Exception ex)
		{
			log.error("Error in getMessageMetadata method:\n" + ex);
			return (null);
		}
	}

	private boolean processMessageEnvelope()
	{
		try
		{
			if (DEBUG_MODE)
				log.notify("processMessageEnvelope: start");
			if (processMetadataGroupHeader(getMessageMetadata(SHConstants.CMDP_GROUP_HEADER)) == false)
			{
				log.error("Error in processMessageEnvelope: ffailed to process " + SHConstants.CMDP_GROUP_HEADER
						+ " metadata");
				// Continue, hopefully the parameters from previous messageEnvelope will do
			}
			fPlayerReady = false;
			// Create and set the client message envelope
			fClientMessageEnvelope = new SHMessageEnvelope(fReceivedMessageIdentifier,
					SHConstants.CMDP_GROUP_MESSAGE_OMITTED);
			fClientMessageEnvelope.initializeAudio();
			fClientMessageEnvelope.setAudioSampleRate(fReceivedAudioSampleRate);
			fClientMessageEnvelope.setMessageState(fReceivedMessageState);
			fClientMessageEnvelopeReady = true;
			synchronized (this)
			{
				notifyAll();
			}
			// This method does not return until all audio is processed by the player
			int tempMessageIdentifier = fClientMessageEnvelope.getMessageIdentifier();
			boolean status = fClientAudioDataHandler.processAudio(fClientMessageEnvelope);
			if (DEBUG_MODE)
				log.notify("processMessageEnvelope: success  return from player " + tempMessageIdentifier);
			fPlayerReady = true;
			synchronized (this)
			{
				notifyAll();
			}
			return (status);
		} catch (Exception ex)
		{
			log.error("Error in processMessageEnvelope method:\n" + ex);
			return (false);
		}
	}

	private boolean processMessageData()
	{
		// Process message audio data once the message envelope has been created
		try
		{
			if (DEBUG_MODE)
				log.notify("processMessageData: start " + fClientMessageEnvelope.getMessageIdentifier());
			String commMessage;
			for (;;)
			{
				commMessage = getMessageAudioData();
				if (commMessage.startsWith(SHConstants.CMDP_METADATA_PREFIX))
				{
					if (commMessage.startsWith(SHConstants.CMDP_GROUP_STATUS_CANCEL_MARKER))
					{
						// Do nothing continue to read
						continue;
					}
					// If server not blocking check for unavailable status
					else if (commMessage.startsWith(SHConstants.CMDP_GROUP_STATUS_UNAVAILABLE_MARKER))
					{
						// Try again
						if (DEBUG_MODE)
							log.notify("processMessageData: unavailable data");
						synchronized (this)
						{
							wait(10);
						}
						continue;
					} else
					{
						// Here check later for error code etc
						log.error("Error in processMessageData method: unexpected metadata:\n" + commMessage);
						continue;
					}
				} else
				{
					// Audio data
					byte[] sample = fBASE64Decoder.decodeBuffer(commMessage);
					/*
					 * If cancel pending do not send audio to the player
					 * But in any case it is essential to send bytes array of length zero to terminate the audio message in the player
					*/
					if ((fCancelPending == false) || (sample.length == 0))
					{
						fClientMessageEnvelope.putAudio(sample);
						if (DEBUG_MODE)
							log.notify("processMessageData: pushing audio to player");
					}
					if (sample.length == 0)
					{
						if (DEBUG_MODE)
							log.notify("processMessageData: success" + fClientMessageEnvelope.getMessageIdentifier());
						return (true);
					}
				}
			}
		} catch (Exception ex)
		{
			log.error("Error in processMessageData method:\n" + StackTrace.stackToString(ex));
			// Return zero bytes array to release player
			try
			{
				fClientMessageEnvelope.putAudio(new byte[0]);
			} catch (InterruptedException e)
			{
			}
			return (false);
		}
	}

	private boolean processMessageEnd()
	{
		// Once the player is ready, Process message end to signal the server that a new message is required
		try
		{
			if (DEBUG_MODE)
				log.notify("processMessageEnd: start " + fClientMessageEnvelope.getMessageIdentifier());
			String commMessage;
			for (;;)
			{
				commMessage = getMessageAudioData();
				if (commMessage.startsWith(SHConstants.CMDP_METADATA_PREFIX))
				{
					if (commMessage.startsWith(SHConstants.CMDP_GROUP_STATUS_END_MARKER))
					{
						if (DEBUG_MODE)
							log.notify("processMessageEnd: success " + fClientMessageEnvelope.getMessageIdentifier());
						return (true);
					} else
					// Unexpected metadata try again
					{
						log.error("Error in processMessageEnd method: unexpected metadata:\n" + commMessage
								+ " trying to read next data");
						continue;
					}
				} else
				{
					log.error("Error in processMessageEnd method: unexpected audio data trying to read next data");
					continue;
				}
			}
		} catch (Exception ex)
		{
			log.error("Error in processMessageEnd method:\n" + StackTrace.stackToString(ex));
			return (false);
		}
	}

	@Override
	public void run()
	{
		for (;;)
		{
			synchronized (this)
			{
				while (fClientMessageEnvelopeReady == false)
					try
					{
						wait();
					} catch (InterruptedException ex)
					{
						log.error("Error in run method:\n" + ex);
					}
			}
			processMessageData();
			processMessageEnd();
			fClientMessageEnvelopeReady = false;
		}
	}

	// Event handling 

	public void messageBegin(String clientIdentifier, int messageIdentifier)
	{
		// Unused
	}

	public void messageEnd(String clientIdentifier, int messageIdentifier)
	{
		// Unused
	}

	public void messagePause(String clientIdentifier, int messageIdentifier)
	{
		// Unused
	}

	public void messageResume(String clientIdentifier, int messageIdentifier)
	{
		// Unused
	}

	public void messageCancel(String clientIdentifier, int messageIdentifier)
	{
		// Unused
	}

	@Override
	public void messageAudioAvailable(String clientIdentifier, int messageIdentifier)
	{
		if (messageIdentifier > fCancelledMessageIdentifier)
			fCancelPending = false;
		fAudioDataAvailable = true;
		synchronized (this)
		{
			notifyAll();
		}
		if (DEBUG_MODE)
			log.notify("messageAudioAvailable: " + messageIdentifier);
	}

	private void synchronizeAudioDataReady()
	{
		synchronized (this)
		{
			while (fAudioDataAvailable == false)
				try
				{
					wait();
				} catch (InterruptedException ex)
				{
					log.error("Error in synchronizeAudioDataReady method:\n" + ex);
				}
		}
	}

	private void synchronizePlayerReady()
	{
		synchronized (this)
		{
			while (fPlayerReady == false)
				try
				{
					wait();
				} catch (InterruptedException ex)
				{
					log.error("Error in synchronizePlayerReady method:\n" + ex);
				}
		}
	}

	private class MessageEnvelopeProcessing implements Runnable
	{

		public MessageEnvelopeProcessing()
		{
			Thread messageEnvelopeProcessingThread = new Thread(this);
			messageEnvelopeProcessingThread.setName(MessageEnvelopeProcessing.class.getName());
			messageEnvelopeProcessingThread.start();
		}

		@Override
		public void run()
		{
			for (;;)
			{
				synchronizeAudioDataReady();
				fAudioDataAvailable = false;
				synchronizePlayerReady();
				processMessageEnvelope();
			}
		}

	}

}
