/*
* Copyright (c) Isaac Porat 2011 - 2013 all rights reserved.
* This file is released under the GNU General Public License, version 2, or (at your option) any later version.
* See the file SpeechHub_License.txt for more details.
*/

package sh.comm;

import java.io.IOException;
import java.util.Arrays;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;

import log.common.CascadingLoggerInterface;
import sh.connections.SHConstants;
import sh.connections.SHConstants.MessageState;
import sh.connections.SHMessageEnvelopeInterface;
import sh.connections.SHAudioDataHandlerInterface;
import sh.connections.SHEventsListenerInterface;
import sun.misc.BASE64Encoder;

/**
 * Server message data communication 
 * Sends message data including metadata and audio data to the client 
 * 
 * @author Isaac Porat
 *
 */
public class SHServerCommMessageData implements SHAudioDataHandlerInterface
{
	private CascadingLoggerInterface log;
	private final static boolean DEBUG_MODE = false;
	private static boolean GET_MESSAGE_DATA_BLOCKED_MODE = true;
	private final static int BYTES_TRANSFER_SIZE = 1024;
	private BlockingQueue<String> fMessageDataBuffer = new LinkedBlockingQueue<String>();
	private BASE64Encoder fBASE64Encoder = new BASE64Encoder();
	private SHMessageEnvelopeInterface fServerMessageEnvelope;
	private boolean fClientRetrievingMessageData = true;
	private SHEventsListenerInterface fEventsListener;
	private int fCurrentMessageIdentifier = 0;
	private int fCancelledMessageIdentifier = 0;
	private boolean fCancelPending = false;

	public SHServerCommMessageData(SHEventsListenerInterface eventsListener, CascadingLoggerInterface logger)
	{
		log = logger.clone(SHServerCommMessageData.class);
		fEventsListener = eventsListener;
	}

	public boolean cancel()
	{
		fCancelledMessageIdentifier = fCurrentMessageIdentifier;
		fCancelPending = true;
		return (true);
	}

	public void indicateSpeak(int messageIdentifier, String message, boolean speakSync)
	{
		fCurrentMessageIdentifier = messageIdentifier;
	}

	private String getMetadataGroupHeader(int messageIdentifier, String message, int sampleRate,
			MessageState messageState)
	{
		String metadataGroupHeader = SHConstants.CMDP_METADATA_PREFIX + SHConstants.CMDP_GROUP_HEADER
				+ SHConstants.CMDP_METADATA_SEPARATOR + getMetadataMessageIdentifier(messageIdentifier)
				+ SHConstants.CMDP_METADATA_SEPARATOR + getMetadataAudioSampleRate(sampleRate)
				+ SHConstants.CMDP_METADATA_SEPARATOR + getMetadataMessageState(messageState);
		return (metadataGroupHeader);
	}

	private String getMetadataMessageIdentifier(int messageIdentifier)
	{
		String metadataMessageIdentifier = Integer.toString(messageIdentifier);
		return (metadataMessageIdentifier);
	}

	private String getMetadataAudioSampleRate(int sampleRate)
	{
		String metadataAudioSampleRate = Integer.toString(sampleRate);
		return (metadataAudioSampleRate);
	}

	private String getMetadataMessageState(MessageState messageState)
	{
		String metadataMessageState = SHConstants.CMDP_GROUP_HEADER_MESSAGESTATE_MIDDLE;
		if (messageState.equals(MessageState.SINGLE))
			metadataMessageState = SHConstants.CMDP_GROUP_HEADER_MESSAGESTATE_SINGLE;
		else if (messageState.equals(MessageState.BEGIN))
			metadataMessageState = SHConstants.CMDP_GROUP_HEADER_MESSAGESTATE_BEGIN;
		else if (messageState.equals(MessageState.END))
			metadataMessageState = SHConstants.CMDP_GROUP_HEADER_MESSAGESTATE_END;
		return (metadataMessageState);
	}

	private String getMetadataGroupMessage(String message)
	{
		String metadataGroupStatus = SHConstants.CMDP_METADATA_PREFIX + SHConstants.CMDP_GROUP_MESSAGE
				+ SHConstants.CMDP_METADATA_SEPARATOR + getMetadataMessage(message);
		return (metadataGroupStatus);
	}

	private String getMetadataMessage(String message)
	{
		if (message.length() == 0)
			message = SHConstants.CMDP_GROUP_MESSAGE_BLANK;
		return (message);
	}

	private String getMetadataGroupStatus(String status)
	{
		String metadataGroupStatus = SHConstants.CMDP_METADATA_PREFIX + SHConstants.CMDP_GROUP_STATUS
				+ SHConstants.CMDP_METADATA_SEPARATOR + status;
		return (metadataGroupStatus);
	}

	private String getEncodedAudioData(byte[] bytes) throws IOException
	{
		String encodedString = fBASE64Encoder.encode(bytes);
		return (encodedString);
	}

	private void addToMessageDataBuffer(String messageData) throws InterruptedException
	{
		fMessageDataBuffer.put(messageData);
	}

	public String getMessageData(String type, String identifier)
	{
		try
		{
			if (type.equalsIgnoreCase(SHConstants.CMDP_MESSAGE_AUDIODATA))
				return (getMessageAudioData());
			return (getMessageMetadata(identifier));
		} catch (Exception ex)
		{
			log.error("Error in getMessageData method:\n" + ex);
			return (null);
		}
	}

	private String getMessageAudioData() throws InterruptedException
	{
		if (GET_MESSAGE_DATA_BLOCKED_MODE == false)
		{
			if (fMessageDataBuffer.isEmpty())
			{
				if (DEBUG_MODE)
					log.notify("getMessageAudioData: unavailable data");
				return (getMetadataGroupStatus(SHConstants.CMDP_GROUP_STATUS_UNAVAILABLE));
			}
		}
		String messageData = fMessageDataBuffer.take();
		if (messageData.startsWith(SHConstants.CMDP_METADATA_PREFIX))
		{
			if (messageData.startsWith(SHConstants.CMDP_GROUP_STATUS_END_MARKER))
			{
				synchronized (this)
				{
					fClientRetrievingMessageData = false;
					notifyAll();
				}
			}
		} else
		// Audio data
		{
			if ((fCancelPending) && (messageData.equals("") == false))
			{
				// Send the cancel marker to save on transmition data
				return (SHConstants.CMDP_GROUP_STATUS_CANCEL_MARKER);
			}
		}
		return (messageData);
	}

	private String getMessageMetadata(String metadataGroupIdentifier)
	{
		synchronized (this)
		{
			while (fServerMessageEnvelope == null)
				try
				{
					wait();
				} catch (InterruptedException ex)
				{
					log.error("Error in getMessageMetadata method:\n" + ex);
				}
		}
		if (metadataGroupIdentifier.equals(SHConstants.CMDP_GROUP_HEADER))
		{
			return (getMetadataGroupHeader(fServerMessageEnvelope.getMessageIdentifier(),
					fServerMessageEnvelope.getMessage(), fServerMessageEnvelope.getAudioSampleRate(),
					fServerMessageEnvelope.getMessageState()));
		}
		if (metadataGroupIdentifier.equals(SHConstants.CMDP_GROUP_MESSAGE))
		{
			return (getMetadataGroupMessage(fServerMessageEnvelope.getMessage()));
		}
		return (null);
	}

	@Override
	public boolean processAudio(byte[] sample)
	{
		// Unused
		return (false);
	}

	@Override
	public boolean processAudio(final SHMessageEnvelopeInterface messageEnvelope)
	{
		try
		{
			fServerMessageEnvelope = messageEnvelope;
			int messageIdentifier = messageEnvelope.getMessageIdentifier();
			if (messageIdentifier > fCancelledMessageIdentifier)
				fCancelPending = false;
			synchronized (this)
			{
				notifyAll();
			}
			fClientRetrievingMessageData = true;
			// Send audio data						
			int counter = 0;
			for (;;)
			{
				byte[] sample = messageEnvelope.takeAudio();
				// Regardless of cancel ensure that zero length byte array is send to terminate player message   
				if ((fCancelPending == false) || (sample.length == 0))
				{
					//addToMessageDataBuffer(getEncodedAudioData(bytes));
					addMessageFragmentsToMessageDataBuffer(sample);
					if (DEBUG_MODE)
						log.notify("processAudio: adding to com buffer");
				}
				counter++;
				if (counter == 1)
				{
					fEventsListener.messageAudioAvailable("unknown client", messageIdentifier);
				}
				if (sample.length == 0)
					break;
			}
			// Add status group 'end'						
			addToMessageDataBuffer(getMetadataGroupStatus(SHConstants.CMDP_GROUP_STATUS_END));
			synchronized (this)
			{
				while (fClientRetrievingMessageData)
					try
					{
						wait();
					} catch (InterruptedException ex)
					{
						log.error("Error in processAudio method:\n" + ex);
					}
			}
			fServerMessageEnvelope = null;
			return (true);
		} catch (Exception ex)
		{
			log.error("Error in processAudio method:\n" + ex);
			return (false);
		}
	}

	private void addMessageFragmentsToMessageDataBuffer(byte[] sample) throws InterruptedException, IOException
	{
		int sampleLength = sample.length;
		if (sampleLength == 0)
		{
			addToMessageDataBuffer(getEncodedAudioData(sample));
			return;
		}
		int sampleFrom = 0;
		int sampleTo = 0;
		for (;;)
		{
			sampleTo = sampleFrom + BYTES_TRANSFER_SIZE;
			if (sampleTo > sampleLength)
				sampleTo = sampleLength;
			byte[] subSample = Arrays.copyOfRange(sample, sampleFrom, sampleTo);
			String encodedData = getEncodedAudioData(subSample);
			addToMessageDataBuffer(encodedData);
			if (sampleTo == sampleLength)
				return;
			sampleFrom += BYTES_TRANSFER_SIZE;
		}
	}
}
