/*
* Copyright (c) Isaac Porat 2011 - 2013 all rights reserved.
* This file is released under the GNU General Public License, version 2, or (at your option) any later version.
* See the file SpeechHub_License.txt for more details.
*/

package sh.comm;

import log.common.CascadingLoggerInterface;
import sh.connections.SHCustomeHandlersInterface;
import sh.connections.SHEngineInterface;
import sh.connections.SHMessageEnvelopeInterface;
import sh.connections.SHModuleInterface;
import sh.connections.SHModuleTimestampInterface;
import sh.connections.SHAudioDataHandlerInterface;
import sh.connections.SHEventsListenerInterface;
import sh.connections.SHTTSInterface;
import sh.connections.SHUpdateParametersInterface;
import core.utilities.StackTrace;

/**
 * TTS comm client emulator
 * Emulating receiving message data including metadata and audio data from the  server 
 * @author Isaac Porat
 *
 */
public class SHTTSClientCommEmulator implements SHTTSInterface, SHModuleInterface, SHEngineInterface,
		SHCommMessageDataHandlerInterface
{
	private CascadingLoggerInterface log;
	private SHTTSServerCommEmulator fServer;
	private SHClientCommMessageData fClientCommAudioData;

	public SHTTSClientCommEmulator(SHTTSServerCommEmulator server, CascadingLoggerInterface logger)
	{
		log = logger.clone(SHTTSClientCommEmulator.class);
		fServer = server;
	}

	@Override
	public boolean configureTiming(SHModuleTimestampInterface moduleTimestamp, boolean masterCancel)
	{
		// Not applicable
		return (false);
	}

	@Override
	public boolean initialize()
	{
		return (fServer.initialize());
	}

	@Override
	public boolean close()
	{
		return (fServer.close());
	}

	@Override
	public boolean setPriority(int priorityLevel, boolean selfCancel)
	{
		// Not applicable
		return (false);
	}

	@Override
	public String[] getEngineIdentifiers()
	{
		try
		{
			return (fServer.getEngineIdentifiers());
		} catch (Exception ex)
		{
			log.error("Error in getEngineIdentifiers method:\n" + StackTrace.stackToString(ex));
			return (null);
		}
	}

	@Override
	public boolean selectEngine(String engineIdentifier)
	{
		try
		{
			return (fServer.selectEngine(engineIdentifier));
		} catch (Exception ex)
		{
			log.error("Error in selectEngine method:\n" + StackTrace.stackToString(ex));
			return (false);
		}
	}

	@Override
	public String getSelectedEngineIdentifier()
	{
		return (fServer.getSelectedEngineIdentifier());
	}

	@Override
	public boolean selectLanguage(String languageCode)
	{
		return (fServer.selectLanguage(languageCode));
	}

	@Override
	public String getSelectedLanguageCode()
	{
		return (fServer.getSelectedLanguageCode());
	}

	@Override
	public String[] getVoiceIdentifiers()
	{
		return (fServer.getVoiceIdentifiers());
	}

	@Override
	public boolean selectVoice(String voiceIdentifier)
	{
		return (fServer.selectVoice(voiceIdentifier));
	}

	@Override
	public String getSelectedVoiceIdentifier()
	{
		return (fServer.getSelectedVoiceIdentifier());
	}

	@Override
	public String[] getVariantIdentifiers()
	{
		return (fServer.getVariantIdentifiers());
	}

	@Override
	public boolean selectVariant(String variantIdentifier)
	{
		return (fServer.selectVariant(variantIdentifier));
	}

	@Override
	public String getSelectedVariantIdentifier()
	{
		return (fServer.getSelectedVariantIdentifier());
	}

	@Override
	public boolean speak(SHMessageEnvelopeInterface messageEnvelope)
	{
		// Not applicable
		return (false);
	}

	@Override
	public boolean speak(int messageIdentifier, String message, boolean speakSync)
	{
		if (fClientCommAudioData != null)
			fClientCommAudioData.indicateSpeak(messageIdentifier, message, speakSync);
		return (fServer.speak(messageIdentifier, message, speakSync));
	}

	@Override
	public boolean speakChar(int messageIdentifier, String message, boolean speakSync)
	{
		return (fServer.speakChar(messageIdentifier, message, speakSync));
	}

	@Override
	public boolean speakKey(int messageIdentifier, String message, boolean speakSync)
	{
		return (fServer.speakKey(messageIdentifier, message, speakSync));
	}

	@Override
	public boolean speakSpell(int messageIdentifier, String message, boolean speakSync)
	{
		return (fServer.speakSpell(messageIdentifier, message, speakSync));
	}

	@Override
	public boolean pause()
	{
		return (fServer.pause());
	}

	@Override
	public boolean resume()
	{
		return (fServer.resume());
	}

	@Override
	public boolean stop()
	{
		return (fServer.stop());
	}

	@Override
	public boolean cancel()
	{
		if (fClientCommAudioData != null)
			fClientCommAudioData.cancel();
		return (fServer.cancel());
	}

	@Override
	public boolean setVolume(float volume)
	{
		return (fServer.setVolume(volume));
	}

	@Override
	public float getVolume()
	{
		return (fServer.getVolume());
	}

	@Override
	public boolean setRate(float rate)
	{
		return (fServer.setRate(rate));
	}

	@Override
	public float getRate()
	{
		return (fServer.getRate());
	}

	@Override
	public boolean setRateBoostMode(boolean mode)
	{
		return (fServer.setRateBoostMode(mode));
	}

	@Override
	public boolean setPitch(float pitch)
	{
		return (fServer.setPitch(pitch));
	}

	@Override
	public float getPitch()
	{
		return (fServer.getPitch());
	}

	@Override
	public boolean setPunctuationLevel(String level)
	{
		return (fServer.setPunctuationLevel(level));
	}

	@Override
	public boolean setCharCapitalIndicator(String charCapitalIndicator, String arg)
	{
		return (fServer.setCharCapitalIndicator(charCapitalIndicator, arg));
	}

	@Override
	public boolean setSilenceBeforeMessage(int silenceMilliseconds)
	{
		return (fServer.setSilenceBeforeMessage(silenceMilliseconds));
	}

	@Override
	public boolean setProperty(String key, String value)
	{
		return (fServer.setProperty(key, value));
	}

	@Override
	public String getProperty(String key)
	{
		return (fServer.getProperty(key));
	}

	@Override
	public boolean setEventsListener(SHEventsListenerInterface eventsListener)
	{
		return (fServer.setEventsListener(eventsListener));
	}

	@Override
	public boolean setAudioDataHandler(SHAudioDataHandlerInterface audioDataHandler)
	{
		fClientCommAudioData = new SHClientCommMessageData(this, audioDataHandler, log);
		fServer.setEventsListener(fClientCommAudioData);
		return (true);
	}

	@Override
	public boolean setCustomeHandlers(SHCustomeHandlersInterface customeHandlers)
	{
		// Not applicable
		return (false);
	}

	@Override
	public boolean updateParametersDirectly(SHUpdateParametersInterface updateParameters)
	{
		// Not applicable
		return (false);
	}

	@Override
	public boolean setUpdateIdentifier(int updateIdentifier)
	{
		// Not applicable
		return (false);
	}

	@Override
	public boolean isUpdateAccepted(int updateIdentifier)
	{
		// Not applicable
		return (false);
	}

	@Override
	public String getMessageData(String type, String identifier)
	{
		try
		{
			return (fServer.getMessageData(type, identifier));
		} catch (Exception ex)
		{
			log.error("Error in getMessageData method:\n" + ex);
			return (null);
		}
	}

}
