/*
* Copyright (c) Isaac Porat 2011 - 2013 all rights reserved.
* This file is released under the GNU General Public License, version 2, or (at your option) any later version.
* See the file SpeechHub_License.txt for more details.
*/

package sh.common;

import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;

import log.common.CascadingLoggerInterface;

import sh.connections.SHMessageEnvelopeInterface;
import sh.connections.SHMessageQueueHandlerInterface;
import sh.connections.SHMessageQueuePeekHandlerInterface;
import sh.connections.SHUpdateParametersInterface;

/**
 * Clients multiplexer message queue
 * including implementation of  priorities handling and scheduling, pause and resume.
 * @author Isaac Porat
 *
 */
public class SHMessageQueue implements SHMessageQueuePeekHandlerInterface
{
	private CascadingLoggerInterface log;
	private List<SHMessageEnvelopeInterface> fMessageEnvelopeList = Collections
			.synchronizedList(new LinkedList<SHMessageEnvelopeInterface>());
	private SHMessageQueueHandlerInterface fHandler;
	private Set<String> fClientPaused = new HashSet<String>();
	private boolean fMessageSpeaking = false;
	private int fCurrentMessagePriorityLevel = 4;

	public SHMessageQueue(SHMessageQueueHandlerInterface handler, CascadingLoggerInterface logger)
	{
		fHandler = handler;
		log = logger.clone(SHMessageQueue.class);
	}

	public void addMessageToQueue(SHMessageEnvelopeInterface messageEnvelope)
	{
		synchronized (fMessageEnvelopeList)
		{
			cancelMessagesOfSamePriority(messageEnvelope);
			addMessageInPriorityOrder(messageEnvelope);
			cancelCurrentMessageBasedOnPriority(messageEnvelope);
			fMessageEnvelopeList.notifyAll();
		}
	}

	public void removeMessageFromQueue()
	{
		SHMessageEnvelopeInterface messageEnvelope = null;
		synchronized (fMessageEnvelopeList)
		{
			int index;
			while ((index = findAvailableMessage()) == -1)
			{
				try
				{
					fMessageEnvelopeList.wait();
				} catch (InterruptedException ex)
				{
				}
			}
			messageEnvelope = fMessageEnvelopeList.remove(index);
		}
		fCurrentMessagePriorityLevel = messageEnvelope.getPriorityLevel();
		fMessageSpeaking = true;
		fHandler.processMessage(messageEnvelope);
		fMessageSpeaking = false;
		synchronized (fMessageEnvelopeList)
		{
			if (fClientPaused.contains(messageEnvelope.getClientIdentifier()))
				fMessageEnvelopeList.add(0, messageEnvelope);
		}
	}

	public void addMessageInPriorityOrder(SHMessageEnvelopeInterface messageEnvelope)
	{
		synchronized (fMessageEnvelopeList)
		{
			try
			{
				if (fMessageEnvelopeList.isEmpty())
				{
					fMessageEnvelopeList.add(messageEnvelope);
					return;
				}
				// Note no need to consider different clients 
				int messagePriorityLevel = messageEnvelope.getPriorityLevel();
				for (int n = 0; n < fMessageEnvelopeList.size(); n++)
				{
					if (messagePriorityLevel > fMessageEnvelopeList.get(n).getPriorityLevel())
					{
						fMessageEnvelopeList.add(n, messageEnvelope);
						return;
					}
				}
				fMessageEnvelopeList.add(messageEnvelope);
			} catch (Exception ex)
			{
				log.error("Error in addMessageInPriorityOrder method:\n" + ex);
			}
		}
	}

	private void cancelMessagesOfSamePriority(SHMessageEnvelopeInterface messageEnvelope)
	{
		synchronized (fMessageEnvelopeList)
		{
			try
			{
				if (fMessageEnvelopeList.isEmpty())
					return;
				if (messageEnvelope.getSelfCancel() == false)
					return;
				String clientIdentifier = messageEnvelope.getClientIdentifier();
				int priorityLevel = messageEnvelope.getPriorityLevel();
				SHMessageEnvelopeInterface listMessageEnvelope;
				for (Iterator<SHMessageEnvelopeInterface> i = fMessageEnvelopeList.iterator(); i.hasNext();)
				{
					listMessageEnvelope = i.next();
					if ((priorityLevel == listMessageEnvelope.getPriorityLevel())
							&& (clientIdentifier.equals(listMessageEnvelope.getClientIdentifier())))
					{
						i.remove();
					}
				}
			} catch (Exception ex)
			{
				log.error("Error in cancelMessagesOfSamePriority method in removing messageEnvelope\n" + ex);
			}
		}
	}

	private void cancelCurrentMessageBasedOnPriority(SHMessageEnvelopeInterface messageEnvelope)
	{
		// Note no need to consider different clients
		try
		{
			if ((fMessageSpeaking) && (messageEnvelope.getPriorityLevel() > fCurrentMessagePriorityLevel))
				fHandler.cancelCurrentMessage();
		} catch (Exception ex)
		{
			log.error("Error in cancelCurrentMessageBasedOnPriority Message method:\n" + ex);
		}
	}

	private int findAvailableMessage()
	{
		if (fMessageEnvelopeList.isEmpty())
			return (-1);
		if (fClientPaused.isEmpty())
			return (0); // Remove from head of queue
		for (int n = 0; n < fMessageEnvelopeList.size(); n++)
		{
			if (fClientPaused.contains(fMessageEnvelopeList.get(n).getClientIdentifier()) == false)
				return (n); // Not Paused message ready for removal found 
		}
		return (-1); // No messages ready for removal found
	}

	public boolean pause(String clientIdentifier)
	{
		fClientPaused.add(clientIdentifier);
		return (true);
	}

	public boolean resume(String clientIdentifier)
	{
		fClientPaused.remove(clientIdentifier);
		synchronized (fMessageEnvelopeList)
		{
			fMessageEnvelopeList.notifyAll();
		}
		return (true);
	}

	public boolean cancel(String clientIdentifier)
	{
		boolean status = true;
		synchronized (fMessageEnvelopeList)
		{
			try
			{
				SHMessageEnvelopeInterface messageEnvelope;
				for (Iterator<SHMessageEnvelopeInterface> i = fMessageEnvelopeList.iterator(); i.hasNext();)
				{
					messageEnvelope = i.next();
					if (clientIdentifier.equals(messageEnvelope.getClientIdentifier()))
						i.remove();
				}
			} catch (Exception ex)
			{
				log.error("Error in cancel method in removing messageEnvelope\n" + ex);
				status = false;
			}
		}
		return (status);
	}

	@Override
	public SHMessageEnvelopeInterface peekNextMessage(Set<String> engineSet)
	{
		if (fMessageEnvelopeList.isEmpty())
			return (null);
		for (int n = 0; n < fMessageEnvelopeList.size(); n++)
		{
			SHMessageEnvelopeInterface messageEnvelope = fMessageEnvelopeList.get(n);
			if (fClientPaused.contains(messageEnvelope.getClientIdentifier()) == false)
			{
				// Not Paused message found 
				SHUpdateParametersInterface updateParameters = messageEnvelope.getUpdateParameters();
				if (engineSet.contains(updateParameters.getEngineIdentifier()))
					return (messageEnvelope);
			}
		}
		return (null); // No messages ready for removal found
	}

}
