/*
* Copyright (c) Isaac Porat 2011 - 2013 all rights reserved.
* This file is released under the GNU General Public License, version 2, or (at your option) any later version.
* See the file SpeechHub_License.txt for more details.
*/

package sh.common;

import sh.connections.SHConstants;
import sh.connections.SHUpdateParametersInterface;

/**
 * Update parameters associated with every message spoken
 * These parameters always form part of the SHClientParameters 
 * and SHMessageEnvelope and thus queued together with a message  
 * 
 * @author Isaac Porat
 *
 */
public class SHUpdateParameters implements SHUpdateParametersInterface
{
	private String fClientIdentifier;
	private String fEngineIdentifier = SHConstants.ENGINE_IDENTIFIER_DEFAULT;
	private String fLanguageCode = SHConstants.ENGINE_LANGUAGE_CODE_DEFAULT;
	private String fVoiceIdentifier = SHConstants.ENGINE_VOICE_IDENTIFIER_DEFAULT;
	private String fVariantIdentifier = SHConstants.ENGINE_VARIANT_IDENTIFIER_DEFAULT;
	private float fVolume = SHConstants.ENGINE_VOLUME_DEFAULT;
	private float fRate = SHConstants.ENGINE_RATE_DEFAULT;
	private float fPitch = SHConstants.ENGINE_PITCH_DEFAULT;
	// Always update if updateIdentifier 0 or > than previous updateIdentifier
	private int fUpdateIdentifier = 0;

	public static SHUpdateParametersInterface getNewDisabledUpdateParameters(String clientIdentifier)
	{
		SHUpdateParametersInterface updateParameters = new SHUpdateParameters(clientIdentifier);
		updateParameters.setEngineIdentifier("");
		updateParameters.setLanguageCode("");
		updateParameters.setVoiceIdentifier("");
		updateParameters.setVariantIdentifier("");
		updateParameters.setVolume(SHConstants.FLOAT_PARAMETER_DISABLED);
		updateParameters.setRate(SHConstants.FLOAT_PARAMETER_DISABLED);
		updateParameters.setPitch(SHConstants.FLOAT_PARAMETER_DISABLED);
		return (updateParameters);
	}

	/*
	 * Convenience methods for checking if update required
	 */

	public static boolean isUpdateRequired(int updateIdentifier, float updateValue, float currentValue)
	{
		// Check for disabled parameter indicator
		if (updateValue == SHConstants.FLOAT_PARAMETER_DISABLED)
			return (false);
		if (updateIdentifier == 0)
			return (true); // Force update
		return (updateValue != currentValue);
	}

	public static boolean isUpdateRequired(int updateIdentifier, String updateValue, String currentValue)
	{
		if (updateValue == null)
			return (false);
		// Check for disabled parameter indicator
		if (updateValue.equals(SHConstants.STRING_PARAMETER_DISABLED))
			return (false);
		if (updateIdentifier == 0)
			return (true); // Force update
		if (currentValue == null)
			return (true);
		return (updateValue.equals(currentValue) == false);
	}

	public SHUpdateParameters(String clientIdentifier)
	{
		fClientIdentifier = clientIdentifier;
	}

	private SHUpdateParameters(String clientIdentifier, String engineIdentifier, String languageCode,
			String voiceIdentifier, String variantIdentifier, float volume, float rate, float pitch,
			int updateIdentifier)
	{
		fClientIdentifier = clientIdentifier;
		fEngineIdentifier = engineIdentifier;
		fLanguageCode = languageCode;
		fVoiceIdentifier = voiceIdentifier;
		fVariantIdentifier = variantIdentifier;
		fVolume = volume;
		fRate = rate;
		fPitch = pitch;
		fUpdateIdentifier = updateIdentifier;
	}

	@Override
	public String getClientIdentifier()
	{
		return (fClientIdentifier);
	}

	@Override
	public String getEngineIdentifier()
	{
		return (fEngineIdentifier);
	}

	@Override
	public void setEngineIdentifier(String engineIdentifier)
	{
		fEngineIdentifier = engineIdentifier;
	}

	@Override
	public String getLanguageCode()
	{
		return (fLanguageCode);
	}

	@Override
	public void setLanguageCode(String languageCode)
	{
		fLanguageCode = languageCode;
	}

	@Override
	public String getVoiceIdentifier()
	{
		return (fVoiceIdentifier);
	}

	@Override
	public void setVoiceIdentifier(String voiceIdentifier)
	{
		fVoiceIdentifier = voiceIdentifier;
	}

	@Override
	public String getVariantIdentifier()
	{
		return (fVariantIdentifier);
	}

	@Override
	public void setVariantIdentifier(String variantIdentifier)
	{
		fVariantIdentifier = variantIdentifier;
	}

	@Override
	public float getVolume()
	{
		return (fVolume);
	}

	@Override
	public void setVolume(float volume)
	{
		fVolume = volume;
	}

	@Override
	public float getRate()
	{
		return (fRate);
	}

	@Override
	public void setRate(float rate)
	{
		fRate = rate;
	}

	@Override
	public float getPitch()
	{
		return (fPitch);
	}

	@Override
	public void setPitch(float pitch)
	{
		fPitch = pitch;
	}

	@Override
	public int getUpdateIdentifier()
	{
		return (fUpdateIdentifier);
	}

	public void markNewUpdate()
	{
		// Standard updateIdentifier updates are accepted when an updateIdentifier is greater than the previous one
		fUpdateIdentifier = SHIdentifiers.getNewUpdateIdentifier();
	}

	public void markInsertUpdate()
	{
		// Zero indicates insert update 
// All following updates are ignored until the updateIdentifier is greater than the previous one 
			fUpdateIdentifier = 0; 
	}	

	public void markResetUpdate()
	{
		// This update is always accepted and the update counter is reset to 0, this is applied when client change 
		// as a client might still be on a perfectly valid lower updateIdentifier
		fUpdateIdentifier = -1;
	}
	
	public SHUpdateParametersInterface getClone()
	{
		return (new SHUpdateParameters(fClientIdentifier, fEngineIdentifier, fLanguageCode, fVoiceIdentifier,
				fVariantIdentifier, fVolume, fRate, fPitch, fUpdateIdentifier));
	}

	public boolean equals(SHUpdateParametersInterface updateParameters)
	{
		if (updateParameters == null)
			return (false);
		return (fUpdateIdentifier == updateParameters.getUpdateIdentifier());
	}

}
