/*
* Copyright (c) Isaac Porat 2011 - 2013 all rights reserved.
* This file is released under the GNU General Public License, version 2, or (at your option) any later version.
* See the file SpeechHub_License.txt for more details.
*/

package sh.connections;

/**
 * Engine connector interface
 * 
 * @author Isaac Porat
 */
public interface SHEngineConnectorInterface
{

	/**
	 * Initialize engine connector
	 * @return true if successful false if not
	 */
	public boolean initialize();

	/**
	 * Close engine connector
	 * @return true if successful false if not
	 */
	public boolean close();

	/**
	 * Get engine identifiers
	 * 
	 * @return engineIdentifierArray
	 */
	public String[] getEngineIdentifiers();

	/**
	 * Select engine based on its identifier
	 * 
	 * @param engineIdentifier Identify the module required
	  * @return true if successful false if not.
	 */
	public boolean selectEngine(String engineIdentifier);

	/**
	 * Select the synthesized language according to language code as specified by RFC 1766 
	 * @param languageCode
	 * @return
	 */
	public boolean selectLanguage(String languageCode);

	/**
		 * Get synthesizer voice identifiers
		 * 
		 * @return Voice identifier array.
		 */
	public String[] getVoiceIdentifiers();

	/**
	 * Select the synthesizer's voice
	 * 
	 * @param voiceIdentifier Identify the voice required
	 * @return true if successful false if not.
	 */
	public boolean selectVoice(String voiceIdentifier);

	/**
	 * Get synthesizer voice variant identifiers
	 * Applies only to Formant type synthesizers
	 * @return variant identifier array.
	 */
	public String[] getVariantIdentifiers();

	/**
	  * Select the synthesizer's voice variant
	 * Applies only to Formant type synthesizers
		 * @param varientIdentifier Identify the variant required
	 * @return true if successful false if not.
	 */
	public boolean selectVariant(String variantIdentifier);

	/**
	* Speak message 
	 * 	  
	 * @param messageIdentifier Identifies message used in server
	 * @param message Message to speak
	 * @param speakSync if true return after completion, if false return immediatly 
	 * @return true if successful false if not
	 * Notes on implementation:
	 * There is no need to implement both speakSync and speakAsync functionality
	 * speakSync is used internally by SH and therefore prefered
	 * Alternatively, speakAsync can be implemented, in which case
	 * The end event must be implemented through the appropriate event listener interface
	 * The property "engine.supports.speak.sync" must be set accordingly for the module stack to work correctly 
	 */
	public boolean speak(int MessageIdentifier, String message, boolean speakSync);

	/**
	 * Pause speech
	 * @return true if successful false if not
	 */
	public boolean pause();

	/**
	 * Resume speech after a pause
	* @return true if successful false if not
	 	 */
	public boolean resume();

	/**
	* Flushes the synthesizer buffer and cancel its current output.
	*   @return true if successful false if not
		*/
	public boolean cancel();

	/**
	 * Set speech volume in the range of 0 - 100.0
	 * 
	 * @param volume Speech volume
	 * 
	 * @return true if successful false if not
	 */
	public boolean setVolume(float volume);

	/**
	 * Set speech rate in the range of -100.0 - 100.0
	 * 
	 * @param rate Speech rate
	 	 * 
	 * @return true if successful false if not
	*/
	public boolean setRate(float rate);

	/**
	* Set pitch in % above and below the base pitch.<br/>
	* 
	* @param pitch Offset  in +-100.0% from base pitch for the voice.
	* 
	* @return true if successful false if not
	*/
	public boolean setPitch(float pitch);

	/**
	* This method is used to get information about engine capabilities and advanced settings
	* @param key property key
	* @return property value if found null if not
	*/
	public String getProperty(String key);

	/**
	 * 	 This method is used to set more advanced engine options such as punctuation levels, spelling modes etc. 
	 * @param key property key
	 * @param value property value
	 * @return true if successful false if not
	 */
	public boolean setProperty(String key, String value);

	/**
	 * Listener for receiving module events
	 * @param eventsListener
	 * @return true if supported by the engine implementation false if not 
	 */
	public boolean setEventsListener(SHEventsListenerInterface eventsListener);

	/**
	 * Set handler to optionally bring upstream audio raw data generated by the synthesizer and process
	 * @param audioDataHandler
	 * @return true if supported by the engine implementation false if not 
	 */
	public boolean setAudioDataHandler(SHAudioDataHandlerInterface audioDataHandler);

	/**
	 * Custome handlers for upstream and global communication
	 * @param customeHandlers
	 * @return true if supported by the engine implementation false if not 
		 */
	public boolean setCustomeHandlers(SHCustomeHandlersInterface customeHandlers);

	/**
	 * Called before other methods call to ensure that engine is initialized
	 * 
	 */
	public void checkEngine();

}
