/*
* Copyright (c) Isaac Porat 2011 - 2013 all rights reserved.
* This file is released under the GNU General Public License, version 2, or (at your option) any later version.
* See the file SpeechHub_License.txt for more details.
*/

package sh.connections;

/**
 * Engine interface  
 * Adds to engine connector interface getters functionality
 * @author Isaac Porat
 */
public interface SHEngineInterface
{
	/**
	 * Initialize engine connector 
	 * @return true if successful false if not
	 */
	public boolean initialize();

	/**
	 * Close engine connector 
	 * @return true if successful false if not
	 */
	public boolean close();

	/**
	 * Get engine identifiers 
	 * 
	 * @return Engine identifier array.
	 */
	public String[] getEngineIdentifiers();

	/**
	 * Select the engine
	 * 
	 * @param engineIdentifier Identify the engine required
	 * @return true if successful false if not
	 */
	public boolean selectEngine(String engineIdentifier);

	/**
	 *  
	 * @return Selected engine identifier
	 */
	public String getSelectedEngineIdentifier();

	/**
	 * Get engine voice identifiers
	 * 
	 * @return Voice identifier array.
	 */
	public String[] getVoiceIdentifiers();

	/**
	 * Select the engine's voice
	 * 
	 * @param voiceIdentifier Identify the voice required
	 * @return true if successful false if not.
	 */
	public boolean selectVoice(String voiceIdentifier);

	/**
	 *  
	 * @return Selected voice identifier
	 */
	public String getSelectedVoiceIdentifier();

	/**
	 * Get synthesizer voice variant identifiers
	 * Applies only to Formant type synthesizers
	 * @return variant identifier array.
	 */
	public String[] getVariantIdentifiers();

	/**
	  * Select the synthesizer's voice variant
	 * Applies only to Formant type synthesizers
		 * @param varientIdentifier Identify the variant required
	 * @return true if successful false if not.
	 */
	public boolean selectVariant(String variantIdentifier);

	/**
	 * Applies only to Formant type synthesizers
	 * @return Selected variant identifier
	 */
	public String getSelectedVariantIdentifier();

	/**
	* Speak message 
	 * 	  
	 * @param messageIdentifier Identifies message used in server
	 * @param message Message to speak
	 * @param speakSync if true return after completion, if false return immediatly 
	 * @return true if successful false if not
	 */
	public boolean speak(int messageIdentifier, String message, boolean speakSync);

	/**
	 * Pause speech
	 * @return true if successful false if not
	 */
	public boolean pause();

	/**
	 * Resume speech after a pause
	* @return true if successful false if not
	 	 */
	public boolean resume();

	/**
	* Flushes the engine buffer and cancel its current output.
	* @return true if successful false if not
	 */
	public boolean cancel();

	/**
	 * 
	 * @return the synthesized language code as specified by RFC 1766 
	 */
	public String getSelectedLanguageCode();

	/**
	 * Select the engine language according to language code as specified by RFC 1766 
	 * @param languageCode
	 * @return
	 */
	public boolean selectLanguage(String languageCode);

	/**
	 * Set speech volume in the range of 0 - 100.0
	 * 
	 * @param volume Speech volume
	 * 
	 * @return true if successful false if not
	 */
	public boolean setVolume(float volume);

	/**
	 * 
	 * @return Volume
	 */
	public float getVolume();

	/**
	 * Set speech rate in the range of 0 - 100.0
	 * 
	 * @param rate Speech rate
	 	 * 
	 * @return true if successful false if not
	*/
	public boolean setRate(float rate);

	/**
	 * 
	 * @return Rate
	 */
	public float getRate();

	/**
	 * Set pitch above and below the base pitch.
	 * 
	 * @param pitch Offset  in -100.0 to -100.0  from base pitch.
	 * 
	 * @return true if successful false if not
	*/
	public boolean setPitch(float pitch);

	/**
	 * 
	 * @return Pitch offset  in -100.0 to  100.0 from base pitch.
	 */
	public float getPitch();

	/**
	 * This method is used to get information about engine capabilities and settings
	 * @param key property key
	 * @return value if found null if not
	 */
	public String getProperty(String key);

	/**
	 * 	 This method is used to set more advanced engine options such as punctuation levels, spelling modes etc. 
	 * @param key property key
	 * @param value property value
	 * @return true if successful false if not
	 */
	public boolean setProperty(String key, String value);

	/**
	 * Listener for receiving module events
	 * @param eventsListener
	 * @return true if supported by the engine implementation false if not 
	 */
	public boolean setEventsListener(SHEventsListenerInterface eventsListener);

	/**
	 * Set handler to optionally bring upstream audio raw data generated by the synthesizer and process
	 * @param audioDataHandler
	 * @return true if supported by the engine implementation false if not 
	 */
	public boolean setAudioDataHandler(SHAudioDataHandlerInterface audioDataHandler);

	/**
	 * Custome handlers for upstream and global communication
	 * @param customeHandlers
	 * @return true if supported by the engine implementation false if not 
		 */
	public boolean setCustomeHandlers(SHCustomeHandlersInterface customeHandlers);

	/**
	 * 
	 * @param updateIdentifier the update identifier of the updateParameter object to check
	 * @return true if ready to accept and update false if not
	 */
	public boolean isUpdateAccepted(int updateIdentifier);

	/**
	 * 
	 * @param updateIdentifier identifier for the latest parameters update
	 * @return true if successful false if not 
	 */
	public boolean setUpdateIdentifier(int updateIdentifier);

}
