/*
* Copyright (c) Isaac Porat 2011 - 2013 all rights reserved.
* This file is released under the GNU General Public License, version 2, or (at your option) any later version.
* See the file SpeechHub_License.txt for more details.
*/

package sh.connections;

/**
 * TTS module interface
 * 
 * @author Isaac Porat
 */
public interface SHModuleInterface
{

	/**
	 * Initialize engine connector 
	 * @return true if successful false if not
	 */
	public boolean initialize();

	/**
	 * Close engine connector 
	 * @return true if successful false if not
	 */
	public boolean close();

	/**
	 * 
	 * @param priorityLevel levels 1 - 5, 5 the highest 
	 * @param selfCancel indicates if a message cancel itself
	 * @return true if successful false if not
	 */
	public boolean setPriority(int priorityLevel, boolean selfCancel);

	/**
	 * Get engine identifiers 
	 * 
	 * @return Engine identifier array.
	 */
	public String[] getEngineIdentifiers();

	/**
	*  
	* @return Selected engine identifier
	*/
	public String getSelectedEngineIdentifier();

	/**
	 * 
	 * @return the synthesized language code as specified by RFC 1766 
	 */
	public String getSelectedLanguageCode();

	/**
	 * Get synthesizer voice identifiers
	 * 
	 * @return Voice identifier array.
	 */
	public String[] getVoiceIdentifiers();

	/**
	 *  
	 * @return Selected voice identifier
	 */
	public String getSelectedVoiceIdentifier();

	/**
	 * Get synthesizer voice variant identifiers
	 * Applies only to Formant type synthesizers
	 * @return variant identifier array.
	 */
	public String[] getVariantIdentifiers();

	/**
	 * Applies only to Formant type synthesizers
	 * @return Selected variant identifier
	 */
	public String getSelectedVariantIdentifier();

	/**
	* Speak message of all types: standard char and key as defined in messageEnvelope
	*  
	 * @param message envelope  object 
	 * @return true if successful false if not
	 */
	public boolean speak(SHMessageEnvelopeInterface messageEnvelope);

	/**
	* Speak standard message 
	 * 	  
	 * @param messageIdentifier Identifies message used in server
	 * @param message Message to speak
	 * @param speakSync if true return after completion, if false return immediatly 
	 * @return true if successful false if not
	 */
	public boolean speak(int messageIdentifier, String message, boolean speakSync);

	/**
	* Speak char message 
	 * 	  
	 * @param messageIdentifier Identifies message used in server
	 * @param message Message to speak
	 * @param speakSync if true return after completion, if false return immediatly 
	 * @return true if successful false if not
	 */
	public boolean speakChar(int messageIdentifier, String message, boolean speakSync);

	/**
	* Speak key message 
	 * 	  
	 * @param messageIdentifier Identifies message used in server
	 * @param message Message to speak
	 * @param speakSync if true return after completion, if false return immediatly 
	 * @return true if successful false if not
	 */
	public boolean speakKey(int messageIdentifier, String message, boolean speakSync);

	/**
	 * Pause speech
	 * @return true if successful false if not
	 	 */
	public boolean pause();

	/**
	 * Resume speech after a pause
	 * @return true if successful false if not
	 */
	public boolean resume();

	/**
	* Flushes the synthesizer last message buffer and cancel its current output.
	* @return true if successful false if not
	 */
	public boolean stop();

	/**
	* Flushes the synthesizer all messages buffer and cancel its current output.
	* @return true if successful false if not
	 */
	public boolean cancel();

	/**
	 * 
	 * @return Volume
	 */
	public float getVolume();

	/**
	 * 
	 * @return Rate
	 */
	public float getRate();

	/**
	 * 
	 * @return Pitch offset  in +-100.0% from base pitch for the voice.
	 */
	public float getPitch();

	/**
	 * This method is used to get information about module and engine capabilities and advanced settings
	 * @param key property key
	 * @return property value if found null if not
	 */
	public String getProperty(String key);

	/**
	 * 	 This method is used to set more advanced module and engine options such as punctuation levels, spelling modes etc. 
	 * @param key property key
	 * @param value property value
	 * @return true if successful false if not
	 */
	public boolean setProperty(String key, String value);

	/**
	 * Listener for receiving module events
	 * @param eventsListener
	 * @return true if successful false if not
	 */
	public boolean setEventsListener(SHEventsListenerInterface eventsListener);

	/**
	 * Set handler to optionally bring upstream audio raw data generated by the synthesizer and process
	 * @param audioDataHandler
	 * @return true if supported by the engine implementation false if not 
	 */
	public boolean setAudioDataHandler(SHAudioDataHandlerInterface audioDataHandler);

	/**
	 * Configure module timing
	 * @param moduleTimestamp module timestamp control object
	 * @param masterCancel true set independenly cancel timing, false rely on supplied cancel timing. 
	 * @return true if successful, false if not
	 */
	public boolean configureTiming(SHModuleTimestampInterface moduleTimestamp, boolean masterCancel);

	/**
	 * Custome handlers for upstream and global communication
	 * @param customeHandlers
	 * @return true if supported by the engine implementation false if not 
		 */
	public boolean setCustomeHandlers(SHCustomeHandlersInterface customeHandlers);

	/**
	 * Transfer parameters to player and engine as necessary directly
	 * bypassing the module stack
	 * Use only for situation that update needs to take effect immediatly
	 * @param updateParameters containing all engine update parameters 
	 * @return true if successful false if not
	 */
	public boolean updateParametersDirectly(SHUpdateParametersInterface updateParameters);

	/**
	 * 
	 * @param updateIdentifier the update identifier of the updateParameter object to check
	 * @return true if ready to accept and update false if not
	 */
	public boolean isUpdateAccepted(int updateIdentifier);

	/**
	 * 
	 * @param updateIdentifier identifier for the latest parameters update
	 * @return true if successful false if not 
	 */
	public boolean setUpdateIdentifier(int updateIdentifier);

}
