/*
* Copyright (c) Isaac Porat 2011 - 2013 all rights reserved.
* This file is released under the GNU General Public License, version 2, or (at your option) any later version.
* See the file SpeechHub_License.txt for more details.
*/

package sh.connections;

/**
 * Server client API 
 * 
 * @author Isaac Porat
 */
public interface SHServerClientInterface
{

	/**
	 * Initialize client 
	 * @return true if successful false if not
	 */
	public boolean initialize(String clientIdentifier);

	/**
	 * Close server client 
	 * @return true if successful false if not
	 */
	public boolean close();

	/**
	 * Not required, optionally send by client to aid identification useful for debugging  
	 * @param clientIdentifierCaption 
	 */
	public void setClientIdentifierCaption(String clientIdentifierCaption);

	/**
	 * 
	 * @param priorityLevel levels 1 - 5, 5 the highest 
	 * @param selfCancel indicates if a message cancel itself
	 * @return true if successful false if not
	 */
	public boolean setPriority(int priorityLevel, boolean selfCancel);

	/**
	 * Get engine identifiers
	 * 
	 * @return engineIdentifierArray
	 */
	public String[] getEngineIdentifiers();

	/**
	 * Select engine based on its identifier
	 * 
	 * @param engineIdentifier Identify the module required
	  * @return true if successfull false if not.
	 */
	public boolean selectEngine(String engineIdentifier);

	/**
	 * 
	 * @return selectedEngineIdentifier
	 */
	public String getSelectedEngineIdentifier();

	/**
	 * Select the synthesized language according to language code as specified by RFC 1766 
	 * @param languageCode
	 * @return
	 */
	public boolean selectLanguage(String languageCode);

	/**
	 * 
	 * @return languageCode as specified by RFC 1766 
	 */
	public String getSelectedLanguageCode();

	/**
		 * Get synthesizer voice identifiers
		 * 
		 * @return Voice identifier array.
		 */
	public String[] getVoiceIdentifiers();

	/**
	 * Select the synthesizer's voice
	 * 
	 * @param voiceIdentifier Identify the voice required
	 * @return true if successfull false if not.
	 */
	public boolean selectVoice(String voiceIdentifier);

	/**
	 * 
	 * @return voiceIdentifier
	 */
	public String getSelectedVoiceIdentifier();

	/**
	 * Get synthesizer voice variant identifiers
	 * Applies only to Formant type synthesizers
	 * @return variant identifier array.
	 */
	public String[] getVariantIdentifiers();

	/**
	  * Select the synthesizer's voice variant
	 * Applies only to Formant type synthesizers
		 * @param varientIdentifier Identify the variant required
	 * @return true if successfull false if not.
	 */
	public boolean selectVariant(String variantIdentifier);

	/**
	 * 
	 * @return variantIdentifier for a Formant synthesizer 
	 */
	public String getSelectedVariantIdentifier();

	/**
	* Speak message and return immediatly 
	 * 	  
	 * @param messageIdentifier Identifies message used in server
	 * @param message Message to speak
	 * @return true if successful false if not
	 */
	public boolean speakAsync(int MessageIdentifier, String message);

	/**
	* Speak char message and return immediatly 
	 * 	  
	 * @param messageIdentifier Identifies message used in server
	 * @param message Message to speak
	 * @return true if successful false if not
	 */
	public boolean speakAsyncChar(int messageIdentifier, String message);

	/**
	* Speak key message and return immediatly 
	 * 	  
	 * @param messageIdentifier Identifies message used in server
	 * @param message Message to speak
	 * @return true if successful false if not
	 */
	public boolean speakAsyncKey(int messageIdentifier, String message);

	/**
	 * Pause speech
	 * @return true if successful false if not
	 */
	public boolean pause();

	/**
	 * Resume speech after a pause
	* @return true if successful false if not
	 	 */
	public boolean resume();

	/**
	* Flushes the synthesizer last message buffer and cancel its current output.
	* @return true if successful false if not
	 */
	public boolean stop();

	/**
	* Flushes the synthesizer buffer and cancel its current output.
	*   @return true if successful false if not
		*/
	public boolean cancel();

	/**
	 * Set speech volume in the range of 0 - 100.0
	 * 
	 * @param volume Speech volume
	 * 
	 * @return true if successfull false if not
	 */
	public boolean setVolume(float volume);

	/**
	 * 
	 * @return volume in the range 0 - 100.0
	 */
	public float getVolume();

	/**
	 * Set speech rate in the range of -100.0 - 100.0
	 * 
	 * @param rate Speech rate
	 	 * 
	 * @return true if successfull false if not
	*/
	public boolean setRate(float rate);

	/**
	 * 
	 * @return rate in the range -100.0 - 100.0
	 */
	public float getRate();

	/**
	* Set pitch in the range -100 - 100.0
	* @return true if successfull false if not
	*/
	public boolean setPitch(float pitch);

	/**
	 * 
	 * @return pitch in the range -100.0 - 100.0
	 */
	public float getPitch();

	/**
	 * 	 This method is intended for use with less frequently called  features 
	 * that perhaps are set only a few times per session for example:
	 *  punctuation levels, spelling modes, punctuation level, spell mode, SSML mode, etc. 
	 * @param key property key
	 * @param value property value
	 * @return true if successful false if not
	 */
	public boolean setProperty(String key, String value);

	/**
	 * Listener for receiving module events
	 * @param eventsListener
	 * @return true if supported by the engine implementation false if not 
	 */
	public boolean setEventsListener(SHEventsListenerInterface eventsListener);

	/**
	 * Set handler to optionally bring upstream audio raw data generated by the synthesizer and process
	 * @param audioDataHandler
	 * @return true if supported by the engine implementation false if not 
	 */
	public boolean setAudioDataHandler(SHAudioDataHandlerInterface audioDataHandler);

}
