/*
* Copyright (c) Isaac Porat 2011 - 2013 all rights reserved.
* This file is released under the GNU General Public License, version 2, or (at your option) any later version.
* See the file SpeechHub_License.txt for more details.
*/

package sh.connections;

/**
 * TTS API 
 * 
 * @author Isaac Porat
 */
public interface SHTTSInterface
{

	/**
	 * Initialize engine connector
	 * @return true if successful false if not
	 */
	public boolean initialize();

	/**
	* Close engine connector
	* @return true if successful false if not
	*/
	public boolean close();

	/**
	 * Get engine identifiers 
	 * 
	 * @return Engine identifier array.
	 */
	public String[] getEngineIdentifiers();

	/**
	 * Select the engine
	 * 
	 * @param engineIdentifier Identify the engine required
	 * @return true if successful false if not
	 */
	public boolean selectEngine(String engineIdentifier);

	/**
	 * 	
	 * @return selectedEngineIdentifier
	 */
	public String getSelectedEngineIdentifier();

	/**
	 * Select the synthesized language 
	 * @param languageCode as specified by RFC 1766 
	 * @return true if successful false if not
	 */
	public boolean selectLanguage(String languageCode);

	/**
	 * 
	 * @return the synthesized language code as specified by RFC 1766 
	 */
	public String getSelectedLanguageCode();

	/**
	 * Get synthesizer voice identifiers
	 * 
	 * @return Voice identifier array.
	 */
	public String[] getVoiceIdentifiers();

	/**
	 * Select the synthesizer's voice
	 * 
	 * @param voiceIdentifier Identify the voice required
	 * @return true if successful false if not.
	 */
	public boolean selectVoice(String voiceIdentifier);

	/**
	 *  
	 * @return Selected voice identifier
	 */
	public String getSelectedVoiceIdentifier();

	/**
	 * Get synthesizer voice variant identifiers
	 * Applies only to Formant type synthesizers
	 * @return variant identifier array.
	 */
	public String[] getVariantIdentifiers();

	/**
	  * Select the synthesizer's voice variant
	 * Applies only to Formant type synthesizers
		 * @param varientIdentifier Identify the variant required
	 * @return true if successful false if not.
	 */
	public boolean selectVariant(String variantIdentifier);

	/**
	 * Applies only to Formant type synthesizers
	 * @return Selected variant identifier
	 */
	public String getSelectedVariantIdentifier();

	/**
	* Speak message 
	 * 	  
	 * @param messageIdentifier Identifies message used in server
	 * @param message Message to speak
	 * @param speakSync if true return after completion, if false return immediatly 
	 * @return true if successful false if not
	 */
	public boolean speak(int messageIdentifier, String message, boolean speakSync);

	/**
	* Speak char message 
	 * 	  
	 * @param messageIdentifier Identifies message used in server
	 * @param message Message to speak
	 * @param speakSync if true return after completion, if false return immediatly 
	 * @return true if successful false if not
	 */
	public boolean speakChar(int messageIdentifier, String message, boolean speakSync);

	/**
	* Speak key message 
	 * 	  
	 * @param messageIdentifier Identifies message used in server
	 * @param message Message to speak
	 * @param speakSync if true return after completion, if false return immediatly 
	 * @return true if successful false if not
	 */
	public boolean speakKey(int messageIdentifier, String message, boolean speakSync);

	/**
	* Speak message spelled 
	 * 	  
	 * @param messageIdentifier Identifies message used in server
	 * @param message Message to spell
	 * @param speakSync if true return after completion, if false return immediatly 
	 * @return true if successful false if not
	 */
	public boolean speakSpell(int messageIdentifier, String message, boolean speakSync);

	/**
		 * Pause speech
		 * @return true if successful false if not
		 	 */
	public boolean pause();

	/**
	 * Resume speech after a pause
	 * @return true if successful false if not
	 */
	public boolean resume();

	/**
	* Flushes the synthesizer last message buffer and cancel its current output.
	* @return true if successful false if not
	 */
	public boolean stop();

	/**
	* Flushes the synthesizer all messages buffer and cancel its current output.
	* @return true if successful false if not
	 */
	public boolean cancel();

	/**
	 * Set speech volume in the range of 0 - 100.0
	 * 
	 * @param volume Speech volume
	 * 
	 * @return true if successful false if not
	 */
	public boolean setVolume(float volume);

	/**
	 * 
	 * @return Volume
	 */
	public float getVolume();

	/**
	 * Set speech rate in the range of -100.0 - 100.0
	 * 
	 * @param rate Speech rate
	 	 * 
	 * @return true if successful false if not
	*/
	public boolean setRate(float rate);

	/**
	 * 
	 * @return Rate
	 */
	public float getRate();

	/**
	 * Switches rate boost on or off using Sonic
	 * @param mode true to switch on false to switch off
	 * @return true if successful false if not
	 */
	public boolean setRateBoostMode(boolean mode);

	/**
	 * Set pitch in % above and below the base pitch.<br/>
	 * 
	 * @param pitch Offset  in +-100.0% from base pitch for the voice.
	 * 
	 * @return true if successful false if not
	*/
	public boolean setPitch(float pitch);

	/**
	 * 
	 * @return Pitch offset  in +-100.0% from base pitch for the voice.
	 */
	public float getPitch();

	/**
	 * Set punctuation level
	 * @param level all, most, some, none 
	 * @return true if successful false if not
	 */
	public boolean setPunctuationLevel(String level);

	/**
	 * 
	 * @param charCapitalIndicator none, pitch, spell, icon
	 * @param arg only pitchLevel is supported with pitch charCapitalIndicator
	 * icon is not supported yet.  In the future when supported arg could be identifier or path to audio icon
	 * @return true if successful false if not
	 */
	public boolean setCharCapitalIndicator(String charCapitalIndicator, String arg);

	/**
	 * Silence before the next speak message
	 * Mainly used for recording audio to provide silence between chunks of text
	 * @param silenceMilliseconds before the next speak message
	 * @return true if successful false if not
	 */
	public boolean setSilenceBeforeMessage(int silenceMilliseconds);

	/**
	 * 	 This method can be used to set advanced module and engine settings
	 * Most settings where applicable wrap this method using dedicated methods
	 * @param key property key
	 * @param value property value
	 * @return true if successful false if not
	 */
	public boolean setProperty(String key, String value);

	/**
		 * This method is used to get information about module and engine configuration and advanced settings
		 * Engine factory properties can be obtained using compound key in the form:
		 * engine.factory::engineIdentifier::enginePropertyKey
		 * For example the key for getting the engine.caption property for espeak:
		 * engine.factory::espeak::engine.caption
		 * @param key property key
		 * @return property value if found null if not
		 */
	public String getProperty(String key);

	/**
	* Listener for receiving module events
	* @param eventsListener
	* @return true if successful false if not
	 */
	public boolean setEventsListener(SHEventsListenerInterface eventsListener);

	/**
	 * Set handler to optionally bring upstream audio raw data generated by the synthesizer and process
	 * @param audioDataHandler
	 * @return true if supported by the engine implementation false if not 
	 */
	public boolean setAudioDataHandler(SHAudioDataHandlerInterface audioDataHandler);

}
