/*
* Copyright (c) Isaac Porat 2011 - 2013 all rights reserved.
* This file is released under the GNU General Public License, version 2, or (at your option) any later version.
* See the file SpeechHub_License.txt for more details.
*/

package sh.connector.maryTTS;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Timer;
import java.util.TimerTask;

import log.common.CascadingLoggerInterface;

import org.w3c.dom.Document;
import org.w3c.dom.Node;

import sh.connections.SHConstants;
import sh.connections.SHCustomeHandlersInterface;
import sh.connections.SHEngineConnectorInterface;
import sh.connections.SHAudioDataHandlerInterface;
import sh.connections.SHEventsListenerInterface;
import sh.engine.SHEngineCommand;
import sh.engine.SHEngineExecLoader;
import sh.settings.SHEngineParameters;
import core.utilities.StackTrace;
import core.utilities.XMLUtils;

/**
 * Implementing MaryTTS client socket
 * @author Bill Cox
 *
 */
public class SHConnectorMaryTTSClientAdaptor implements SHEngineConnectorInterface
{
	private final static int PCM16BE = 3;
	private static final long SERVER_TIMEOUT_DELAY = 6000;
	// Note 30 seconds on very slow computers to ensure engine loading
	private static final long SERVER_INITIAL_TIMEOUT_DELAY = 30000; // We allow the engine to take a while to load
	private static final long TTS_TIMEOUT_DELAY = 30000;
	private static final long TTS_INITIAL_TIMEOUT_DELAY = 90000; // We allow the engine to take a while to load

	private CascadingLoggerInterface log;
	private InputStream fIn;
	private PrintStream fOut;
	private SHEngineExecLoader fEnginesLoader;
	private SHAudioDataHandlerInterface fAudioHandler;
	private int fSampleRate = 16000;
	private boolean fCancel = false;
	private String fVoiceName;
	private Timer fIOMonitor;
	private IOMonitorTask fIOMonitorTask;
	private boolean fJustStarted;
	private boolean fDebugMode = false;
	private boolean fServerMode = false;
	private SHEngineParameters fEngineParameters;

	public SHConnectorMaryTTSClientAdaptor(CascadingLoggerInterface logger)
	{
		log = logger.clone(SHConnectorMaryTTSClientAdaptor.class);
		fEnginesLoader = new SHEngineExecLoader(false, logger);
		fEngineParameters = SHEngineParameters.getInstance(logger);
		fIOMonitor = new Timer("IOMonitor");
	}

	private static int intAt(byte buf[], int index)
	{
		int x = index << 2;

		return ((buf[x] & 0xff) << 24) | ((buf[x + 1] & 0xff) << 16) | ((buf[x + 2] & 0xff) << 8) | (buf[x + 3] & 0xff);
	}

	private boolean readAuHeader(byte buf[])
	{
		// We're expecting a 24 byte AU header.
		int magic = intAt(buf, 0);
		int encoding = intAt(buf, 3);
		fSampleRate = intAt(buf, 4);
		int channels = intAt(buf, 5);
		if (magic != 0x2e736e64 || encoding != PCM16BE || channels != 1)
		{
			return false;
		}
		return true;
	}

	// Send text, and end it with a dot on a line by itself.  If the line starts with a dot, prepend another one.
	private void sendMessage(String message)
	{
		// Work around marytts bug where it hangs if the message starts with punctuation that ends phrases.
		int position = 0;
		char c;
		while (position < message.length()
				&& ((c = message.charAt(position)) == '.' || c == ',' || c == ';' || c == ':' || c == '?' || c == '!'))
		{
			position++;
		}
		if (position > 0)
		{
			message = message.substring(position);
			if (message.length() == 0)
				message = " ";
		}
		message = message.trim();
		if (message.length() == 0)
		{
			// Better than crashing!  Can we read it as punctuation?
			message = "dot";
		}
		String[] lines = message.split("\n");
		for (String line : lines)
		{
			line = line.trim();
			if (line.startsWith("."))
			{
				line = "." + line;
			}
			println(line);
		}
		println(".");
	}

	// Read a byte with a timeout.  This is needed because open mary sometimes just never returns any
	// data for a speak method.  The timeout is in milliseconds.
	private byte read() throws IOException
	{
		return (byte) fIn.read();
	}

	// Returns null if end of input.
	private byte[] readBytes() throws IOException
	{
		byte[] buf = new byte[1024];
		byte low, high;
		int position = 0;

		while (true)
		{
			high = read();
			low = read();
			if (high == (byte) 0x80 && low == (byte) 0)
			{
				if (position == 0)
				{
					if (fDebugMode)
						log.error("Got empty data block");
					restartTimer();
					return (null);
				}
				if (fDebugMode)
					log.notify("Got " + position + " bytes");
				restartTimer();
				fJustStarted = false;
				return Arrays.copyOf(buf, position);
			}
			if (position == buf.length)
			{
				byte[] newBuf = new byte[position << 1];
				System.arraycopy(buf, 0, newBuf, 0, position);
				buf = newBuf;
			}
			buf[position + 1] = high;
			buf[position] = low;
			position += 2;
		}
	}

	@Override
	public boolean speak(int MessageIdentifier, String message, boolean speakSync)
	{
		startTimer();
		try
		{
			fCancel = false;
			print("MARY IN=TEXT OUT=AUDIO LOCALE=en-us AUDIO=STREAMING_AU");
			if (fVoiceName != null && !fVoiceName.equals(""))
			{
				print(" VOICE=" + fVoiceName);
			}
			println("");
			sendMessage(message);
			byte[] auHeader = new byte[24];
			read(auHeader);
			if (readAuHeader(auHeader))
			{
				byte[] buf;
				do
				{
					buf = readBytes();
					if (buf != null)
					{
						if (!fAudioHandler.processAudio(buf))
						{
							fCancel = true;
							println("false");
						} else
						{
							println("true");
						}
					} else
					{
						println("true");
					}
				} while (!fCancel && buf != null);
			}
		} catch (IOException ex)
		{
			log.error("Open mary timed out on speak.");
			return false;
		}
		stopTimer();
		return true;
	}

	@Override
	public boolean cancel()
	{
		return true;
	}

	@Override
	public String[] getEngineIdentifiers()
	{
		// Not implemented, single engine module 	
		String[] engineIdentifiers =
		{ SHConstants.ENGINE_IDENTIFIER_MARYTTS };
		return (engineIdentifiers);
	}

	@Override
	public String getProperty(String key)
	{
		if (key == null)
			return (null);
		if (key.equals(SHConstants.PROP_ENGINE_SAMPLE_RATE))
			return (Integer.toString(fSampleRate));
		return null;
	}

	private String readLine() throws IOException
	{
		String line = "";
		char c;
		while ((c = (char) fIn.read()) != '\n')
		{
			line += c;
		}
		if (fDebugMode)
			log.notify("Got: " + line);
		fJustStarted = false;
		restartTimer();
		return line.trim();
	}

	@Override
	public String[] getVoiceIdentifiers()
	{
		/*
		 * Obtaining voiceIdentifiers list from engine replaced by getting the same information
		 * from installed manifest XML file so as to avoid loading the engine
		 * just for this purpose
		 * Note the voiceIdentifiers array is cached upstream
		 */
		//String[] voiceArray = getVoiceIdentifiersFromEngine(); 
		String[] voiceArray = getVoiceIdentifiersFromInstalledManifests();
		String voiceAlias;
		for (int n = 0; n < voiceArray.length; n++)
		{
			voiceAlias = fEngineParameters.getEngineVoicesFactoryProperty(SHConstants.ENGINE_IDENTIFIER_MARYTTS,
					voiceArray[n]);
			if (voiceAlias == null)
				continue;
			voiceArray[n] = voiceAlias;
		}
		return (voiceArray);
	}

	@SuppressWarnings("unused")
	private String[] getVoiceIdentifiersFromEngine()
	{
		startTimer();
		List<String> list = new ArrayList<String>();
		println("MARY LIST VOICES");
		try
		{
			String response = readLine();
			while (!response.isEmpty())
			{
				String fields[] = response.split(" ");
				list.add(fields[0] + "," + fields[1].toLowerCase().replace('_', '-'));
				response = readLine();
			}
		} catch (IOException e)
		{
			// Must have timed out
			return new String[0];
		}
		String voices[] = new String[list.size()];
		stopTimer();
		return list.toArray(voices);
	}

	@Override
	public boolean initialize()
	{
		if (!fEnginesLoader.loadEngine(SHConstants.ENGINE_IDENTIFIER_MARYTTS))
		{
			return (false);
		}
		fIn = fEnginesLoader.getStdinRaw();
		fOut = fEnginesLoader.getStdout();
		fJustStarted = true;
		return (true);
	}

	@Override
	public boolean close()
	{
		if (!fEnginesLoader.unloadActiveEngine())
		{
			return (false);
		}
		stopTimer();
		fIn = null;
		fOut = null;
		return (true);
	}

	@Override
	public boolean pause()
	{
		// Implemented by the player
		return false;
	}

	@Override
	public boolean resume()
	{
		// Implemented by the player
		return false;
	}

	@Override
	public boolean selectEngine(String engineIdentifier)
	{
		// One engine client
		return true;
	}

	@Override
	public boolean selectVoice(String voiceIdentifier)
	{
		// Remove locale if exists (separated by spaces)
		voiceIdentifier = voiceIdentifier.replaceAll("[, ][]a-z ]{5,5}$", "");
		// Reverse voice aliases if exists
		String voiceName = fEngineParameters.getEngineVoicesFactoryProperty(SHConstants.ENGINE_IDENTIFIER_MARYTTS,
				voiceIdentifier);
		if (voiceName == null)
			fVoiceName = voiceIdentifier;
		else
			fVoiceName = voiceName;
		return true;
	}

	@Override
	public String[] getVariantIdentifiers()
	{
		// Not implemented, variants not supported 
		return (new String[0]);
	}

	@Override
	public boolean selectVariant(String variantIdentifier)
	{
		// Not implemented, variants not supported 
		return (true); // true to prevent generating an error
	}

	@Override
	public boolean setEventsListener(SHEventsListenerInterface eventsListener)
	{
		// Not required
		return false;
	}

	@Override
	public boolean setAudioDataHandler(SHAudioDataHandlerInterface audioDataHandler)
	{
		fAudioHandler = audioDataHandler;
		return true;
	}

	@Override
	public boolean setCustomeHandlers(SHCustomeHandlersInterface customeHandlers)
	{
		// Implemented by SHPlayerEngine upstream 
		return (false);
	}

	@Override
	public boolean selectLanguage(String languageCode)
	{
		// TODO: Filter voices by the language code.
		return true;
	}

	@Override
	public boolean setPitch(float pitch)
	{
		// Implemented by Sonic
		return false;
	}

	@Override
	public boolean setProperty(String key, String value)
	{
		if (key.equals(SHConstants.PROP_COMPONENTS_CONFIGURATION))
		{
			if (value.equals(SHConstants.COMPONENTS_CONFIGURATION_SERVER))
				fServerMode = true;
			else
				fServerMode = false;
			return (true);
		}
		return (false);
	}

	@Override
	public boolean setRate(float rate)
	{
		// Implemented by Sonic
		return false;
	}

	@Override
	public boolean setVolume(float volume)
	{
		// Implemented by Sonic
		return false;
	}

	@Override
	public void checkEngine()
	{
		// Implemented in SHMaryTTSConnector
	}

	private void println(String line)
	{
		if (fDebugMode)
			log.notify("Sent: " + line);
		fOut.print(line + '\n');
		restartTimer();
	}

	private void print(String text)
	{
		if (fDebugMode)
			log.notify("Sent: " + text);
		fOut.print(text);
		restartTimer();
	}

	private void read(byte[] array) throws IOException
	{
		fIn.read(array);
		fJustStarted = false;
		restartTimer();
	}

	private class IOMonitorTask extends TimerTask
	{
		@Override
		public void run()
		{
			try
			{
				if (fIn != null)
					fIn.close();
			} catch (IOException e)
			{
				e.printStackTrace();
			}
			if (fOut != null)
				fOut.close();
		}
	}

	private void startTimer()
	{
		if (fIOMonitorTask == null)
		{
			fIOMonitorTask = new IOMonitorTask();
			if (fJustStarted)
			{
				if (fServerMode)
					fIOMonitor.schedule(fIOMonitorTask, SERVER_INITIAL_TIMEOUT_DELAY);
				else
					fIOMonitor.schedule(fIOMonitorTask, TTS_INITIAL_TIMEOUT_DELAY);
			} else if (fServerMode)
				fIOMonitor.schedule(fIOMonitorTask, SERVER_TIMEOUT_DELAY);
			else
				fIOMonitor.schedule(fIOMonitorTask, TTS_TIMEOUT_DELAY);
		}
	}

	private void restartTimer()
	{
		stopTimer();
		startTimer();
	}

	private void stopTimer()
	{
		if (fIOMonitorTask != null)
		{
			fIOMonitorTask.cancel();
			fIOMonitorTask = null;
		}
	}

	private String[] getVoiceIdentifiersFromInstalledManifests()
	{
		try
		{
			SHEngineCommand engineCommand = new SHEngineCommand(log);
			String enginePath = engineCommand.getEngineDirectoryPath(SHConstants.ENGINE_IDENTIFIER_MARYTTS);
			String installedPath = enginePath + "/stdio/installed";
			File installedFile = new File(installedPath);
			if (installedFile.exists() == false)
			{
				log.error("Error in getVoiceIdentifiersFromInstalledManifests method, directory does not exist: "
						+ installedPath);
				return (new String[0]);
			}
			File[] manifestsFile = installedFile.listFiles();
			String voiceIdentifier = null;
			List<String> voiceIdentifiersList = new ArrayList<String>();
			for (int n = 0; n < manifestsFile.length; n++)
			{
				voiceIdentifier = extractVoiceIdentifier(manifestsFile[n].getAbsolutePath());
				if (voiceIdentifier == null)
					continue;
				voiceIdentifiersList.add(voiceIdentifier);
			}
			return (voiceIdentifiersList.toArray(new String[0]));
		} catch (Exception ex)
		{
			log.error("Error in getVoiceIdentifiersFromInstalledManifests method:\n" + StackTrace.stackToString(ex));
			return (new String[0]);
		}
	}

	private String extractVoiceIdentifier(String manifestPath)
	{
		try
		{
			Document document = XMLUtils.parseXmlFile(manifestPath);
			Node node = XMLUtils.extractNode(document, "/marytts-install/voice");
			if (node == null)
				return (null); // Not all manifests are of a voice
			String name = XMLUtils.getAttributeValue(node, "name");
			if (name == null)
			{
				log.error("error in extractVoiceIdentifier method: name attribute null in file\n" + manifestPath);
				return (null);
			}
			String locale = XMLUtils.getAttributeValue(node, "locale");
			if (locale == null)
			{
				log.error("error in extractVoiceIdentifier method: local attribute null in file\n" + manifestPath);
				return (null);
			}
			String voiceIdentifier = name + "," + locale.toLowerCase();
			return (voiceIdentifier);
		} catch (Exception ex)
		{
			log.error("Error in extractVoiceIdentifier method:\n" + StackTrace.stackToString(ex));
			return (null);
		}
	}

}
