/*
* Copyright (c) Bill Cox and Isaac Porat 2011 - 2013 all rights reserved.
* This file is released under the GNU General Public License, version 2, or (at your option) any later version.
* See the file SpeechHub_License.txt for more details.
*/

package sh.connector.multiNative;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;

import log.common.BasicLogger;
import log.common.CascadingLoggerInterface;
import sh.connections.SHConstants;
import sh.connections.SHCustomeHandlersInterface;
import sh.connections.SHEngineConnectorInterface;
import sh.connections.SHAudioDataHandlerInterface;
import sh.connections.SHEventsListenerInterface;
import sh.settings.SHEngineParameters;
import core.utilities.StackTrace;

/**
 * Multi engine native connector
 * @author Bill Cox
 * @author Isaac Porat
 *
 */
public class SHConnectorMultiNativeAdaptor implements SHEngineConnectorInterface
{
	private CascadingLoggerInterface log;
	private SHEngineConnectorInterface fEngine;
	private SHEngineParameters fEngineParameters;
	private boolean fEngineInitialized = false;
	private boolean fEngineSelected = false;
	private String fSelectedEngineIdentifier = SHConstants.ENGINE_IDENTIFIER_DEFAULT;
	private boolean fDebugMode = false;
	private String fSelectedLanguageCode = "";
	private Map<String, String[]> fVoiceIdentifiersCache;

	public SHConnectorMultiNativeAdaptor(CascadingLoggerInterface logger)
	{
		log = logger.clone(SHConnectorMultiNativeAdaptor.class);
		fVoiceIdentifiersCache = new HashMap<String, String[]>();
		fEngineParameters = SHEngineParameters.getInstance(logger);
		loadEngineUserProperties();
		fEngine = new SHConnectorMultiNativeClientAdaptor(log);
	}

	/**
	 * Initialize engine connector.
	 * @return true if successful false if not
	 */
	@Override
	public boolean initialize()
	{
		// This method is unused, instead individual engines are initialized on selection 
		return true;
	}

	/**
	 * Initialize individual engine connectors.  This should be called after selecting the engine.  We
	 * automate this by detecting that a call has been made that requires the engine to be initialized.
	 * @return true if successful false if not
	 */
	private boolean initializeEngines()
	{
		try
		{
			if (fEngineInitialized)
			{
				debug("Skipping re-initialization");
				return true;
			}
			debug("Initializing");
			fEngineInitialized = fEngine.initialize();
			return (fEngineInitialized);
		} catch (Exception ex)
		{
			log.error("Error in initializeEngines method:\n" + StackTrace.stackToString(ex));
			return (false);
		}
	}

	/**
	* Get engine identifiers
	* 
	* @return engineIdentifierArray
	*/
	@Override
	public String[] getEngineIdentifiers()
	{
		try
		{
			debug("getEngineIdentifiers method called");
			return (fEngine.getEngineIdentifiers());
		} catch (Exception ex)
		{
			log.error("Error in getEngineIdentifiers method:\n" + ex);
			return (null);
		}
	}

	/**
	 * Select engine based on its identifier
	 * 
	 * @param engineIdentifier Identify the module required
	 * @return true if successful false if not.
	 */
	@Override
	public boolean selectEngine(String engineIdentifier)
	{
		try
		{
			if (engineIdentifier.equals(SHConstants.ENGINE_IDENTIFIER_DEFAULT))
				engineIdentifier = getProperty(SHConstants.PROP_ENGINE_IDENTIFIER_DEFAULT);
			if ((engineIdentifier.equals(fSelectedEngineIdentifier)) && (fEngineSelected))
			{
				debug("Engine " + engineIdentifier + " already selected... skipping selectEngine");
				return true;
			}
			debug("Selecting engine: " + engineIdentifier);
			fEngineInitialized = false;
			if (fEngine.selectEngine(engineIdentifier) == false)
				return false;
			fEngineSelected = true;
			fSelectedEngineIdentifier = engineIdentifier;
			return (true);
		} catch (Exception ex)
		{
			log.error("Error in selectEngine method: engine|Identifier: " + engineIdentifier + "\n"
					+ StackTrace.stackToString(ex));
			return (false);
		}
	}

	/**
	 * Check the engine initialization status and initialize if not done. 
	 */
	@Override
	public void checkEngine()
	{
		if (!fEngineSelected)
		{
			selectEngine(fSelectedEngineIdentifier);
		}
		if (!fEngineInitialized)
		{
			initializeEngines();
		}
	}

	/**
	* Close engine connector
	* @return true if successful false if not
	*/
	@Override
	public boolean close()
	{
		try
		{
			/*
			 * Close the engine connector which closes all open engines
			 */
			if (fEngineInitialized == false)
			{
				debug("Skipping closing, since we're not initialized");
				return true;
			}
			debug("Closing");
			fEngineInitialized = false;
			fEngineSelected = false;
			return fEngine.close(); // true if successful
		} catch (Exception ex)
		{
			log.error("Error in close method:\n" + ex);
			return (false);
		}
	}

	/**
	 * Set the synthesized language according to language code as specified by RFC 1766 
	 * @param languageCode
	 * @return
	 */
	@Override
	public boolean selectLanguage(String languageCode)
	{
		// Determine default voice selection
		fSelectedLanguageCode = languageCode;
		return (true);
	}

	/**
	     * Get synthesizer voice identifiers
	     * 
	     * @return Voice identifier array.
	     */
	@Override
	public String[] getVoiceIdentifiers()
	{
		try
		{
			String[] voiceIdentifiersArray = fVoiceIdentifiersCache.get(fSelectedEngineIdentifier);
			if (voiceIdentifiersArray != null)
				return (voiceIdentifiersArray);
			String[] voiceArray = fEngine.getVoiceIdentifiers();
			debug("Available voice identifiers:");
			if (fDebugMode)
			{
				for (int n = 0; n < voiceArray.length; n++)
				{
					debug(voiceArray[n]);
				}
			}
			// Now just filter out the voices that match the current language
			ArrayList<String> list = new ArrayList<String>();
			for (int i = 0; i < voiceArray.length; i++)
			{
				String fields[] = voiceArray[i].split(",");
				String name = fields[0];
				String subFields[] = fields[1].split("-");
				String language = subFields[0];
				String dialect = "none";
				if (subFields.length == 2)
				{
					dialect = subFields[1];
				}
				list.add(name + " " + language + " " + dialect);
			}
			voiceIdentifiersArray = list.toArray(new String[list.size()]);
			fVoiceIdentifiersCache.put(fSelectedEngineIdentifier, voiceIdentifiersArray);
			return (voiceIdentifiersArray);
		} catch (Exception ex)
		{
			log.error("Error in getVoiceIdentifiers method:\n" + ex);
			return (null);
		}
	}

	/**
	 * Select the synthesizer's voice
	 * 
	 * @param voiceIdentifier Identify the voice required
	 * @return true if successful false if not.
	 */
	@Override
	public boolean selectVoice(String voiceIdentifier)
	{
		try
		{
			// Remove language code if exists
			voiceIdentifier = voiceIdentifier.replaceAll(" [a-zA-Z]+ [a-zA-Z]+$", "");
			debug("Selecting voice: " + voiceIdentifier);
			if (voiceIdentifier.equalsIgnoreCase(SHConstants.ENGINE_VOICE_IDENTIFIER_DEFAULT))
			{
				// Select a default (first voice in the specified language)
				String[] voiceIdentifiersArray = getVoiceIdentifiers();
				String codePattern = " " + fSelectedLanguageCode + " ";
				for (int n = 0; n < voiceIdentifiersArray.length; n++)
				{
					if (voiceIdentifiersArray[n].contains(codePattern))
					{
						voiceIdentifier = voiceIdentifiersArray[n];
						break;
					}
				}
			}
			return fEngine.selectVoice(voiceIdentifier); // true if successful
		} catch (Exception ex)
		{
			log.error("Error in selectVoice method:\n" + ex);
			return (false);
		}
	}

	/**
	 * Get synthesizer voice variant identifiers
	 * Applies only to Formant type synthesizers
	 * @return variant identifier array.
	 */
	@Override
	public String[] getVariantIdentifiers()
	{
		try
		{
			return (fEngine.getVariantIdentifiers());
		} catch (Exception ex)
		{
			log.error("Error in getVariantIdentifiers method:\n" + ex);
			return (null);
		}
	}

	/**
	 * Select the synthesizer's voice variant
	* Applies only to Formant type synthesizers
		 * @param varientIdentifier Identify the variant required
	* @return true if successful false if not.
	*/
	@Override
	public boolean selectVariant(String variantIdentifier)
	{
		try
		{
			// Here connect to method to select a voice based on its identifier
			return (fEngine.selectVariant(variantIdentifier)); // true if successful
		} catch (Exception ex)
		{
			log.error("Error in selectVariant method:\n" + ex);
			return (false);
		}
	}

	/**
	* Speak message 
	 *       
	 * @param messageIdentifier Identifies message used in server
	 * @param message Message to speak
	 * @param speakSync if true return after completion, if false return immediatly 
	 * @return true if successful false if not
	 * Notes on implementation:
	 * There is no need to implement both speakSync and speakAsync functionality
	 * speakSync is used internally by SH and therefore prefered
	 * Alternatively, speakAsync can be implemented, in which case
	 * The end event must be implemented through the appropriate event listener interface
	 * The property "engine.supports.speak.sync" must be set accordingly for the module stack to work correctly 
	 */
	@Override
	public boolean speak(int MessageIdentifier, String message, boolean speakSync)
	{
		try
		{
			/*
			 * Note that this connector is configured with the relevant property
			 * so that only speakSync is supported
			 * There is no need to check for speakAsync 
			 */
			debug("Speaking message: " + message);
			debug("speakSync flag is set to: " + speakSync);
			return (fEngine.speak(MessageIdentifier, message, speakSync));
		} catch (Exception ex)
		{
			log.error("Error in speak method:\n" + StackTrace.stackToString(ex));
			return (false);
		}
	}

	/**
	 * Pause speech
	 * 
	 * @return true if successful false if not
	  */
	@Override
	public boolean pause()
	{
		try
		{
			debug("Pausing speech");
			return (fEngine.pause());
		} catch (Exception ex)
		{
			log.error("Error in pause method:\n" + ex);
			return (false);
		}
	}

	/**
	 * Resume speech after a pause
	 */
	@Override
	public boolean resume()
	{
		try
		{
			debug("Resuming speech");
			return (fEngine.resume());
		} catch (Exception ex)
		{
			log.error("Error in resume method:\n" + ex);
			return (false);
		}
	}

	/**
	* Flushes the synthesizer buffer and cancel its current output.
	* @return true if successful false if not
	        */
	@Override
	public boolean cancel()
	{
		try
		{
			if (!fEngineInitialized)
			{
				debug("Skipping cancel speech since we're not initialized.");
				return true;
			}
			debug("Cancelling speech");
			return (fEngine.cancel());
		} catch (Exception ex)
		{
			log.error("Error in cancel method:\n" + ex);
			return (false);
		}
	}

	/**
	 * Set speech volume in the range of 0 - 100.0
	 * 
	 * @param volume Speech volume
	 * 
	 * @return true if successful false if not
	 */
	@Override
	public boolean setVolume(float volume)
	{
		try
		{
			// Here scale as necessary and connect to the engine's volume method
			debug("Volume: " + volume);
			return fEngine.setVolume(volume); // true if successful false if not 
		} catch (Exception ex)
		{
			log.error("Error in setVolume method:\n" + ex);
			return (false);
		}
	}

	/**
	 * Set speech rate in the range of -100.0 to 100.0
	 * 
	 * @param rate Speech rate
	      * 
	 * @return true if successful false if not
	*/
	@Override
	public boolean setRate(float rate)
	{
		try
		{
			// Here scale as necessary and connect to the engine's rate method
			debug("Rate: " + rate);
			return fEngine.setRate(rate); // true if successful false if not
		} catch (Exception ex)
		{
			log.error("Error in setRate method:\n" + ex);
			return (false);
		}
	}

	/**
	* Set pitch above and below the base pitch.<br/>
	* 
	* @param pitch Offset  in -100.0 to +100 from base pitch for the voice.
	* 
	* @return true if successful false if not
	*/
	@Override
	public boolean setPitch(float pitch)
	{
		try
		{
			// Here scale as necessary and connect to the engine's pitch method
			debug("Pitch: " + pitch);
			return fEngine.setPitch(pitch); // true if successful false if not
		} catch (Exception ex)
		{
			log.error("Error in setPitch method:\n" + ex);
			return (false);
		}
	}

	/**
	* This method is used to get information about engine capabilities and advanced settings
	* @param key property key
	* @return property value if found null if not
	*/
	@Override
	public String getProperty(String key)
	{
		try
		{
			if (key == null)
				return (null);
			if (key.equals(SHConstants.PROP_ENGINE_SUPPORTS_SPEAK_SYNC))
				return ("true"); // true if engine supports synchronized speech 
			if (key.equals(SHConstants.PROP_ENGINE_REQUIRES_ISOLATION_RESTART))
				return ("true");
			if (key.equals(SHConstants.PROP_ENGINE_SUPPORTS_AUDIO_DATA))
				return ("true"); // All engines returns audio data supported by SHEnginePlayer
			if (key.equals(SHConstants.PROP_ENGINE_SUPPORTS_PAUSE_RESUME))
				return ("true"); // Always supported by SHEnginePlayer 
			if (key.equals(SHConstants.PROP_ENGINE_IDENTIFIER_DEFAULT))
			{
				/*
				 * For a multi engines connector return the default engine loaded on startup
				 * This is important as it is used (together with other parameters)
				 * to point to the engine settings directory 
				 */
				String engineIdentifier = SHConstants.ENGINE_IDENTIFIER_ESPEAK;
				return (engineIdentifier);
			}
			// Check 'Factory set' engine properties
			String enginePropertiesIdentifier = fSelectedEngineIdentifier;
			if (enginePropertiesIdentifier.equals(SHConstants.ENGINE_IDENTIFIER_DEFAULT))
			{
				// Call this method again to obtain property
				enginePropertiesIdentifier = getProperty(SHConstants.PROP_ENGINE_IDENTIFIER_DEFAULT);
			}
			if (key.equals(SHConstants.PROP_ENGINE_SUPPORTS_PUNCTUATION))
				return (fEngineParameters.getEngineFactoryProperty(enginePropertiesIdentifier, key));
			if (key.equals(SHConstants.PROP_ENGINE_SUPPORTS_SSML))
				return (fEngineParameters.getEngineFactoryProperty(enginePropertiesIdentifier, key));
			if (key.equals(SHConstants.PROP_ENGINE_USE_SONIC))
				return (fEngineParameters.getEngineFactoryProperty(enginePropertiesIdentifier, key));
			checkEngine();
			return fEngine.getProperty(key);
		} catch (Exception ex)
		{
			log.error("Error in getProperty method:\n" + ex);
			return (null);
		}
	}

	/**
	*      This method is used to set more advanced engine options such as punctuation levels, spelling modes etc. 
	* @param key property key
	* @param value property value
	* @return true if successful false if not
	*/
	@Override
	public boolean setProperty(String key, String value)
	{
		try
		{
			if ((key == null) || (value == null))
				return (false);
			debug("set property key: " + key + " value: " + value);
			return (fEngine.setProperty(key, value));
		} catch (Exception ex)
		{
			log.error("Error in setProperty method:\n" + ex);
			return (false);
		}
	}

	/**
	 * Listener for receiving module events
	 * @param eventsListener
	 * @return true if supported by the engine's implementation false if not 
	 */
	@Override
	public boolean setEventsListener(SHEventsListenerInterface eventsListener)
	{
		// SpeakSync implemented no need for this functionality
		debug("set event listener");
		return (true); // true if successful false if not
	}

	/**
	 * Set handler to optionally bring upstream audio raw data generated by the engine and process
	 * @param audioHandler
	 * @return true if supported by the engine's implementation false if not 
	 */
	@Override
	public boolean setAudioDataHandler(SHAudioDataHandlerInterface audioDataHandler)
	{
		try
		{
			debug("set audio data handler");
			return fEngine.setAudioDataHandler(audioDataHandler);
		} catch (Exception ex)
		{
			log.error("Error in setAudioDataHandler method:\n" + ex);
			return (false);
		}
	}

	/**
	 * Custome handlers for upstream and global communication
	 * @param customeHandlers
	 * @return true if supported by the engine implementation false if not 
		 */
	@Override
	public boolean setCustomeHandlers(SHCustomeHandlersInterface customeHandlers)
	{
		try
		{
			debug("set custome handlers");
			return fEngine.setCustomeHandlers(customeHandlers);
		} catch (Exception ex)
		{
			log.error("Error in setCustomeHandlers method:\n" + ex);
			return (false);
		}
	}

	private void loadEngineUserProperties()
	{
		try
		{
			fDebugMode = false;
			String debugModeConnectorIdentifier = fEngineParameters
					.getEngineCommonProperty(SHConstants.PROP_ENGINE_DEBUG_MODE_CONNECTOR_IDENTIFIER);
			if ((debugModeConnectorIdentifier == null) || debugModeConnectorIdentifier.equals("none"))
				fDebugMode = false;
			else if ((debugModeConnectorIdentifier.equals("all"))
					|| (debugModeConnectorIdentifier.equals(SHConstants.ENGINE_CONNECTOR_IDENTIFIER_MULTI_NATIVE)))
				fDebugMode = true;
		} catch (Exception ex)
		{
			log.error("Error in loadEngineUserProperties method:\n" + ex);
		}
	}

	private void debug(String message)
	{
		if (fDebugMode)
		{
			BasicLogger.notifyPrintln(message);
		}
	}

}
