/*
* Copyright (c) Bill Cox 2011 - 2013 all rights reserved.
* This file is released under the GNU General Public License, version 2, or (at your option) any later version.
* See the file SpeechHub_License.txt for more details.
*/

package sh.connector.multiNative;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.PrintStream;

import core.utilities.StackTrace;

import log.common.CascadingLoggerInterface;
import sh.connections.SHConstants;
import sh.connections.SHCustomeHandlersInterface;
import sh.connections.SHEngineConnectorInterface;
import sh.connections.SHAudioDataHandlerInterface;
import sh.connections.SHEventsListenerInterface;
import sh.engine.SHEngineExecLoader;
import sh.settings.SHEngineParameters;

/**
 * 
 * @author Bill Cox
 *
 */
public class SHConnectorMultiNativeClientAdaptor implements SHEngineConnectorInterface
{
	private CascadingLoggerInterface log;
	private SHEngineParameters fEngineParameters;
	private SHEngineExecLoader fEnginesLoader;
	private BufferedReader fIn;
	private PrintStream fOut;
	private String[] fVoiceIdentifiers = null;
	private SHAudioDataHandlerInterface fAudioHandler = null;
	private String fEngineIdentifier;

	public SHConnectorMultiNativeClientAdaptor(CascadingLoggerInterface logger)
	{
		log = logger.clone(SHConnectorMultiNativeClientAdaptor.class);
		fEngineParameters = SHEngineParameters.getInstance(logger);
		fEnginesLoader = new SHEngineExecLoader(true, logger);
		fEngineIdentifier = null;
	}

	private boolean readBoolean() throws IOException
	{
		String response = readln();
		if (response.equalsIgnoreCase("true"))
		{
			return true;
		} else if (response.equalsIgnoreCase("false"))
		{
			return false;
		}
		throw (new IOException());
	}

	private int readInt() throws IOException
	{
		String response = readln();
		return Integer.parseInt(response);
	}

	public String[] getEngineIdentifiers()
	{
		return fEngineParameters.getProtocolEngineIdentifiers(SHConstants.ENGINE_PROTOCOL_NATIVE, "1.0");
	}

	/**
	 * Ask the engine what version of the stdio protocol it uses.  For now, we expect it to be 1.
	 */
	private int getVersion()
	{
		try
		{
			writeln("get version");
			return readInt();
		} catch (Exception ex)
		{
			// This will be reported by caller.
		}
		return -1;
	}

	/**
	 * Load engine
	 * Return true if successful false if not
	 */
	public boolean selectEngine(String engineIdentifier)
	{
		String previousEngineIdentifier = fEngineIdentifier;
		fVoiceIdentifiers = null;
		if (fEnginesLoader.loadEngine(engineIdentifier))
		{
			fIn = fEnginesLoader.getStdin();
			fOut = fEnginesLoader.getStdout();
			// See if the engine is alive
			if (getVersion() != 1)
			{
				log.error("Unable to comunicate with engine.  Shutting it down.");
				close();
				if (previousEngineIdentifier != null)
					selectEngine(previousEngineIdentifier);
				return (false);
			}
			// Load voice identifiers up-front so we wont interrupt speech later on. 
			getVoiceIdentifiers();
			fEngineIdentifier = engineIdentifier;
			return (true);
		}
		return (false);
	}

	public boolean cancel()
	{
		// This method should never be needed.
		return (true);
	}

	// Close the client, terminating the process.
	public boolean close()
	{
		if (fEnginesLoader.unloadActiveEngine())
		{
			fIn = null;
			fOut = null;
			fVoiceIdentifiers = null;
			fEngineIdentifier = null;
			return (true);
		}
		return (false);
	}

	public String getProperty(String key)
	{
		if (key == null)
			return (null);
		if (key.equals(SHConstants.PROP_ENGINE_SAMPLE_RATE))
		{
			try
			{
				writeln("get samplerate");
				return Integer.toString(readInt());
			} catch (Exception ex)
			{
				log.error("Error in getSampleRate method.  Is the TTS engine working properly?");
			}
		}
		return (null);
	}

	private void loadVoiceIdentifiers() throws IOException
	{
		if (fVoiceIdentifiers == null)
		{
			writeln("get voices");
			int numStrings = readInt();
			fVoiceIdentifiers = new String[numStrings];
			for (int i = 0; i < numStrings; i++)
			{
				fVoiceIdentifiers[i] = readln();
			}
		}
	}

	/*
	 * We have the engine return a list of strings in pairs, where the
	 * first is the voice name, and the second is it's language/dialect code.
	*	 This method just filters out the voice identifiers based on the current
	 * language.
	*/
	public String[] getVoiceIdentifiers()
	{
		try
		{
			loadVoiceIdentifiers();
			return fVoiceIdentifiers;
		} catch (Exception ex)
		{
			log.error("Error on client socket getVoiceIdentifiers: " + StackTrace.stackToString(ex));
			return null; // Is this the best return value on error?
		}
	}

	// We initialize the engine when we select it, so there's nothing to do here.
	public boolean initialize()
	{
		return true;
	}

	public boolean pause()
	{
		// Not used, implemented by SHEnginePlayer
		return false;
	}

	public boolean resume()
	{
		// Not used, implemented by SHEnginePlayer
		return false;
	}

	public boolean selectVoice(String voiceIdentifier)
	{
		try
		{
			writeln("set voice " + voiceIdentifier);
			return readBoolean();
		} catch (Exception ex)
		{
			log.error("Error on client socket selectVoice: " + StackTrace.stackToString(ex));
			return false;
		}
	}

	@Override
	public String[] getVariantIdentifiers()
	{
		try
		{
			writeln("get variants");
			int numStrings = readInt();
			String variantIdentifiers[] = new String[numStrings];
			for (int i = 0; i < numStrings; i++)
			{
				variantIdentifiers[i] = readln();
			}
			return variantIdentifiers;
		} catch (Exception ex)
		{
			log.error("Error on client socket getVariantIdentifiers: " + StackTrace.stackToString(ex));
			return (null);
		}
	}

	@Override
	public boolean selectVariant(String variantIdentifier)
	{
		try
		{
			writeln("set variant " + variantIdentifier);
			return readBoolean();
		} catch (Exception ex)
		{
			log.error("Error on client socket selectVariant: " + StackTrace.stackToString(ex));
			return false;
		}
	}

	@Override
	public boolean setEventsListener(SHEventsListenerInterface eventsListener)
	{
		// Not required
		return false;
	}

	@Override
	public boolean setAudioDataHandler(SHAudioDataHandlerInterface audioDataHandler)
	{
		fAudioHandler = audioDataHandler;
		return true;
	}

	@Override
	public boolean setCustomeHandlers(SHCustomeHandlersInterface customeHandlers)
	{
		// Implemented by SHPlayerEngine upstream 
		return (false);
	}

	public boolean selectLanguage(String languageCode)
	{
		return true;
	}

	public boolean setPitch(float pitch)
	{
		try
		{
			writeln("set pitch " + pitch);
			return readBoolean();
		} catch (Exception ex)
		{
			log.error("Error on client socket setPitch: " + StackTrace.stackToString(ex));
			return false;
		}
	}

	public boolean setProperty(String key, String value)
	{
		try
		{
			if (key.equals(SHConstants.PROP_ENGINE_SSML_MODE))
			{
				writeln("set ssml " + value);
				return readBoolean();
			}
			if (key.equals(SHConstants.PROP_ENGINE_PUNCTUATION_LEVEL))
			{
				writeln("set punctuation " + value);
				return readBoolean();
			}
		} catch (Exception ex)
		{
			log.error("Error on client socket setProperty: key: " + key + " value: " + value + "\n"
					+ StackTrace.stackToString(ex));
			return false;
		}
		return false;
	}

	public boolean setRate(float rate)
	{
		try
		{
			writeln("set speed " + rate);
			return readBoolean();
		} catch (Exception ex)
		{
			log.error("Error on client socket setRate: " + StackTrace.stackToString(ex));
			return false;
		}
	}

	public boolean setVolume(float volume)
	{
		// Do we want to use other than max no-distortion volume? 
		return (true);
	}

	// Convert hexadecimal strings to a byte array.
	private byte[] decodeHex(String hexString) throws IOException
	{
		String upper = hexString.toUpperCase();
		int length = upper.length();
		byte data[] = new byte[length / 2];
		for (int i = 0; i < length >> 2; i++)
		{
			short value = 0;
			for (int j = 0; j < 4; j++)
			{
				value <<= 4;
				char c = upper.charAt((i << 2) + j);
				if (c >= 'A' && c <= 'F')
				{
					value |= (short) (10 + (int) (c) - (int) 'A');
				} else if (c >= '0' && c <= '9')
				{
					value |= (short) ((int) (c) - (int) '0');
				} else
				{
					throw new IOException();
				}
			}
			// We store in little-endian, right?
			data[(i << 1) + 0] = (byte) (value & 0xff);
			data[(i << 1) + 1] = (byte) ((value >> 8) & 0xff);
		}
		return data;
	}

	// Speak the message.  Send the "speak" command, and the message one
	// line at a time.  If a line starts with ".", add another "." to
	// insure we don't accidentally end the message.  End with "." on a
	// line by itself.  Then, read synthesized samples until we read "done".
	public boolean speak(int MessageIdentifier, String message, boolean speakSync)
	{
		try
		{
			writeln("speak");
			String lines[] = message.split("\n");
			for (int i = 0; i < lines.length; i++)
			{
				String line = lines[i];
				if (line.startsWith("."))
				{
					writeln("." + line);
				} else
				{
					writeln(line);
				}
			}
			writeln(".");
			String response = readln();
			while (response != null && !response.equals("true") && !response.equals("false"))
			{
				byte samples[] = decodeHex(response);
				if (fAudioHandler.processAudio(samples))
				{
					writeln("true");
				} else
				{
					writeln("false");
				}
				response = readln();
			}
			return response != null && response.equals("true");
		} catch (Exception ex)
		{
			log.error("Error on client socket speak: " + StackTrace.stackToString(ex));
			return false;
		}
	}

	public void checkEngine()
	{
		// Implemented in SHMultiNativeConnector
	}

	public boolean watchdog(boolean reset)
	{
		// This method is never called, functionality is handled by SHEnginePlayer 
		return (true);
	}

	private void writeln(String text)
	{
		//log.error("Writing: " + text);
		fOut.print(text + "\n");
	}

	private String readln() throws IOException
	{
		String response = fIn.readLine();
		//log.error("Read: " + response);
		return response;
	}

}
