/*
* Copyright (c) Isaac Porat 2011 - 2013 all rights reserved.
* This file is released under the GNU General Public License, version 2, or (at your option) any later version.
* See the file SpeechHub_License.txt for more details.
*/

package sh.connector.sapi5;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import log.common.CascadingLoggerInterface;
import sh.connections.SHConstants;
import sh.connections.SHAudioDataHandlerInterface;

import com.jacob.activeX.ActiveXComponent;
import com.jacob.com.ComThread;
import com.jacob.com.Dispatch;
import com.jacob.com.DispatchEvents;
import com.jacob.com.Variant;

import core.utilities.StackTrace;
import core.windows.registry.WinRegistry;

/**
 * SAPI5 com wrapper with audio plugin for feeding back audio data to the SH player
 *  
 * @author Isaac Porat
 *
 */
public class SHSAPI5ComWrapperAudioPlugin
{
	private CascadingLoggerInterface log;
	private ActiveXComponent fTTSComponent;
	private Dispatch fTTSDispatch;
	private ActiveXComponent fOutputStreamComponent;
	private String fVoiceIdentifier = "Daniel";
	private int fVolume = 50;
	private int fRate = 0;
	private int fPitchOffset = 0;
	private int fSpeakFlags = SVSFlagsAsync;
	//private static final int SVSFDefault = 0;
	private static final int SVSFlagsAsync = 1;
	@SuppressWarnings("unused")
	private static final int SVSFPurgeBeforeSpeak = 2;
	private static int SVSFNLPSpeakPunc = 64;

	private static int SAFT8kHz16BitMono = 6;
	private static int SAFT11kHz16BitMono = 10;
	private static int SAFT12kHz16BitMono = 14;
	private static int SAFT16kHz16BitMono = 18;
	private static int SAFT22kHz16BitMono = 22;
	private static int SAFT24kHz16BitMono = 26;
	private static int SAFT32kHz16BitMono = 30;
	private static int SAFT44kHz16BitMono = 34;
	private static int SAFT48kHz16BitMono = 38;

	private static int SEEK_RELATIVE_START = 0;
	private static int SEEK_RELATIVE_CURRENT = 1;
	private static int SEEK_RELATIVE_END = 2;
	// Events - unused
	private static int SVEStartInputStream = 2;
	private static int SVEEndInputStream = 4;

	private int fSampleRate = 0;
	private int readIndex = 0;
	private int fRequiredChunk = 1024;
	private SHAudioDataHandlerInterface fAudioDataHandler;
	// DispatchEvent field is used directly by jacob using reflection and is not refered to by this class
	@SuppressWarnings("unused")
	private DispatchEvents fSAPI5DispatchEvents;
	private String fMessage;
	private SHEngineLanguageFactoryProperties fLanguageProperties;
	private final static String SAPI_SPVOICE_PROGID = "Sapi.SpVoice";
	private static String fSAPISpVoiceClsid = null;
	private final static String SAPI_SPMEMORYSTREAM_PROGID = "Sapi.SpMemoryStream";
	private static String fSAPISpMemoryStreamClsid = null;

	public SHSAPI5ComWrapperAudioPlugin(CascadingLoggerInterface logger)
	{
		log = logger.clone(SHSAPI5ComWrapperAudioPlugin.class);
		fLanguageProperties = new SHEngineLanguageFactoryProperties(SHConstants.ENGINE_IDENTIFIER_SAPI5, logger);
	}

	public static boolean isAvailable(CascadingLoggerInterface logger)
	{
		try
		{
			if ((fSAPISpVoiceClsid != null) && (fSAPISpMemoryStreamClsid != null))
				return (true);
			String sapiSpVoiceClsidKey = SAPI_SPVOICE_PROGID + "\\CLSID";
			fSAPISpVoiceClsid = WinRegistry.readString(WinRegistry.HKEY_LOCAL_MACHINE, "software\\classes\\"
					+ sapiSpVoiceClsidKey, "");
			if (fSAPISpVoiceClsid == null)
			{
				logger.notify("SAPI SPVoice CLSID not found, probably SAPI not installed");
				return (false);
			}
			fSAPISpVoiceClsid = "clsid:" + fSAPISpVoiceClsid;

			String sapiSpMemoryStreamClsidKey = SAPI_SPMEMORYSTREAM_PROGID + "\\CLSID";
			fSAPISpMemoryStreamClsid = WinRegistry.readString(WinRegistry.HKEY_LOCAL_MACHINE, "software\\classes\\"
					+ sapiSpMemoryStreamClsidKey, "");
			if (fSAPISpMemoryStreamClsid == null)
			{
				logger.notify("SAPI SPMemoryStream CLSID not found, probably SAPI not installed");
				return (false);
			}
			fSAPISpMemoryStreamClsid = "clsid:" + fSAPISpMemoryStreamClsid;

			return (true);
		} catch (Exception ex)
		{
			logger.error("Error in ComWrapper isAvailable method:\n" + StackTrace.stackToString(ex));
			return (false);
		}
	}

	public void initialize()
	{
		/*
		 * Using clsid directly created problems with JFrame component not initializing (obtaining focus)
		 * properly and therefore abandond
		 * in favour of using progid instead    
		 	// clsid obtained from progid using isAvailable method which must be called first	
		// By obtaining the clsid once hopefully Jacob does not need to go to the registry each time an object is created
		fTTSComponent = new ActiveXComponent(fSAPISpVoiceClsid);
		*/
		fTTSComponent = new ActiveXComponent(SAPI_SPVOICE_PROGID);
		fTTSDispatch = fTTSComponent.getObject();
		//fTTSComponent.setProperty("AllowAudioOutputFormatChangesOnNextSet", false);
		configureOutputStream();
	}

	/**
	 * Events never worked and the same aim was achieved by polling waitUntilDone
	 * This was perhaps due to fault with 1.16 1.15 might work
	 * Events problems are perhaps due to code error but more probably
	 * because of failure of using Jacob correctly for this purpose
	 * It could be that SAPI5 does not implement certain event interface (see Jacob docs on web main page)
	 * in which case a different constructor might be required.
	 * Events use was never debugged properly one possibility is to switch debug on in Jacob.  
	 */
	@SuppressWarnings("unused")
	private void configureEvents()
	{
		fSAPI5DispatchEvents = new DispatchEvents(fTTSDispatch, new SHSAPI5ComEvents());
		fTTSComponent.setProperty("EventInterests", (SVEStartInputStream | SVEEndInputStream));
	}

	private void configureOutputStream()
	{
		/*
		 * Using clsid directly created problems with JFrame component not initializing (obtaining focus)
		 * properly and therefore abandond
		 * in favour of using progid instead    
		 // clsid obtained from progid using isAvailable method which must be called first	
		// By obtaining the clsid once hopefully Jacob does not need to go to the registry each time an object is created
		fOutputStreamComponent = new ActiveXComponent(fSAPISpMemoryStreamClsid);
		*/
		fOutputStreamComponent = new ActiveXComponent(SAPI_SPMEMORYSTREAM_PROGID);

		Dispatch outputStreamDispatch = fOutputStreamComponent.getObject();
		Dispatch.putRef(fTTSDispatch, "AudioOutputStream", outputStreamDispatch);

		/*
		 * The following set manually the format
		 * Microsoft Anna has noise but the reason for this was not found
		 */
		/*
		ActiveXComponent formatComponent = fOutputStreamComponent.getPropertyAsComponent("format");
		formatComponent.setProperty("type", SAFT22kHz16BitMono);
		Dispatch formatDispatch = formatComponent.getObject();
		Dispatch.putRef(fOutputStreamComponent, "format", formatDispatch);
		*/
	}

	/**
	 * Unused, uses SPCustomStream, no need
	 * for this, SPMemoryStream can be used directly
	 */
	@SuppressWarnings("unused")
	private void configureOutputStreamCustom()
	{
		ActiveXComponent customStreamComponent = new ActiveXComponent("speech.SPCustomStream");
		Dispatch customStreamDispatch = customStreamComponent.getObject();
		ActiveXComponent memoryStreamComponent = new ActiveXComponent(SAPI_SPMEMORYSTREAM_PROGID);
		Dispatch memoryStreamDispatch = memoryStreamComponent.getObject();
		Dispatch.putRef(customStreamDispatch, "BaseStream", memoryStreamDispatch);
		Dispatch.putRef(fTTSDispatch, "AudioOutputStream", customStreamDispatch);
		fOutputStreamComponent = memoryStreamComponent;
	}

	@SuppressWarnings("unused")
	private void speakWithAttributes(String string) throws InterruptedException
	{
		String msSpeech = "<voice required=\"Name=" + fVoiceIdentifier + "\">" + "<volume level = \"" + fVolume
				+ "\"><rate absspeed =\"" + fRate + "\">" + string + "</rate></volume></voice>";
		speak(msSpeech);
	}

	public void speak(String message) throws InterruptedException
	{
		fMessage = message;
		/*
		// Engine simulated crash during speech for SHEngineWatchdog class test
		if (message.contains("kK"))
		{
			throw new InterruptedException("Simulating engine crash");
		}
		*/
		// Newline needs to be removed for correct prosity in Speech Platform
		message = message.replaceAll("\n", " ").trim();

		message = "<pitch absmiddle=\"" + fPitchOffset + "\">" + message + "</pitch>";
		resetBuffer();

		fTTSComponent.invoke("speak", message, fSpeakFlags);
	}

	private boolean waitUntilDone(int timeout)
	{
		Variant result = fTTSComponent.invoke("WaitUntilDone", timeout);
		return (result.changeType(Variant.VariantBoolean).getBoolean());
	}

	private void obtainAudioFormat()
	{
		ActiveXComponent formatComponent = fOutputStreamComponent.getPropertyAsComponent("format");
		Variant typeVariant = formatComponent.getProperty("type");
		int type = typeVariant.changeType(Variant.VariantInt).getInt();
		if (type == SAFT8kHz16BitMono)
			fSampleRate = 8000;
		else if (type == SAFT11kHz16BitMono)
			fSampleRate = 11025;
		else if (type == SAFT12kHz16BitMono)
			fSampleRate = 12000;
		else if (type == SAFT16kHz16BitMono)
			fSampleRate = 16000;
		else if (type == SAFT22kHz16BitMono)
			fSampleRate = 22050;
		else if (type == SAFT24kHz16BitMono)
			fSampleRate = 24000;
		else if (type == SAFT32kHz16BitMono)
			fSampleRate = 32000;
		else if (type == SAFT44kHz16BitMono)
			fSampleRate = 44100;
		else if (type == SAFT48kHz16BitMono)
			fSampleRate = 48000;
		else
			log.error("Voice audio format type: " + type + " not recognized");
	}

	public int getSampleRate()
	{
		obtainAudioFormat();
		return (fSampleRate);
	}

	public void processAudio()
	{
		readIndex = 0;
		int snapshotEndMessageIndex = 0;
		int snapshotPreviousEndMessageIndex = 0;
		int delay = 10;
		if (fMessage.length() > 10)
			delay = 200;
		for (int n = 0; n < 1000; n++)
		{
			if (waitUntilDone(delay))
				break;
			Variant snapshotEndMessageIndexVariant = fOutputStreamComponent.invoke("seek", 0, SEEK_RELATIVE_CURRENT);
			snapshotEndMessageIndex = snapshotEndMessageIndexVariant.changeType(Variant.VariantInt).getInt();
			if (snapshotEndMessageIndex > snapshotPreviousEndMessageIndex)
			{
				if (processAudioAsync(snapshotEndMessageIndex) == false) // Message cancelled
					return;
				snapshotPreviousEndMessageIndex = snapshotEndMessageIndex;
			}
		}
		// Complete message
		//Variant endMessageIndexVariant = fOutputStreamComponent.invoke("seek", 0, SEEK_RELATIVE_CURRENT);
		Variant endMessageIndexVariant = fOutputStreamComponent.invoke("seek", 0, SEEK_RELATIVE_END);
		int endMessageIndex = endMessageIndexVariant.changeType(Variant.VariantInt).getInt();
		if (endMessageIndex > snapshotEndMessageIndex)
		{
			processAudioAsync(endMessageIndex);
		}
	}

	@SuppressWarnings("unused")
	private void processAudioAfterDone()
	{
		waitUntilDone(10000);
		Variant endMessageIndexVariant = fOutputStreamComponent.invoke("seek", 0, SEEK_RELATIVE_CURRENT);
		int endMessageIndex = endMessageIndexVariant.changeType(Variant.VariantInt).getInt() - fRequiredChunk; // Point to last actual data to avoid overrun to previous message held in  the buffer   
		processAudioAsync(endMessageIndex);
	}

	@SuppressWarnings("unused")
	private boolean processAudioSync(int endMessageIndex)
	{
		fOutputStreamComponent.invoke("seek", readIndex, SEEK_RELATIVE_START);
		for (;;)
		{
			Variant samples = new Variant(Variant.VariantArray, true);
			int actualChunk = fRequiredChunk;
			if ((readIndex + fRequiredChunk) > endMessageIndex)
				actualChunk = endMessageIndex - readIndex;
			Variant count = fOutputStreamComponent.invoke("read", samples, new Variant(actualChunk));
			int readBytes = count.changeType(Variant.VariantInt).getInt();
			readIndex += readBytes;
			byte[] bytes = samples.toSafeArray().toByteArray();
			if (bytes.length == 0) // Avoid false audio termination indicator
				return (true);
			if (fAudioDataHandler.processAudio(bytes) == false) // Message cancelled
				return (false);
			if (readIndex >= endMessageIndex)
				return (true);
		}
	}

	private boolean processAudioAsync(int endMessageIndex)
	{
		Variant samplesVariant = Dispatch.call(fOutputStreamComponent, "getData");
		byte[] bytes = samplesVariant.toSafeArray().toByteArray();
		for (;;)
		{
			int actualChunk = fRequiredChunk;
			if ((readIndex + fRequiredChunk) > endMessageIndex)
				actualChunk = endMessageIndex - readIndex;
			byte[] bytesChunk = readArray(bytes, actualChunk);
			readIndex += bytesChunk.length;
			if (bytesChunk.length == 0) // Avoid false audio termination indicator
				return (true);
			if (fAudioDataHandler.processAudio(bytesChunk) == false) // Message cancelled
				return (false);
			if (readIndex >= endMessageIndex)
				return (true);
		}
	}

	private byte[] readArray(byte[] bytes, int actualChunk)
	{
		byte[] bytesChunk = new byte[actualChunk];
		System.arraycopy(bytes, readIndex, bytesChunk, 0, actualChunk);
		return (bytesChunk);
	}

	private void resetBuffer()
	{
		/*
		// Example of using setData method with no data- unused; it does not reset buffer
		// Left for potential future use
		SafeArray bytesSafeArray = new SafeArray(Variant.VariantByte, 0);
		bytesSafeArray.fromByteArray(new byte[0]);
		Variant bytesVariant = new Variant();
		bytesVariant.putSafeArrayRef(bytesSafeArray);
		fOutputStreamComponent.invoke("setData", bytesVariant);
		*/

		// For debugging
		//Variant index = fOutputStreamComponent.invoke("seek", 0, SEEK_RELATIVE_END);
		//int bufferEnd = index.changeType(Variant.VariantInt).getInt();
		/*
				* To avoid memory errors it is important to follow the following sequence
				* GC is not used with Jacob, release memory to avoid leaks
				 * To ensure that the references to the components are not removed it is important
				 * to make sure that the speak Thread is different from the initialize thread.
				 * This is true for normal operation and restart but it is on the same thread
				 * automatic shutdown of the engine which is therefore kept alive
		*/
		ComThread.Release();
		// Configuring output stream seems to be the only reliable method to allow flawless operation with all voices
		configureOutputStream();
		fOutputStreamComponent.invoke("seek", 0, SEEK_RELATIVE_START);
	}

	public void volume(int volume)
	{
		// Range 0 to 100
		fTTSComponent.setProperty("volume", volume);
		fVolume = volume;
	}

	public void rate(int rate)
	{
		// Range -10 to 10
		fTTSComponent.setProperty("rate", rate);
		fRate = rate;
	}

	public void pitch(int pitchOffset)
	{
		// Range -10 to 10
		fPitchOffset = pitchOffset;
	}

	public String[] getAllVoicesWithLanguageCode()
	{
		List<String> voicesList = new ArrayList<String>();
		Dispatch voices = Dispatch.call(fTTSDispatch, "getVoices").toDispatch();
		Variant itemCount = Dispatch.get(voices, "Count");
		for (int i = 0; i < itemCount.changeType(Variant.VariantInt).getInt(); i++)
		{
			Dispatch voiceDispatch = Dispatch.call(voices, "Item", new Variant(i)).toDispatch();
			Variant voiceName = Dispatch.call(voiceDispatch, "GetDescription");
			String voiceIdentifier = voiceName.toString();
			String rfcLanguageCode = getRFCCombinedCode(voiceDispatch);
			if (rfcLanguageCode == null)
				continue;
			// Avoid Microsoft sample voices
			if (voiceIdentifier.startsWith("SampleTTSVoice"))
				continue;
			// Replace spaces with double underline
			voiceIdentifier = voiceIdentifier.replaceAll(" ", "__");
			voicesList.add(voiceIdentifier + rfcLanguageCode);
		}
		return (voicesList.toArray(new String[voicesList.size()]));
	}

	private String getRFCCombinedCode(Dispatch voiceDispatch)
	{
		try
		{
			if (voiceDispatch == null)
				return (null);
			Variant languageAttributeVariant = Dispatch.call(voiceDispatch, "GetAttribute", "language");
			if (languageAttributeVariant == null)
				return (null);
			String languageCode = languageAttributeVariant.toString();
			if (languageCode == null)
				return (null);
			if (languageCode.contains(";"))
				languageCode = languageCode.replaceAll(";[a-zA-Z0-9]*$", "");
			languageCode = languageCode.toLowerCase().trim();
			return (getCombinedRFCCode(languageCode));
		} catch (Exception ex)
		{
			log.error("Error in getRFCCombinedCode method:\n" + ex);
			return (null);
		}
	}

	private String getCombinedRFCCode(String microsoftCombinedCode) throws IOException
	{
		microsoftCombinedCode = microsoftCombinedCode.toLowerCase();
		if (microsoftCombinedCode.length() == 3)
			microsoftCombinedCode = "0" + microsoftCombinedCode;
		String rfcCombinedCode = getRFCLanguageCountryCode(microsoftCombinedCode);
		if (rfcCombinedCode != null)
			return (" " + rfcCombinedCode);
		// Combined language country code not found, try language code only
		String rfcLanguageCode = getRFCLanguageCode(microsoftCombinedCode);
		if (rfcLanguageCode != null)
			return (" " + rfcLanguageCode + " none");
		// Language not found	
		log.error("Language code: " + microsoftCombinedCode + " not specified");
		return (" none none");
	}

	private String getRFCLanguageCode(String microsoftCombinedCode) throws IOException
	{
		String microsoftLanguageCode = microsoftCombinedCode.substring(2);
		return (fLanguageProperties.getLanguageCodeProperty(microsoftLanguageCode));
	}

	private String getRFCLanguageCountryCode(String microsoftCombinedCountryCode) throws IOException
	{
		return (fLanguageProperties.getLanguageCountryCodeProperty(microsoftCombinedCountryCode));
	}

	private boolean filterVoice(Dispatch voiceDispatch)
	{
		try
		{
			if (voiceDispatch == null)
				return (false);
			Variant languageAttributeVariant = Dispatch.call(voiceDispatch, "GetAttribute", "language");
			if (languageAttributeVariant == null)
				return (false);
			String languageCode = languageAttributeVariant.toString();
			if (languageCode == null)
				return (false);
			if (languageCode.contains(";"))
				languageCode = languageCode.replaceAll(";[a-zA-Z0-9]*$", "");
			String rfcCode = getCombinedRFCCode(languageCode);
			if (rfcCode.startsWith(" none"))
				return (false);
			return (true);
		} catch (Exception ex)
		{
			log.error("Error in filterVoice method:\n" + ex);
			return (false);
		}
	}

	public String getVoice()
	{
		ActiveXComponent voiceComponent = fTTSComponent.getPropertyAsComponent("voice");
		Variant voiceName = voiceComponent.invoke("getDescription");
		return (voiceName.toString() + "\n\n");
	}

	public String getVoiceIdentifier()
	{
		return (fVoiceIdentifier);
	}

	public boolean setVoice(String voiceIdentifier)
	{
		if (setSAPIVoice(voiceIdentifier))
		{
			obtainAudioFormat();
			return (true);
		}
		return (false);
	}

	private boolean setSAPIVoice(String voiceIdentifier)
	{
		if (voiceIdentifier == null)
			return (false);
		// Replace double underline with space
		voiceIdentifier = voiceIdentifier.replaceAll("__", " ");
		voiceIdentifier = voiceIdentifier.toLowerCase();
		Dispatch voices = Dispatch.call(fTTSDispatch, "getVoices").toDispatch();
		Variant itemCount = Dispatch.get(voices, "Count");
		for (int i = 0; i < itemCount.changeType(Variant.VariantInt).getInt(); i++)
		{
			// Jacob 1.16 fails on next line with NVDA
			// Error of method not found
			// OK with Jacob 1.15
			Dispatch voiceDispatch = Dispatch.call(voices, "Item", i).toDispatch();
			if (filterVoice(voiceDispatch))
			{
				Variant voiceNameVariant = Dispatch.call(voiceDispatch, "GetDescription");
				String voiceName = voiceNameVariant.toString();
				if (voiceName != null)
				{
					if (voiceName.toLowerCase().contains(voiceIdentifier))
					{
						fVoiceIdentifier = voiceName;
						Dispatch.putRef(fTTSDispatch, "voice", voiceDispatch);
						return (true);
					}
				}
			}
		}
		return (false);
	}

	public void pause()
	{
		fTTSComponent.invoke("pause");
	}

	public void resume()
	{
		fTTSComponent.invoke("resume");
	}

	public void cancelAll()
	{
		// Unused functionality implemented by processAudio return false from player
		//fTTSComponent.invoke("speak", "", (SVSFPurgeBeforeSpeak));
	}

	public void setPunctuation(boolean mode)
	{
		if (mode)
			fSpeakFlags = SVSFNLPSpeakPunc | SVSFlagsAsync;
		else
			fSpeakFlags = SVSFlagsAsync;
	}

	public void close()
	{
		// No method for shutting the voice was found
		ComThread.Release();
		// Small delay is recommended before the system is shutdown
		try
		{
			Thread.sleep(100);
		} catch (InterruptedException ex)
		{
		}
		fTTSComponent = null;
		fTTSDispatch = null;
		fOutputStreamComponent = null;
	}

	public boolean setAudioDataHandler(SHAudioDataHandlerInterface audioDataHandler)
	{
		fAudioDataHandler = audioDataHandler;
		return (true);
	}

}
