/*
* Copyright (c) Isaac Porat 2011 - 2013 all rights reserved.
* This file is released under the GNU General Public License, version 2, or (at your option) any later version.
* See the file SpeechHub_License.txt for more details.
*/

package sh.connector.speechPlatform;

import core.utilities.StackTrace;
import log.common.CascadingLoggerInterface;
import sh.connections.SHConstants;
import sh.connections.SHCustomeHandlersInterface;
import sh.connections.SHEngineConnectorInterface;
import sh.connections.SHAudioDataHandlerInterface;
import sh.connections.SHEventsListenerInterface;

/**
 * Microsoft Speech Platform engine connector adaptor with audio plugin
 * @author Isaac Porat
 */
public class SHConnectorSpeechPlatformAdaptorAudioPlugin implements SHEngineConnectorInterface
{
	private CascadingLoggerInterface log;
	private final static String VOICE_IDENTIFIER_DEFAULT = "Microsoft"; // Any microsoft voice
	private final static int SINGLE_CHAR_InCREASED_TTS_RATE = 3;
	private SHSpeechPlatformComWrapperAudioPlugin fSCW;
	private String fPunctuationLevel = "none";
	private String[] fVoiceIdentifiersCache;
	private SHAudioDataHandlerInterface fAudioDataHandler;
	private int fTTSRate = 0;
	private int fTTSSingleCharRate = 0;
	private boolean fSingleCharMode = false;

	public SHConnectorSpeechPlatformAdaptorAudioPlugin(CascadingLoggerInterface logger)
	{
		log = logger.clone(SHConnectorSpeechPlatformAdaptorAudioPlugin.class);
		/*
		 * Debug mode was provided and disabled but lines kept for future implementation
		if (fBundleConfig != null)
		{
			String debugModeString = fBundleConfig.getProperty("debugMode");
			if ((debugModeString != null) && (debugModeString.toLowerCase().equals("true")))
				fDebugMode = true;
				}
		*/
	}

	@Override
	public boolean initialize()
	{
		try
		{
			if (fSCW == null)
			{
				fSCW = new SHSpeechPlatformComWrapperAudioPlugin(log);
				fSCW.initialize();
				setAudioDataHandler(fAudioDataHandler);
			}
			return (true);
		} catch (Exception ex)
		{
			log.error("Error in initialize method:\n" + ex);
			return (false);
		}
	}

	@Override
	public boolean close()
	{
		try
		{
			if (fSCW != null)
			{
				fSCW.close();
				fSCW = null;
			}
			return (true);
		} catch (Exception ex)
		{
			log.error("Error in close method:\n" + ex);
			return (false);
		}
	}

	@Override
	public void checkEngine()
	{
		if (fSCW == null)
			initialize();
	}

	@Override
	public String[] getEngineIdentifiers()
	{
		// Not implemented, single engine module 	
		String[] engineIdentifiers =
		{ SHConstants.ENGINE_IDENTIFIER_MSP };
		return (engineIdentifiers);
	}

	@Override
	public boolean selectEngine(String engineIdentifier)
	{
		// Not implemented, single engine module 	
		return (true);
	}

	@Override
	public String[] getVoiceIdentifiers()
	{
		try
		{
			if (fVoiceIdentifiersCache == null)
				fVoiceIdentifiersCache = fSCW.getAllVoicesWithLanguageCode();
			return (fVoiceIdentifiersCache);
		} catch (Exception ex)
		{
			log.error("Error in getVoiceIdentifiers method:\n" + ex);
			return (null);
		}
	}

	@Override
	public boolean selectVoice(String voiceIdentifier)
	{
		try
		{
			// Remove language code if exists
			voiceIdentifier = voiceIdentifier.replaceAll(" [a-zA-Z]+ [a-zA-Z]+$", "");
			if (voiceIdentifier.equalsIgnoreCase(SHConstants.ENGINE_VOICE_IDENTIFIER_DEFAULT))
				voiceIdentifier = VOICE_IDENTIFIER_DEFAULT;
			return (fSCW.setVoice(voiceIdentifier));
		} catch (Exception ex)
		{
			log.error("Error in selectVoice method:\n" + ex);
			return (false);
		}
	}

	@Override
	public String[] getVariantIdentifiers()
	{
		/* 
		* Not implemented - applies to Formant synthesizers only
		  */
		return (new String[0]);
	}

	@Override
	public boolean selectVariant(String variantIdentifier)
	{
		/* 
		* Not implemented - applies to Formant synthesizers only
		  */
		return (true); // true to avoid generating error
	}

	@Override
	public boolean speak(int messageIdentifier, String message, boolean speakSync)
	{
		try
		{
			if (message.length() == 1) // single char
			{
				if (fSingleCharMode == false) // Transition
					fSCW.rate(fTTSSingleCharRate);
				fSingleCharMode = true;
			} else
			// Longer message
			{
				if (fSingleCharMode) // Transition	
					fSCW.rate(fTTSRate);
				fSingleCharMode = false;
			}
			// messageIdentifier unused 
			return (speak(message, speakSync));
		} catch (Exception ex)
		{
			log.error("Error in speakMessage method:\n" + ex);
			return (false);
		}
	}

	/**
		* Sends string to synthesizer and wait for completion if required<br/>
		* 
		* @param message message to speak
		* @param speakSync true - wait for completion, false - return immediatly.
		*/
	private boolean speak(String message, boolean speakSync)
	{
		/*
		*  Convert special characters  to entities to ensures no '<' and '>'
		* enter the system as if tags not closed they cause Speech Platform to crash
		*/
		try
		{
			if (message.contains("&"))
			{
				// First replace entities back to characters
				message = message.replaceAll("&[lL][tT];", "<");
				message = message.replaceAll("&[gG][tT];", ">");
				message = message.replaceAll("&[aA][mM][pP];", "&");
				// Convert back to entities
				if (message.contains("&"))
					message = message.replaceAll("&", "&amp;");
			}
			if (message.contains("<"))
				message = message.replaceAll("<", "&lt;");
			if (message.contains(">"))
				message = message.replaceAll(">", "&gt;");
			fSCW.speak(message);
			fSCW.processAudio();
			return (true);
		} catch (Exception ex)
		{
			log.error("Error in speak method:\n" + StackTrace.stackToString(ex));
			return (false);
		}
	}

	@Override
	public boolean pause()
	{
		try
		{
			fSCW.pause();
			return (true);
		} catch (Exception ex)
		{
			log.error("Error in pause method:\n" + ex);
			return (false);
		}
	}

	@Override
	public boolean resume()
	{
		try
		{
			fSCW.resume();
			return (true);
		} catch (Exception ex)
		{
			log.error("Error in resume method:\n" + ex);
			return (false);
		}
	}

	@Override
	public boolean cancel()
	{
		try
		{
			if (fSCW != null)
				fSCW.cancelAll();
			return (true);
		} catch (Exception ex)
		{
			log.error("Error in cancel method:\n" + ex);
			return (false);
		}
	}

	@Override
	public boolean setPitch(float pitch)
	{
		// Range -100.0 to 100.0
		try
		{
			int ttsPitch = (int) (pitch / 10);
			fSCW.pitch(ttsPitch);
			return (true);
		} catch (Exception ex)
		{
			log.error("Error in setPitch: " + ex);
			return (false);
		}
	}

	@Override
	public boolean setRate(float rate)
	{
		// Range -100.0 to 100.0
		try
		{
			fTTSRate = (int) (rate / 10);
			// Apply increase rate for single characters if Sonic is not used
			if (fTTSRate > 0)
			{
				fTTSSingleCharRate = fTTSRate + SINGLE_CHAR_InCREASED_TTS_RATE;
				if (fTTSSingleCharRate > 10)
					fTTSSingleCharRate = 10;
			} else
			{
				// Probably Sonic, do not increase rate for single chars to avoid distortion 
				fTTSSingleCharRate = 0;
			}
			fSCW.rate(fTTSRate);
			return (true);
		} catch (Exception ex)
		{
			log.error("Error in setRate: " + ex);
			return (false);
		}
	}

	@Override
	public boolean selectLanguage(String languageCode)
	{
		// Unused, default voice determined by non language specific method 
		return (true);
	}

	@Override
	public boolean setVolume(float volume)
	{
		// Range 0 to 100.0
		try
		{
			fSCW.volume((int) volume);
			return (true);
		} catch (Exception ex)
		{
			log.error("Error in setVolume: " + ex);
			return (false);
		}
	}

	@Override
	public String getProperty(String key)
	{
		try
		{
			if (key == null)
				return (null);
			if (key.equals(SHConstants.PROP_ENGINE_SUPPORTS_SPEAK_SYNC))
				return ("true");
			if (key.equals(SHConstants.PROP_ENGINE_REQUIRES_ISOLATION_RESTART))
				return ("true");
			if (key.equals(SHConstants.PROP_ENGINE_SUPPORTS_PUNCTUATION))
				return ("true");
			if (key.equals(SHConstants.PROP_ENGINE_PUNCTUATION_LEVEL))
				return (fPunctuationLevel);
			if (key.equals(SHConstants.PROP_ENGINE_SUPPORTS_PAUSE_RESUME))
				return ("true");
			if (key.equals(SHConstants.PROP_ENGINE_SUPPORTS_SSML))
			{
				// Provide support in due course for SSML in SpeechPlatform 
				return ("false");
			}
			if (key.equals(SHConstants.PROP_ENGINE_SUPPORTS_AUDIO_DATA))
				return ("true");
			if (key.equals(SHConstants.PROP_ENGINE_SAMPLE_RATE))
				return (Integer.toString(fSCW.getSampleRate()));
			if (key.equals(SHConstants.PROP_ENGINE_IDENTIFIER_DEFAULT))
				return (SHConstants.ENGINE_IDENTIFIER_MSP);
			return (null);
		} catch (Exception ex)
		{
			log.error("Error in getProperty method:\n" + StackTrace.stackToString(ex));
			return (null);
		}
	}

	@Override
	public boolean setProperty(String key, String value)
	{
		try
		{
			if ((key == null) || (value == null))
				return (false);
			if (key.equals(SHConstants.PROP_ENGINE_PUNCTUATION_LEVEL))
			{
				if ((value.equals(SHConstants.ENGINE_PUNCTUATION_LEVEL_SOME))
						|| (value.equals(SHConstants.ENGINE_PUNCTUATION_LEVEL_MOST))
						|| (value.equals(SHConstants.ENGINE_PUNCTUATION_LEVEL_ALL)))
				{
					// Note only 'some' punctuation level  is supported
					fSCW.setPunctuation(true);
					fPunctuationLevel = SHConstants.ENGINE_PUNCTUATION_LEVEL_SOME;
				} else
				{
					fSCW.setPunctuation(false);
					fPunctuationLevel = SHConstants.ENGINE_PUNCTUATION_LEVEL_NONE;
				}
				return (true);
			}
			return (false);
		} catch (Exception ex)
		{
			log.error("Error in setProperty method:\n" + StackTrace.stackToString(ex));
			return (false);
		}
	}

	@Override
	public boolean setEventsListener(SHEventsListenerInterface ttsModuleEventsListener)
	{
		// Not implemented
		return (false);
	}

	@Override
	public boolean setAudioDataHandler(SHAudioDataHandlerInterface audioDataHandler)
	{
		fAudioDataHandler = audioDataHandler;
		return (fSCW.setAudioDataHandler(audioDataHandler));
	}

	@Override
	public boolean setCustomeHandlers(SHCustomeHandlersInterface customeHandlers)
	{
		// Not required 
		return (false);
	}

	public static boolean isAvailable(CascadingLoggerInterface logger)
	{
		return (SHSpeechPlatformComWrapperAudioPlugin.isAvailable(logger));
	}

}
