/*
* Copyright (c) Isaac Porat 2011 - 2013 all rights reserved.
* This file is released under the GNU General Public License, version 2, or (at your option) any later version.
* See the file SpeechHub_License.txt for more details.
*/

package sh.connector.speechPlatform;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import sh.connections.SHConstants;

import log.common.CascadingLoggerInterface;

import com.jacob.activeX.ActiveXComponent;
import com.jacob.com.ComThread;
import com.jacob.com.Dispatch;
import com.jacob.com.Variant;

import core.utilities.StackTrace;
import core.windows.registry.WinRegistry;

/**
 * Microsoft Speech Platform com wrapper with self audio
 * It seems that Speech Platform is backward compatible with SAPI5 although the author did not see documentation to that effect
 * This class is almost identical to the equivalent SAPI5 com wrapper
 * One difference is the progid of the component
 * Note that in SAPI5 it is necessary to initialize before a voice change
 *   this does not seems to be necessary in MSP see setVoice method
 * Although this class could have been reduced to a subclass on the SAPI equivalent, it was decided to keep them separate for potential branching out in the future in case that compatibility with SAPI5 is broken
 *  
 * @author Isaac Porat
 *
 */
public class SHSpeechPlatformComWrapperSelfAudio
{
	private CascadingLoggerInterface log;
	private ActiveXComponent fTTSComponent;
	private String fVoiceIdentifier = "Daniel";
	private int fVolume = 50;
	private int fRate = 0;
	private int fPitchOffset = 0;
	private int fSpeakFlags = SVSFlagsAsync;
	//private static final int SVSFDefault = 0;
	private static final int SVSFlagsAsync = 1;
	private static final int SVSFPurgeBeforeSpeak = 2;
	private static int SVSFNLPSpeakPunc = 64;

	/*
	Other constants for future reference
	SVSFIsFilename = 4 
	SVSFIsXML = 8
	SVSFIsNotXML = 16
	SVSFPersistXML = 32
	*/
	private SHEngineLanguageFactoryProperties fLanguageProperties;
	private final static String MSP_SPVOICE_PROGID = "Speech.SpVoice";
	private static String fMSPSpVoiceClsid = null;

	public SHSpeechPlatformComWrapperSelfAudio(CascadingLoggerInterface logger)
	{
		log = logger.clone(SHSpeechPlatformComWrapperSelfAudio.class);
		fLanguageProperties = new SHEngineLanguageFactoryProperties(SHConstants.ENGINE_IDENTIFIER_MSP, logger);
	}

	public static boolean isAvailable(CascadingLoggerInterface logger)
	{
		try
		{
			if (fMSPSpVoiceClsid != null)
				return (true);
			String mspSpVoiceClsidKey = MSP_SPVOICE_PROGID + "\\CLSID";
			fMSPSpVoiceClsid = WinRegistry.readString(WinRegistry.HKEY_LOCAL_MACHINE, "software\\classes\\"
					+ mspSpVoiceClsidKey, "");
			if (fMSPSpVoiceClsid == null)
			{
				logger.notify("MSP SPVoice CLSID not found, probably MSP not installed");
				return (false);
			}
			fMSPSpVoiceClsid = "clsid:" + fMSPSpVoiceClsid;

			return (true);
		} catch (Exception ex)
		{
			logger.error("Error in ComWrapper isAvailable method:\n" + StackTrace.stackToString(ex));
			return (false);
		}
	}

	public void initialize()
	{
		/*
		 * Using clsid directly created problems with JFrame component not initializing (obtaining focus)
		 * properly and therefore abandond
		 * in favour of using progid instead    
		 	// clsid obtained from progid using isAvailable method which must be called first	
		// By obtaining the clsid once hopefully Jacob does not need to go to the registry each time an object is created
		fTTSComponent = new ActiveXComponent(fMSPSpVoiceClsid);
		*/
		fTTSComponent = new ActiveXComponent(MSP_SPVOICE_PROGID);
		//		fTTSComponent.setProperty("AllowAudioOutputFormatChangesOnNextSet", false);
	}

	@SuppressWarnings("unused")
	private void speakWithAttributes(String message) throws InterruptedException
	{
		String msSpeech = "<voice required=\"Name=" + fVoiceIdentifier + "\">" + "<volume level = \"" + fVolume
				+ "\"><rate absspeed =\"" + fRate + "\">" + message + "</rate></volume></voice>";
		speak(msSpeech);
		//fTTSComponent.invoke ("speak", msSpeech);    
	}

	public void speak(String message) throws InterruptedException
	{
		/*
		// Engine simulated crash during speech for SHEngineWatchdog class test
		if (message.contains("kK"))
		{
			throw new InterruptedException("Simulating engine crash");
		}
		*/
		// Newline needs to be removed for correct prosity in Speech Platform
		message = message.replaceAll("\n", " ").trim();

		message = "<pitch absmiddle=\"" + fPitchOffset + "\">" + message + "</pitch>";
		fTTSComponent.invoke("speak", message, fSpeakFlags);
	}

	public void waitUntilDone()
	{
		fTTSComponent.invoke("WaitUntilDone", 10000);
	}

	public void volume(int volume)
	{
		// Range 0 to 100
		fTTSComponent.setProperty("volume", volume);
		fVolume = volume;
	}

	public void rate(int rate)
	{
		// Range -10 to 10
		fTTSComponent.setProperty("rate", rate);
		fRate = rate;
	}

	public void pitch(int pitchOffset)
	{
		// Range -10 to 10
		fPitchOffset = pitchOffset;
	}

	public String[] getAllVoicesWithLanguageCode()
	{
		List<String> voicesList = new ArrayList<String>();
		Dispatch ttsDispatch = fTTSComponent.getObject();
		Dispatch voices = Dispatch.call(ttsDispatch, "getVoices").toDispatch();
		Variant itemCount = Dispatch.get(voices, "Count");
		for (int i = 0; i < itemCount.getInt(); i++)
		{
			Dispatch voiceDispatch = Dispatch.call(voices, "Item", new Variant(i)).toDispatch();
			Variant voiceName = Dispatch.call(voiceDispatch, "GetDescription");
			String voiceIdentifier = voiceName.toString();
			String rfcLanguageCode = getRFCCombinedCode(voiceDispatch);
			if (rfcLanguageCode == null)
				continue;
			// Avoid Microsoft sample voices
			if (voiceIdentifier.startsWith("SampleTTSVoice"))
				continue;
			// Replace spaces with double underline
			voiceIdentifier = voiceIdentifier.replaceAll(" ", "__");
			voicesList.add(voiceIdentifier + rfcLanguageCode);
		}
		return (voicesList.toArray(new String[voicesList.size()]));
	}

	private String getRFCCombinedCode(Dispatch voiceDispatch)
	{
		try
		{
			if (voiceDispatch == null)
				return (null);
			Variant languageAttributeVariant = Dispatch.call(voiceDispatch, "GetAttribute", "language");
			if (languageAttributeVariant == null)
				return (null);
			String languageCode = languageAttributeVariant.toString();
			if (languageCode == null)
				return (null);
			if (languageCode.contains(";"))
				languageCode = languageCode.replaceAll(";[a-zA-Z0-9]*$", "");
			languageCode = languageCode.toLowerCase().trim();
			return (getCombinedRFCCode(languageCode));
		} catch (Exception ex)
		{
			log.error("Error in getRFCCombinedCode method:\n" + ex);
			return (null);
		}
	}

	private String getCombinedRFCCode(String microsoftCombinedCode) throws IOException
	{
		microsoftCombinedCode = microsoftCombinedCode.toLowerCase();
		if (microsoftCombinedCode.length() == 3)
			microsoftCombinedCode = "0" + microsoftCombinedCode;
		String rfcCombinedCode = getRFCLanguageCountryCode(microsoftCombinedCode);
		if (rfcCombinedCode != null)
			return (" " + rfcCombinedCode);
		// Combined language country code not found, try language code only
		String rfcLanguageCode = getRFCLanguageCode(microsoftCombinedCode);
		if (rfcLanguageCode != null)
			return (" " + rfcLanguageCode + " none");
		// Language not found	
		log.error("Language code: " + microsoftCombinedCode + " not specified");
		return (" none none");
	}

	private String getRFCLanguageCode(String microsoftCombinedCode) throws IOException
	{
		String microsoftLanguageCode = microsoftCombinedCode.substring(2);
		return (fLanguageProperties.getLanguageCodeProperty(microsoftLanguageCode));
	}

	private String getRFCLanguageCountryCode(String microsoftCombinedCountryCode) throws IOException
	{
		return (fLanguageProperties.getLanguageCountryCodeProperty(microsoftCombinedCountryCode));
	}

	private boolean filterVoice(Dispatch voiceDispatch)
	{
		try
		{
			if (voiceDispatch == null)
				return (false);
			Variant languageAttributeVariant = Dispatch.call(voiceDispatch, "GetAttribute", "language");
			if (languageAttributeVariant == null)
				return (false);
			String languageCode = languageAttributeVariant.toString();
			if (languageCode == null)
				return (false);
			if (languageCode.contains(";"))
				languageCode = languageCode.replaceAll(";[a-zA-Z0-9]*$", "");
			String rfcCode = getCombinedRFCCode(languageCode);
			if (rfcCode.startsWith(" none"))
				return (false);
			return (true);
		} catch (Exception ex)
		{
			log.error("Error in filterVoice method:\n" + ex);
			return (false);
		}
	}

	public String getVoice()
	{
		ActiveXComponent voiceComponent = fTTSComponent.getPropertyAsComponent("voice");
		Variant voiceName = voiceComponent.invoke("getDescription");
		return (voiceName.toString() + "\n\n");
	}

	public String getVoiceIdentifier()
	{
		return (fVoiceIdentifier);
	}

	public boolean setVoice(String voiceIdentifier)
	{
		// In SAPI5 it is necessary at this point to call initialize
		// to prevent some voices from not sounding correctly when one voice change to another
		// It seems that this is not required in MSP and thus commented out
		// initialize();
		if (setSpeechPlatformVoice(voiceIdentifier))
			return (true);
		return (false);
	}

	public boolean setSpeechPlatformVoice(String voiceIdentifier)
	{
		if (voiceIdentifier == null)
			return (false);
		// Replace double underline with space
		voiceIdentifier = voiceIdentifier.replaceAll("__", " ");
		voiceIdentifier = voiceIdentifier.toLowerCase();
		Dispatch ttsDispatch = fTTSComponent.getObject();
		Dispatch voices = Dispatch.call(ttsDispatch, "getVoices").toDispatch();
		Variant itemCount = Dispatch.get(voices, "Count");
		for (int i = 0; i < itemCount.getInt(); i++)
		{
			Dispatch voiceDispatch = Dispatch.call(voices, "Item", new Variant(i)).toDispatch();
			if (filterVoice(voiceDispatch))
			{
				Variant voiceNameVariant = Dispatch.call(voiceDispatch, "GetDescription");
				String voiceName = voiceNameVariant.toString();
				if (voiceName != null)
				{
					if (voiceName.toLowerCase().contains(voiceIdentifier))
					{
						fVoiceIdentifier = voiceName;
						Dispatch.putRef(ttsDispatch, "voice", voiceDispatch);
						return (true);
					}
				}
			}
		}
		return (false);
	}

	public void pause()
	{
		fTTSComponent.invoke("pause");
	}

	public void resume()
	{
		fTTSComponent.invoke("resume");
	}

	public void cancelAll()
	{
		fTTSComponent.invoke("speak", "", (SVSFPurgeBeforeSpeak));
	}

	public void setPunctuation(boolean mode)
	{
		if (mode)
			fSpeakFlags = SVSFNLPSpeakPunc | SVSFlagsAsync;
		else
			fSpeakFlags = SVSFlagsAsync;
	}

	public void close()
	{
		// No method for shutting the voice was found
		ComThread.Release();
		// Small delay is recommended before the system is shutdown
		try
		{
			Thread.sleep(100);
		} catch (InterruptedException ex)
		{
		}
		fTTSComponent = null;
	}

}
