/*
* Copyright (c) Isaac Porat 2011 - 2013 all rights reserved.
* This file is released under the GNU General Public License, version 2, or (at your option) any later version.
* See the file SpeechHub_License.txt for more details.
*/

package sh.dev;

import log.common.CascadingLoggerInterface;
import sh.connections.SHConstants;
import sh.connections.SHCustomeHandlersInterface;
import sh.connections.SHEngineConnectorInterface;
import sh.connections.SHAudioDataHandlerInterface;
import sh.connections.SHEventsListenerInterface;

/**
 * Engine connector implementation template
 * Temporary printed messages are provided for initial development
 * All outputs to STDOUT or STDERR
 * 
 * @author Isaac Porat
 */
public class SHEngineConnectorTemplate implements SHEngineConnectorInterface
{
	private CascadingLoggerInterface log;

	public SHEngineConnectorTemplate(CascadingLoggerInterface logger)
	{
		log = logger.clone(SHEngineConnectorTemplate.class);
	}

	/**
	 * Initialize engine connector 
	 * @return true if successful false if not
	 */
	@Override
	public boolean initialize()
	{
		try
		{
			// Here connect to methods to initialize engine connector 
			log.notify("Initializing");
			return (true); // true if successful
		} catch (Exception ex)
		{
			log.error("Error in initialize method:\n" + ex);
			return (false);
		}
	}

	/**
	 * Close engine connector
	 * @return true if successful false if not
	 */
	@Override
	public boolean close()
	{
		try
		{
			// Here connect to methods to close engine connector and release any resources associated with it
			log.notify("Closing");
			return (true); // true if successful
		} catch (Exception ex)
		{
			log.error("Error in close method:\n" + ex);
			return (false);
		}
	}

	/**
	 * Get engine identifiers
	 * 
	 * @return engineIdentifierArray
	 */
	@Override
	public String[] getEngineIdentifiers()
	{
		try
		{
			log.notify("getEngineIdentifiers method called");
			// Here connect to method to obtain engine identifier list
			String[] engineArray =
			{ SHConstants.ENGINE_IDENTIFIER_ESPEAK, SHConstants.ENGINE_IDENTIFIER_IBMTTS };
			return (engineArray);
		} catch (Exception ex)
		{
			log.error("Error in getEngineIdentifiers method:\n" + ex);
			return (null);
		}
	}

	/**
	 * Select engine based on its identifier
	 * 
	 * @param engineIdentifier Identify the module required
	 * @return true if successful false if not.
	 */
	@Override
	public boolean selectEngine(String engineIdentifier)
	{
		try
		{
			// Here connect to method to select engine based on its identifier
			log.notify("Selecting engine: " + engineIdentifier);
			return (true); // true if successful false if not
		} catch (Exception ex)
		{
			log.error("Error in selectEngine method:\n" + ex);
			return (false);
		}
	}

	/**
	 * Set the synthesized language according to language code as specified by RFC 1766 
	 * @param languageCode
	 * @return
	 */
	@Override
	public boolean selectLanguage(String languageCode)
	{
		try
		{
			// Here connect to method to set engine language
			log.notify("Language: " + languageCode);
			return (true); // true if successful
		} catch (Exception ex)
		{
			log.error("Error in setLanguage method:\n" + ex);
			return (false);
		}
	}

	/**
		 * Get synthesizer voice identifiers
		 * 
		 * @return Voice identifier array.
		 */
	@Override
	public String[] getVoiceIdentifiers()
	{
		try
		{
			log.notify("getVoiceIdentifiers method called");
			// Here connect to method to obtain engine list of voices
			String[] voiceArray =
			{ "Voicename 1", "Voicename 2", "Voicename 3" };
			return (voiceArray);
		} catch (Exception ex)
		{
			log.error("Error in getVoiceIdentifiers method:\n" + ex);
			return (null);
		}
	}

	/**
	 * Select the synthesizer's voice
	 * 
	 * @param voiceIdentifier Identify the voice required
	 * @return true if successful false if not.
	 */
	@Override
	public boolean selectVoice(String voiceIdentifier)
	{
		try
		{
			// If "default" identifier replace with the default identifier for the engine 
			if (voiceIdentifier.equalsIgnoreCase(SHConstants.ENGINE_VOICE_IDENTIFIER_DEFAULT))
				voiceIdentifier = "YourEngineDefaultVoiceIdentifier";
			// Here connect to method to select a voice based on its identifier
			log.notify("Selecting voice: " + voiceIdentifier);
			return (true); // true if successful
		} catch (Exception ex)
		{
			log.error("Error in selectVoice method:\n" + ex);
			return (false);
		}
	}

	/**
	 * Get synthesizer voice variant identifiers
	 * Applies only to Formant type synthesizers
	 * @return variant identifier array.
	 */
	@Override
	public String[] getVariantIdentifiers()
	{
		try
		{
			log.notify("getVariantIdentifiers method called");
			// Here connect to method to obtain engine list of variants
			String[] variantArray =
			{ "Variantname 1", "Variantname 2", "Variantname 3" };
			return (variantArray);
		} catch (Exception ex)
		{
			log.error("Error in getVariantIdentifiers method:\n" + ex);
			return (null);
		}
	}

	/**
	 * Select the synthesizer's voice variant
	* Applies only to Formant type synthesizers
		 * @param varientIdentifier Identify the variant required
	* @return true if successful false if not.
	*/
	@Override
	public boolean selectVariant(String variantIdentifier)
	{
		try
		{
			// Here connect to method to select a variant based on its identifier
			log.notify("Selecting variant: " + variantIdentifier);
			return (true); // true if successful
		} catch (Exception ex)
		{
			log.error("Error in selectVariant method:\n" + ex);
			return (false);
		}
	}

	/**
	* Speak message 
	 * 	  
	 * @param messageIdentifier Identifies message used in server
	 * @param message Message to speak
	 * @param speakSync if true return after completion, if false return immediatly 
	 * @return true if successful false if not
	 * Notes on implementation:
	 * There is no need to implement both speakSync and speakAsync functionality
	 * speakSync is used internally by SH and therefore prefered
	 * Alternatively, speakAsync can be implemented, in which case
	 * The end event must be implemented through the appropriate event listener interface
	 * The property "engine.supports.speak.sync" must be set accordingly for the module stack to work correctly 
	 */
	@Override
	public boolean speak(int MessageIdentifier, String message, boolean speakSync)
	{
		try
		{
			/* 
			* Here connect to method to speak
			* Note that provided that the property in the notes above is specified
			* there is no need to test the speakSync argument when this method is called
			*/
			log.notify("Speaking message: " + message);
			log.notify("speakSync flag is set to: " + speakSync);
			return (true); // true is successful
		} catch (Exception ex)
		{
			log.error("Error in speak method:\n" + ex);
			return (false);
		}
	}

	/**
	 * Pause speech
	 * 
	 * @return true if successful false if not
	  */
	@Override
	public boolean pause()
	{
		try
		{
			/* 
			* Here connect to the engine's pause method if available
			* Note - indicate availability with the engine.supports.pause.resume property
			*/
			log.notify("Pausing speech");
			return (true); // true if successful
		} catch (Exception ex)
		{
			log.error("Error in pause method:\n" + ex);
			return (false);
		}
	}

	/**
	 * Resume speech after a pause
	 */
	@Override
	public boolean resume()
	{
		try
		{
			/* 
			* Here connect to the engine's resume method if available
			* Note - indicate availability with the engine.supports.pause.resume property
			*/
			log.notify("Resuming speech");
			return (true); // true if successful
		} catch (Exception ex)
		{
			log.error("Error in resume method:\n" + ex);
			return (false);
		}
	}

	/**
	* Flushes the synthesizer buffer and cancel its current output.
	* @return true if successful false if not
		*/
	@Override
	public boolean cancel()
	{
		try
		{
			// Here connect to the engine's cancel method
			log.notify("Cancelling speech");
			return (true); // true if successeful false if not
		} catch (Exception ex)
		{
			log.error("Error in cancel method:\n" + ex);
			return (false);
		}
	}

	/**
	 * Set speech volume in the range of 0 - 100.0
	 * 
	 * @param volume Speech volume
	 * 
	 * @return true if successful false if not
	 */
	@Override
	public boolean setVolume(float volume)
	{
		// Range 0 to 100.0
		try
		{
			// Here scale as necessary and connect to the engine's volume method
			log.notify("Volume: " + volume);
			return (true); // true if successful false if not 
		} catch (Exception ex)
		{
			log.error("Error in setVolume method:\n" + ex);
			return (false);
		}
	}

	/**
	 * Set speech rate in the range of -100.0 - 100.0
	 * 
	 * @param rate Speech rate
	 	 * 
	 * @return true if successful false if not
	*/
	@Override
	public boolean setRate(float rate)
	{
		// Range -100.0 to 100.0
		try
		{
			// Here scale as necessary and connect to the engine's rate method
			log.notify("Rate: " + rate);
			return (true); // true if successful false if not
		} catch (Exception ex)
		{
			log.error("Error in setRate method:\n" + ex);
			return (false);
		}
	}

	/**
	* Set pitch in % above and below the base pitch.<br/>
	* 
	* @param pitch Offset  in +-100.0% from base pitch for the voice.
	* 
	* @return true if successful false if not
	*/
	@Override
	public boolean setPitch(float pitch)
	{
		// Range -100.0 to 100.0
		try
		{
			// Here scale as necessary and connect to the engine's pitch method
			log.notify("Pitch: " + pitch);
			return (true); // true if successful false if not
		} catch (Exception ex)
		{
			log.error("Error in setPitch method:\n" + ex);
			return (false);
		}
	}

	/**
	* This method is used to get information about engine capabilities and advanced settings
	* @param key property key
	* @return property value if found null if not
	*/
	@Override
	public String getProperty(String key)
	{
		try
		{
			if (key == null)
				return (null);
			if (key.equals(SHConstants.PROP_ENGINE_SUPPORTS_SPEAK_SYNC))
				return ("true"); // true if engine supports synchronized speech (see note with speak method)
			if (key.equals(SHConstants.PROP_ENGINE_REQUIRES_ISOLATION_RESTART))
				return ("true");
			if (key.equals(SHConstants.PROP_ENGINE_SUPPORTS_PUNCTUATION))
				return ("true"); // true if engine supports punctuation
			if (key.equals(SHConstants.PROP_ENGINE_PUNCTUATION_LEVEL))
			{
				// Here obtain the punctuation level and return as String 
				// Example of "most" below
				String punctuationLevel = "most";
				log.notify("get property  key: " + key);
				log.notify("Value: " + punctuationLevel);
				return (punctuationLevel);
			}
			if (key.equals(SHConstants.PROP_ENGINE_SUPPORTS_PAUSE_RESUME))
				return ("true"); // true if engine supports pause / resume
			if (key.equals(SHConstants.PROP_ENGINE_SUPPORTS_SSML))
				return ("true"); // true if engine supports sSML
			if (key.equals(SHConstants.PROP_ENGINE_SUPPORTS_AUDIO_DATA))
				return ("true"); // true if engine returns audio data
			if (key.equals(SHConstants.PROP_ENGINE_IDENTIFIER_DEFAULT))
			{
				/*
				 * For a single engine connector return the engine identifier
				 * For a multi engines connector return the default engine loaded on startup
				 * This is important as it is used (together with other parameters)
				 * to point to the engine settings directory 
				 */
				String engineIdentifier = SHConstants.ENGINE_IDENTIFIER_SAPI5;
				return (engineIdentifier);
			}
			return (null);
		} catch (Exception ex)
		{
			log.error("Error in getProperty method:\n" + ex);
			return (null);
		}
	}

	/**
	 * 	 This method is used to set more advanced engine options such as punctuation levels, spelling modes etc. 
	 * @param key property key
	 * @param value property value
	 * @return true if successful false if not
	 */
	@Override
	public boolean setProperty(String key, String value)
	{
		try
		{
			if ((key == null) || (value == null))
				return (false);
			if (key.equals(SHConstants.PROP_ENGINE_PUNCTUATION_LEVEL))
			{
				/* 
				* Here connect to method to set engine punctuation levels if available
				* Although module handles this by default this feature is used if
				* a language is not supported yet by the module
					*/
				log.notify("set property key: " + key + " value: " + value);
				return (true);
			}
			if (key.equals(SHConstants.PROP_ENGINE_SSML_MODE))
			{
				/* 
				* Here connect to method to set engine SSML mode if available
				* SSML processing is not supported by the module, only by the engine if available
					*/
				log.notify("set property key: " + key + " value: " + value);
				return (true);
			}
			return (false);
		} catch (Exception ex)
		{
			log.error("Error in setProperty method:\n" + ex);
			return (false);
		}
	}

	/**
	 * Listener for receiving module events
	 * @param eventsListener
	 * @return true if supported by the engine's implementation false if not 
	 */
	@Override
	public boolean setEventsListener(SHEventsListenerInterface eventsListener)
	{
		try
		{
			/*
					* Here provide a method to set the engine's event listener
					* There is no need to implement this method if
					* asynchronized speech method is implemented (see note in the speak method) 
					* If speak asynchronized is implemented Only the end message event is required
					*/
			log.notify("set event listener");
			return (true); // true if successful false if not
		} catch (Exception ex)
		{
			log.error("Error in setEventsListener method:\n" + ex);
			return (false);
		}
	}

	/**
	 * Set handler to optionally bring upstream audio raw data generated by the engine and process
	 * @param audioDataHandler
	 * @return true if supported by the engine's implementation false if not 
	 */
	@Override
	public boolean setAudioDataHandler(SHAudioDataHandlerInterface audioDataHandler)
	{
		try
		{
			/*
			* Here provide a method to set the engine's audio data handler
			* This can be implemented if the engine supports this feature 
			*/
			log.notify("set audio data handler");
			return (true); // true if successful false if not
		} catch (Exception ex)
		{
			log.error("Error in setAudioDataHandler method:\n" + ex);
			return (false);
		}
	}

	/**
	 * Custome handlers for upstream and global communication
	 * @param customeHandlers
	 * @return true if supported by the engine implementation false if not 
		 */
	@Override
	public boolean setCustomeHandlers(SHCustomeHandlersInterface customeHandlers)
	{
		try
		{
			/*
			* Here provide a method to set the engine's custome handlers
			* This can be implemented if the engine supports this feature 
			*/
			log.notify("set custome handlers");
			return (true); // true if successful false if not
		} catch (Exception ex)
		{
			log.error("Error in setCustomeHandlers method:\n" + ex);
			return (false);
		}
	}

	public void checkEngine()
	{
		/*
		* Optionally implemented to ensure engine initialization
		* Called automatically when wrapped with SHWrapper
		*/
	}

	public boolean watchdog(boolean reset)
	{
		return (true); // Watchdog disabled
	}

}
