/*
* Copyright (c) Isaac Porat 2011 - 2013 all rights reserved.
* This file is released under the GNU General Public License, version 2, or (at your option) any later version.
* See the file SpeechHub_License.txt for more details.
*/

package sh.dev.test;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.nio.MappedByteBuffer;
import java.nio.channels.FileChannel;
import java.nio.charset.Charset;

import core.utilities.StackTrace;

import log.common.BasicLogger;
import log.common.CascadingLoggerInterface;
import log.common.SimpleLogger;

import sh.audio.SHWaveListener;
import sh.connections.SHConstants;
import sh.connections.SHEngineConnectorInterface;
import sh.connections.SHEngineInterface;
import sh.connector.multiNative.SHConnectorMultiNativeAdaptor;
import sh.engine.SHEngineFilterAdaptor;
import sh.independent.SHConfigIndependent;

/**
 * Direct multiNative type engine implementation test from the command line 
 * Note - To work, this test must include SHConnectorAudioPlayer in the connector stack 
 * (not engaged by default)
* @author Bill Cox
 *
 */
public class SHNativeEngineCliTest
{

	private static void usage(String format, Object... args)
	{
		BasicLogger.notifyPrintf(format, args);
		BasicLogger.notifyPrintln("Usage: sh-say [parameter]... [text]\n" + "Parameters can be:\n"
				+ "-s speed:  Set the speed of speech.  -100 is slowest, 100 is fastest.\n"
				+ "-p pitch:  Set pitch.  -100 is lowest, 100 is highest.\n" + "-f file:   Speak the text in file.\n"
				+ "-e engine: Switch to the engine, one of: espeak, ibmtts\n"
				+ "-v voice:  Switch to the named voice\n" + "-l:        List supported voices\n"
				+ "-w file:   Write speech output to a file rather than speak it\n"
				+ "--:        Indicate end of options so the next parameter will be taken as text\n");
		System.exit(1);
	}

	public static String readFile(String path)
	{
		try
		{
			FileInputStream stream = new FileInputStream(new File(path));
			try
			{
				FileChannel fc = stream.getChannel();
				MappedByteBuffer bb = fc.map(FileChannel.MapMode.READ_ONLY, 0, fc.size());
				/* Instead of using default, pass in a decoder. */
				return Charset.defaultCharset().decode(bb).toString();
			} finally
			{
				stream.close();
			}
		} catch (IOException e)
		{
			usage("Unable to read file %s\n", path);
		}
		return null; // Dummy return
	}

	public static float stringToFloat(String s)
	{
		try
		{
			float f = Float.valueOf(s.trim()).floatValue();
			return f;
		} catch (NumberFormatException nfe)
		{
			usage("Expected floating point number after flag");
		}
		return 0.0f; // Dummy return;
	}

	public static void main(String[] args)
	{
		float speed = 0.0f;
		float pitch = 0.0f;
		String voiceName = null;
		int xArg = 0;
		char c;
		boolean listVoices = false;
		String outputFile = null;

		String text = "Hello, world!";
		String engineName = SHConstants.ENGINE_IDENTIFIER_ESPEAK;

		while (xArg < args.length && args[xArg].charAt(0) == '-')
		{
			c = args[xArg].charAt(1);
			if (c == '-')
			{
				xArg++;
				break;
			}
			switch (c)
			{
			case 'f':
				if (++xArg == args.length)
				{
					usage("Expected file name after -f parameter\n");
				}
				text = readFile(args[xArg]);
				break;
			case 'l':
				listVoices = true;
				break;
			case 'p':
				if (++xArg == args.length)
				{
					usage("Expected pitch after -p parameter\n");
				}
				pitch = stringToFloat(args[xArg]);
				break;
			case 's':
				if (++xArg == args.length)
				{
					usage("Expected speed after -s parameter\n");
				}
				speed = stringToFloat(args[xArg]);
				break;
			case 'e':
				if (++xArg == args.length)
				{
					usage("Expected engine name after -e parameter\n");
				}
				engineName = args[xArg];
				break;
			case 'v':
				if (++xArg == args.length)
				{
					usage("Expected voice name after -v parameter\n");
				}
				voiceName = args[xArg];
				break;
			case 'w':
				if (++xArg == args.length)
				{
					usage("Expected wav file name after -w parameter\n");
				}
				outputFile = args[xArg];
				break;
			default:
				usage("Unknown option %c\n", c);
			}
			xArg++;
		}
		if (xArg + 1 == args.length)
		{
			text = args[xArg];
		} else if (xArg != args.length)
		{
			usage("Only one non-option parameter allowed.");
		}
		int logLevel = 2;
		CascadingLoggerInterface log = new SimpleLogger(SHNativeEngineCliTest.class, logLevel);

		SHEngineInterface engine;
		try
		{
			// Configure locations - settings and  engines
			SHConfigIndependent.configureLocations(log);

			SHEngineConnectorInterface engineConnector = new SHConnectorMultiNativeAdaptor(log);
			engine = new SHEngineFilterAdaptor(engineConnector, log);
		} catch (Exception ex)
		{
			log.error("Error in main method:\n" + StackTrace.stackToString(ex));
			return;
		}

		engine.selectEngine(engineName);
		engine.initialize();
		if (listVoices)
		{
			String voices[] = engine.getVoiceIdentifiers();
			for (int i = 0; i < voices.length; i++)
			{
				BasicLogger.notifyPrintln(voices[i]);
			}
		} else
		{
			engine.setRate(speed);
			engine.setPitch(pitch);
			if (voiceName != null)
			{
				engine.selectVoice(voiceName);
			}
			SHWaveListener listener = null;
			if (outputFile != null)
			{
				int sampleRate = Integer.parseInt(engine.getProperty(SHConstants.PROP_ENGINE_SAMPLE_RATE));
				listener = new SHWaveListener(outputFile, sampleRate, log);
				engine.setAudioDataHandler(listener);
			}
			engine.speak(0, text, false);
			if (outputFile != null)
			{
				listener.save();
			}
		}
		engine.close();
		System.exit(0);
	}

}
