/*
* Copyright (c) Isaac Porat and Bill Cox 2011 - 2013 all rights reserved.
* This file is released under the GNU General Public License, version 2, or (at your option) any later version.
* See the file SpeechHub_License.txt for more details.
*/

package sh.engine;

import java.io.IOException;

import log.common.CascadingLoggerInterface;
import sh.common.SHConfig;
import sh.connections.SHConstants;
import sh.settings.SHEngineFactoryProperties;
import sh.settings.SHEngineParameters;
import core.utilities.OS;

/**
 * Engine starting command utilities 
 * 
 * @author Isaac Porat
 * @author Bill Cox
 *
 */
public class SHEngineCommand
{
	private CascadingLoggerInterface log;
	private SHEngineParameters fEngineParameters;

	public SHEngineCommand(CascadingLoggerInterface logger)
	{
		log = logger.clone(SHEngineCommand.class);
		fEngineParameters = SHEngineParameters.getInstance(logger);
	}

	/**
	 * Get stdio command array if the engine provides it
	 * @param engineIdentifier
	 * @return commandArray
	 */
	public String getEngineDirectoryPath(String engineIdentifier)
	{
		return (SHConfig.getEnginesBaseDirectory() + "/engine/" + fEngineParameters
				.getEngineDirectory(engineIdentifier));
	}

	/**
	 * Get stdio command array if specified
	 * @param engineIdentifier
	 * @return commandArray
	 */
	public String[] getStartStdioCommand(String engineIdentifier)
	{
		String engineDirectoryPath = getEngineDirectoryPath(engineIdentifier);
		String key = SHConstants.PROP_ENGINE_COMMAND_STDIO_WINDOWS;
		if (OS.isWindows() == false)
			key = SHConstants.PROP_ENGINE_COMMAND_STDIO_LINUX;
		String command = fEngineParameters.getEngineFactoryProperty(engineIdentifier, key);
		if (command == null)
			return (null);
		// Replace EQUALS with = to restore command
		command = command.replaceAll("EQUALS", "=");
		// Split before substituting engine directory path in case it has a space, which it most likely will in Windows.
		String[] cmdArray = command.split(" ");
		for (int i = 0; i < cmdArray.length; i++)
		{
			// Replace engineDirectoryPath with actual path
			String cmd = cmdArray[i];
			cmd = cmd.replaceAll("engineDirectoryPath", engineDirectoryPath);
			cmdArray[i] = cmd;
		}
		return cmdArray;
	}

	/**
	 * Gets socket inet server port if specified 
	 * @param engineIdentifier
	 * @return socket inet server port in String form 
	 * @throws IOException
	 */
	public String getSocketInetServerPort(String engineIdentifier) throws IOException
	{
		SHEngineFactoryProperties engineProperties = new SHEngineFactoryProperties(log);
		return (engineProperties.getProperty(SHConstants.ENGINE_IDENTIFIER_FREETTS,
				SHConstants.PROP_ENGINE_SERVER_SOCKET_PORT));
	}

	/**
	 * Get socket inet command array if specified
	 * @param engineIdentifier
	 * @return commandArray
	 */
	public String[] getStartSocketInetCommand(String engineIdentifier, String serverPort)
	{
		String engineDirectoryPath = getEngineDirectoryPath(engineIdentifier);
		String key = SHConstants.PROP_ENGINE_COMMAND_SOCKET_INET_WINDOWS;
		if (OS.isWindows() == false)
			key = SHConstants.PROP_ENGINE_COMMAND_SOCKET_INET_LINUX;
		String command = fEngineParameters.getEngineFactoryProperty(engineIdentifier, key);
		if (command == null)
			return (null);
		// Add the port argument
		command += " " + serverPort;
		// Replace EQUALS with = to restore command
		command = command.replaceAll("EQUALS", "=");
		// Split before substituting engine directory path in case it has a space, which it most likely will in Windows.
		String[] cmdArray = command.split(" ");
		for (int i = 0; i < cmdArray.length; i++)
		{
			// Replace engineDirectoryPath with actual path
			String cmd = cmdArray[i];
			cmd = cmd.replaceAll("engineDirectoryPath", engineDirectoryPath);
			cmdArray[i] = cmd;
		}
		return cmdArray;
	}

}
