/*
* Copyright (c) Isaac Porat and Bill Cox 2011 - 2013 all rights reserved.
* This file is released under the GNU General Public License, version 2, or (at your option) any later version.
* See the file SpeechHub_License.txt for more details.
*/

package sh.engine;

import hrpc.common.HRTransportInterface;
import hrpc.transport.HRTransportStdioClient;

import java.io.BufferedReader;
import java.io.InputStream;
import java.io.PrintStream;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Timer;
import java.util.TimerTask;

import log.common.CascadingLoggerInterface;
import sh.common.SHConfig;
import sh.connections.SHConstants;
import sh.settings.SHEngineParameters;
import core.utilities.StackTrace;

/**
 * Manage load / unload of tts engine executables
 * @author Isaac Porat
 * @author Bill Cox
 */
public class SHEngineExecLoader
{
	private final static int ENGINE_CLOSING_DELAY = 60000; // ms
	private CascadingLoggerInterface log;
	private Map<String, HRTransportInterface> fTransportMap;
	private String fLoadedEngineIdentifier = "";
	private SHEngineParameters fEngineParameters;
	private SHEngineCommand fEngineCommand;
	private boolean fMultiEngineMode = true;
	private Timer fEngineTimer;
	private EngineClosingTask fEngineClosingTask;
	private boolean fEngineInUse = false;

	public SHEngineExecLoader(boolean multiEngineMode, CascadingLoggerInterface logger)
	{
		log = logger.clone(SHEngineExecLoader.class);
		fMultiEngineMode = multiEngineMode;
		fTransportMap = new HashMap<String, HRTransportInterface>();
		fEngineParameters = SHEngineParameters.getInstance(logger);
		fEngineCommand = new SHEngineCommand(logger);
		if (fMultiEngineMode)
			fEngineTimer = new Timer();
	}

	public BufferedReader getStdin()
	{
		HRTransportInterface transport = fTransportMap.get(fLoadedEngineIdentifier);
		if (transport != null)
			return (transport.getStdin());
		return (null);
	}

	public InputStream getStdinRaw()
	{
		HRTransportInterface transport = fTransportMap.get(fLoadedEngineIdentifier);
		if (transport != null)
			return (transport.getStdinRaw());
		return (null);
	}

	public PrintStream getStdout()
	{
		HRTransportInterface transport = fTransportMap.get(fLoadedEngineIdentifier);
		if (transport != null)
			return (transport.getStdout());
		return (null);
	}

	/**
	 * Load engine
	 * Return true if successful false if not
	 */
	public boolean loadEngine(String engineIdentifier)
	{
		synchronized (fTransportMap)
		{
			try
			{
				if (engineIdentifier == fLoadedEngineIdentifier)
					return (true);
				// Check if transport is already in the map and if so no need to loadit
				if (fTransportMap.get(engineIdentifier) != null)
				{
					fLoadedEngineIdentifier = engineIdentifier;
					return (true);
				}
				notifyEngineInUse();
				String libPath = fEngineCommand.getEngineDirectoryPath(engineIdentifier);
				/**
				* Check if there is a pre-defined command to start this engine.
				*/
				String[] cmdArray = fEngineCommand.getStartStdioCommand(engineIdentifier);
				if (cmdArray == null)
				{
					/*
					 * I did not consider 64 bits for now
					 * and did not investigate how to detect if required
					 * Is this correct?
					 */
					String lib = null;
					/*
					 * Note selection based on JVM architecture not OS itself
					 * Not aware of direct way of finding OS architecture in Java
					 * Applies at this stage only to Linux
					 */
					String jvmArch = core.utilities.OS.getJVMArchitecture();
					if (core.utilities.OS.isWindows())
						lib = "stdio-w";
					else
					// Assume at this stage Linux
					{
						if (jvmArch.equals("64"))
							lib = "stdio-l64";
						else
							lib = "stdio-l";
					}
					// Here consider testing for other implemented OS if required

					libPath += "/" + lib;
					String synthdataPath = SHConfig.getEnginesBaseDirectory() + "/engine/"
							+ fEngineParameters.getEngineDirectory(engineIdentifier) + "/synthdata";
					String property = fEngineParameters.getEngineFactoryProperty(engineIdentifier,
							SHConstants.PROP_ENGINE_FORCE_SLASH_PATHSEP);
					boolean forceSlashPathsep = property != null && property.equals("true");
					if (core.utilities.OS.isWindows() && !forceSlashPathsep)
						synthdataPath = synthdataPath.replace("/", "\\");
					String serverPath = libPath + "/sh_" + engineIdentifier;
					cmdArray = new String[]
					{ serverPath, synthdataPath };
				}
				String format = "UTF-8";
				String formatProperty = fEngineParameters.getEngineFactoryProperty(engineIdentifier,
						SHConstants.PROP_ENGINE_TEXT_FORMAT);
				if (formatProperty != null)
					format = formatProperty;
				HRTransportInterface transport = new HRTransportStdioClient(cmdArray, libPath, engineIdentifier
						+ " engine", format, log);
				transport.open();
				fTransportMap.put(engineIdentifier, transport);
				fLoadedEngineIdentifier = engineIdentifier;
				return true;
			} catch (Exception ex)
			{
				log.error("Error in loadEngine method:\n" + StackTrace.stackToString(ex));
				return false;
			}
		}
	}

	/**
	 * Unload and terminate active tts engine process
	 * @return true if successful false if not
	 */
	public boolean unloadActiveEngine()
	{
		synchronized (fTransportMap)
		{
			boolean status = true;
			try
			{
				HRTransportInterface transport = fTransportMap.get(fLoadedEngineIdentifier);
				if (transport != null)
				{
					transport.getStdout().print("quit\n");
					Thread.sleep(250);
					transport.close();
				}
				fTransportMap.remove(fLoadedEngineIdentifier);
				fLoadedEngineIdentifier = "";
			} catch (Exception ex)
			{
				log.error("Error in unloadActiveEngine method: Can't close " + fLoadedEngineIdentifier + "\n"
						+ StackTrace.stackToString(ex));
				status = false;
			}
			return (status);
		}
	}

	/**
	 * Unload and terminate non selected tts engine processes
	 */
	public void unloadUnselectedEngines()
	{
		synchronized (fTransportMap)
		{
			String engineIdentifier = "none";
			for (Iterator<Map.Entry<String, HRTransportInterface>> iterator = fTransportMap.entrySet().iterator(); iterator
					.hasNext();)
			{
				try
				{
					Map.Entry<String, HRTransportInterface> entry = iterator.next();
					engineIdentifier = entry.getKey();
					if (engineIdentifier.equals(fLoadedEngineIdentifier) == false) // Not selected engine close and remove
					{
						String engineKeepAlive = fEngineParameters.getEngineFactoryProperty(engineIdentifier,
								SHConstants.PROP_ENGINE_KEEP_ALIVE);
						if ((engineKeepAlive != null) && (engineKeepAlive.equals("true")))
							continue; // Do not close engine
						HRTransportInterface transport = entry.getValue();
						if (transport != null)
						{
							transport.getStdout().print("quit\n");
							Thread.sleep(250);
							transport.close();
						}
						iterator.remove();
					}
				} catch (Exception ex)
				{
					log.error("Error on client socket close: Can't close " + engineIdentifier + "\n"
							+ StackTrace.stackToString(ex));
				}
			}
		}
	}

	private void notifyEngineInUse()
	{
		if ((fMultiEngineMode) && (fEngineClosingTask == null))
		{
			fEngineClosingTask = new EngineClosingTask();
			fEngineTimer.schedule(fEngineClosingTask, ENGINE_CLOSING_DELAY);
		}
		fEngineInUse = true;
	}

	private class EngineClosingTask extends TimerTask
	{

		@Override
		public void run()
		{
			try
			{
				/*
				 * The following arrangement ensures that new TimerTask is only created at the end of timeout
				 * but engine can notify in use at any time
				 * Effective timeout not less than 1x and not greater of 2x nominal timeout 
				 */
				if (fEngineInUse)
				{
					if (fEngineClosingTask != null)
						fEngineClosingTask.cancel();
					fEngineClosingTask = new EngineClosingTask();
					fEngineTimer.schedule(fEngineClosingTask, ENGINE_CLOSING_DELAY);
					fEngineInUse = false;
					return;
				}
				unloadUnselectedEngines();
				if (fEngineClosingTask != null)
					fEngineClosingTask.cancel();
				fEngineClosingTask = null;
			} catch (Exception ex)
			{
				log.error("Error in run method:\n" + StackTrace.stackToString(ex));
				if (fEngineClosingTask != null)
					fEngineClosingTask.cancel();
				fEngineClosingTask = null;
			}
		}

	}

}
