/*
* Copyright (c) Isaac Porat 2011 - 2013 all rights reserved.
* This file is released under the GNU General Public License, version 2, or (at your option) any later version.
* See the file SpeechHub_License.txt for more details.
*/

package sh.engine;

import java.util.Timer;
import java.util.TimerTask;

import log.common.CascadingLoggerInterface;
import sh.connections.SHAudioDataHandlerInterface;
import sh.connections.SHConstants;
import sh.connections.SHCustomeHandlersInterface;
import sh.connections.SHEngineConnectorInterface;
import sh.connections.SHEngineInterface;
import sh.connections.SHEventsListenerInterface;
import sh.settings.SHEngineParameters;
import core.utilities.StackTrace;

/**
 * Engine filter adapter
 * Provides support to engine connector;
 * Additional getters, parameters initialization support 
 * duplicate parameters filtering and engine life cycle managment
 *  
 * @author Isaac Porat
 */
public class SHEngineFilterAdaptor implements SHEngineInterface
{
	private CascadingLoggerInterface log;
	private final static int ENGINE_CLOSING_DELAY = 60000; // ms
	private SHEngineConnectorInterface fEngine;
	private int fUpdateIdentifier = 0;
	private String fSelectedEngineIdentifier = SHConstants.ENGINE_IDENTIFIER_DEFAULT;
	private String fSelectedLanguageCode = SHConstants.ENGINE_LANGUAGE_CODE_DEFAULT;
	private String fSelectedVoiceIdentifier = SHConstants.ENGINE_VOICE_IDENTIFIER_DEFAULT;
	private String fSelectedVariantIdentifier = SHConstants.ENGINE_VARIANT_IDENTIFIER_DEFAULT;
	private float fSetVolume = SHConstants.ENGINE_VOLUME_DEFAULT;
	private float fSetRate = SHConstants.ENGINE_RATE_DEFAULT;
	private float fSetPitch = SHConstants.ENGINE_PITCH_DEFAULT;
	private String fSetPropPunctuationLevel = SHConstants.ENGINE_PUNCTUATION_LEVEL_NONE;
	private String fSetPropSSMLMode = "false";
	private boolean fSelectEngineRequired = true;
	private boolean fSetLanguageRequired = true;
	private boolean fSelectVoiceRequired = true;
	private boolean fSelectVariantRequired = true;
	private boolean fSetVolumeRequired = true;
	private boolean fSetRateRequired = true;
	private boolean fSetPitchRequired = true;
	private boolean fSetPropPunctuationLevelRequired = true;
	private boolean fSetPropSSMLModeRequired = true;

	private boolean fParametersInitialized = false;
	private SHEngineParameters fEngineParameters;
	private Timer fEngineTimer;
	private EngineClosingTask fEngineClosingTask;
	private boolean fEngineInUse = false;

	public SHEngineFilterAdaptor(SHEngineConnectorInterface engine, CascadingLoggerInterface logger)
	{
		log = logger.clone(SHEngineFilterAdaptor.class);
		fEngine = engine;
		fEngineParameters = SHEngineParameters.getInstance(logger);
		fEngineTimer = new Timer();
	}

	@Override
	public synchronized boolean initialize()
	{
		boolean status = fEngine.initialize();
		if (status)
		{
			fSelectEngineRequired = true;
			fSetLanguageRequired = true;
			fSelectVoiceRequired = true;
			fSelectVariantRequired = true;
			fSetVolumeRequired = true;
			fSetRateRequired = true;
			fSetPitchRequired = true;
			fSetPropPunctuationLevelRequired = true;
			fSetPropSSMLModeRequired = true;
			fParametersInitialized = false;
		}
		return (status);
	}

	private void initializeParameters()
	{
		/* 
		*Best effor to initialize, errors if any reported by engine connector
			*/
		selectEngine(fSelectedEngineIdentifier);
		selectLanguage(fSelectedLanguageCode);
		selectVoice(fSelectedVoiceIdentifier);
		setVolume(fSetVolume);
		setRate(fSetRate);
		setPitch(fSetPitch);
		// Initialize only once	
		fParametersInitialized = true;
	}

	@Override
	public synchronized boolean close()
	{
		fSelectEngineRequired = true;
		return (fEngine.close());
	}

	@Override
	public String[] getEngineIdentifiers()
	{
		/*
		 * Check engine is discouraged here as most connector can return the list of engines
		 * without having to initialize any engine 
		 */
		return (fEngine.getEngineIdentifiers());
	}

	@Override
	public boolean selectEngine(String engineIdentifier)
	{
		if ((fSelectEngineRequired == false) && (engineIdentifier.equals(fSelectedEngineIdentifier)))
			return (true);
		/*
		 * Connector of single engine do not require checkEngine call
		 * 		This is enefficient as the engine might be initialized unnecessarily (waste of resources)
		 * Therefore connectors with multi engines need to ensure checkEngine functionality in the connector itself
		 * note also notifyEngineInUse method called directly  
		 */
		notifyEngineInUse();
		boolean status = fEngine.selectEngine(engineIdentifier);
		if (status)
		{
			fSelectedEngineIdentifier = engineIdentifier;
			fSelectEngineRequired = false;
			fSetLanguageRequired = true;
			fSelectVoiceRequired = true;
			fSelectVariantRequired = true;
			fSetVolumeRequired = true;
			fSetRateRequired = true;
			fSetPitchRequired = true;
			fSetPropPunctuationLevelRequired = true;
			fSetPropSSMLModeRequired = true;
			fParametersInitialized = false;
		}
		return (status);
	}

	@Override
	public String getSelectedEngineIdentifier()
	{
		checkEngine();
		return (fSelectedEngineIdentifier);
	}

	@Override
	public String[] getVoiceIdentifiers()
	{
		checkEngine();
		return (fEngine.getVoiceIdentifiers());
	}

	@Override
	public boolean selectVoice(String voiceIdentifier)
	{
		if ((fSelectVoiceRequired == false) && (voiceIdentifier.equals(fSelectedVoiceIdentifier)))
			return (true);
		checkEngine();
		boolean status = fEngine.selectVoice(voiceIdentifier);
		if (status)
		{
			fSelectedVoiceIdentifier = voiceIdentifier;
			fSelectVoiceRequired = false;
		}
		return (status);
	}

	@Override
	public String getSelectedVoiceIdentifier()
	{
		checkEngine();
		return (fSelectedVoiceIdentifier);
	}

	@Override
	public String[] getVariantIdentifiers()
	{
		checkEngine();
		return (fEngine.getVariantIdentifiers());
	}

	@Override
	public boolean selectVariant(String variantIdentifier)
	{
		if ((fSelectVariantRequired == false) && (variantIdentifier.equals(fSelectedVariantIdentifier)))
			return (true);
		checkEngine();
		boolean status = fEngine.selectVariant(variantIdentifier);
		if (status)
		{
			fSelectedVariantIdentifier = variantIdentifier;
			fSelectVariantRequired = false;
		}
		return (status);
	}

	@Override
	public String getSelectedVariantIdentifier()
	{
		checkEngine();
		return (fSelectedVariantIdentifier);
	}

	@Override
	public boolean speak(int messageIdentifier, String message, boolean speakSync)
	{
		checkEngine();
		/* 
		*The following arrangement ensures that parameters initialization is done 
		*using the same thread (some engines require this)
		*If some parameters initialized already, they are simply set again
		*/
		if (fParametersInitialized == false)
			initializeParameters();
		return (fEngine.speak(messageIdentifier, message, speakSync));
	}

	public boolean pause()
	{
		checkEngine();
		return (fEngine.pause());
	}

	public boolean resume()
	{
		checkEngine();
		return (fEngine.resume());
	}

	@Override
	public boolean cancel()
	{
		checkEngine();
		return (fEngine.cancel());
	}

	@Override
	public String getSelectedLanguageCode()
	{
		return (fSelectedLanguageCode);
	}

	@Override
	public boolean selectLanguage(String languageCode)
	{
		if ((fSetLanguageRequired == false) && (languageCode.equals(fSelectedLanguageCode)))
			return (true);
		/*
		 * Language selection is used for
		 * selecting punctuations and replacements tables
		 * at the module level above this class
		 * It has no effect on the list of voice identifiers returned 
		 * which always contain all languages  
		 * Set language also on some engines determine the default voice selected
		 */
		checkEngine();
		boolean status = fEngine.selectLanguage(languageCode);
		if (status)
		{
			fSelectedLanguageCode = languageCode;
			fSetLanguageRequired = false;
		}
		return (status);
	}

	@Override
	public boolean setVolume(float volume)
	{
		if ((fSetVolumeRequired == false) && (volume == fSetVolume))
			return (true);
		// Range 0 to 100.0
		if (volume > SHConstants.ENGINE_VOLUME_MAX)
			volume = SHConstants.ENGINE_VOLUME_MAX;
		else if (volume < SHConstants.ENGINE_VOLUME_MIN)
			volume = SHConstants.ENGINE_VOLUME_MIN;
		checkEngine();
		boolean status = fEngine.setVolume(volume);
		if (status)
		{
			fSetVolume = volume;
			fSetVolumeRequired = false;
		}
		return (status);
	}

	@Override
	public float getVolume()
	{
		return (fSetVolume);
	}

	@Override
	public boolean setRate(float rate)
	{
		if ((fSetRateRequired == false) && (rate == fSetRate))
			return (true);
		// Range -100.0 to 100.0
		if (rate > SHConstants.ENGINE_RATE_MAX)
			rate = SHConstants.ENGINE_RATE_MAX;
		else if (rate < SHConstants.ENGINE_RATE_MIN)
			rate = SHConstants.ENGINE_RATE_MIN;
		checkEngine();
		boolean status = fEngine.setRate(rate);
		if (status)
		{
			fSetRate = rate;
			fSetRateRequired = false;
		}
		return (status);
	}

	@Override
	public float getRate()
	{
		return (fSetRate);
	}

	@Override
	public boolean setPitch(float pitch)
	{
		if ((fSetPitchRequired == false) && (pitch == fSetPitch))
			return (true);
		// Range -100.0 to 100.0 
		if (pitch > SHConstants.ENGINE_PITCH_MAX)
			pitch = SHConstants.ENGINE_PITCH_MAX;
		else if (pitch < SHConstants.ENGINE_PITCH_MIN)
			pitch = SHConstants.ENGINE_PITCH_MIN;
		checkEngine();
		boolean status = fEngine.setPitch(pitch);
		if (status)
		{
			fSetPitch = pitch;
			fSetPitchRequired = false;
		}
		return (status);
	}

	@Override
	public float getPitch()
	{
		return (fSetPitch);
	}

	@Override
	public boolean setProperty(String key, String value)
	{
		if ((key == null) || (value == null))
			return (false);
		checkEngine();
		if (key.equals(SHConstants.PROP_ENGINE_PUNCTUATION_LEVEL))
		{
			if ((fSetPropPunctuationLevelRequired == false) && (value.equals(fSetPropPunctuationLevel)))
				return (true);
			boolean status = fEngine.setProperty(key, value);
			if (status)
			{
				fSetPropPunctuationLevel = value;
				fSetPropPunctuationLevelRequired = false;
				return (status);
			}
		}
		if (key.equals(SHConstants.PROP_ENGINE_SSML_MODE))
		{
			if ((fSetPropSSMLModeRequired == false) && (value.equals(fSetPropSSMLMode)))
				return (true);
			boolean status = fEngine.setProperty(key, value);
			if (status)
			{
				fSetPropSSMLMode = value;
				fSetPropSSMLModeRequired = false;
				return (status);
			}
		}
		return (fEngine.setProperty(key, value));
	}

	@Override
	public String getProperty(String key)
	{
		/*
		 * Many properties do not require engine check, therefore check engine is implemented by
		 * the connector if necessary
		* note also notifyEngineInUse method called directly  
		 */
		notifyEngineInUse();
		return (fEngine.getProperty(key));
	}

	public boolean setEventsListener(SHEventsListenerInterface eventsListener)
	{
		return (fEngine.setEventsListener(eventsListener));
	}

	@Override
	public boolean setAudioDataHandler(SHAudioDataHandlerInterface audioDataHandler)
	{
		return (fEngine.setAudioDataHandler(audioDataHandler));
	}

	@Override
	public boolean setCustomeHandlers(SHCustomeHandlersInterface customeHandlers)
	{
		return (fEngine.setCustomeHandlers(customeHandlers));
	}

	@Override
	public boolean setUpdateIdentifier(int updateIdentifier)
	{
		fUpdateIdentifier = updateIdentifier;
		return (true);
	}

	@Override
	public boolean isUpdateAccepted(int updateIdentifier)
	{
		if (updateIdentifier == 0) // Insert update
			return (true);
		if (updateIdentifier == -1) // Reset update
		{
			fUpdateIdentifier = 0; // Ensure that after this update the engine updates again with the last before this update identifier
			return (true);
		}
		return (updateIdentifier > fUpdateIdentifier); // Standard update
	}

	private void checkEngine()
	{
		notifyEngineInUse();
		if (fSelectEngineRequired)
		{
			selectEngine(fSelectedEngineIdentifier);
			return;
		}
		fEngine.checkEngine();
	}

	private void notifyEngineInUse()
	{
		if (fEngineClosingTask == null)
		{
			fEngineClosingTask = new EngineClosingTask();
			fEngineTimer.schedule(fEngineClosingTask, ENGINE_CLOSING_DELAY);
		}
		fEngineInUse = true;
	}

	private class EngineClosingTask extends TimerTask
	{

		@Override
		public void run()
		{
			try
			{
				/*
				 * The following arrangement ensures that new TimerTask is only created at the end of timeout
				 * but engine can notify in use at any time
				 * Effective timeout not less than 1x and not greater of 2x nominal timeout 
				 */
				if (fEngineInUse)
				{
					if (fEngineClosingTask != null)
						fEngineClosingTask.cancel();
					fEngineClosingTask = new EngineClosingTask();
					fEngineTimer.schedule(fEngineClosingTask, ENGINE_CLOSING_DELAY);
					fEngineInUse = false;
					return;
				}
				String engineIdentifier = fSelectedEngineIdentifier;
				if (engineIdentifier.equals(SHConstants.ENGINE_IDENTIFIER_DEFAULT))
					engineIdentifier = getProperty(SHConstants.PROP_ENGINE_IDENTIFIER_DEFAULT);
				String engineKeepAlive = fEngineParameters.getEngineFactoryProperty(engineIdentifier,
						SHConstants.PROP_ENGINE_KEEP_ALIVE);
				if ((engineKeepAlive == null) || (engineKeepAlive.equals("false")))
					close();
				if (fEngineClosingTask != null)
					fEngineClosingTask.cancel();
				fEngineClosingTask = null;
			} catch (Exception ex)
			{
				log.error("Error in run method:\n" + StackTrace.stackToString(ex));
				if (fEngineClosingTask != null)
					fEngineClosingTask.cancel();
				fEngineClosingTask = null;
			}
		}

	}

}
