/*
* Copyright (c) Isaac Porat 2011 - 2013 all rights reserved.
* This file is released under the GNU General Public License, version 2, or (at your option) any later version.
* See the file SpeechHub_License.txt for more details.
*/

package sh.engine;

import java.util.Timer;
import java.util.TimerTask;

import core.utilities.StackTrace;

import log.common.CascadingLoggerInterface;

import sh.connections.SHClientEngineVoiceSelectorHandlerInterface;
import sh.connections.SHWatchdogListenerInterface;
import sh.settings.SHEngineParameters;

/**
 * Watchdog is triggered by a false return from a speak method
 * The engine is closed, selected again and The dispatch thread is restarted 
 * If the speak method fails three timesin succession , the engine is disabled for the duration of the session 
 * Another engine is selected and the thread is restarted
 * 
 * @author Isaac Porat
 *
 */
public class SHEngineWatchdog
{
	private CascadingLoggerInterface log;
	private final static int DELAY = 10; // ms
	private SHWatchdogListenerInterface fModule;
	private Timer fWatchdogTimer;
	private MonitoringTask fMonitoringTask;
	private boolean fMonitoringTaskRunning = false;
	private int fEngineFailCounter = 0;
	@SuppressWarnings("unused")
	private Thread fDispatchThread;
	private String fCurrentEngineIdentifier;
private String fCurrentVoiceIdentifier;
	private String fSelectedEngineIdentifier = "";
	private String fErrorMessage = "";

	public SHEngineWatchdog(SHWatchdogListenerInterface module, CascadingLoggerInterface logger)
	{
		log = logger.clone(SHEngineWatchdog.class);
		fModule = module;
		fWatchdogTimer = new Timer();
	}

	public void resetEngineFailCounter()
	{
		fEngineFailCounter = 0;
	}

	public void start(String engineIdentifier, String voiceIdentifier, String errorMessage, Thread dispatchThread)
	{
		fCurrentEngineIdentifier = engineIdentifier;
		fCurrentVoiceIdentifier = voiceIdentifier;
		fErrorMessage = errorMessage;
		fDispatchThread = dispatchThread;
		if (fMonitoringTaskRunning)
			return;
		if (fMonitoringTask != null)
			fMonitoringTask.cancel();
		fMonitoringTask = new MonitoringTask();
		fWatchdogTimer.schedule(fMonitoringTask, DELAY);
	}

	private class MonitoringTask extends TimerTask
	{

		@Override
		public void run()
		{
			try
			{
				fMonitoringTaskRunning = true;
				/*
				 * Thread passed for future possible processing based on thread state
				 Enum<Thread.State> threadState = fDispatchThread.getState();
				log.notify(fDispatchThread.getName() + " is " + threadState.name());
				*/
				if (fCurrentEngineIdentifier == null)
					fCurrentEngineIdentifier = "";
				// Reset failure counter on engine change
				if (fCurrentEngineIdentifier != fSelectedEngineIdentifier)
				{
					fEngineFailCounter = 0;
					fSelectedEngineIdentifier = fCurrentEngineIdentifier;
				}
				if (fEngineFailCounter < 3)
				{
					fEngineFailCounter++;
					// Try to restart engine
					log.error("Watchdog detected engine: " + fSelectedEngineIdentifier + " voice: " + fCurrentVoiceIdentifier + " crash, due to "
							+ fErrorMessage + "\n" + "Restarting engine");
					fModule.interruptThread();
					fModule.close();
					Thread.sleep(50); // Allow engine to unload
					fModule.initialize();
					fModule.selectEngineDirectly(fCurrentEngineIdentifier);
					fModule.startThread();
					fMonitoringTaskRunning = false;
					return;
				}
				// More than 3 attempts try to select  available default engine
				log.error("Watchdog detected engine: " + fSelectedEngineIdentifier + " voice: " + fCurrentVoiceIdentifier + " more than 3 crashes, trying to select another engine");
				String clientIdentifier = fModule.getClientIdentifier();
				if (clientIdentifier == null)
				{
					log.error("Error in run method, clientIdentifier null");
					fMonitoringTaskRunning = false;
					return;
				}
				SHClientEngineVoiceSelectorHandlerInterface selector = fModule.getClientEngineVoiceSelectorHandler();
				if (selector == null)
				{
					log.error("Error in run method: selector null for clientIdentifier " + clientIdentifier);
					fMonitoringTaskRunning = false;
					return;
				}
				SHEngineParameters engineParameters = SHEngineParameters.getInstance(log);
				engineParameters.disableEngine(fSelectedEngineIdentifier);
				String engineIdentifier = engineParameters.getAvailableDefaultEngineIdentifier();
				String voiceIdentifier = engineParameters.getAvailableDefaultVoiceIdentifier();
				log.notify("Selecting engine: " + engineIdentifier + " voice: " + voiceIdentifier);
				if (selector.selectClientEngineVoice(clientIdentifier, engineIdentifier, voiceIdentifier) == false)
				{
					// Error logged by method
					// Here consider generating error for client - there is no mechanism for this yet
					fMonitoringTaskRunning = false;
					return;
				}
				// Start the thread to release any methods blocked (indirectly through notifyAll)
				fModule.interruptThread();
				fModule.startThread();
				// Reset failure counter
				fEngineFailCounter = 0;
				fMonitoringTaskRunning = false;
			} catch (Exception ex)
			{
				log.error("Error in run method:\n" + StackTrace.stackToString(ex));
				fMonitoringTaskRunning = false;
			}
		}

	}

}
