/*
* Copyright (c) Isaac Porat 2011 - 2013 all rights reserved.
* This file is released under the GNU General Public License, version 2, or (at your option) any later version.
* See the file SpeechHub_License.txt for more details.
*/

package sh.format;

import java.text.NumberFormat;
import java.text.ParseException;
import java.util.Currency;
import java.util.Locale;
import java.util.regex.Pattern;

import core.utilities.RegexpMatchGroupInterface;
import core.utilities.RegexpUtilities;

import log.common.CascadingLoggerInterface;

/**
 * Currency formatting
 * Implemented so far for English only and includes USD, EUR and GBP
 * 
 * @author Isaac Porat
 *
 */
public class SHCurrencyFormat
{
	private CascadingLoggerInterface log;
	// Currency code ISO 4217
	// Currency symbols either char if available or unicode representation if above decimal 127
	private static final String CURRANCY_USD_PATTERN = "(\\$[\\d\\.\\,]*[\\d])";
	private static final String CURRANCY_EUR_PATTERN = "(\u20ac[\\d\\.\\,]*[\\d])";
	private static final String CURRANCY_GBP_PATTERN = "([\\d\\.\\,]*[\\d])";
	private Pattern fCurrencyUSDPattern;
	private Pattern fCurrencyEURPattern;
	private Pattern fCurrencyGBPPattern;

	public SHCurrencyFormat(CascadingLoggerInterface logger)
	{
		log = logger.clone(SHCurrencyFormat.class);
		fCurrencyUSDPattern = Pattern.compile(CURRANCY_USD_PATTERN);
		fCurrencyEURPattern = Pattern.compile(CURRANCY_EUR_PATTERN);
		fCurrencyGBPPattern = Pattern.compile(CURRANCY_GBP_PATTERN);
	}

	public String process(String string, String languageCode)
	{
		try
		{
			if (languageCode.startsWith("en") == false) // Not handled at present
				return (string);
			if (string.length() <= 1)
				return (string);
			String stringFormatted = null;
			if ((string.contains("$")) || (string.contains("USD")))
			{
				stringFormatted = string.replaceAll("USD", "$");
				stringFormatted = process(stringFormatted, fCurrencyUSDPattern, "$", "USD", "dollar", "dollars",
						"cent", "cents");
			}
			if (stringFormatted == null) // Formatting failed
				stringFormatted = string;
			if ((string.contains("\u20ac")) || (string.contains("EUR")))
			{
				stringFormatted = stringFormatted.replaceAll("EUR", "\u20ac");
				stringFormatted = process(stringFormatted, fCurrencyEURPattern, "\u20ac", "EUR", "euro", "euros",
						"cent", "cents");
			}
			if (stringFormatted == null) // Formatting failed
				stringFormatted = string;
			if ((string.contains("")) || (string.contains("GBP")))
			{
				stringFormatted = stringFormatted.replaceAll("GBP", "");
				stringFormatted = process(stringFormatted, fCurrencyGBPPattern, "", "GBP", "pound", "pounds", "pence",
						"pence");
			}
			if (stringFormatted == null) // Formatting failed
				stringFormatted = string;
			return (stringFormatted);
		} catch (Exception ex)
		{
			log.error("Error in process method:\n" + ex);
			return (string);
		}
	}

	private String process(String string, Pattern pattern, String currencySymbol, String currencyCode,
			String majorSingle, String majorPlural, String minorSingle, String minorPlural)
	{
		try
		{
			if (RegexpUtilities.contains(string, pattern) == false)
				return (string);
			StringBuilder stringBuilder = new StringBuilder(string);
			RegexpMatchGroupInterface[] groups = RegexpUtilities.findGroupMatches(stringBuilder, pattern, 1);
			// Start from the end to ensure correct replacement
			for (int n = groups.length - 1; n >= 0; n--)
			{
				RegexpMatchGroupInterface group = groups[n];
				String groupString = group.getGroup();
				// Format currency			
				String currencySpeech = getSpeechFormatCommonEnglish(groupString, currencySymbol, currencyCode,
						majorSingle, majorPlural, minorSingle, minorPlural);
				if (currencySpeech == null) // Unable to format
					currencySpeech = getSpeechFormatUnmodifiedEnglish(groupString, currencySymbol, currencyCode,
							majorSingle);
				if (currencySpeech != null)
					groupString = currencySpeech;
				stringBuilder.replace(group.getGroupStart(), group.getGroupEnd(), groupString);
			}
			return (stringBuilder.toString());
		} catch (Exception ex)
		{
			log.error("Error in process method:\n" + ex);
			return (null);
		}
	}

	/**
	 * 
	 * @param currencyFormat
	 * This method is unused as it was easier to format for speech from scratch
	 * Method retained as a development aid and future use
	 * The number.format() method is useful to determine correct format for various currency
	 * The number.parse() method  requires care as it accepts symbols for the correct locale and currencyCode for different currency for the same locale
	 */
	@SuppressWarnings("unused")
	private void parse(String currencyFormat)
	{
		Locale locale = Locale.US;
		//Locale locale = Locale.getDefault();		
		NumberFormat numberFormat = NumberFormat.getCurrencyInstance(locale);
		Currency newCurrency = Currency.getInstance("USD");

		numberFormat.setCurrency(newCurrency);
		try
		{
			//log.notify(numberFormat.format(12.99));
			Number number = numberFormat.parse(currencyFormat);
			if (number instanceof Long)
			{
				log.notify(number.toString());
			} else
			// double
			{
				log.notify(number.toString());
			}
		} catch (ParseException ex)
		{
			log.notify(ex.toString());
		}
	}

	private String getSpeechFormatUnmodifiedEnglish(String currencyFormat, String currencySymbol, String currencyCode,
			String majorSingle)
	{
		try
		{
			String currencySpeech = currencyFormat.replaceAll("\\" + currencySymbol, "");
			currencySpeech = currencySpeech.replaceAll(currencyCode, "");
			currencySpeech = currencySpeech.replaceAll("\\,", "");
			currencySpeech = currencySpeech.replaceAll("[\\.]+$", "");
			currencySpeech = majorSingle + " " + currencySpeech;
			return (currencySpeech);
		} catch (Exception ex)
		{
			log.notify("Error in getSpeechFormatUnmodifiedEnglish method:\n" + ex);
			return (null);
		}
	}

	private String getSpeechFormatCommonEnglish(String currencyFormat, String currencySymbol, String currencyCode,
			String majorSingle, String majorPlural, String minorSingle, String minorPlural)
	{
		try
		{
			String currencySpeech = currencyFormat.replaceAll("\\" + currencySymbol, "");
			currencySpeech = currencySpeech.replaceAll("\\,", "");
			if (currencySpeech.contains("."))
			{
				String[] parts = currencySpeech.split("\\.");
				if (parts.length > 2) // Unknown number format
					return (null);
				if (parts[1].length() > 2) // wrong format
					return (null);
				if (parts[1].length() == 1)
					parts[1] += "0";
				String major = majorPlural;
				String minor = minorPlural;
				parts[0] = parts[0].replaceFirst("^0", "");
				if (parts[0].equals("1"))
					major = majorSingle;
				parts[1] = parts[1].replaceFirst("^0", "");
				if ((parts[1].equals("1")) || (parts[1].equals("0")))
					minor = minorSingle;
				return (parts[0] + " " + major + " " + parts[1] + " " + minor);
			} else
			{
				String major = majorPlural;
				currencySpeech = currencySpeech.replaceFirst("^0", "");
				if (currencySpeech.equals("1"))
					major = majorSingle;
				return (currencySpeech + " " + major);
			}
		} catch (Exception ex)
		{
			log.error("Error in getSpeechFormatCommonEnglish method:\n" + ex);
			return (null);
		}
	}

}
