/*
* Copyright (c) Isaac Porat 2011 - 2013 all rights reserved.
* This file is released under the GNU General Public License, version 2, or (at your option) any later version.
* See the file SpeechHub_License.txt for more details.
*/

package sh.format;

import java.util.regex.Pattern;

import log.common.CascadingLoggerInterface;

import core.utilities.RegexpMatchGroupInterface;
import core.utilities.RegexpUtilities;

/**
 * Number simple speech formatting in accordance with observed practice
 * If a number starts with 0 the number is broken to its digits
 * If a number does not start with a zero it is left to the engine to pronounce
 * 
 * @author Isaac Porat
 *
 */
public class SHNumberFormat
{
	private CascadingLoggerInterface log;
	private static final String NUMBER_PATTERN = "([\\d\\.\\,]{2,})";
	private static final String NUMBER_SINGLE_PATTERN = "([\\d])";
	private Pattern fNumberPattern;
	private Pattern fNumberSinglePattern;

	public SHNumberFormat(CascadingLoggerInterface logger)
	{
		log = logger.clone(SHNumberFormat.class);
		fNumberPattern = Pattern.compile(NUMBER_PATTERN);
		fNumberSinglePattern = Pattern.compile(NUMBER_SINGLE_PATTERN);
	}

	public String process(String string)
	{
		try
		{
			if (string.length() <= 1)
				return (string);
			if (RegexpUtilities.contains(string, fNumberPattern) == false)
				return (string);
			StringBuilder stringBuilder = new StringBuilder(string);
			RegexpMatchGroupInterface[] groups = RegexpUtilities.findGroupMatches(stringBuilder, fNumberPattern, 1);
			// Start from the end to ensure correct replacement
			for (int n = groups.length - 1; n >= 0; n--)
			{
				RegexpMatchGroupInterface group = groups[n];
				String groupString = group.getGroup();
				// Remove dot if exists at the end of string  
				groupString = groupString.replaceAll("\\.$", "");
				// If pattern contains dot or commas leave to engine to pronounce				
				if ((groupString.contains(".") || (groupString.contains(","))))
					continue;
				// JAWS behaviour
				// If less than 5 digits and does not start with 0 let engine pronounce
				//if ((groupString.length() < 5) && (groupString.startsWith("0") == false))
				// Common engine behaviour if starts with 0 speak individually
				if (groupString.startsWith("0") == false)
					continue;
				// Split to individual digits
				groupString = RegexpUtilities.replaceAll(groupString, fNumberSinglePattern, "$1 ");
				stringBuilder.replace(group.getGroupStart(), group.getGroupEnd(), groupString);
			}
			return (stringBuilder.toString());
		} catch (Exception ex)
		{
			log.error("Error in process method:\n" + ex);
			return (string);
		}
	}

}
