/*
* Copyright (c) Isaac Porat 2011 - 2013 all rights reserved.
* This file is released under the GNU General Public License, version 2, or (at your option) any later version.
* See the file SpeechHub_License.txt for more details.
*/

package sh.independent;

import java.util.ArrayList;

import java.util.Arrays;
import java.util.List;

import core.utilities.OS;
import core.utilities.StackTrace;

import log.common.CascadingLoggerInterface;
import sh.connections.SHConstants;
import sh.connections.SHEngineConnectorInterface;
import sh.connections.SHEngineInterface;
import sh.connections.SHModuleInterface;
import sh.connector.freeTTS.SHFreeTTSStdioClientSetup;
import sh.connector.maryTTS.SHConnectorMaryTTSAdaptor;
import sh.connector.multiNative.SHConnectorMultiNativeAdaptor;
import sh.connector.sapi5.SHConnectorSAPI5AdaptorAudioPlugin;
import sh.connector.sapi5.SHSAPI5ComWrapperAudioPlugin;
import sh.connector.sapi5Common.SHConnectorSAPI5AdaptorSelfAudio;
import sh.connector.sapi5Common.SHSAPI5ComWrapperSelfAudio;
import sh.connector.speechPlatform.SHConnectorSpeechPlatformAdaptorAudioPlugin;
import sh.connector.speechPlatform.SHSpeechPlatformComWrapperAudioPlugin;
import sh.engine.SHEngineFilterAdaptor;
import sh.module.server.SHModuleServerEngineWrapper;
import sh.services.SHTTSModulesControllerServiceInterface;
import sh.settings.SHEngineParameters;
import sh.tts.SHTTSModuleAdaptor;

/**
 * Module controller for the independent SH version
 * (without TCF - OSGi support)
 * @author Isaac Porat
 *
 */
public class SHIndependentModulesController implements SHTTSModulesControllerServiceInterface
{
	private CascadingLoggerInterface log;
	private final static String DEFAULT_CLIENT_IDENTIFIER = "DefaultClient";
	private SHEngineParameters fEngineParameters;
	private List<ModuleContainerInterface> fModuleContainerList;
	private SHModuleInterface fSelectedModule;
	private String fSelectedModuleEngineIdentifier = "";
	private boolean fServerMode = false;

	public SHIndependentModulesController(boolean serverMode, CascadingLoggerInterface logger)
	{
		log = logger.clone(SHIndependentModulesController.class);
		fServerMode = serverMode;
		fEngineParameters = SHEngineParameters.getInstance(logger);
		fModuleContainerList = new ArrayList<ModuleContainerInterface>();
		// Add available connectors
		fModuleContainerList.add(new MultiNativeModuleContainer());
		if (OS.isWindows())
		{
			if (SHSAPI5ComWrapperAudioPlugin.isAvailable(log))
				fModuleContainerList.add(new SAPI5ModuleContainer());
			if (SHSAPI5ComWrapperSelfAudio.isAvailable(log))
				fModuleContainerList.add(new SAPI5CommonModuleContainer());
			if (SHSpeechPlatformComWrapperAudioPlugin.isAvailable(log))
				fModuleContainerList.add(new SpeechPlatformModuleContainer());
		}
		fModuleContainerList.add(new MaryTTSModuleContainer());
		fModuleContainerList.add(new FreettsModuleContainer());
	}

	public String[] getTTSEngineIdentifiers()
	{
		List<String> engineIdentifiersList = new ArrayList<String>();
		for (int n = 0; n < fModuleContainerList.size(); n++)
			engineIdentifiersList.addAll(Arrays.asList(fModuleContainerList.get(n).getEngineIdentifiers()));
		return (engineIdentifiersList.toArray(new String[0]));
	}

	public boolean selectTTSModule(String engineIdentifier)
	{
		if (engineIdentifier == null)
			return (false);
		// Ensure that engine has not been disabled, and if it has, get another engine
		if (fEngineParameters.isEngineAvailable(engineIdentifier) == false)
			engineIdentifier = fEngineParameters.getAvailableDefaultEngineIdentifier();
		for (int n = 0; n < fModuleContainerList.size(); n++)
		{
			ModuleContainerInterface moduleContainer = fModuleContainerList.get(n);
			String[] engineIdentifiers = moduleContainer.getEngineIdentifiers();
			for (int m = 0; m < engineIdentifiers.length; m++)
			{
				if (engineIdentifier.equals(engineIdentifiers[m]))
				{
					SHModuleInterface selectedModule = moduleContainer.getModule(engineIdentifier);
					if (selectedModule != null)
					{
						selectedModule.setProperty(SHConstants.PROP_MODULE_ENGINE_IDENTIFIER, engineIdentifier);
						fSelectedModule = selectedModule;
						fSelectedModuleEngineIdentifier = engineIdentifier;
						return (true);
					}
					return (false);
				}
			}
		}
		return (false);
	}

	public String getSelectedTTSModuleEngineIdentifier()
	{
		return (fSelectedModuleEngineIdentifier);
	}

	public SHModuleInterface getSelectedTTSModule()
	{
		return (fSelectedModule);
	}

	private interface ModuleContainerInterface
	{

		public String[] getEngineIdentifiers();

		public SHModuleInterface getModule(String engineIdentifier);

	}

	private SHModuleInterface createModule(SHEngineInterface engine)
	{
		if (fServerMode)
			return (new SHModuleServerEngineWrapper(engine, log));
		return (new SHTTSModuleAdaptor(engine, DEFAULT_CLIENT_IDENTIFIER, log));
	}

	private class MultiNativeModuleContainer implements ModuleContainerInterface
	{
		private SHModuleInterface fModule;

		public String[] getEngineIdentifiers()
		{
			// List available engine with the required comms protocol
			return (SHEngineParameters.getInstance(log).getProtocolEngineIdentifiers(
					SHConstants.ENGINE_PROTOCOL_NATIVE, "1.0"));
		}

		public SHModuleInterface getModule(String engineIdentifier)
		{
			try
			{
				if (fModule == null)
				{
					SHEngineConnectorInterface engineConnector = new SHConnectorMultiNativeAdaptor(log);
					SHEngineInterface engine = new SHEngineFilterAdaptor(engineConnector, log);
					fModule = createModule(engine);
					fModule.initialize();
				}
				return (fModule);
			} catch (Exception ex)
			{
				log.error("Error in getModule method: engineIdentifier: " + engineIdentifier + "\n"
						+ StackTrace.stackToString(ex));
				return (null);
			}
		}
	}

	private class FreettsModuleContainer implements ModuleContainerInterface
	{
		private SHModuleInterface fModule;

		public String[] getEngineIdentifiers()
		{
			// List available engine with the required comms protocol
			return (SHEngineParameters.getInstance(log).getProtocolEngineIdentifiers(
					SHConstants.ENGINE_PROTOCOL_FREETTS, "1.0"));
		}

		public SHModuleInterface getModule(String engineIdentifier)
		{
			try
			{
				if (fModule == null)
				{
					SHEngineConnectorInterface engineConnector = (new SHFreeTTSStdioClientSetup(log))
							.getEngineConnector();
					SHEngineInterface engine = new SHEngineFilterAdaptor(engineConnector, log);
					fModule = createModule(engine);
					fModule.initialize();
				}
				return (fModule);
			} catch (Exception ex)
			{
				log.error("Error in getModule method: engineIdentifier: " + engineIdentifier + "\n"
						+ StackTrace.stackToString(ex));
				return (null);
			}
		}

	}

	private class SAPI5ModuleContainer implements ModuleContainerInterface
	{
		private SHModuleInterface fModule;

		public String[] getEngineIdentifiers()
		{
			// List available engine with the required comms protocol
			return (SHEngineParameters.getInstance(log).getProtocolEngineIdentifiers(SHConstants.ENGINE_PROTOCOL_SAPI5,
					"1.0"));
		}

		public SHModuleInterface getModule(String engineIdentifier)
		{
			try
			{
				if (fModule == null)
				{
					SHEngineConnectorInterface engineConnector = new SHConnectorSAPI5AdaptorAudioPlugin(log);
					SHEngineInterface engine = new SHEngineFilterAdaptor(engineConnector, log);
					fModule = createModule(engine);
					fModule.initialize();
				}
				return (fModule);
			} catch (Exception ex)
			{
				log.error("Error in getModule method: engineIdentifier: " + engineIdentifier + "\n"
						+ StackTrace.stackToString(ex));
				return (null);
			}
		}

	}

	private class SAPI5CommonModuleContainer implements ModuleContainerInterface
	{
		private SHModuleInterface fModule;

		public String[] getEngineIdentifiers()
		{
			// List available engine with the required comms protocol
			return (SHEngineParameters.getInstance(log).getProtocolEngineIdentifiers(
					SHConstants.ENGINE_PROTOCOL_SAPI5_COMMON, "1.0"));
		}

		public SHModuleInterface getModule(String engineIdentifier)
		{
			try
			{
				if (fModule == null)
				{
					SHEngineConnectorInterface engineConnector = new SHConnectorSAPI5AdaptorSelfAudio(log);
					SHEngineInterface engine = new SHEngineFilterAdaptor(engineConnector, log);
					fModule = createModule(engine);
					fModule.initialize();
				}
				return (fModule);
			} catch (Exception ex)
			{
				log.error("Error in getModule method: engineIdentifier: " + engineIdentifier + "\n"
						+ StackTrace.stackToString(ex));
				return (null);
			}
		}

	}

	private class SpeechPlatformModuleContainer implements ModuleContainerInterface
	{
		private SHModuleInterface fModule;

		public String[] getEngineIdentifiers()
		{
			// List available engine with the required comms protocol
			return (SHEngineParameters.getInstance(log).getProtocolEngineIdentifiers(SHConstants.ENGINE_PROTOCOL_MSP,
					"1.0"));
		}

		public SHModuleInterface getModule(String engineIdentifier)
		{
			try
			{
				if (fModule == null)
				{
					SHEngineConnectorInterface engineConnector = new SHConnectorSpeechPlatformAdaptorAudioPlugin(log);
					SHEngineInterface engine = new SHEngineFilterAdaptor(engineConnector, log);
					fModule = createModule(engine);
					fModule.initialize();
				}
				return (fModule);
			} catch (Exception ex)
			{
				log.error("Error in getModule method: engineIdentifier: " + engineIdentifier + "\n"
						+ StackTrace.stackToString(ex));
				return (null);
			}
		}

	}

	private class MaryTTSModuleContainer implements ModuleContainerInterface
	{
		private SHModuleInterface fModule;

		public String[] getEngineIdentifiers()
		{
			// List available engine with the required comms protocol
			return (SHEngineParameters.getInstance(log).getProtocolEngineIdentifiers(
					SHConstants.ENGINE_PROTOCOL_MARYTTS, "1.0"));
		}

		public SHModuleInterface getModule(String engineIdentifier)
		{
			try
			{
				if (fModule == null)
				{
					SHEngineConnectorInterface engineConnector = new SHConnectorMaryTTSAdaptor(log);
					SHEngineInterface engine = new SHEngineFilterAdaptor(engineConnector, log);
					fModule = createModule(engine);
					fModule.initialize();
				}
				return (fModule);
			} catch (Exception ex)
			{
				log.error("Error in getModule method: engineIdentifier: " + engineIdentifier + "\n"
						+ StackTrace.stackToString(ex));
				return (null);
			}
		}

	}

}
