/*
* Copyright (c) Isaac Porat 2011 - 2013 all rights reserved.
* This file is released under the GNU General Public License, version 2, or (at your option) any later version.
* See the file SpeechHub_License.txt for more details.
*/

package sh.module;

import java.util.HashSet;
import java.util.Set;

import log.common.CascadingLoggerInterface;
import sh.audio.SHAudioSpeechCache;
import sh.audio.SHAudioSpeechPlayerParallel;
import sh.common.SHUpdateParameters;
import sh.connections.SHAudioDataHandlerInterface;
import sh.connections.SHConstants;
import sh.connections.SHConstants.MessageType;
import sh.connections.SHCustomeHandlersInterface;
import sh.connections.SHEventsListenerInterface;
import sh.connections.SHMessageEnvelopeInterface;
import sh.connections.SHMessageQueuePeekHandlerInterface;
import sh.connections.SHModuleInterface;
import sh.connections.SHUpdateParametersHandlerInterface;
import sh.connections.SHUpdateParametersInterface;
import core.utilities.StackTrace;

/**
 *Module audio speech player adaptor
 *controlls the player and process and pass parameters to an engine
 *
 * @Author Isaac Porat
 */
public class SHModuleAudioSpeechPlayerAdaptor extends SHModuleBase implements SHModuleInterface,
		SHUpdateParametersHandlerInterface
{
	private SHAudioSpeechPlayerParallel fPlayer;
	private SHAudioSpeechCache fAudioSpeechCache;
	private SHAudioDataHandlerInterface fAudioDataHandler;
	private Set<String> fConnectorEngineSet;
	private Set<String> fClientUseSonicSet;
	private SHMessageQueuePeekHandlerInterface fMessagePeekHandler;

	private String fCurrentClientIdentifier = "unknown";
	private String fCurrentEngineIdentifier = SHConstants.STRING_PARAMETER_DISABLED;
	private String fCurrentVoiceIdentifier = SHConstants.STRING_PARAMETER_DISABLED;
	private String fCurrentVariantIdentifier = SHConstants.STRING_PARAMETER_DISABLED;
	private float fCurrentVolume = SHConstants.FLOAT_PARAMETER_DISABLED;
	private float fCurrentRate = SHConstants.FLOAT_PARAMETER_DISABLED;
	private boolean fCurrentClientUseSonic = false;
	private float fCurrentPitch = SHConstants.FLOAT_PARAMETER_DISABLED;
	private float fCharCapitalIndicatorPitchOffset = SHConstants.MODULE_CHAR_CAPITAL_INDICATOR_PITCH_LEVEL_DEFAULT;
	private int fInsertUpdates = 0;

	public SHModuleAudioSpeechPlayerAdaptor(SHModuleInterface module, CascadingLoggerInterface logger)
	{
		log = logger.clone(SHModuleAudioSpeechPlayerAdaptor.class);
		fModule = module;
		fAudioSpeechCache = new SHAudioSpeechCache(log);
		fPlayer = new SHAudioSpeechPlayerParallel(this, log);
		fClientUseSonicSet = new HashSet<String>();
	}

	@Override
	public boolean initialize()
	{
		try
		{
			if (!fModule.initialize())
			{
				return false;
			}
			getEngineIdentifiers(); // This sets the engine map required for peek operation
			fPlayer.initialize(fModule, fAudioSpeechCache);
			// Note if in the future the player is going to be implemented globally perhaps the next line registration
			// with the module needs to be disabled before another module is engaged
			fModule.setAudioDataHandler(fPlayer);
			if (fAudioDataHandler != null)
				fPlayer.setAudioDataHandler(fAudioDataHandler);
			if (fMessagePeekHandler != null)
				fPlayer.setMessagePeekHandler(fMessagePeekHandler);
			fPlayer.setConnectorEngineSet(fConnectorEngineSet);
			return (true);
		} catch (Exception ex)
		{
			log.error("Error in initialize method:\n" + ex);
			return (false);
		}
	}

	@Override
	public boolean close()
	{
		try
		{
			if (fModule.close() == false)
				return false;
			return (fPlayer.close());
		} catch (Exception ex)
		{
			log.error("Error in close method:\n" + ex);
			return (false);
		}
	}

	@Override
	public String[] getEngineIdentifiers()
	{
		try
		{
			String[] engineArray = fModule.getEngineIdentifiers();
			if (fConnectorEngineSet == null)
				fConnectorEngineSet = new HashSet<String>();
			if (fConnectorEngineSet.isEmpty())
			{
				for (int n = 0; n < engineArray.length; n++)
					fConnectorEngineSet.add(engineArray[n]);
			}
			return (engineArray);
		} catch (Exception ex)
		{
			log.error("Error in getEngineIdentifiers method:\n" + ex);
			return (new String[0]);
		}
	}

	@Override
	public boolean speak(SHMessageEnvelopeInterface messageEnvelope)
	{
		try
		{
			return (fPlayer.speak(messageEnvelope));
		} catch (Exception ex)
		{
			log.error("Error in speak method:\n" + StackTrace.stackToString(ex));
			return (false);
		}
	}

	@Override
	public boolean cancel()
	{
		try
		{
			return (fPlayer.cancel());
		} catch (Exception ex)
		{
			log.error("Error in cancel method:\n" + ex);
			return (false);
		}
	}

	@Override
	public boolean pause()
	{
		try
		{
			return (fPlayer.pause());
		} catch (Exception ex)
		{
			log.error("Error in pause method:\n" + ex);
			return (false);
		}
	}

	@Override
	public boolean resume()
	{
		try
		{
			return (fPlayer.resume());
		} catch (Exception ex)
		{
			log.error("Error in resume method:\n" + ex);
			return (false);
		}
	}

	@Override
	public boolean setAudioDataHandler(SHAudioDataHandlerInterface audioDataHandler)
	{
		fAudioDataHandler = audioDataHandler;
		return (fPlayer.setAudioDataHandler(fAudioDataHandler));
	}

	@Override
	public boolean setCustomeHandlers(SHCustomeHandlersInterface customeHandlers)
	{
		try
		{
			if (customeHandlers == null)
			{
				log.error("Error in setCustomeHandlers method: customHandlers null");
				return (false);
			}
			fMessagePeekHandler = customeHandlers.getMessageQueuePeekHandler();
			if (fMessagePeekHandler == null)
			{
				log.error("Error in setCustomeHandlers method: messagePeekHandler null");
				return (false);
			}
			fPlayer.setMessagePeekHandler(fMessagePeekHandler);
			return (true);
		} catch (Exception ex)
		{
			log.error("Error in setCustomeHandlers method:\n" + ex);
			return (false);
		}
	}

	@Override
	public boolean setEventsListener(SHEventsListenerInterface eventsListener)
	{
		// Unused, handled higher in the stack
		return false;
	}

	private boolean configureSonic(boolean mode)
	{
		try
		{
			if (mode)
			{
				// Send ahead set 0 rate command to ensure engine does not control rate
				SHUpdateParametersInterface engineUpdateParameters = SHUpdateParameters
						.getNewDisabledUpdateParameters(fCurrentClientIdentifier);
				engineUpdateParameters.setRate(0);
				// updateIdentifier by default 0 ensures unmarked update followed correctly by updateIdentifier in the next envelope
				if (fModule.updateParametersDirectly(engineUpdateParameters) == false)
					return (false);
				if (fPlayer.setClientUseSonic(true) == false)
					return (false);
				if (fPlayer.setRate(fCurrentRate) == false)
					return (false);
			} else
			{
				if (fPlayer.setClientUseSonic(false) == false)
					return (false);
				// Send ahead set rate command to restore engine rate control
				SHUpdateParametersInterface engineUpdateParameters = SHUpdateParameters
						.getNewDisabledUpdateParameters(fCurrentClientIdentifier);
				engineUpdateParameters.setRate(fCurrentRate);
				// updateIdentifier by default 0 ensures unmarked update followed correctly by updateIdentifier in the next envelope
				if (fModule.updateParametersDirectly(engineUpdateParameters) == false)
					return (false);
			}
			fCurrentClientUseSonic = mode;
			if (fCurrentClientUseSonic)
				fClientUseSonicSet.add(fCurrentClientIdentifier);
			else
				fClientUseSonicSet.remove(fCurrentClientIdentifier);
			return (true);
		} catch (Exception ex)
		{
			log.error("Error in configureSonic method:\n" + ex);
			return (false);
		}
	}

	@Override
	public boolean setProperty(String key, String value)
	{
		try
		{
			if ((key == null) || (value == null))
				return (false);
			if (key.equals(SHConstants.PROP_MODULE_CHAR_CAPITAL_INDICATOR_PITCH_LEVEL))
				fCharCapitalIndicatorPitchOffset = Float.parseFloat(value);
			else if (key.equals(SHConstants.PROP_ENGINE_USE_SONIC))
			{
				String engineUseSonic = fModule.getProperty(SHConstants.PROP_ENGINE_USE_SONIC);
				// Force Sonic true if engine requires it
				boolean mode = value.equals("true") || ((engineUseSonic != null) && (engineUseSonic.equals("true")));
				return (configureSonic(mode));
			}
			return (fModule.setProperty(key, value));
		} catch (Exception ex)
		{
			log.error("Error in setProperty method at key = " + key + " value = " + value + "\n" + ex);
			return (false);
		}
	}

	@Override
	public boolean updateParameters(SHMessageEnvelopeInterface messageEnvelope)
	{
		try
		{
			SHUpdateParametersInterface updateParameters = messageEnvelope.getUpdateParameters();
			if (messageEnvelope.getMessageType().equals(MessageType.CHAR))
			{
				if (messageEnvelope.getCapitalIndicator().equals(SHConstants.CapitalIndicator.PITCH))
				{
					// Only adjust pitch if current pitch and updatePitch are not disabled
					float updatePitch = updateParameters.getPitch();
					if ((fCurrentPitch != SHConstants.FLOAT_PARAMETER_DISABLED)
							&& (updatePitch != SHConstants.FLOAT_PARAMETER_DISABLED))
					{
						updateParameters.setPitch(updatePitch + fCharCapitalIndicatorPitchOffset);
						fInsertUpdates = 2;
					}
				}
			}
			if (fInsertUpdates > 0)
			{
				updateParameters.setEngineIdentifier(SHConstants.STRING_PARAMETER_DISABLED);
				updateParameters.setVoiceIdentifier(SHConstants.STRING_PARAMETER_DISABLED);
				updateParameters.setVariantIdentifier(SHConstants.STRING_PARAMETER_DISABLED);
				updateParameters.setVolume(SHConstants.FLOAT_PARAMETER_DISABLED);
				updateParameters.setRate(SHConstants.FLOAT_PARAMETER_DISABLED);
				updateParameters.markInsertUpdate();
				if (fPlayer.isPlayerHandlesPitch())
					fInsertUpdates--;
				else
				{
					if (fModule.getPitch() == updateParameters.getPitch())
						fInsertUpdates--;
				}
			}
			return (updateParameters(updateParameters));
		} catch (Exception ex)
		{
			log.error("Error in updateParameters method:\n" + StackTrace.stackToString(ex));
			return (false);
		}
	}

	private boolean updateParameters(SHUpdateParametersInterface updateParameters)
	{
		try
		{
			String clientIdentifier = updateParameters.getClientIdentifier();
			if (fCurrentClientIdentifier.equals(clientIdentifier) == false)
				updateParameters.markResetUpdate();
			int updateIdentifier = updateParameters.getUpdateIdentifier();
			if (fModule.isUpdateAccepted(updateIdentifier) == false)
				return (true);

			String engineIdentifier = updateParameters.getEngineIdentifier();
			if (SHUpdateParameters.isUpdateRequired(updateIdentifier, engineIdentifier, fCurrentEngineIdentifier))
			{
				// Send ahead selectEngine command
				SHUpdateParametersInterface engineUpdateParameters = SHUpdateParameters
						.getNewDisabledUpdateParameters(clientIdentifier);
				engineUpdateParameters.setEngineIdentifier(engineIdentifier);
				// updateIdentifier by default 0 ensures unmarked update followed correctly by updateIdentifier in the next envelope
				if (fModule.updateParametersDirectly(engineUpdateParameters))
				{
					if (fPlayer.notifyEngineChanged(engineIdentifier))
					{
						if (fCurrentEngineIdentifier.equals(engineIdentifier) == false)
							fAudioSpeechCache.clearCache();
						fCurrentEngineIdentifier = engineIdentifier;
					} else
					{
						// errors logged by method
						return (false);
					}
				}
			}
			String voiceIdentifier = updateParameters.getVoiceIdentifier();
			if (SHUpdateParameters.isUpdateRequired(updateIdentifier, voiceIdentifier, fCurrentVoiceIdentifier))
			{
				if (fCurrentVoiceIdentifier.equals(voiceIdentifier) == false)
					fAudioSpeechCache.clearCache();
				fCurrentVoiceIdentifier = voiceIdentifier;
			}
			String variantIdentifier = updateParameters.getVariantIdentifier();
			if (SHUpdateParameters.isUpdateRequired(updateIdentifier, variantIdentifier, fCurrentVariantIdentifier))
			{
				if (fCurrentVariantIdentifier.equals(variantIdentifier) == false)
					fAudioSpeechCache.clearCache();
				fCurrentVariantIdentifier = variantIdentifier;
			}
			float volume = updateParameters.getVolume();
			if (SHUpdateParameters.isUpdateRequired(updateIdentifier, volume, fCurrentVolume))
			{
				fPlayer.setVolume(volume);
				// Ignore errors
				fCurrentVolume = volume;
			}
			float rate = updateParameters.getRate();
			if (SHUpdateParameters.isUpdateRequired(updateIdentifier, rate, fCurrentRate))
			{
				fPlayer.setRate(rate);
				// Ignore errors
				fCurrentRate = rate;
			}
			float pitch = updateParameters.getPitch();
			if (SHUpdateParameters.isUpdateRequired(updateIdentifier, pitch, fCurrentPitch))
			{
				fPlayer.setPitch(pitch);
				// Ignore errors
				fCurrentPitch = pitch;
			}
			// Disabling parameters needs to be set regardless if the actual parameter was changed
			// Always disable volume with the use of the player
			updateParameters.setVolume(SHConstants.FLOAT_PARAMETER_DISABLED);
			if (fPlayer.isPlayerHandlesRate())
				updateParameters.setRate(SHConstants.FLOAT_PARAMETER_DISABLED);
			if (fPlayer.isPlayerHandlesPitch())
				updateParameters.setPitch(SHConstants.FLOAT_PARAMETER_DISABLED);
			fAudioSpeechCache.setPitch(updateParameters.getPitch());
			// Check if Sonic use needs to change with change of client
			if (fCurrentClientIdentifier.equals(clientIdentifier) == false)
			{
				boolean clientUseSonic = fClientUseSonicSet.contains(clientIdentifier);
				if (fCurrentClientUseSonic != clientUseSonic) // Use Sonic changed
					configureSonic(clientUseSonic);
			}
			fCurrentClientIdentifier = clientIdentifier;
			return (true);
		} catch (Exception ex)
		{
			log.error("Error in updateParameters method:\n" + StackTrace.stackToString(ex));
			return (false);
		}
	}

	@Override
	public boolean updateParametersDirectly(SHUpdateParametersInterface updateParameters)
	{
		updateParameters(updateParameters);
		return (fModule.updateParametersDirectly(updateParameters));
	}

}
