/*
* Copyright (c) Isaac Porat 2011 - 2013 all rights reserved.
* This file is released under the GNU General Public License, version 2, or (at your option) any later version.
* See the file SpeechHub_License.txt for more details.
*/

package sh.module;

import core.utilities.StackTrace;
import log.common.CascadingLoggerInterface;
import sh.connections.SHConstants;
import sh.connections.SHCustomeHandlersInterface;
import sh.connections.SHMessageEnvelopeInterface;
import sh.connections.SHModuleInterface;
import sh.connections.SHModuleTimestampInterface;
import sh.connections.SHAudioDataHandlerInterface;
import sh.connections.SHEventsListenerInterface;
import sh.connections.SHUpdateParametersInterface;

/**
 * module base 
 * 
 * @author Isaac Porat
 */
public abstract class SHModuleBase implements SHModuleInterface
{
	protected CascadingLoggerInterface log;
	protected SHModuleInterface fModule;
	protected SHModuleTimestampInterface fModuleTimestamp;
	protected boolean fMasterCancel = true;

	@Override
	public boolean configureTiming(SHModuleTimestampInterface moduleTimestamp, boolean masterCancel)
	{
		try
		{
			fModuleTimestamp = moduleTimestamp;
			fMasterCancel = masterCancel;
			return (fModule.configureTiming(moduleTimestamp, masterCancel));
		} catch (Exception ex)
		{
			log.error("Error in configure method:\n" + ex);
			return (false);
		}
	}

	@Override
	public boolean initialize()
	{
		try
		{
			return (fModule.initialize());
		} catch (Exception ex)
		{
			log.error("Error in initialize method:\n" + ex);
			return (false);
		}
	}

	@Override
	public boolean close()
	{
		try
		{
			return (fModule.close());
		} catch (Exception ex)
		{
			log.error("Error in close method:\n" + ex);
			return (false);
		}
	}

	@Override
	public boolean setPriority(int priorityLevel, boolean selfCancel)
	{
		try
		{
			return (fModule.setPriority(priorityLevel, selfCancel));
		} catch (Exception ex)
		{
			log.error("Error in setPriority method:\n" + ex);
			return (false);
		}
	}

	@Override
	public String[] getEngineIdentifiers()
	{
		try
		{
			return (fModule.getEngineIdentifiers());
		} catch (Exception ex)
		{
			log.error("Error in getEngineIdentifiers method:\n" + ex);
			return (null);
		}
	}

	@Override
	public String getSelectedEngineIdentifier()
	{
		try
		{
			return (fModule.getSelectedEngineIdentifier());
		} catch (Exception ex)
		{
			log.error("Error in getSelectedEngineIdentifier Priority method:\n" + ex);
			return (null);
		}
	}

	@Override
	public String getSelectedLanguageCode()
	{
		try
		{
			return (fModule.getSelectedLanguageCode());
		} catch (Exception ex)
		{
			log.error("Error in getLanguage method:\n" + ex);
			return (null);
		}
	}

	@Override
	public String[] getVoiceIdentifiers()
	{
		try
		{
			return (fModule.getVoiceIdentifiers());
		} catch (Exception ex)
		{
			log.error("Error in getVoiceIdentifiers method:\n" + ex);
			return (null);
		}
	}

	@Override
	public String getSelectedVoiceIdentifier()
	{
		try
		{
			return (fModule.getSelectedVoiceIdentifier());
		} catch (Exception ex)
		{
			log.error("Error in selectVoiceIdentifier method:\n" + ex);
			return (null);
		}
	}

	@Override
	public String[] getVariantIdentifiers()
	{
		try
		{
			return (fModule.getVariantIdentifiers());
		} catch (Exception ex)
		{
			log.error("Error in getVariantIdentifiers method:\n" + ex);
			return (null);
		}
	}

	@Override
	public String getSelectedVariantIdentifier()
	{
		try
		{
			return (fModule.getSelectedVariantIdentifier());
		} catch (Exception ex)
		{
			log.error("Error in getSelectedVariantIdentifier method:\n" + ex);
			return (null);
		}
	}

	@Override
	// define in every Adaptor and Wrapper
	public abstract boolean speak(SHMessageEnvelopeInterface messageEnvelope);

	@Override
	public boolean speak(int messageIdentifier, String message, boolean speakSync)
	{
		// Override only in Buffer (entry) Adaptor and relevant Wrapper
		return (false);
	}

	@Override
	public boolean speakChar(int messageIdentifier, String message, boolean speakSync)
	{
		// Override only in Buffer (entry) Adaptor and relevant Wrapper
		return (false);
	}

	@Override
	public boolean speakKey(int messageIdentifier, String message, boolean speakSync)
	{
		// Override only in Buffer (entry) Adaptor and relevant Wrapper
		return (false);
	}

	@Override
	public boolean pause()
	{
		try
		{
			return (fModule.pause());
		} catch (Exception ex)
		{
			log.error("Error in pause method:\n" + ex);
			return (false);
		}
	}

	@Override
	public boolean resume()
	{
		try
		{
			return (fModule.resume());
		} catch (Exception ex)
		{
			log.error("Error in resume method:\n" + ex);
			return (false);
		}
	}

	@Override
	public boolean stop()
	{
		try
		{
			return (fModule.stop());
		} catch (Exception ex)
		{
			log.error("Error in stop method:\n" + ex);
			return (false);
		}
	}

	@Override
	public boolean cancel()
	{
		try
		{
			return (fModule.cancel());
		} catch (Exception ex)
		{
			log.error("Error in cancel method:\n" + ex);
			return (false);
		}
	}

	@Override
	public float getVolume()
	{
		try
		{
			return (fModule.getVolume());
		} catch (Exception ex)
		{
			log.error("Error in getVolume method:\n" + ex);
			return (SHConstants.FLOAT_PARAMETER_ERROR);
		}
	}

	@Override
	public float getRate()
	{
		try
		{
			return (fModule.getRate());
		} catch (Exception ex)
		{
			log.error("Error in getRate method:\n" + ex);
			return (SHConstants.FLOAT_PARAMETER_ERROR);
		}
	}

	@Override
	public float getPitch()
	{
		try
		{
			return (fModule.getPitch());
		} catch (Exception ex)
		{
			log.error("Error in getPitch method:\n" + ex);
			return (SHConstants.FLOAT_PARAMETER_ERROR);
		}
	}

	@Override
	public boolean setProperty(String key, String value)
	{
		try
		{
			return (fModule.setProperty(key, value));
		} catch (Exception ex)
		{
			log.error("Error in setProperty method:\n" + ex);
			return (false);
		}
	}

	@Override
	public String getProperty(String key)
	{
		try
		{
			return (fModule.getProperty(key));
		} catch (Exception ex)
		{
			log.error("Error in getProperty method:\n" + ex);
			return (null);
		}
	}

	@Override
	public boolean setEventsListener(SHEventsListenerInterface eventsListener)
	{
		try
		{
			return (fModule.setEventsListener(eventsListener));
		} catch (Exception ex)
		{
			log.error("Error in setEventsListener method:\n" + ex);
			return (false);
		}
	}

	@Override
	public boolean setAudioDataHandler(SHAudioDataHandlerInterface audioDataHandler)
	{
		try
		{
			return (fModule.setAudioDataHandler(audioDataHandler));
		} catch (Exception ex)
		{
			log.error("Error in setAudioDataHandler method:\n" + ex);
			return (false);
		}
	}

	@Override
	public boolean setCustomeHandlers(SHCustomeHandlersInterface customeHandlers)
	{
		try
		{
			return (fModule.setCustomeHandlers(customeHandlers));
		} catch (Exception ex)
		{
			log.error("Error in setAudioDataHandler method:\n" + ex);
			return (false);
		}
	}

	@Override
	public boolean updateParametersDirectly(SHUpdateParametersInterface updateParameters)
	{
		try
		{
			return (fModule.updateParametersDirectly(updateParameters));
		} catch (Exception ex)
		{
			log.error("Error in updateParametersDirectly method:\n" + StackTrace.stackToString(ex));
			return (false);
		}
	}

	@Override
	public boolean setUpdateIdentifier(int updateIdentifier)
	{
		try
		{
			return (fModule.setUpdateIdentifier(updateIdentifier));
		} catch (Exception ex)
		{
			log.error("Error in setUpdateIdentifier method:\n" + ex);
			return (false);
		}
	}

	@Override
	public boolean isUpdateAccepted(int updateIdentifier)
	{
		try
		{
			return (fModule.isUpdateAccepted(updateIdentifier));
		} catch (Exception ex)
		{
			log.error("Error in isUpdateAccepted method:\n" + ex);
			return (false);
		}
	}

}
