/*
* Copyright (c) Isaac Porat 2011 - 2013 all rights reserved.
* This file is released under the GNU General Public License, version 2, or (at your option) any later version.
* See the file SpeechHub_License.txt for more details.
*/

package sh.module;

import java.util.HashMap;
import java.util.Map;

import core.utilities.StackTrace;

import log.common.CascadingLoggerInterface;

import sh.connections.SHConstants;
import sh.connections.SHCustomeHandlersInterface;
import sh.connections.SHEngineInterface;
import sh.connections.SHMessageEnvelopeInterface;
import sh.connections.SHModuleInterface;
import sh.connections.SHModuleTimestampInterface;
import sh.connections.SHAudioDataHandlerInterface;
import sh.connections.SHEventsListenerInterface;
import sh.connections.SHUpdateParametersInterface;
import sh.connections.SHConstants.MessageType;
import sh.settings.SHEngineParameters;

/**
 * Module to engine adaptor 
 * 
 * @author Isaac Porat
 */
public class SHModuleEngineAdaptor extends SHModuleBase implements SHModuleInterface
{
	private SHEngineInterface fEngine;
	private SHEngineParameters fEngineParameters;
	private boolean fInitialized = false;
	private Map<String, String> fModuleProperties;
	private boolean fEngineSupportsPunctuation = false;
	private boolean fEngineSupportsSSML = false;
	private boolean fModuleHandlesPunctuation = true;

	public SHModuleEngineAdaptor(SHEngineInterface engine, CascadingLoggerInterface logger)
	{
		log = logger.clone(SHModuleEngineAdaptor.class);
		fEngine = engine;
		fEngineParameters = SHEngineParameters.getInstance(log);
		fModuleProperties = new HashMap<String, String>();
		updateFromEngineProperties();
	}

	private void updateFromEngineProperties()
	{
		String property = getProperty(SHConstants.PROP_ENGINE_SUPPORTS_PUNCTUATION, "false");
		fEngineSupportsPunctuation = property.equalsIgnoreCase("true");
		property = getProperty(SHConstants.PROP_ENGINE_SUPPORTS_SSML, "false");
		fEngineSupportsSSML = property.equalsIgnoreCase("true");
	}

	@Override
	public boolean configureTiming(SHModuleTimestampInterface moduleTimestamp, boolean masterCancel)
	{
		// Not used by engine
		return (false);
	}

	@Override
	public boolean initialize()
	{
		if (fInitialized)
			return (true);
		fInitialized = fEngine.initialize();
		return (fInitialized);
	}

	@Override
	public boolean close()
	{
		boolean status = fEngine.close();
		fInitialized = !status;
		return (status);
	}

	@Override
	public boolean setPriority(int priorityLevel, boolean selfCancel)
	{
		// Not applicable - not implemented by engine
		return (false);
	}

	@Override
	public String[] getEngineIdentifiers()
	{
		return (fEngine.getEngineIdentifiers());
	}

	private boolean selectEngine(String engineIdentifier)
	{
		boolean status = fEngine.selectEngine(engineIdentifier);
		if (status)
			updateFromEngineProperties();
		return (status);
	}

	@Override
	public String getSelectedEngineIdentifier()
	{
		return (fEngine.getSelectedEngineIdentifier());
	}

	@Override
	public String getSelectedLanguageCode()
	{
		return (fEngine.getSelectedLanguageCode());
	}

	@Override
	public String[] getVoiceIdentifiers()
	{
		return (fEngine.getVoiceIdentifiers());
	}

	@Override
	public String getSelectedVoiceIdentifier()
	{
		return (fEngine.getSelectedVoiceIdentifier());
	}

	@Override
	public String[] getVariantIdentifiers()
	{
		return (fEngine.getVariantIdentifiers());
	}

	@Override
	public String getSelectedVariantIdentifier()
	{
		return (fEngine.getSelectedVariantIdentifier());
	}

	@Override
	public boolean speak(SHMessageEnvelopeInterface messageEnvelope)
	{
		try
		{
			if (updateParametersToEngine(messageEnvelope.getUpdateParameters()) == false)
			{
				// Error logged by method
				return (false);
			}
			int messageIdentifier = messageEnvelope.getMessageIdentifier();
			String message = messageEnvelope.getMessage();
			MessageType messageType = messageEnvelope.getMessageType();
			if (messageType.equals(MessageType.SSML))
			{
				if (fEngineSupportsSSML == false)
				{
					// The following regexp can be made more efficient not sure if it justifies the extra complexity
					// Strip XML tags
					message = message.replaceAll("<(.|\n)+?>", "");
					// Replace '>' and '<' enteties
					message = message.replaceAll("\\&gt\\;", ">");
					message = message.replaceAll("\\&lt\\;", "<");
				}
			}
			if (messageEnvelope.getMessageSynchronized())
				return (processSync(messageIdentifier, message, messageType));
			return (fEngine.speak(messageIdentifier, message, false));
		} catch (Exception ex)
		{
			log.error("Error in speak method:\n" + ex);
			return (false);
		}
	}

	private boolean processSync(int messageIdentifier, String message, MessageType messageType)
	{
		if (message.length() == 0) // Block to ensure correct operation but do not call 
		{
			synchronized (this)
			{
				try
				{
					wait(1);
				} catch (InterruptedException e)
				{
				}
			}
			return (true);
		}
		return (fEngine.speak(messageIdentifier, message, true));
	}

	@Override
	public boolean pause()
	{
		return (fEngine.pause());
	}

	@Override
	public boolean resume()
	{
		return (fEngine.resume());
	}

	@Override
	public boolean stop()
	{
		// Not implemented, functionality achieved using cancel upstream
		return (false);
	}

	@Override
	public boolean cancel()
	{
		return (fEngine.cancel());
	}

	@Override
	public float getVolume()
	{
		return (fEngine.getVolume());
	}

	@Override
	public float getRate()
	{
		return (fEngine.getRate());
	}

	@Override
	public float getPitch()
	{
		return (fEngine.getPitch());
	}

	@Override
	public String getProperty(String key)
	{
		if (key == null)
			return (null);
		if (key.equals(SHConstants.PROP_MODULE_CLIENT_IDENTIFIER))
			return (fModuleProperties.get(key));
		String property = fModuleProperties.get(key);
		if (property != null)
			return (property);
		return (fEngine.getProperty(key));
	}

	private String getProperty(String key, String defaultValue)
	{
		String value = getProperty(key);
		return (value != null ? value : defaultValue);
	}

	@Override
	public boolean setProperty(String key, String value)
	{
		if ((key == null) || (value == null))
			return (false);
		if (key.equals(SHConstants.PROP_MODULE_CLIENT_IDENTIFIER))
		{
			fModuleProperties.put(key, value);
			return (true);
		}
		if (fInitialized == false)
		{
			if (initialize() == false)
				return (false);
		}
		if (key.equals(SHConstants.PROP_MODULE_HANDLES_PUNCTUATION))
		{
			fModuleHandlesPunctuation = value.equalsIgnoreCase("true");
			// If supported and handled by module set engine punctuation to none
			if ((fEngineSupportsPunctuation) && (fModuleHandlesPunctuation))
				fEngine.setProperty(SHConstants.PROP_ENGINE_PUNCTUATION_LEVEL,
						SHConstants.ENGINE_PUNCTUATION_LEVEL_NONE);
		} else if (key.equalsIgnoreCase(SHConstants.PROP_MODULE_PUNCTUATION_LEVEL))
		{
			if (fEngineSupportsPunctuation)
			{
				if (fModuleHandlesPunctuation)
					fEngine.setProperty(SHConstants.PROP_ENGINE_PUNCTUATION_LEVEL,
							SHConstants.ENGINE_PUNCTUATION_LEVEL_NONE);
				else
					fEngine.setProperty(SHConstants.PROP_ENGINE_PUNCTUATION_LEVEL, value);
			}
		} else if (key.equals(SHConstants.PROP_MODULE_SSML_MODE))
		{
			if (fEngineSupportsSSML)
				fEngine.setProperty(SHConstants.PROP_ENGINE_SSML_MODE, value);
		}
		// Store module properties 
		if (key.startsWith("module"))
		{
			fModuleProperties.put(key, value);
			return (true);
		}
		// Store components properties
		if (key.startsWith("components"))
		{
			fModuleProperties.put(key, value);
		}
		return (fEngine.setProperty(key, value));
	}

	@Override
	public boolean setEventsListener(SHEventsListenerInterface eventsListener)
	{
		return (fEngine.setEventsListener(eventsListener));
	}

	@Override
	public boolean setAudioDataHandler(SHAudioDataHandlerInterface audioDataHandler)
	{
		return (fEngine.setAudioDataHandler(audioDataHandler));
	}

	@Override
	public boolean setCustomeHandlers(SHCustomeHandlersInterface customeHandlers)
	{
		return (fEngine.setCustomeHandlers(customeHandlers));
	}

	@Override
	public boolean updateParametersDirectly(SHUpdateParametersInterface updateParameters)
	{
		return (updateParametersToEngine(updateParameters));
	}

	@Override
	public boolean setUpdateIdentifier(int updateIdentifier)
	{
		return (fEngine.setUpdateIdentifier(updateIdentifier));
	}

	@Override
	public boolean isUpdateAccepted(int updateIdentifier)
	{
		return (fEngine.isUpdateAccepted(updateIdentifier));
	}

	private synchronized boolean updateParametersToEngine(SHUpdateParametersInterface updateParameters)
	{
		try
		{
			/* 
			* Check if engine parameters update required
			 * Update always if updateIdentifier 0 (insertIdentifier) or -1 (reset identifier) or > than previous updateIdentifier
			 */
			int updateIdentifier = updateParameters.getUpdateIdentifier();
			if (isUpdateAccepted(updateIdentifier) == false)
				return (true);
			/*
			 * Parameters efficiency update are rare compared to speech messages
			 * therefore parameters transfer flexibility is of
			 * greater importance than transfer efficiency  
			* Parameters which are transfered to the engine as duplicates
			*  of a value of a previous update 
			* are filtered out by the engine filter layer
			*/
			String engineIdentifier = updateParameters.getEngineIdentifier();
			if (engineIdentifier.equals(SHConstants.STRING_PARAMETER_DISABLED) == false)
			{
				if (selectEngine(engineIdentifier) == false)
				{
					fEngineParameters.disableEngine(updateParameters.getEngineIdentifier());
					log.error("Error in updateParametersToEngine method:\nModule: "
							+ updateParameters.getEngineIdentifier() + " selection failed");
					return (false);
				}
			}
			String languageCode = updateParameters.getLanguageCode();
			if (languageCode.equals(SHConstants.STRING_PARAMETER_DISABLED) == false)
			{
				if (fEngine.selectLanguage(languageCode) == false)
				{
					log.error("Error in updateParametersToEngine method:\n set language failed");
					return (false);
				}
			}
			String voiceIdentifier = updateParameters.getVoiceIdentifier();
			if (voiceIdentifier.equals(SHConstants.STRING_PARAMETER_DISABLED) == false)
			{
				if (fEngine.selectVoice(voiceIdentifier) == false)
				{
					log.error("Error in updateParametersToEngine method:\nModule: "
							+ updateParameters.getEngineIdentifier() + " Voice: "
							+ updateParameters.getVoiceIdentifier() + " selection failed");
					return (false);
				}
			}
			String variantIdentifier = updateParameters.getVariantIdentifier();
			if (variantIdentifier.equals(SHConstants.STRING_PARAMETER_DISABLED) == false)
			{
				if (fEngine.selectVariant(variantIdentifier) == false)
				{
					log.error("Error in updateParametersToEngine method:\nModule: "
							+ updateParameters.getEngineIdentifier() + " Variant: "
							+ updateParameters.getVariantIdentifier() + " selection failed");
					// Ignore error, not critical
				}
			}
			// Best effort to update the following parameters no returned status check
			float volume = updateParameters.getVolume();
			if (volume != SHConstants.FLOAT_PARAMETER_DISABLED)
			{
				fEngine.setVolume(volume);
			}
			float rate = updateParameters.getRate();
			if (rate != SHConstants.FLOAT_PARAMETER_DISABLED)
				fEngine.setRate(rate);
			float pitch = updateParameters.getPitch();
			if (pitch != SHConstants.FLOAT_PARAMETER_DISABLED)
				fEngine.setPitch(pitch);
			// Do not mark identifier if 0
			if (updateIdentifier > 0)
				setUpdateIdentifier(updateParameters.getUpdateIdentifier());
			return (true);
		} catch (Exception ex)
		{
			log.error("Error in updateParametersToEngine method:\n" + StackTrace.stackToString(ex));
			return (false);
		}
	}

}
