/*
* Copyright (c) Isaac Porat 2011 - 2013 all rights reserved.
* This file is released under the GNU General Public License, version 2, or (at your option) any later version.
* See the file SpeechHub_License.txt for more details.
*/

package sh.module;

import java.util.Iterator;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;

import log.common.CascadingLoggerInterface;

import sh.common.SHMessageEnvelope;
import sh.connections.SHClientParametersInterface;
import sh.connections.SHConstants;
import sh.connections.SHMessageEnvelopeInterface;
import sh.connections.SHModuleInterface;
import sh.connections.SHEventsListenerInterface;
import sh.connections.SHConstants.CapitalIndicator;
import sh.connections.SHConstants.MessageState;
import sh.connections.SHConstants.MessageType;

/**
 * Module entry  
 * Initialize messageEnvelope  and handles messages buffer
 * 
 * @author Isaac Porat
  */
public class SHModuleEntryAdaptor extends SHModuleBase implements SHModuleInterface, Runnable
{
	private BlockingQueue<SHMessageEnvelopeInterface> fMessageEnvelopeList = new LinkedBlockingQueue<SHMessageEnvelopeInterface>();
	private boolean fMessageSynchronizedInProcess = false;
	private int fMessageSynchronizedIdentifier = 0;
	private int fMessagePriorityLevel = 4;
	private boolean fMessageSelfCancel = false;
	/*
	* Indicates if messages of one char long are routed automatically to char handling 
	* useful for capital indication
	*/
	private boolean fModuleRouteMessageToChar = true;
	private CapitalIndicator fModuleCharCapitalIndicator = CapitalIndicator.PITCH;
	private boolean fModuleSpellMode = false;
	private boolean fModuleSSMLMode = false;
	private SHEventsListenerInterface fEntryEventsListener;
	private SHEventsListenerInterface fEventsListener;
	private SHClientParametersInterface fClientParameters;
	private int fSilenceBeforeMessage = 0; // In ms

	public SHModuleEntryAdaptor(SHModuleInterface module, SHClientParametersInterface clientParameters,
			CascadingLoggerInterface logger)
	{
		log = logger.clone(SHModuleEntryAdaptor.class);
		fModule = module;
		fClientParameters = clientParameters;
		fEntryEventsListener = new EntryEvents();
		fModule.setEventsListener(fEntryEventsListener);
		Thread dispatchThread = new Thread(this);
		dispatchThread.setName(SHModuleEntryAdaptor.class.getName());
		dispatchThread.start();
	}

	@Override
	public boolean speak(int messageIdentifier, String message, boolean speakSync)
	{
		// Message is never discarded to ensure events integrity
		if (message == null)
			message = "";
		MessageType messageType;
		CapitalIndicator capitalIndicator = CapitalIndicator.NONE;
		if (fModuleSSMLMode)
			messageType = MessageType.SSML;
		else if ((fModuleRouteMessageToChar) && (message.length() == 1) && (fModuleSpellMode == false)) // Check if message to char routing is required and not in spell mode
		{
			messageType = MessageType.CHAR;
			capitalIndicator = fModuleCharCapitalIndicator;
		} else if (fModuleSpellMode)
		{
			messageType = MessageType.SPELL;
			capitalIndicator = fModuleCharCapitalIndicator;
		} else
			messageType = MessageType.STANDARD;
		if (speakSync)
			return (processSync(messageType, capitalIndicator, messageIdentifier, message));
		return (processAsync(messageType, capitalIndicator, messageIdentifier, message));
	}

	@Override
	public boolean speakChar(int messageIdentifier, String message, boolean speakSync)
	{
		if (speakSync)
			return (processSync(MessageType.CHAR, fModuleCharCapitalIndicator, messageIdentifier, message));
		return (processAsync(MessageType.CHAR, fModuleCharCapitalIndicator, messageIdentifier, message));
	}

	@Override
	public boolean speakKey(int messageIdentifier, String message, boolean speakSync)
	{
		if (speakSync)
			return (processSync(MessageType.KEY, CapitalIndicator.NONE, messageIdentifier, message));
		return (processAsync(MessageType.KEY, CapitalIndicator.NONE, messageIdentifier, message));
	}

	private boolean processSync(MessageType messageType, CapitalIndicator capitalIndicator, int messageIdentifier,
			String message)
	{
		synchronized (this)
		{
			messageToQueue(messageType, capitalIndicator, messageIdentifier, message, true);
			while (fMessageSynchronizedInProcess)
			{
				try
				{
					wait();
				} catch (InterruptedException ex)
				{
					return (false);
				}
			}
		}
		return (true);
	}

	private boolean processAsync(MessageType messageType, CapitalIndicator capitalIndicator, int messageIdentifier,
			String message)
	{
		messageToQueue(messageType, capitalIndicator, messageIdentifier, message, false);
		return (true);
	}

	@Override
	public boolean speak(SHMessageEnvelopeInterface messageEnvelope)
	{
		if (messageEnvelope.getMessageSynchronized())
			return (processSync(messageEnvelope));
		return (processAsync(messageEnvelope));
	}

	private boolean processSync(SHMessageEnvelopeInterface messageEnvelope)
	{
		synchronized (this)
		{
			messageToQueue(messageEnvelope, true);
			while (fMessageSynchronizedInProcess)
			{
				try
				{
					wait();
				} catch (InterruptedException ex)
				{
					return (false);
				}
			}
		}
		return (true);
	}

	private boolean processAsync(SHMessageEnvelopeInterface messageEnvelope)
	{
		messageToQueue(messageEnvelope, false);
		return (true);
	}

	@Override
	public boolean setPriority(int priorityLevel, boolean selfCancel)
	{
		fMessagePriorityLevel = priorityLevel;
		fMessageSelfCancel = selfCancel;
		return (true);
	}

	@Override
	public boolean cancel()
	{
		if (fMasterCancel)
			fModuleTimestamp.setCancelTimestampToNow();
		boolean status1 = true;
		synchronized (fMessageEnvelopeList)
		{
			try
			{
				for (Iterator<SHMessageEnvelopeInterface> i = fMessageEnvelopeList.iterator(); i.hasNext();)
				{
					i.next();
					i.remove();
				}
			} catch (Exception ex)
			{
				log.error("Error in cancel method in removing messageEnvelope\n" + ex);
				status1 = false;
			}
		}
		boolean status2 = fModule.cancel();
		fMessageSynchronizedInProcess = false;
		notifySpeakSync();
		return (status1 && status2);
	}

	private void messageToQueue(MessageType messageType, CapitalIndicator capitalIndicator, int messageIdentifier,
			String message, boolean messageSynchronized)
	{
		if (messageSynchronized)
			fMessageSynchronizedInProcess = true;
		SHMessageEnvelopeInterface messageEnvelope = new SHMessageEnvelope(messageIdentifier, message);
		messageEnvelope.setTimestamp(fModuleTimestamp.getTimestamp());
		messageEnvelope.setMessageType(messageType);
		messageEnvelope.setMessageState(MessageState.SINGLE);
		messageEnvelope.setCapitalIndicator(capitalIndicator);
		messageEnvelope.setMessageSynchronized(messageSynchronized);
		messageEnvelope.setPriorityLevel(fMessagePriorityLevel);
		messageEnvelope.setSelfCancel(fMessageSelfCancel);
		if (fSilenceBeforeMessage > 0)
		{
			messageEnvelope.setSilenceBeforeMessage(fSilenceBeforeMessage);
			fSilenceBeforeMessage = 0; // Silence for one message only
		}
		if (fClientParameters != null)
			messageEnvelope.setUpdateParameters(fClientParameters.getClonedUpdateParameters());
		try
		{
			fMessageEnvelopeList.put(messageEnvelope);
		} catch (InterruptedException ex)
		{
			log.error("Error in messageToQueue method:\n" + ex);
		}

	}

	private void messageToQueue(SHMessageEnvelopeInterface messageEnvelope, boolean messageSynchronized)
	{
		/*
		 * It is assumed that the envelope already contains all parameters but for 
		 * messageSynchronized, priorityLevel and selfCancel
		 */
		messageEnvelope.setPriorityLevel(fMessagePriorityLevel);
		messageEnvelope.setSelfCancel(fMessageSelfCancel);
		synchronized (fMessageEnvelopeList)
		{
			if (messageSynchronized)
				fMessageSynchronizedInProcess = true;
			messageEnvelope.setMessageSynchronized(messageSynchronized);
			try
			{
				fMessageEnvelopeList.put(messageEnvelope);
			} catch (InterruptedException ex)
			{
				log.error("Error in messageToQueue method:\n" + ex);
			}
		}
	}

	private void queueToModule()
	{
		SHMessageEnvelopeInterface messageEnvelope = null;
		try
		{
			messageEnvelope = fMessageEnvelopeList.take();
		} catch (InterruptedException ex)
		{
			log.error("Error in queueToModule method:\n" + ex);
		}
		fModule.speak(messageEnvelope);
		boolean messageSynchronized = messageEnvelope.getMessageSynchronized();
		if (messageSynchronized)
			fMessageSynchronizedIdentifier = messageEnvelope.getMessageIdentifier();
		else
			fMessageSynchronizedIdentifier = -1;
	}

	private void notifySpeakSync()
	{
		synchronized (this)
		{
			notifyAll();
		}
	}

	@Override
	public void run()
	{
		for (;;)
		{
			queueToModule();
		}
	}

	@Override
	public boolean setProperty(String key, String value)
	{
		try
		{
			if ((key == null) || (value == null))
				return (false);
			if (key.equals(SHConstants.PROP_MODULE_ROUTE_MESSAGE_TO_CHAR))
				fModuleRouteMessageToChar = value.equalsIgnoreCase("true");
			else if (key.equals(SHConstants.PROP_MODULE_CHAR_CAPITAL_INDICATOR))
			{
				// Possible values: none, pitch, icon, spell  
				// Audio icon not supported yet in SH
				if (value.equals(SHConstants.MODULE_CHAR_CAPITAL_INDICATOR_SPELL))
					fModuleCharCapitalIndicator = CapitalIndicator.SPELL;
				else if (value.equals(SHConstants.MODULE_CHAR_CAPITAL_INDICATOR_PITCH))
					fModuleCharCapitalIndicator = CapitalIndicator.PITCH;
				else if (value.equals(SHConstants.MODULE_CHAR_CAPITAL_INDICATOR_ICON))
					fModuleCharCapitalIndicator = CapitalIndicator.ICON;
				else
					fModuleCharCapitalIndicator = CapitalIndicator.NONE;
			} else if (key.equals(SHConstants.PROP_MODULE_SPELL_MODE))
				fModuleSpellMode = value.equalsIgnoreCase("true");
			else if (key.equals(SHConstants.PROP_MODULE_SSML_MODE))
				fModuleSSMLMode = value.equals("true");
			else if (key.equals(SHConstants.PROP_MODULE_SILENCE_BEFORE_MESSAGE))
				fSilenceBeforeMessage = Integer.valueOf(value);
			return (fModule.setProperty(key, value));
		} catch (Exception ex)
		{
			log.error("Error in setProperty method:\n" + ex);
			return (false);
		}
	}

	@Override
	public boolean setEventsListener(SHEventsListenerInterface eventsListener)
	{
		fEventsListener = eventsListener;
		return (true);
	}

	private class EntryEvents implements SHEventsListenerInterface
	{

		@Override
		public void messageBegin(String clientIdentifier, int messageIdentifier)
		{
			if (fEventsListener != null)
				fEventsListener.messageBegin(clientIdentifier, messageIdentifier);
		}

		@Override
		public void messageEnd(String clientIdentifier, int messageIdentifier)
		{
			if (messageIdentifier >= fMessageSynchronizedIdentifier)
			{
				fMessageSynchronizedInProcess = false;
				notifySpeakSync();
			}
			if (fEventsListener != null)
				fEventsListener.messageEnd(clientIdentifier, messageIdentifier);
		}

		@Override
		public void messagePause(String clientIdentifier, int messageIdentifier)
		{
			if (fEventsListener != null)
				fEventsListener.messagePause(clientIdentifier, messageIdentifier);
		}

		@Override
		public void messageResume(String clientIdentifier, int messageIdentifier)
		{
			if (fEventsListener != null)
				fEventsListener.messageResume(clientIdentifier, messageIdentifier);
		}

		@Override
		public void messageCancel(String clientIdentifier, int messageIdentifier)
		{
			if (fEventsListener != null)
				fEventsListener.messageCancel(clientIdentifier, messageIdentifier);
		}

		@Override
		public void messageAudioAvailable(String clientIdentifier, int messageIdentifier)
		{
			// Irrelevant generated at the server's client level	
		}

	}

}
