/*
* Copyright (c) Isaac Porat 2011 - 2013 all rights reserved.
* This file is released under the GNU General Public License, version 2, or (at your option) any later version.
* See the file SpeechHub_License.txt for more details.
*/

package sh.module;

import java.util.Date;

import log.common.CascadingLoggerInterface;

import sh.connections.SHConstants;
import sh.connections.SHMessageEnvelopeInterface;
import sh.connections.SHModuleInterface;
import sh.connections.SHEventsListenerInterface;
import sh.connections.SHConstants.MessageState;

/**
 * Module Providing full message events - this class must be included in the adaptors chain.
 * @author Isaac Porat
 *
 */
public class SHModuleEventsAdaptor extends SHModuleBase
{
	private SHEventsListenerInterface fEventsListener;
	private SHMessageEnvelopeInterface fLastMessageEnvelope;
	private boolean fPaused = false;
	private boolean fMessagePauseEventRequired = false;
	private boolean fModuleHandlesPauseResume = true;

	public SHModuleEventsAdaptor(SHModuleInterface module, CascadingLoggerInterface logger)
	{
		fModule = module;
		log = logger.clone(SHModuleEventsAdaptor.class);
	}

	@Override
	public boolean setEventsListener(SHEventsListenerInterface eventsListener)
	{
		fEventsListener = eventsListener;
		/*
			* The event handler is not plugged further
			* It is assumed that engines do not handle events in SH for consistancy
			* This may change in future implementation if required
			*/
		return (true);
	}

	@Override
	public boolean speak(SHMessageEnvelopeInterface messageEnvelope)
	{
		// This method must be called with messageSynchronized true	
		return (processSync(messageEnvelope));
	}

	private boolean processSync(SHMessageEnvelopeInterface messageEnvelope)
	{
		fLastMessageEnvelope = messageEnvelope;
		return (speakWithEvents(messageEnvelope));
	}

	@Override
	public boolean pause()
	{
		if (fPaused)
		{
			// Not an error but no point to proceed
			return (true);
		}
		fMessagePauseEventRequired = true;
		if (fModuleHandlesPauseResume == false)
		{
			if (fModule.pause()) // Command to engine
			{
				fPaused = true;
				return (true);
			}
			return (false);
		}
		return (true);
	}

	@Override
	public boolean resume()
	{
		if (fPaused == false)
		{
			log.error("Can't resume, TTS not paused");
			return (false);
		}
		// Fire the event first to ensure correct timing
		if (fEventsListener != null)
		{
			if (fLastMessageEnvelope != null)
				fEventsListener.messageResume(fLastMessageEnvelope.getClientIdentifier(),
						fLastMessageEnvelope.getMessageIdentifier());
		}
		if (fModuleHandlesPauseResume == false)
		{
			if (fModule.resume()) // Command to engine
			{
				fPaused = false;
				return (true);
			}
			return (false);
		}
		return (true);
	}

	@Override
	public boolean stop()
	{
		fModuleTimestamp.setCancelTimestampJustAfter(fLastMessageEnvelope.getTimestamp());
		return (fModule.cancel());
	}

	private boolean speakWithEvents(SHMessageEnvelopeInterface messageEnvelope)
	{
		Date timestamp = messageEnvelope.getTimestamp();
		MessageState messageState = messageEnvelope.getMessageState();
		int messageIdentifier = messageEnvelope.getMessageIdentifier();
		String clientIdentifier = messageEnvelope.getClientIdentifier();
		String message = messageEnvelope.getMessage();
		boolean messageCancelled = fModuleTimestamp.isMessageCancelled(timestamp, false);
		/* 
				* Note that the logic of which event notification is actually  fired is handled by the Event handler in the server
				*/
		if (fEventsListener != null)
		{
			if ((messageState.equals(MessageState.SINGLE)) || (messageState.equals(MessageState.BEGIN)))
			{
				if (messageCancelled == false)
					fEventsListener.messageBegin(clientIdentifier, messageIdentifier);
			}
		}
		boolean status = false;
		if (message != null)
		{
			if (messageCancelled == false)
				status = fModule.speak(messageEnvelope);
		}
		/*
		* Even if speakSync failed process the events - they may not be actually sent
		*/
		if (fEventsListener != null)
		{
			/*
			 *  It is important to call the pause event in this order to ensure correct operation
			* The logic is determined also if the events are actually enabled 			 
			*/
			if (fMessagePauseEventRequired)
				fEventsListener.messagePause(clientIdentifier, messageIdentifier);
			fMessagePauseEventRequired = false;
			/*
			 * Note resume event is fired in the resume method
			 */
			if ((messageState.equals(MessageState.SINGLE)) || (messageState.equals(MessageState.END)))
			{
				/* 
				* Note if not cancelled anymore reenable 
				* for efficiency - can only be done in the last END event
				*/
				if (fModuleTimestamp.isMessageCancelled(timestamp, true) == false)
					fEventsListener.messageEnd(clientIdentifier, messageIdentifier);
				else
					fEventsListener.messageCancel(clientIdentifier, messageIdentifier);
			}
		}
		return (status);
	}

	@Override
	public boolean setProperty(String key, String value)
	{
		if ((key == null) || (value == null))
			return (false);
		key = key.toLowerCase();
		if (key.equals(SHConstants.PROP_MODULE_HANDLES_PAUSE_RESUME))
			fModuleHandlesPauseResume = value.equalsIgnoreCase("true");
		return (fModule.setProperty(key, value));
	}

}
