/*
* Copyright (c) Isaac Porat 2011 - 2013 all rights reserved.
* This file is released under the GNU General Public License, version 2, or (at your option) any later version.
* See the file SpeechHub_License.txt for more details.
*/

package sh.module;

import java.util.concurrent.BlockingQueue;

import java.util.concurrent.LinkedBlockingQueue;

import core.utilities.StackTrace;

import log.common.CascadingLoggerInterface;

import sh.common.SHMethodEnvelope;
import sh.common.SHUpdateParameters;
import sh.connections.SHClientEngineVoiceSelectorHandlerInterface;
import sh.connections.SHCustomeHandlersInterface;
import sh.connections.SHMethodEnvelopeInterface;
import sh.connections.SHConstants;
import sh.connections.SHMessageEnvelopeInterface;
import sh.connections.SHModuleInterface;
import sh.connections.SHAudioDataHandlerInterface;
import sh.connections.SHEventsListenerInterface;
import sh.connections.SHUpdateParametersInterface;
import sh.connections.SHWatchdogListenerInterface;
import sh.engine.SHEngineWatchdog;

/**
 * Module isolation and engine restart
 * Provides engine isolation from the module main thread from most of the engine's methods 
 * This protects SH in both server and tts modes in the event of engine crash.
 * Engine crash is detected and using SHWatchdog attempt is made to restart the engine. 
 * While the engine is restarted, relevant threads are blocked
 * The functionality is achieved by Converting multi threads to a single thread 
 * Configured to work with engines supporting both speakSync and speakAsync with messageEnd event
 * This class requires to support only methods implemented by the  connector clients
 * The methods initialize, close, cancel, pause and resume needs to be implemented in a thread safe way
 * This functionality is achieved with some engine using SH audio player placed downstream 
 * Method get property is not accessing the engine itself as it takes its values from the engine's properties files
 * (but for the property engine.sample.rate which is called directly by the player so not called from this class)
 * All the methods    above are allowed to pass through this adaptor to ensure other methods interuption
 * Method setEventsListener also needs to be implemented in a thread safe way 
 * as it is called by the constructor
 * This method is only required if the engineConnector implementation 
 * supports only speakAsync (events listener required)
 *  
 * @author Isaac Porat
 *
 */
public class SHModuleIsolationRestartAdaptor extends SHModuleBase implements SHModuleInterface, Runnable,
		SHWatchdogListenerInterface
{
	private BlockingQueue<SHMethodEnvelopeInterface> fMessageList = new LinkedBlockingQueue<SHMethodEnvelopeInterface>();
	private SHEventsListenerInterface fEvents;
	private SHEngineWatchdog fWatchdog;
	private boolean fSpeakSyncRequired = false;
	private SHMessageEnvelopeInterface fCurrentMessageEnvelope;

	private boolean fGetEngineIdentifiersEnded = false;
	private String[] fGetEngineIdentifiersResult;

	private boolean fGetVoiceIdentifiersEnded = false;
	private String[] fGetVoiceIdentifiersResult;

	private boolean fGetVariantIdentifiersEnded = false;
	private String[] fGetVariantIdentifiersResult;

	private boolean fSpeakEnded = false;
	private boolean fSpeakStatus = false;

	private boolean fSetPropertyEnded = false;
	private boolean fSetPropertyStatus = false;

	private boolean fSetAudioDataHandlerEnded = false;
	private boolean fSetAudioDataHandlerStatus = false;

	private boolean fUpdateParametersDirectlyEnded = false;
	private boolean fUpdateParametersDirectlyStatus = false;

	private Thread fDispatchThread;
	private String fClientIdentifier = "unknown";
	private boolean fThreadForceInterrupted = false;
	private SHClientEngineVoiceSelectorHandlerInterface fClientEngineVoiceSelectorHandler;

	public SHModuleIsolationRestartAdaptor(SHModuleInterface module, boolean speakSyncRequired,
			CascadingLoggerInterface logger)
	{
		log = logger.clone(SHModuleIsolationRestartAdaptor.class);
		fModule = module;
		fSpeakSyncRequired = speakSyncRequired;
		if (fSpeakSyncRequired)
		{
			// Note - engine must provide speakAsync with messageEnd event for this to work
			fEvents = new Events();
			// Monitor messageEnd only  from downstream
			fModule.setEventsListener(fEvents);
		}
		fWatchdog = new SHEngineWatchdog(this, log);
		startThread();
	}

	@Override
	public void startThread()
	{
		notifySpeakSync();
		fDispatchThread = new Thread(this);
		fDispatchThread.setName(SHModuleIsolationRestartAdaptor.class.getName());
		fDispatchThread.start();
	}

	@Override
	public void interruptThread()
	{
		if (fDispatchThread != null)
		{
			fThreadForceInterrupted = true;
			fDispatchThread.interrupt();
		}
	}

	@Override
	public boolean setEventsListener(SHEventsListenerInterface eventsListener)
	{
		// Do nothing, disable this method  
		return (false);
	}

	public String[] getEngineIdentifiers()
	{
		SHMethodEnvelopeInterface methodEnvelope = new SHMethodEnvelope();
		methodEnvelope.setCommandType(SHConstants.MethodName.GET_ENGINE_IDENTIFIERS);
		synchronized (this)
		{
			fGetEngineIdentifiersEnded = false;
			commandToQueue(methodEnvelope);
			try
			{
				while (fGetEngineIdentifiersEnded == false)
					wait();
			} catch (InterruptedException e)
			{
			}
		}
		return (fGetEngineIdentifiersResult);
	}

	@Override
	public boolean speak(SHMessageEnvelopeInterface messageEnvelope)
	{
		// This method must be called with messageSynchronized true	
		synchronized (this)
		{
			fSpeakEnded = false;
			messageEnvelope.setMessageSynchronized(!fSpeakSyncRequired);
			commandToQueue(messageEnvelope);
			try
			{
				while (fSpeakEnded == false)
					wait();
			} catch (InterruptedException e)
			{
			}
			return (fSpeakStatus);
		}
	}

	private void notifySpeakSync()
	{
		fSpeakEnded = true;
		synchronized (this)
		{
			notifyAll();
		}
	}

	@Override
	public String[] getVoiceIdentifiers()
	{
		SHMethodEnvelopeInterface methodEnvelope = new SHMethodEnvelope();
		methodEnvelope.setCommandType(SHConstants.MethodName.GET_VOICE_IDENTIFIERS);
		synchronized (this)
		{
			fGetVoiceIdentifiersEnded = false;
			commandToQueue(methodEnvelope);
			try
			{
				while (fGetVoiceIdentifiersEnded == false)
					wait();
			} catch (InterruptedException e)
			{
			}
		}
		return (fGetVoiceIdentifiersResult);
	}

	@Override
	public String[] getVariantIdentifiers()
	{
		SHMethodEnvelopeInterface methodEnvelope = new SHMethodEnvelope();
		methodEnvelope.setCommandType(SHConstants.MethodName.GET_VARIANT_IDENTIFIERS);
		synchronized (this)
		{
			fGetVariantIdentifiersEnded = false;
			commandToQueue(methodEnvelope);
			try
			{
				while (fGetVariantIdentifiersEnded == false)
					wait();
			} catch (InterruptedException e)
			{
			}
		}
		return (fGetVariantIdentifiersResult);
	}

	@Override
	public boolean setProperty(String key, String value)
	{
		if (key == null)
			return (false);
		if (key.equals(SHConstants.PROP_MODULE_CLIENT_IDENTIFIER))
			fClientIdentifier = value;
		SHMethodEnvelopeInterface methodEnvelope = new SHMethodEnvelope();
		methodEnvelope.setCommandType(SHConstants.MethodName.SET_PROPERTY);
		methodEnvelope.setArgString1(key);
		methodEnvelope.setArgString2(value);
		synchronized (this)
		{
			fSetPropertyEnded = false;
			commandToQueue(methodEnvelope);
			try
			{
				while (fSetPropertyEnded == false)
					wait();
			} catch (InterruptedException e)
			{
			}
		}
		return (fSetPropertyStatus);
	}

	@Override
	public boolean updateParametersDirectly(SHUpdateParametersInterface updateParameters)
	{
		SHMethodEnvelopeInterface methodEnvelope = new SHMethodEnvelope();
		methodEnvelope.setCommandType(SHConstants.MethodName.UPDATE_PARAMETERS_DIRECTLY);
		methodEnvelope.setUpdateParametersDirectly(updateParameters);
		synchronized (this)
		{
			fUpdateParametersDirectlyEnded = false;
			commandToQueue(methodEnvelope);
			try
			{
				while (fUpdateParametersDirectlyEnded == false)
					wait();
			} catch (InterruptedException e)
			{
			}
		}
		return (fUpdateParametersDirectlyStatus);
	}

	@Override
	public boolean setAudioDataHandler(SHAudioDataHandlerInterface audioDataHandler)
	{
		SHMethodEnvelopeInterface methodEnvelope = new SHMethodEnvelope();
		methodEnvelope.setCommandType(SHConstants.MethodName.SET_AUDIO_DATA_HANDLER);
		methodEnvelope.setAudioDataHandler(audioDataHandler);
		synchronized (this)
		{
			fSetAudioDataHandlerEnded = false;
			commandToQueue(methodEnvelope);
			try
			{
				while (fSetAudioDataHandlerEnded == false)
					wait();
			} catch (InterruptedException e)
			{
			}
		}
		return (fSetAudioDataHandlerStatus);
	}

	public void commandToQueue(SHMethodEnvelopeInterface methodEnvelope)
	{
		try
		{
			fMessageList.put(methodEnvelope);
		} catch (InterruptedException ex)
		{
			log.error("Error in commandToQueue method:\n" + ex);
		}
	}

	public boolean queueToModule()
	{
		try
		{
			SHMethodEnvelopeInterface methodEnvelope = null;
			methodEnvelope = fMessageList.take();
			SHConstants.MethodName commandType = methodEnvelope.getCommandType();
			if (commandType.equals(SHConstants.MethodName.SPEAK))
			{
				fCurrentMessageEnvelope = (SHMessageEnvelopeInterface) methodEnvelope;
				fSpeakStatus = fModule.speak(fCurrentMessageEnvelope);
				if (fSpeakStatus)
					fWatchdog.resetEngineFailCounter();
				else
				{
					// Start the watchdog indicating failure
					fWatchdog.start(fCurrentMessageEnvelope.getUpdateParameters().getEngineIdentifier(),
							fCurrentMessageEnvelope.getUpdateParameters().getVoiceIdentifier(), "speech method failed",
							fDispatchThread);
					synchronized (this)
					{
						// Block thread to allow the timer to start properly
						wait(100);
					}
				}
				if (fSpeakSyncRequired == false)
					notifySpeakSync();
				return (true);
			} else if (commandType.equals(SHConstants.MethodName.GET_ENGINE_IDENTIFIERS))
			{
				fGetEngineIdentifiersResult = fModule.getEngineIdentifiers();
				fGetEngineIdentifiersEnded = true;
			} else if (commandType.equals(SHConstants.MethodName.GET_VOICE_IDENTIFIERS))
			{
				fGetVoiceIdentifiersResult = fModule.getVoiceIdentifiers();
				fGetVoiceIdentifiersEnded = true;
			} else if (commandType.equals(SHConstants.MethodName.GET_VARIANT_IDENTIFIERS))
			{
				fGetVariantIdentifiersResult = fModule.getVariantIdentifiers();
				fGetVariantIdentifiersEnded = true;
			} else if (commandType.equals(SHConstants.MethodName.SET_PROPERTY))
			{
				fSetPropertyStatus = fModule
						.setProperty(methodEnvelope.getArgString1(), methodEnvelope.getArgString2());
				fSetPropertyEnded = true;
			} else if (commandType.equals(SHConstants.MethodName.UPDATE_PARAMETERS_DIRECTLY))
			{
				fUpdateParametersDirectlyStatus = fModule.updateParametersDirectly(methodEnvelope
						.getUpdateParametersDirectly());
				fUpdateParametersDirectlyEnded = true;
			} else if (commandType.equals(SHConstants.MethodName.SET_AUDIO_DATA_HANDLER))
			{
				fSetAudioDataHandlerStatus = fModule.setAudioDataHandler(methodEnvelope.getAudioDataHandler());
				fSetAudioDataHandlerEnded = true;
			}
		} catch (InterruptedException ex)
		{
			if (fThreadForceInterrupted == false)
				log.error("Error in queueToModule method:\n" + ex);
			fThreadForceInterrupted = false;
			// Start the watchdog indicating failure
			fWatchdog.start(fCurrentMessageEnvelope.getUpdateParameters().getEngineIdentifier(),
					fCurrentMessageEnvelope.getUpdateParameters().getVoiceIdentifier(),
					"exception in queueToModule method", fDispatchThread);
			return (false);
		} catch (Exception ex)
		{
			log.error("Error in queueToModule method:\n" + StackTrace.stackToString(ex));
			// Start the watchdog indicating failure
			fWatchdog.start(fCurrentMessageEnvelope.getUpdateParameters().getEngineIdentifier(),
					fCurrentMessageEnvelope.getUpdateParameters().getVoiceIdentifier(),
					"exception in queueToModule method", fDispatchThread);
			return (true); // Do not interrupt the thread 
		}

		synchronized (this)
		{
			notifyAll();
		}
		return (true);
	}

	@Override
	public void run()
	{
		for (;;)
		{
			if (queueToModule() == false)
				return;
		}
	}

	public String getClientIdentifier()
	{
		return (fClientIdentifier);
	}

	public boolean selectEngineDirectly(String engineIdentifier)
	{
		SHUpdateParametersInterface engineUpdateParameters = SHUpdateParameters
				.getNewDisabledUpdateParameters(fClientIdentifier);
		engineUpdateParameters.setEngineIdentifier(engineIdentifier);
		// updateIdentifier by default 0 ensures unmarked update followed correctly by updateIdentifier  			fModule.updateParametersDirectly(engineUpdateParameters);
		return (fModule.updateParametersDirectly(engineUpdateParameters));
	}

	@Override
	public boolean setCustomeHandlers(SHCustomeHandlersInterface customeHandlers)
	{
		if (customeHandlers == null)
		{
			log.error("Error in setCustomeHandlers method: customHandlers null");
			return (false);
		}
		fClientEngineVoiceSelectorHandler = customeHandlers.getClientEngineVoiceSelectorHandler();
		return (fModule.setCustomeHandlers(customeHandlers));
	}

	public SHClientEngineVoiceSelectorHandlerInterface getClientEngineVoiceSelectorHandler()
	{
		return (fClientEngineVoiceSelectorHandler);
	}

	private class Events implements SHEventsListenerInterface
	{

		@Override
		public void messageBegin(String clientIdentifier, int messageIdentifier)
		{
			// Unused - not required	
		}

		@Override
		public void messageEnd(String clientIdentifier, int messageIdentifier)
		{
			notifySpeakSync();
		}

		@Override
		public void messagePause(String clientIdentifier, int messageIdentifier)
		{
			// Unused - not generated by engine		
		}

		@Override
		public void messageResume(String clientIdentifier, int messageIdentifier)
		{
			// Unused - not generated by engine		
		}

		@Override
		public void messageCancel(String clientIdentifier, int messageIdentifier)
		{
			// Unused - not generated by engine		
		}

		@Override
		public void messageAudioAvailable(String clientIdentifier, int messageIdentifier)
		{
			// Unused not generated by engine 
		}

	}

}
