/*
* Copyright (c) Isaac Porat 2011 - 2013 all rights reserved.
* This file is released under the GNU General Public License, version 2, or (at your option) any later version.
* See the file SpeechHub_License.txt for more details.
*/

package sh.module;

import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import log.common.CascadingLoggerInterface;

import sh.common.SHMessageEnvelope;
import sh.connections.SHConstants;
import sh.connections.SHItemReplacementInterface;
import sh.connections.SHMessageEnvelopeInterface;
import sh.connections.SHModuleInterface;
import sh.connections.SHConstants.CapitalIndicator;
import sh.connections.SHConstants.MessageState;
import sh.connections.SHConstants.MessageType;
import sh.connections.SHConstants.PunctuationLevel;
import sh.connections.SHUpdateParametersInterface;
import sh.settings.SHEnhancedPropertiesList;
import sh.settings.SHPropertiesFileLocations;
import sh.settings.SHReplacement;

/**
 * Module providing message splitting and text processing
 * @author Isaac Porat
 *
 */
public class SHModuleTextProcessingAdaptor extends SHModuleBase implements SHModuleInterface
{
	/*
	A paragraph is a contiguous sequence of non-blank lines separated by
	one or more blank lines. The lines in a paragraph can be terminated with any of the legal line termination character sequences:
	 \r, \r\n, or \n. 
	Examples:
	"a\r\rb"; // Mac
	"a\r\n\r\nb";          // Windows
	"a\n\nb";              // Unix
	*/
	private final static int MAX_MESSAGE_LENGTH = 500;
	private final static String PARAGRAPH_SPLIT_PATTERN = "(?:(?:\r\n|\r|\n)[ \\t]*){2,}";
	private final static String ELLIPSIS_INDICATOR = "qQf_";
	private final static String SENTENCE_SEPARATOR = "qQdot";
	// Note ignore 2 - 3 letter abbriviations and sentence indicators  not followed by space 
	private final static String SENTENCE_PREPROC_PATTERN = "([a-zA-Z]{4,}[\\.\\?\\!]+)[\\s]+";
	private final static String SENTENCE_PREPROC_REPLACEMENT = "$1" + SENTENCE_SEPARATOR;
	private final static String SENTENCE_SPLIT_PATTERN = SENTENCE_SEPARATOR;
	private Date fMessageTimestamp;
	private MessageType fMessageType = MessageType.STANDARD;
	private int fMessageIdentifier = 0;
	private int fMessagePriorityLevel = 4;
	private boolean fMessageSelfCancel = false;
	private SHUpdateParametersInterface fUpdateParameters;
	private int fMessageFragmentIdentifierCounter = 0;
	private Pattern fParagraphSplitPattern;
	private Pattern fSentencePreprocPattern;
	private Pattern fSentenceSplitPattern;
	private boolean fModuleHandlesSplit = true;
	private SHPropertiesFileLocations fPropertiesFileLocations;
	private boolean fModuleHandlesPunctuation = true;
	private boolean fModuleHandlesReplacement = true;
	private SHReplacement fReplacement;
	private String fSelectedEngineIdentifier = "";
	private SHItemReplacementInterface[] fPunctuationReplacementArray;
	private SHItemReplacementInterface[] fStandardReplacementArray;
	private SHItemReplacementInterface[] fCharReplacementArray;
	private SHItemReplacementInterface[] fKeyReplacementArray;
	private String fLanguageCode = SHConstants.ENGINE_LANGUAGE_CODE_DEFAULT;
	private PunctuationLevel fModulePunctuationLevel = PunctuationLevel.NONE;
	private CapitalIndicator fModuleCharCapitalIndicator = CapitalIndicator.NONE;
	private CapitalIndicator fCapitalIndicator = CapitalIndicator.NONE;
	private int fSilenceBeforeMessage = 0; // In ms

	public SHModuleTextProcessingAdaptor(SHModuleInterface module, CascadingLoggerInterface logger)
	{
		fModule = module;
		log = logger.clone(SHModuleTextProcessingAdaptor.class);
		fParagraphSplitPattern = Pattern.compile(PARAGRAPH_SPLIT_PATTERN, Pattern.MULTILINE);
		fSentencePreprocPattern = Pattern.compile(SENTENCE_PREPROC_PATTERN, Pattern.MULTILINE);
		fSentenceSplitPattern = Pattern.compile(SENTENCE_SPLIT_PATTERN, Pattern.MULTILINE);
		fPropertiesFileLocations = new SHPropertiesFileLocations(log);
	}

	private void initializeTextProcessing()
	{
		try
		{
			fReplacement = new SHReplacement();
			// Obtain punctuation message replacement 
			SHEnhancedPropertiesList propertiesList = new SHEnhancedPropertiesList();
			String[] propertiesFilePathArray = fPropertiesFileLocations.getPunctuationPropertiesFilesPath(
					fLanguageCode, fModulePunctuationLevel);
			for (int n = 0; n < propertiesFilePathArray.length; n++)
			{
				fReplacement.loadProperties(propertiesList, propertiesFilePathArray[n]);
			}
			fPunctuationReplacementArray = fReplacement.makeReplacementArray(propertiesList);
			// Obtain standard message replacement 
			propertiesList.clear();
			propertiesFilePathArray = fPropertiesFileLocations.getStandardPropertiesFilesPath(
					fSelectedEngineIdentifier, fLanguageCode);
			for (int n = 0; n < propertiesFilePathArray.length; n++)
			{
				fReplacement.loadProperties(propertiesList, propertiesFilePathArray[n]);
			}
			fStandardReplacementArray = fReplacement.makeReplacementArray(propertiesList);
			// Obtain char message replacement 
			propertiesList.clear();
			propertiesFilePathArray = fPropertiesFileLocations.getCharPropertiesFilesPath(fSelectedEngineIdentifier,
					fLanguageCode);
			for (int n = 0; n < propertiesFilePathArray.length; n++)
			{
				fReplacement.loadProperties(propertiesList, propertiesFilePathArray[n]);
			}
			fCharReplacementArray = fReplacement.makeReplacementArray(propertiesList);
			// Obtain key message replacement 
			propertiesList.clear();
			propertiesFilePathArray = fPropertiesFileLocations.getKeyPropertiesFilesPath(fSelectedEngineIdentifier,
					fLanguageCode);
			for (int n = 0; n < propertiesFilePathArray.length; n++)
			{
				fReplacement.loadProperties(propertiesList, propertiesFilePathArray[n]);
			}
			fKeyReplacementArray = fReplacement.makeReplacementArray(propertiesList);
		} catch (Exception ex)
		{
			log.error("Error in initialize method:\n" + ex);
		}
	}

	private void setLanguage(String languageCode)
	{
		if (languageCode.equals(fLanguageCode))
			return; // Processed already
		/* 
		* language code according to standard is case insensitive but 
		* in SH processing is always in lower case
		* Language dependent replacement tables must be defined in lower case
		*/
		languageCode = languageCode.toLowerCase();
		fLanguageCode = languageCode;
		if (fPropertiesFileLocations.isLanguageReplacementDirectoryAvailable(fLanguageCode) == false)
		{
			// Language directory not available module cannot handle punctuation 
			setProperty(SHConstants.PROP_MODULE_HANDLES_PUNCTUATION, "false");
		}
		if (fSelectedEngineIdentifier.equals("") == false)
			initializeTextProcessing();
	}

	@Override
	public boolean speak(SHMessageEnvelopeInterface messageEnvelope)
	{
		// This method must always be called with messageSynchronized true
		return (processSync(messageEnvelope));
	}

	private boolean processSync(SHMessageEnvelopeInterface messageEnvelope)
	{
		String message = messageEnvelope.getMessage();
		fMessageType = messageEnvelope.getMessageType();
		fCapitalIndicator = CapitalIndicator.NONE;
		fModuleCharCapitalIndicator = messageEnvelope.getCapitalIndicator();
		fMessageTimestamp = messageEnvelope.getTimestamp();
		fMessageIdentifier = messageEnvelope.getMessageIdentifier();
		fMessagePriorityLevel = messageEnvelope.getPriorityLevel();
		fMessageSelfCancel = messageEnvelope.getSelfCancel();
		fSilenceBeforeMessage = messageEnvelope.getSilenceBeforeMessage();
		fUpdateParameters = messageEnvelope.getUpdateParameters();
		fMessageFragmentIdentifierCounter = 0;
		setLanguage(fUpdateParameters.getLanguageCode());
		processMessage(message);
		return (true);
	}

	private void messageFragmentToQueue(MessageState messageState, String message, boolean forceUpdate)
	{
		SHMessageEnvelopeInterface messageEnvelope = new SHMessageEnvelope(fMessageIdentifier, message);
		messageEnvelope.setTimestamp(fMessageTimestamp);
		messageEnvelope.setMessageState(messageState);
		messageEnvelope.setMessageType(fMessageType);
		messageEnvelope.setCapitalIndicator(fCapitalIndicator);
		messageEnvelope.setSilenceBeforeMessage(fSilenceBeforeMessage);
		messageEnvelope.setMessageFragmentIdentifier(fMessageFragmentIdentifierCounter++);
		messageEnvelope.setPriorityLevel(fMessagePriorityLevel);
		messageEnvelope.setSelfCancel(fMessageSelfCancel);
		SHUpdateParametersInterface updateParameters = fUpdateParameters.getClone();
		if (forceUpdate)
			updateParameters.markNewUpdate();
		messageEnvelope.setUpdateParameters(updateParameters);
		fModule.speak(messageEnvelope);
	}

	private void processMessage(String message)
	{
		// Messages are never discarded to preserve events integrity
		if ((message == null) || (fModuleTimestamp.isMessageCancelled(fMessageTimestamp, false)))
		{
			messageFragmentToQueue(MessageState.SINGLE, message, false);
			return;
		}
		message = message.trim();
		if (message.equals(""))
		{
			messageFragmentToQueue(MessageState.SINGLE, message, false);
			return;
		} else if (fMessageType.equals(MessageType.STANDARD))
		{
			boolean split = false;
			if (message.length() > MAX_MESSAGE_LENGTH)
			{
				split = true;
				log.notify("Message length beyond limit, ensured split to paragraphs");
				// Insert paragraph dividers at no longer than  MAX_MESSAGE_LENGTH
				message = ensureMessageLength(message);
			}
			if ((fModuleHandlesSplit) || split)
				splitMessageToParagraphs(message);
			else
				processStandardMessage(MessageState.SINGLE, message);
		} else if (fMessageType.equals(MessageType.SSML))
		{
			// No processing send directly to queue
			messageFragmentToQueue(MessageState.SINGLE, message, false);
			return;
		} else if (fMessageType.equals(MessageType.CHAR))
		{
			message = processCharMessage(message);
			messageFragmentToQueue(MessageState.SINGLE, message, false);
			return;
		} else if (fMessageType.equals(MessageType.KEY))
		{
			message = processKeyMessage(message);
			messageFragmentToQueue(MessageState.SINGLE, message, false);
			return;
		} else if (fMessageType.equals(MessageType.SPELL))
		{
			processSpellMessage(message);
			return;
		}
	}

	private void splitMessageToParagraphs(String message)
	{
		String[] paragraphs = fParagraphSplitPattern.split(message, 0);
		for (int pCounter = 0; pCounter < paragraphs.length; pCounter++)
		{
			splitParagraphToSentences(pCounter, paragraphs.length, paragraphs[pCounter]);
		}
	}

	private void splitParagraphToSentences(int pCounter, int pLength, String message)
	{
		if (message.contains("..."))
			message = message.replaceAll("\\.\\.\\.", ELLIPSIS_INDICATOR);
		Matcher matcher = fSentencePreprocPattern.matcher(message);
		message = matcher.replaceAll(SENTENCE_PREPROC_REPLACEMENT);
		String[] sentences = fSentenceSplitPattern.split(message, 0);
		String sentence = "";
		MessageState messageState;
		for (int sCounter = 0; sCounter < sentences.length; sCounter++)
		{
			if ((pLength == 1) && (sentences.length == 1))
				messageState = MessageState.SINGLE;
			else if ((pCounter == 0) && (sCounter == 0))
				messageState = MessageState.BEGIN;
			else if ((pCounter == pLength - 1) && (sCounter == sentences.length - 1))
				messageState = MessageState.END;
			else
				messageState = MessageState.MIDDLE;
			if ((messageState.equals(MessageState.MIDDLE))
					&& (fModuleTimestamp.isMessageCancelled(fMessageTimestamp, false)))
				continue; // Middle fragments are not important for events integrity
			sentence = sentences[sCounter];
			if (sentence.contains("qQ"))
				sentence = sentence.replaceAll(ELLIPSIS_INDICATOR, "...");
			processStandardMessage(messageState, sentence);
		}
	}

	private void processStandardMessage(MessageState messageState, String message)
	{
		message = replaceText(message);
		messageFragmentToQueue(messageState, message, false);
	}

	private String replaceText(String message)
	{
		if (fReplacement == null)
		{
			log.error("Error in replaceText method, replacement structure null.\n  Probably because the module was not initialized correctly");
			return (message);
		}
		if (fModuleHandlesPunctuation)
			message = fReplacement.replaceAll(fPunctuationReplacementArray, message);
		if (fModuleHandlesReplacement)
			message = fReplacement.replaceAll(fStandardReplacementArray, message);
		return (message);
	}

	public String processCharMessage(String message)
	{
		if (fReplacement == null)
		{
			log.error("Error in processCharMessage method, replacement structure null.\n  Probably because the module was not initialized correctly");
			return (message);
		}
		// Check for capital letter
		if (message.matches("[A-Z]"))
		{
			if (fModuleCharCapitalIndicator.equals(CapitalIndicator.SPELL))
				message = "CAP_STRING " + message;
			fCapitalIndicator = fModuleCharCapitalIndicator;
			// Note: pitch capital indicator is handled in SHModuleAudioSpeechPlayerAdaptor 
		}
		message = fReplacement.replaceAll(fCharReplacementArray, message);
		return (message);
	}

	public String processKeyMessage(String message)
	{
		if (fReplacement == null)
		{
			log.error("Error in processKeyMessage method, replacement structure null.\n  Probably because the module was not initialized correctly");
			return (message);
		}
		message = fReplacement.replaceAll(fKeyReplacementArray, message);
		return (message);
	}

	private void processSpellMessage(String message)
	{
		try
		{
			fMessageType = MessageType.CHAR;
			MessageState messageState;
			String charString;
			int messageLength = message.length();
			for (int n = 0; n < messageLength; n++)
			{
				char ch = message.charAt(n);
				charString = processCharMessage(String.valueOf(ch));
				if (n == 0)
				{
					if (messageLength > 1)
						messageState = MessageState.BEGIN;
					else
						messageState = MessageState.SINGLE;
				} else if (n == messageLength - 1)
					messageState = MessageState.END;
				else
					messageState = MessageState.MIDDLE;
				fCapitalIndicator = CapitalIndicator.NONE;
				if (charString.matches("[A-Z]")) // Check for capital letter
					fCapitalIndicator = fModuleCharCapitalIndicator;
				/*
				 * Spell mode (which includes delay) applies to all characters
				 * Actual delay duration between messages comprises of the time taken for the first audio data to be available after text to audio conversion together with pause between messages specified below
				 * The delay below specified as 1ms to disable message peek functionality in the player
				 * As the delay in this case is a function of text to audio conversion which in turn relates to an extent  on rate
				 * it is somewhat proportional to rate which is a feature required  
				 */
				// Negligible delay; a value above 0 is required to disable message peek functionality in the player
				fSilenceBeforeMessage = 1;
				messageFragmentToQueue(messageState, charString, true);
			}
		} catch (Exception ex)
		{
			log.error("Error in processSpellMessage:\n" + ex);
		} finally
		{
			fCapitalIndicator = CapitalIndicator.NONE;
			fSilenceBeforeMessage = 0;
		}
	}

	@Override
	public boolean setProperty(String key, String value)
	{
		if ((key == null) || (value == null))
			return (false);
		if (key.equals(SHConstants.PROP_MODULE_ENGINE_IDENTIFIER))
		{
			if (value.equals(fSelectedEngineIdentifier) == false)
			{
				fSelectedEngineIdentifier = value;
				initializeTextProcessing();
			}
		} else if (key.equals(SHConstants.PROP_MODULE_HANDLES_SPLIT))
			fModuleHandlesSplit = value.equalsIgnoreCase("true");
		else if (key.equals(SHConstants.PROP_MODULE_HANDLES_PUNCTUATION))
		{
			fModuleHandlesPunctuation = value.equalsIgnoreCase("true");
			if (fModuleHandlesPunctuation)
			{
				// Check if language directory exists
				if (fPropertiesFileLocations.isLanguageReplacementDirectoryAvailable(fLanguageCode) == false)
				{
					// Language directory not available module cannot handle punctuation 
					fModuleHandlesPunctuation = false;
					value = "false";
				}
			}
		} else if (key.equals(SHConstants.PROP_MODULE_PUNCTUATION_LEVEL))
		{
			// Possible values: none, some, most, all 
			if (value.equalsIgnoreCase(SHConstants.ENGINE_PUNCTUATION_LEVEL_SOME))
				fModulePunctuationLevel = PunctuationLevel.SOME;
			else if (value.equalsIgnoreCase(SHConstants.ENGINE_PUNCTUATION_LEVEL_MOST))
				fModulePunctuationLevel = PunctuationLevel.MOST;
			else if (value.equalsIgnoreCase(SHConstants.ENGINE_PUNCTUATION_LEVEL_ALL))
				fModulePunctuationLevel = PunctuationLevel.ALL;
			else
				fModulePunctuationLevel = PunctuationLevel.NONE;
			if (fSelectedEngineIdentifier.equals("") == false)
				initializeTextProcessing();
		} else if (key.equals(SHConstants.PROP_MODULE_HANDLES_REPLACEMENT))
			fModuleHandlesReplacement = value.equalsIgnoreCase("true");
		return (fModule.setProperty(key, value));
	}

	private String ensureMessageLength(String message)
	{
		// This method assumes that fragments can be formed by splitting around spaces otherwise it would not work
		StringBuilder messageBuilder = new StringBuilder();
		List<String> messageList = new ArrayList<String>();
		String[] paragraphs = fParagraphSplitPattern.split(message, 0);
		for (int p = 0; p < paragraphs.length; p++)
		{
			String paragraph = paragraphs[p];
			if (paragraph.length() < MAX_MESSAGE_LENGTH)
			{
				messageBuilder.append(paragraph + "\n\n");
				break;
			}
			/*
			* Paragraph too long insert paragraph dividers
			* Split to fragments divided by spaces
			* As these long messages are rare no compiled pattern used
			*/
			String[] fragments = paragraphs[p].split("[ \n]+");
			StringBuilder paraBuilder = new StringBuilder();
			for (int f = 0; f < fragments.length; f++)
			{
				if (paraBuilder.length() + fragments[f].length() > (MAX_MESSAGE_LENGTH - 2)) // Allow for paragraph divider
				{
					messageList.add(paraBuilder.toString() + "\n\n");
					paraBuilder.setLength(0);
				}
				paraBuilder.append(fragments[f] + " ");
			}
			if (paraBuilder.length() > 0)
			{
				messageList.add(paraBuilder.toString() + "\n\n");
				paraBuilder.setLength(0);
			}
		}
		log.notify("Number of paragraphs generated: " + messageList.size());
		for (int m = 0; m < messageList.size(); m++)
		{
			messageBuilder.append(messageList.get(m));
		}
		return (messageBuilder.toString());
	}

}
