/*
* Copyright (c) Isaac Porat 2011 - 2013 all rights reserved.
* This file is released under the GNU General Public License, version 2, or (at your option) any later version.
* See the file SpeechHub_License.txt for more details.
*/

package sh.module.server;

import java.util.HashMap;
import java.util.Map;

import log.common.CascadingLoggerInterface;

import sh.connections.SHConstants;
import sh.connections.SHCustomeHandlersInterface;
import sh.connections.SHMessageEnvelopeInterface;
import sh.connections.SHModuleInterface;
import sh.connections.SHModuleTimestampInterface;
import sh.connections.SHAudioDataHandlerInterface;
import sh.connections.SHEventsListenerInterface;
import sh.connections.SHUpdateParametersInterface;
import sh.module.SHModuleBase;
import sh.server.SHMultiplexer;

/**
 * Server  TTS module multiplexer adaptor
 * 
 * @author Isaac Porat
 */
public class SHModuleMultiplexerAdaptor extends SHModuleBase implements SHModuleInterface
{
	@SuppressWarnings("unused")
	private CascadingLoggerInterface log;
	private String fClientIdentifier;
	private SHMultiplexer fMultiplexer;
	private Map<String, String> fModuleProperties;
	private boolean fModuleHandlesPauseResume = true;

	public SHModuleMultiplexerAdaptor(String clientIdentifier, SHMultiplexer multiplexer,
			CascadingLoggerInterface logger)
	{
		log = logger.clone(SHModuleMultiplexerAdaptor.class);
		fClientIdentifier = clientIdentifier;
		fMultiplexer = multiplexer;
		fModuleProperties = new HashMap<String, String>();
	}

	@Override
	public boolean configureTiming(SHModuleTimestampInterface moduleTimestamp, boolean masterCancel)
	{
		// Not applicable
		return (false);
	}

	@Override
	public boolean initialize()
	{
		// Not applicable	
		return (false);
	}

	@Override
	public boolean close()
	{
		// Not applicable	
		return (false);
	}

	@Override
	public boolean setPriority(int priorityLevel, boolean selfCancel)
	{
		// Not applicable, priority information embedded already in messageEnvelope
		return (false);
	}

	@Override
	public String[] getEngineIdentifiers()
	{
		return (fMultiplexer.getEngineIdentifiers(fClientIdentifier));
	}

	@Override
	public String getSelectedEngineIdentifier()
	{
		return (fMultiplexer.getSelectedEngineIdentifier(fClientIdentifier));
	}

	@Override
	public String getSelectedLanguageCode()
	{
		return (fMultiplexer.getLanguageCode(fClientIdentifier));
	}

	@Override
	public String[] getVoiceIdentifiers()
	{
		return (fMultiplexer.getVoiceIdentifiers(fClientIdentifier));
	}

	@Override
	public String getSelectedVoiceIdentifier()
	{
		return (fMultiplexer.getSelectedVoiceIdentifier(fClientIdentifier));
	}

	@Override
	public String[] getVariantIdentifiers()
	{
		return (fMultiplexer.getVariantIdentifiers(fClientIdentifier));
	}

	@Override
	public String getSelectedVariantIdentifier()
	{
		return (fMultiplexer.getSelectedVariantIdentifier(fClientIdentifier));
	}

	@Override
	public boolean speak(SHMessageEnvelopeInterface messageEnvelope)
	{
		// This method must be called with messageSynchronized true	
		return (processSync(messageEnvelope));
	}

	public boolean processSync(SHMessageEnvelopeInterface messageEnvelope)
	{
		messageEnvelope.setClientIdentifier(fClientIdentifier);
		return (fMultiplexer.processSync(fClientIdentifier, messageEnvelope));
	}

	@Override
	public boolean pause()
	{
		return (fMultiplexer.pause(fClientIdentifier, fModuleHandlesPauseResume));
	}

	@Override
	public boolean resume()
	{
		return (fMultiplexer.resume(fClientIdentifier, fModuleHandlesPauseResume));
	}

	@Override
	public boolean stop()
	{
		return (fMultiplexer.stop(fClientIdentifier));
	}

	@Override
	public boolean cancel()
	{
		return (fMultiplexer.cancel(fClientIdentifier));
	}

	@Override
	public float getVolume()
	{
		return (fMultiplexer.getVolume(fClientIdentifier));
	}

	@Override
	public float getRate()
	{
		return (fMultiplexer.getRate(fClientIdentifier));
	}

	@Override
	public float getPitch()
	{
		return (fMultiplexer.getPitch(fClientIdentifier));
	}

	@Override
	public String getProperty(String key)
	{
		// Properties map may need to move in future to multiplexer if properties forwarding is required
		return (fModuleProperties.get(key));
	}

	@Override
	public boolean setProperty(String key, String value)
	{
		if ((key == null) || (value == null))
			return (false);
		key = key.toLowerCase();
		if (key.equals(SHConstants.PROP_MODULE_HANDLES_PAUSE_RESUME))
			fModuleHandlesPauseResume = value.equalsIgnoreCase("true");
		// Transfer required properties to multiplexer for settings of client parameters
		if ((key.equals(SHConstants.PROP_MODULE_HANDLES_PUNCTUATION))
				|| (key.equals(SHConstants.PROP_MODULE_PUNCTUATION_LEVEL))
				|| (key.equals(SHConstants.PROP_MODULE_SSML_MODE)) || (key.equals(SHConstants.PROP_ENGINE_USE_SONIC))
				|| (key.equals(SHConstants.PROP_MODULE_CHAR_CAPITAL_INDICATOR_PITCH_LEVEL)))
		{
			fMultiplexer.modulePropertiesToClientParameters(fClientIdentifier, key, value);
		}
		fModuleProperties.put(key, value);
		return (true);
	}

	@Override
	public boolean setEventsListener(SHEventsListenerInterface eventsListener)
	{
		// Not applicable, done by the multiplexer 
		return (false);
	}

	@Override
	public boolean setAudioDataHandler(SHAudioDataHandlerInterface audioDataHandler)
	{
		// Not applicable, done at the engine level
		return (false);
	}

	@Override
	public boolean setCustomeHandlers(SHCustomeHandlersInterface customeHandlers)
	{
		// Not applicable, custome handlers set by module wrapper
		return (true);
	}

	@Override
	public boolean updateParametersDirectly(SHUpdateParametersInterface updateParameters)
	{
		return (fMultiplexer.updateParametersDirectly(fClientIdentifier, updateParameters));
	}

	@Override
	public boolean isUpdateAccepted(int updateIdentifier)
	{
		// Not applicable
		return (false);
	}

	@Override
	public boolean setUpdateIdentifier(int updateIdentifier)
	{
		// Not applicable
		return (false);
	}

}
