/*
* Copyright (c) Isaac Porat 2011 - 2013 all rights reserved.
* This file is released under the GNU General Public License, version 2, or (at your option) any later version.
* See the file SpeechHub_License.txt for more details.
*/

package sh.module.server;

import log.common.CascadingLoggerInterface;
import sh.common.SHCustomeHandlersGlobal;
import sh.connections.SHConstants;
import sh.connections.SHMessageEnvelopeInterface;
import sh.connections.SHModuleInterface;
import sh.connections.SHModuleTimestampInterface;
import sh.module.SHModuleBase;
import sh.module.SHModuleEntryAdaptor;
import sh.module.SHModuleTextProcessingAdaptor;
import sh.server.SHMultiplexer;
import sh.settings.SHEnhancedPropertiesList;
import sh.settings.SHModuleConfiguration;
import sh.settings.SHPropertiesFileLocations;

/**
 * Server  client TTS module wrapper 
 * This class is used to wrap all adapttors required per client 
 * @author Isaac Porat
 */
public class SHModuleServerClientWrapper extends SHModuleBase implements SHModuleInterface
{
	private String fClientIdentifier;
	private SHMultiplexer fTMMultiplexer;

	public SHModuleServerClientWrapper(String clientIdentifier, SHMultiplexer ttsMultiplexer,
			SHModuleTimestampInterface moduleTimestamp, String settingsBaseDirectory, CascadingLoggerInterface logger)
	{
		log = logger.clone(SHModuleServerClientWrapper.class);
		fClientIdentifier = clientIdentifier;
		fTMMultiplexer = ttsMultiplexer;
		fModuleTimestamp = moduleTimestamp;
		wrap();
		initializeSettings();
	}

	private void initializeSettings()
	{
		try
		{
			setProperty(SHConstants.PROP_COMPONENTS_CONFIGURATION, SHConstants.COMPONENTS_CONFIGURATION_SERVER);
			configureTiming(fModuleTimestamp, true);
			SHModuleConfiguration configuration = new SHModuleConfiguration();
			SHPropertiesFileLocations propertiesFileLocations = new SHPropertiesFileLocations(log);
			SHEnhancedPropertiesList propertiesList = new SHEnhancedPropertiesList();
			String[] propertiesFilePathArray = propertiesFileLocations
					.getConfigurationModuleCommonPropertiesFilesPath();
			for (int n = 0; n < propertiesFilePathArray.length; n++)
			{
				configuration.loadProperties(propertiesList, propertiesFilePathArray[n]);
			}
			configuration.transferPropertiesToModule(this, propertiesList);
			// Set custome handlers
			fModule.setCustomeHandlers(SHCustomeHandlersGlobal.getInstance());
		} catch (Exception ex)
		{
			log.error("Error in initializeSettings method:\n" + ex);
		}
	}

	private void wrap()
	{
		// Client multiplexer
		SHModuleInterface moduleAdaptor = new SHModuleMultiplexerAdaptor(fClientIdentifier, fTMMultiplexer, log);
		// Text processing including splitting to paragraphs, sentences and text replacement			
		moduleAdaptor = new SHModuleTextProcessingAdaptor(moduleAdaptor, log);
		// Module entry implementing messages buffer
		fModule = new SHModuleEntryAdaptor(moduleAdaptor, fTMMultiplexer.getClientParameters(fClientIdentifier), log);
	}

	@Override
	public boolean speak(SHMessageEnvelopeInterface messageEnvelope)
	{
		try
		{
			return (fModule.speak(messageEnvelope));
		} catch (Exception ex)
		{
			log.error("Error in speak method:\n" + ex);
			return (false);
		}
	}

	@Override
	public boolean speak(int messageIdentifier, String message, boolean speakSync)
	{
		try
		{
			return (fModule.speak(messageIdentifier, message, speakSync));
		} catch (Exception ex)
		{
			log.error("Error in speak method:\n" + ex);
			return (false);
		}
	}

	@Override
	public boolean speakChar(int messageIdentifier, String message, boolean speakSync)
	{
		try
		{
			return (fModule.speakChar(messageIdentifier, message, speakSync));
		} catch (Exception ex)
		{
			log.error("Error in speakChar method:\n" + ex);
			return (false);
		}
	}

	@Override
	public boolean speakKey(int messageIdentifier, String message, boolean speakSync)
	{
		try
		{
			return (fModule.speakKey(messageIdentifier, message, speakSync));
		} catch (Exception ex)
		{
			log.error("Error in speakKey method:\n" + ex);
			return (false);
		}
	}

}
