/*
* Copyright (c) Isaac Porat 2011 - 2013 all rights reserved.
* This file is released under the GNU General Public License, version 2, or (at your option) any later version.
* See the file SpeechHub_License.txt for more details.
*/

package sh.module.server;

import core.utilities.StackTrace;
import log.common.CascadingLoggerInterface;
import sh.common.SHCustomeHandlersGlobal;
import sh.connections.SHConstants;
import sh.connections.SHEngineInterface;
import sh.connections.SHMessageEnvelopeInterface;
import sh.connections.SHModuleInterface;
import sh.connections.SHEventsListenerInterface;
import sh.module.SHModuleAudioSpeechPlayerAdaptor;
import sh.module.SHModuleBase;
import sh.module.SHModuleEngineAdaptor;
import sh.module.SHModuleEventsAdaptor;
import sh.module.SHModuleIsolationRestartAdaptor;
import sh.module.SHModuleSpeakSyncAdaptor;
import sh.settings.SHEnhancedPropertiesList;
import sh.settings.SHModuleConfiguration;
import sh.settings.SHPropertiesFileLocations;

/**
 * Server  TTS engine module wrapper 
 * This class is used to wrap all adapttors required in a server module
 * @author Isaac Porat
 */
public class SHModuleServerEngineWrapper extends SHModuleBase implements SHModuleInterface

{
	private SHEngineInterface fEngine;
	private String fEngineIdentifier;
	private SHEventsListenerInterface fEventsListener;
	private String fModuleClientIdentifier;

	public SHModuleServerEngineWrapper(SHEngineInterface engine, CascadingLoggerInterface logger)
	{
		fEngine = engine;
		log = logger.clone(SHModuleServerEngineWrapper.class);
		wrap();
		initializeSettings();
	}

	private void initializeSettings()
	{
		try
		{
			setProperty(SHConstants.PROP_COMPONENTS_CONFIGURATION, SHConstants.COMPONENTS_CONFIGURATION_SERVER);
			SHModuleConfiguration configuration = new SHModuleConfiguration();
			SHPropertiesFileLocations propertiesFileLocations = new SHPropertiesFileLocations(log);
			SHEnhancedPropertiesList propertiesList = new SHEnhancedPropertiesList();
			String[] propertiesFilePathArray = propertiesFileLocations
					.getConfigurationModuleCommonPropertiesFilesPath();
			for (int n = 0; n < propertiesFilePathArray.length; n++)
			{
				configuration.loadProperties(propertiesList, propertiesFilePathArray[n]);
			}
			fEngineIdentifier = fEngine.getProperty(SHConstants.PROP_ENGINE_IDENTIFIER_DEFAULT);
			if (fEngineIdentifier == null)
			{
				log.error("Error in wrap method:\n engineConnector property null: "
						+ SHConstants.PROP_ENGINE_IDENTIFIER_DEFAULT
						+ "\nThis property must be defined for the engine to access its configuration properties\n");
			}
			configuration.transferPropertiesToModule(this, propertiesList);
			// Set custome handlers
			fModule.setCustomeHandlers(SHCustomeHandlersGlobal.getInstance());
		} catch (Exception ex)
		{
			log.error("Error in initializeSettings method:\n" + ex);
		}
	}

	private void wrap()
	{
		// Engine wrapper 
		SHModuleInterface moduleAdaptor = new SHModuleEngineAdaptor(fEngine, log);

		// Audio player
		String property = moduleAdaptor.getProperty(SHConstants.PROP_ENGINE_SUPPORTS_AUDIO_DATA);
		if ((property != null) && (property.equals("true")))
		{
			moduleAdaptor = new SHModuleAudioSpeechPlayerAdaptor(moduleAdaptor, log);
		}

		// Note - speakSync support requires the engine to provide speakAsync with messageEnd event
		property = moduleAdaptor.getProperty(SHConstants.PROP_ENGINE_SUPPORTS_SPEAK_SYNC);
		boolean speakSyncRequired = ((property == null) || (property.equals("false")));

		property = moduleAdaptor.getProperty(SHConstants.PROP_ENGINE_REQUIRES_ISOLATION_RESTART);
		boolean isolationRestartRequired = (property != null) && (property.equals("true"));

		if (isolationRestartRequired == false)
		{
			if (speakSyncRequired)
			{
				moduleAdaptor = new SHModuleSpeakSyncAdaptor(moduleAdaptor, log);
			}
		} else
		// Isolation restart required
		{
			moduleAdaptor = new SHModuleIsolationRestartAdaptor(moduleAdaptor, speakSyncRequired, log);
		}
		// Above this line speakSync must be provided
		// Events support
		fModule = new SHModuleEventsAdaptor(moduleAdaptor, log);
	}

	@Override
	public boolean speak(SHMessageEnvelopeInterface messageEnvelope)
	{
		try
		{
			return (fModule.speak(messageEnvelope));
		} catch (Exception ex)
		{
			log.error("Error in speak method:\n" + StackTrace.stackToString(ex));
			return (false);
		}
	}

	@Override
	public boolean setEventsListener(SHEventsListenerInterface eventsListener)
	{
		try
		{
			// For server-engine module events router listener must be inserted only once
			if (fEventsListener != null)
				return (true);
			boolean status = fModule.setEventsListener(eventsListener);
			if (status)
				fEventsListener = eventsListener;
			return (status);
		} catch (Exception ex)
		{
			log.error("Error in setEventsListener method:\n" + ex);
			return (false);
		}
	}

	@Override
	public boolean setProperty(String key, String value)
	{
		if (key == null)
			return (false);
		// Property used often hence local variable
		if (key.equals(SHConstants.PROP_MODULE_CLIENT_IDENTIFIER))
			fModuleClientIdentifier = value;
		return (fModule.setProperty(key, value));
	}

	@Override
	public String getProperty(String key)
	{
		// Property used often hence local variable
		if (key == null)
			return (null);
		if (key.equals(SHConstants.PROP_MODULE_CLIENT_IDENTIFIER))
			return (fModuleClientIdentifier);
		return (fModule.getProperty(key));
	}

}
