/*
* Copyright (c) Isaac Porat 2011 - 2013 all rights reserved.
* This file is released under the GNU General Public License, version 2, or (at your option) any later version.
* See the file SpeechHub_License.txt for more details.
*/

package sh.module.tts;

import log.common.CascadingLoggerInterface;
import sh.common.SHMessageQueue;
import sh.connections.SHConstants;
import sh.connections.SHCustomeHandlersInterface;
import sh.connections.SHMessageEnvelopeInterface;
import sh.connections.SHMessageQueueHandlerInterface;
import sh.connections.SHModuleInterface;
import sh.module.SHModuleBase;

/**
 * Implements message queue and pause resume functionality
 * Used for standalone TTS implementations
 * In multi client application this functionality is implemented in the multiplexer 
 * @author Isaac Porat
 *
 */
public class SHModulePauseResumeQueueAdaptor extends SHModuleBase implements Runnable, SHMessageQueueHandlerInterface
{
	private final static String DEFAULT_CLIENT_IDENTIFIER = "default-client";
	private SHMessageQueue fMessageQueue;
	private boolean fModuleHandlesPauseResume = true;

	public SHModulePauseResumeQueueAdaptor(SHModuleInterface module, CascadingLoggerInterface logger)
	{
		log = logger.clone(SHModulePauseResumeQueueAdaptor.class);
		fModule = module;
		fMessageQueue = new SHMessageQueue(this, log);
		Thread dispatchThread = new Thread(this);
		dispatchThread.setName(SHModulePauseResumeQueueAdaptor.class.getName());
		//dispatchThread.setPriority(DISPATCH_THREAD_PRIORITY);
		dispatchThread.start();
	}

	@Override
	public boolean speak(SHMessageEnvelopeInterface messageEnvelope)
	{
		// This method must be called with messageSynchronized true	
		messageEnvelope.setClientIdentifier(DEFAULT_CLIENT_IDENTIFIER);
		fMessageQueue.addMessageToQueue(messageEnvelope);
		return (true);
	}

	@Override
	public boolean pause()
	{
		if (fModuleHandlesPauseResume)
		{
			if (fMessageQueue.pause(DEFAULT_CLIENT_IDENTIFIER))
			{
				fModule.cancel();
				return (true);
			}
			return (false);
		} else
			return (fModule.pause());
	}

	@Override
	public boolean resume()
	{
		if (fModuleHandlesPauseResume)
		{
			if (fMessageQueue.resume(DEFAULT_CLIENT_IDENTIFIER))
			{
				return (fModule.cancel());
			}
			return (false);
		} else
			return (fModule.resume());
	}

	@Override
	public boolean cancel()
	{
		boolean status1 = false;
		status1 = fMessageQueue.cancel(DEFAULT_CLIENT_IDENTIFIER);
		boolean status2 = fModule.cancel();
		return (status1 && status2);

	}

	@Override
	public void processMessage(SHMessageEnvelopeInterface messageEnvelope)
	{
		fModule.speak(messageEnvelope);
	}

	@Override
	public boolean cancelCurrentMessage()
	{
		return (fModule.cancel());
	}

	@Override
	public void run()
	{
		for (;;)
		{
			fMessageQueue.removeMessageFromQueue();
		}
	}

	@Override
	public boolean setProperty(String key, String value)
	{
		if ((key == null) || (value == null))
			return (false);
		key = key.toLowerCase();
		if (key.equals(SHConstants.PROP_MODULE_HANDLES_PAUSE_RESUME))
			fModuleHandlesPauseResume = value.equalsIgnoreCase("true");
		return (fModule.setProperty(key, value));
	}

	@Override
	public boolean setCustomeHandlers(SHCustomeHandlersInterface customeHandlers)
	{
		customeHandlers.setMessageQueuePeekHandler(fMessageQueue);
		return (fModule.setCustomeHandlers(customeHandlers));
	}

	@Override
	public boolean parametersToModuleEngine(SHMessageEnvelopeInterface messageEnvelope)
	{
		// Not applicable in the context of full stack module
		return (true);
	}

}
