/*
* Copyright (c) Isaac Porat 2011 - 2013 all rights reserved.
* This file is released under the GNU General Public License, version 2, or (at your option) any later version.
* See the file SpeechHub_License.txt for more details.
*/

package sh.module.tts;

import log.common.CascadingLoggerInterface;
import sh.common.SHCustomeHandlersGlobal;
import sh.common.SHCustomeHandlersTTS;
import sh.connections.SHClientEngineVoiceSelectorHandlerInterface;
import sh.connections.SHClientParametersInterface;
import sh.connections.SHConstants;
import sh.connections.SHCustomeHandlersInterface;
import sh.connections.SHEngineInterface;
import sh.connections.SHMessageEnvelopeInterface;
import sh.connections.SHModuleInterface;
import sh.module.SHModuleAudioSpeechPlayerAdaptor;
import sh.module.SHModuleBase;
import sh.module.SHModuleEngineAdaptor;
import sh.module.SHModuleEntryAdaptor;
import sh.module.SHModuleEventsAdaptor;
import sh.module.SHModuleIsolationRestartAdaptor;
import sh.module.SHModuleSpeakSyncAdaptor;
import sh.module.SHModuleTextProcessingAdaptor;
import sh.settings.SHEngineParameters;
import sh.settings.SHModuleConfiguration;
import sh.settings.SHPropertiesFileLocations;
import sh.settings.SHEnhancedPropertiesList;

/**
 * TTS wrapper 
 * This class is used to wrap all adapttors required to implement an independent TTS without a server 
 * @author Isaac Porat
 */
public class SHModuleTTSWrapper extends SHModuleBase implements SHModuleInterface
{
	private SHEngineInterface fEngine;
	private String fEngineIdentifier;
	private SHClientParametersInterface fClientParameters;
	private SHCustomeHandlersInterface fCustomeHandlers;
	private int fMessageIdentifierCounter = 0;

	public SHModuleTTSWrapper(SHEngineInterface engine, SHClientParametersInterface clientParameters,
			CascadingLoggerInterface logger)
	{
		log = logger.clone(SHModuleTTSWrapper.class);
		fEngine = engine;
		fClientParameters = clientParameters;
		wrap();
		initializeSettings();
	}

	private void initializeSettings()
	{
		try
		{
			setProperty(SHConstants.PROP_COMPONENTS_CONFIGURATION, SHConstants.COMPONENTS_CONFIGURATION_TTS);
			configureTiming(fClientParameters.getModuleTimestamp(), true);
			fEngineIdentifier = fEngine.getProperty(SHConstants.PROP_ENGINE_IDENTIFIER_DEFAULT);
			if (fEngineIdentifier == null)
			{
				log.error("Error in initializeSettings method:\n engineIdentifier property null: "
						+ SHConstants.PROP_ENGINE_IDENTIFIER_DEFAULT
						+ "\nThis property must be defined for the engine to access its configuration properties\n");
			}
			setProperty(SHConstants.PROP_MODULE_ENGINE_IDENTIFIER, fEngineIdentifier);
			SHModuleConfiguration configuration = new SHModuleConfiguration();
			SHPropertiesFileLocations propertiesFileLocations = new SHPropertiesFileLocations(log);
			SHEnhancedPropertiesList propertiesList = new SHEnhancedPropertiesList();
			String[] propertiesFilePathArray = propertiesFileLocations
					.getConfigurationModuleCommonPropertiesFilesPath();
			for (int n = 0; n < propertiesFilePathArray.length; n++)
			{
				configuration.loadProperties(propertiesList, propertiesFilePathArray[n]);
			}
			configuration.transferPropertiesToModule(this, propertiesList);
			// Create a new (empty) SHCustomeHandlers object
			fCustomeHandlers = new SHCustomeHandlersTTS();
			/*
			 * The SHClientEngineVoiceSelectorHandlerInterface is used by the watchdog to switch to alternative engine / voice in case of failure 
			* transfer the selector handler object registered by SHTTSClient or any other class implementing the SHClientEngineVoiceSelectorHandlerInterface to the local custom handler 
			*/
			SHClientEngineVoiceSelectorHandlerInterface selectorHandler = SHCustomeHandlersGlobal.getInstance()
					.getClientEngineVoiceSelectorHandler();
			fCustomeHandlers.setClientEngineVoiceSelectorHandler(selectorHandler);
			// Note the message peek handler is registered by SHModulePauseResumeQueueAdaptor down the stack 
			fModule.setCustomeHandlers(fCustomeHandlers);
		} catch (Exception ex)
		{
			log.error("Error in initializeSettings method:\n" + ex);
		}
	}

	private void wrap()
	{
		// Engine wrapper 
		SHModuleInterface moduleAdaptor = new SHModuleEngineAdaptor(fEngine, log);

		// Audio player
		String property = moduleAdaptor.getProperty(SHConstants.PROP_ENGINE_SUPPORTS_AUDIO_DATA);
		if ((property != null) && (property.equals("true")))
		{
			moduleAdaptor = new SHModuleAudioSpeechPlayerAdaptor(moduleAdaptor, log);
		}

		// Note - speakSync support requires the engine to provide speakAsync with messageEnd event
		property = moduleAdaptor.getProperty(SHConstants.PROP_ENGINE_SUPPORTS_SPEAK_SYNC);
		boolean speakSyncRequired = ((property == null) || (property.equals("false")));

		property = moduleAdaptor.getProperty(SHConstants.PROP_ENGINE_REQUIRES_ISOLATION_RESTART);
		boolean isolationRestartRequired = (property != null) && (property.equals("true"));

		if (isolationRestartRequired == false)
		{
			if (speakSyncRequired)
			{
				moduleAdaptor = new SHModuleSpeakSyncAdaptor(moduleAdaptor, log);
			}
		} else
		// Isolation restart required 
		{
			moduleAdaptor = new SHModuleIsolationRestartAdaptor(moduleAdaptor, speakSyncRequired, log);
		}

		// Above this line speakSync must be provided
		// Events support
		moduleAdaptor = new SHModuleEventsAdaptor(moduleAdaptor, log);
		// Single client pause and resume support
		moduleAdaptor = new SHModulePauseResumeQueueAdaptor(moduleAdaptor, log);
		// Text processing including splitting to paragraphs, sentences and text replacement			
		moduleAdaptor = new SHModuleTextProcessingAdaptor(moduleAdaptor, log);
		// Module entry implementing messages buffer
		fModule = new SHModuleEntryAdaptor(moduleAdaptor, fClientParameters, log);
	}

	@Override
	public boolean speak(SHMessageEnvelopeInterface messageEnvelope)
	{
		try
		{
			return (fModule.speak(messageEnvelope));
		} catch (Exception ex)
		{
			log.error("Error in speak method:\n" + ex);
			return (false);
		}
	}

	@Override
	public boolean speak(int messageIdentifier, String message, boolean speakSync)
	{
		try
		{
			// Generate message identifier
			return (fModule.speak(fMessageIdentifierCounter++, message, speakSync));
		} catch (Exception ex)
		{
			log.error("Error in speak method:\n" + ex);
			return (false);
		}
	}

	@Override
	public boolean speakChar(int messageIdentifier, String message, boolean speakSync)
	{
		try
		{
			// Generate message identifier
			return (fModule.speakChar(fMessageIdentifierCounter++, message, speakSync));
		} catch (Exception ex)
		{
			log.error("Error in speakChar method:\n" + ex);
			return (false);
		}
	}

	@Override
	public boolean speakKey(int messageIdentifier, String message, boolean speakSync)
	{
		try
		{
			// Generate message identifier
			return (fModule.speakKey(fMessageIdentifierCounter++, message, speakSync));
		} catch (Exception ex)
		{
			log.error("Error in speakKey method:\n" + ex);
			return (false);
		}
	}

	@Override
	public String getProperty(String key)
	{
		try
		{
			if (key == null)
				return (null);
			// Check for engine factory property
			if (key.startsWith(SHConstants.COMPOUND_PROP_MARKER_ENGINE_FACTORY + SHConstants.COMPOUND_PROP_SEPARATOR))
			{
				// Split the compound key to components
				String[] parts = key.split(SHConstants.COMPOUND_PROP_SEPARATOR);
				return (SHEngineParameters.getInstance(log).getEngineFactoryProperty(parts[1], parts[2]));
			}
			return (fModule.getProperty(key));
		} catch (Exception ex)
		{
			log.error("Error in getProperty method:\n" + ex);
			return (null);
		}
	}

}
