/*
* Copyright (c) Isaac Porat 2011 - 2013 all rights reserved.
* This file is released under the GNU General Public License, version 2, or (at your option) any later version.
* See the file SpeechHub_License.txt for more details.
*/

package sh.server;

import java.util.HashMap;
import java.util.Map;

import core.utilities.StackTrace;

import log.common.CascadingLoggerInterface;

import sh.common.SHConfig;
import sh.common.SHCustomeHandlersGlobal;
import sh.common.SHMessageQueue;
import sh.connections.SHClientParametersInterface;
import sh.connections.SHConstants;
import sh.connections.SHMessageEnvelopeInterface;
import sh.connections.SHMessageQueueHandlerInterface;
import sh.connections.SHModuleInterface;
import sh.connections.SHModuleTimestampInterface;
import sh.connections.SHClientEngineVoiceSelectorHandlerInterface;
import sh.connections.SHAudioDataHandlerInterface;
import sh.connections.SHEventsListenerInterface;
import sh.connections.SHUpdateParametersInterface;
import sh.services.SHTTSModulesControllerServiceInterface;

public class SHMultiplexer implements Runnable, SHMessageQueueHandlerInterface,
		SHClientEngineVoiceSelectorHandlerInterface
{
	private static SHMultiplexer fMultiplexer = null;
	/*
	* Note, This thread must be of higher priority than
	*  the running thread to make sure synthesizer started after first message is added to queue produced
	 */
	private final static int DISPATCH_THREAD_PRIORITY = 7;
	private CascadingLoggerInterface log;
	private SHMessageQueue fMessageQueue;
	private SHTTSModulesControllerServiceInterface fTTSModulesControllerService;
	private SHModuleInterface fSelectedModuleEngine;
	private Map<String, SHClientParametersInterface> fClientParametersMap;
	private Map<String, SHEventsListenerInterface> fEventListenersMap;
	private SHEventsListenerInterface fEventsRouter;
	private Map<String, SHAudioDataHandlerInterface> fAudioDataHandlersMap;
	private SHAudioDataHandlerInterface fAudioDataRouter;
	private String fCurrentClientIdentifier = "";
	private SHServerAutoShutdown fServerAutoShutdown;

	private SHMultiplexer(CascadingLoggerInterface logger)
	{
		log = logger.clone(SHMultiplexer.class);
		fMessageQueue = new SHMessageQueue(this, log);
		SHCustomeHandlersGlobal.getInstance().setMessageQueuePeekHandler(fMessageQueue);
		fClientParametersMap = new HashMap<String, SHClientParametersInterface>();
		fEventListenersMap = new HashMap<String, SHEventsListenerInterface>();
		fEventsRouter = new EventsRouter();
		fAudioDataHandlersMap = new HashMap<String, SHAudioDataHandlerInterface>();
		fAudioDataRouter = new AudioDataRouter();
		fTTSModulesControllerService = SHConfig.getTTSModulesControllerService();
		SHCustomeHandlersGlobal.getInstance().setClientEngineVoiceSelectorHandler(this);
		fServerAutoShutdown = SHServerAutoShutdown.getInstance(logger);
		Thread dispatchThread = new Thread(this);
		dispatchThread.setName(SHMultiplexer.class.getName());
		dispatchThread.setPriority(DISPATCH_THREAD_PRIORITY);
		dispatchThread.start();
	}

	public synchronized static SHMultiplexer getInstance(CascadingLoggerInterface logger)
	{
		if (fMultiplexer == null)
			fMultiplexer = new SHMultiplexer(logger);
		return (fMultiplexer);
	}

	public synchronized void releaseClient(String clientIdentifier)
	{
		String clientIdentifierCaption = "unknown";
		try
		{
			clientIdentifierCaption = fClientParametersMap.get(clientIdentifier).getClientIdentifierCaption();
		} catch (Exception ex)
		{
			// Do nothing, defined as unknown
		}
		fClientParametersMap.remove(clientIdentifier);
		fServerAutoShutdown.setClientsNumber(clientIdentifier, clientIdentifierCaption, fClientParametersMap.size());
		fEventListenersMap.remove(clientIdentifier);
		fAudioDataHandlersMap.remove(clientIdentifier);
	}

	public synchronized void setClientParameters(String clientIdentifier, SHClientParametersInterface clientParameters)
	{
		fClientParametersMap.put(clientIdentifier, clientParameters);
		fServerAutoShutdown.setClientsNumber(clientIdentifier, clientParameters.getClientIdentifierCaption(),
				fClientParametersMap.size());
	}

	public synchronized SHClientParametersInterface getClientParameters(String clientIdentifier)
	{
		return (fClientParametersMap.get(clientIdentifier));
	}

	public void modulePropertiesToClientParameters(String clientIdentifier, String key, String value)
	{
		/*
		* 		Note, some of these properties modified by the module
		* Therefore for consistancy all properties which require transfer to the engine
		* are passed through the module 
		*/
		getClientParameters(clientIdentifier).setProperty(key, value);
	}

	@Override
	public void processMessage(SHMessageEnvelopeInterface messageEnvelope)
	{
		if (parametersToModuleEngine(messageEnvelope))
			fSelectedModuleEngine.speak(messageEnvelope);

		synchronized (this)
		{
			notifyAll();
		}
	}

	@Override
	public boolean cancelCurrentMessage()
	{
		return (fSelectedModuleEngine.cancel());
	}

	@Override
	public synchronized boolean parametersToModuleEngine(SHMessageEnvelopeInterface messageEnvelope)
	{
		String clientIdentifier = messageEnvelope.getClientIdentifier();
		SHClientParametersInterface clientParameters = getClientParameters(clientIdentifier);
		if (clientParameters == null)
			return (true); // Client may have gone away.
		SHUpdateParametersInterface updateParameters = messageEnvelope.getUpdateParameters();
		return (conditionalSelectModule(clientIdentifier, clientParameters, updateParameters));
	}

	private boolean conditionalSelectModule(String clientIdentifier, SHClientParametersInterface clientParameters,
			SHUpdateParametersInterface updateParameters)
	{
		if ((fSelectedModuleEngine == null)
				|| (updateParameters.getEngineIdentifier().equalsIgnoreCase(
						fTTSModulesControllerService.getSelectedTTSModuleEngineIdentifier()) == false))
		{
			SHEventsListenerInterface eventsListener = clientParameters.getEventsListener();
			SHAudioDataHandlerInterface audioDataHandler = clientParameters.getAudioDataHandler();
			SHModuleTimestampInterface moduleTimestamp = clientParameters.getModuleTimestamp();
			String engineIdentifier = updateParameters.getEngineIdentifier();
			if (selectModule(clientIdentifier, engineIdentifier, moduleTimestamp, eventsListener, audioDataHandler) == false)
			{
				log.error("Error in conditionalSelectModule method:\n TTS Module null,selection failed");
				return (false);
			}
			fSelectedModuleEngine.setProperty(SHConstants.PROP_MODULE_CLIENT_IDENTIFIER, clientIdentifier);
			propertiesToModuleEngine(clientParameters);
			clientParameters.resetPropertiesUpdateRequired();
			// Repeat updateIdentifier to ensure that engine is selected	
			updateParameters.markInsertUpdate();
		}
		// Engines might not have changed but voice, variant or another parameter  have
		if (fCurrentClientIdentifier.equals(clientIdentifier) == false)
		{
			// Repeat updateIdentifier to ensure engine update
			updateParameters.markInsertUpdate();
			fCurrentClientIdentifier = clientIdentifier;
		}
		if (clientParameters.isPropertiesUpdateRequired())
		{
			propertiesToModuleEngine(clientParameters);
			clientParameters.resetPropertiesUpdateRequired();
		}
		return (true);
	}

	private void propertiesToModuleEngine(SHClientParametersInterface clientParameters)
	{
		// It is assumed that these properties needs to be updated when an engine is changed
		fSelectedModuleEngine.setProperty(SHConstants.PROP_MODULE_HANDLES_PUNCTUATION,
				clientParameters.getProperty(SHConstants.PROP_MODULE_HANDLES_PUNCTUATION));
		fSelectedModuleEngine.setProperty(SHConstants.PROP_MODULE_PUNCTUATION_LEVEL,
				clientParameters.getProperty(SHConstants.PROP_MODULE_PUNCTUATION_LEVEL));
		fSelectedModuleEngine.setProperty(SHConstants.PROP_MODULE_SSML_MODE,
				clientParameters.getProperty(SHConstants.PROP_MODULE_SSML_MODE));
		fSelectedModuleEngine.setProperty(SHConstants.PROP_ENGINE_USE_SONIC,
				clientParameters.getProperty(SHConstants.PROP_ENGINE_USE_SONIC));
		fSelectedModuleEngine.setProperty(SHConstants.PROP_MODULE_CHAR_CAPITAL_INDICATOR_PITCH_LEVEL,
				clientParameters.getProperty(SHConstants.PROP_MODULE_CHAR_CAPITAL_INDICATOR_PITCH_LEVEL));
	}

	private boolean isEngineIdentifiedWithClient(String clientIdentifier)
	{
		if (fSelectedModuleEngine == null)
		{
			// This is not necessarily an error and always occurs before an engine is selected
			return (false);
		}
		String clientIdentifierEngine = fSelectedModuleEngine.getProperty(SHConstants.PROP_MODULE_CLIENT_IDENTIFIER);
		if ((clientIdentifierEngine == null) || (clientIdentifierEngine.equalsIgnoreCase(clientIdentifier) == false))
			return (false);
		return (true);
	}

	@Override
	public void run()
	{
		for (;;)
		{
			fMessageQueue.removeMessageFromQueue();
		}
	}

	/*
	 *This method is available globaly and intended
	 *to be used by SHEngineWatchdog to select alternative available engine 
	 *when the watchdog failed to restart a crashed engine 
	 */
	@Override
	public boolean selectClientEngineVoice(String clientIdentifier, String engineIdentifier, String voiceIdentifier)
	{
		SHClientParametersInterface clientParameters = getClientParameters(clientIdentifier);
		if (clientParameters == null)
		{
			log.error("Error in selectClientEngineVoice method: clientParameters for " + clientIdentifier + " null");
			return (false);
		}
		clientParameters.setEngineIdentifier(engineIdentifier);
		clientParameters.setVoiceIdentifier(voiceIdentifier);
		if (selectModule(clientIdentifier, engineIdentifier, clientParameters.getModuleTimestamp(),
				clientParameters.getEventsListener(), clientParameters.getAudioDataHandler()) == false)
		{
			log.error("Error in selectClientEngineVoice method: call to selectModule method for " + clientIdentifier
					+ " failed");
			return (false);
		}
		return (true);
	}

	public boolean selectModule(String clientIdentifier, String engineIdentifier, SHModuleTimestampInterface tmGlobal,
			SHEventsListenerInterface eventsListener, SHAudioDataHandlerInterface audioDataHandler)
	{
		try
		{
			if (fTTSModulesControllerService.selectTTSModule(engineIdentifier))
			{
				fSelectedModuleEngine = fTTSModulesControllerService.getSelectedTTSModule();
				// Module configured with masterCancel as false (cancel timing established by client)
				fSelectedModuleEngine.configureTiming(tmGlobal, false);
				fSelectedModuleEngine.setProperty(SHConstants.PROP_MODULE_CLIENT_IDENTIFIER, clientIdentifier);
				fSelectedModuleEngine.setEventsListener(fEventsRouter);
				if (fEventListenersMap.get(clientIdentifier) == null)
					fEventListenersMap.put(clientIdentifier, eventsListener);
				if (audioDataHandler != null)
				{
					fSelectedModuleEngine.setAudioDataHandler(fAudioDataRouter);
					if (fAudioDataHandlersMap.get(clientIdentifier) == null)
						fAudioDataHandlersMap.put(clientIdentifier, audioDataHandler);
				} else
				// Audio handler null
				{
					fSelectedModuleEngine.setAudioDataHandler(null);
				}
				return (true);
			}
			return (false); // Error logged in service
		} catch (Exception ex)
		{
			log.error("Error in selectModule method: engineIdentifier: " + engineIdentifier + "\n"
					+ StackTrace.stackToString(ex));
			return (false);
		}
	}

	public String[] getEngineIdentifiers(String clientIdentifier)
	{
		if (fTTSModulesControllerService == null)
			return (new String[0]);
		return (fTTSModulesControllerService.getTTSEngineIdentifiers());
	}

	public boolean selectEngine(String clientIdentifier, String engineIdentifier)
	{
		// This method selects engine from the currently selected module
		try
		{
			SHClientParametersInterface clientParameters = getClientParameters(clientIdentifier);
			if (clientParameters == null)
				return (false); // Client might have gone away
			clientParameters.setEngineIdentifier(engineIdentifier);
			// It is assumed that a voice is always selected following this method
			// This would happen automatically with a message envelope
			// Voice and variant selection disabled to allow simple switching between modules where no voice and variant changed occured
			clientParameters.setVoiceIdentifier(SHConstants.STRING_PARAMETER_DISABLED);
			clientParameters.setVariantIdentifier(SHConstants.STRING_PARAMETER_DISABLED);
			return (fSelectedModuleEngine.updateParametersDirectly(clientParameters.getClonedUpdateParameters()));
		} catch (Exception ex)
		{
			log.error("Error in selectEngine method:\n" + StackTrace.stackToString(ex));
			return (false);
		}
	}

	public String getSelectedEngineIdentifier(String clientIdentifier)
	{
		// Obtain the last value from clientParameters
		SHClientParametersInterface clientParameters = getClientParameters(clientIdentifier);
		if (clientParameters == null)
			return (SHConstants.STRING_PARAMETER_ERROR); // client might have gone away
		return (clientParameters.getEngineIdentifier());
	}

	public String getLanguageCode(String clientIdentifier)
	{
		// Obtain the last value from clientParameters
		SHClientParametersInterface clientParameters = getClientParameters(clientIdentifier);
		if (clientParameters == null)
			return (SHConstants.STRING_PARAMETER_ERROR); // client might have gone away
		return (clientParameters.getLanguageCode());
	}

	public String[] getVoiceIdentifiers(String clientIdentifier)
	{
		SHClientParametersInterface clientParameters = getClientParameters(clientIdentifier);
		if (clientParameters == null) // Not defined in message envelope
			return (new String[0]); // Client may have gone away.
		if (conditionalSelectModule(clientIdentifier, clientParameters, clientParameters.getClonedUpdateParameters()))
		{
			if (fSelectedModuleEngine.updateParametersDirectly(clientParameters.getClonedUpdateParameters()))
				return (fSelectedModuleEngine.getVoiceIdentifiers());
		}
		log.error("Error in getVoiceIdentifiers method");
		return (new String[0]);
	}

	public String getSelectedVoiceIdentifier(String clientIdentifier)
	{
		// Obtain the last value from clientParameters
		SHClientParametersInterface clientParameters = getClientParameters(clientIdentifier);
		if (clientParameters == null)
			return (SHConstants.STRING_PARAMETER_ERROR); // client might have gone away
		return (clientParameters.getVoiceIdentifier());
	}

	public String[] getVariantIdentifiers(String clientIdentifier)
	{
		SHClientParametersInterface clientParameters = getClientParameters(clientIdentifier);
		if (clientParameters == null) // Not defined in message envelope
			return (new String[0]); // Client may have gone away.
		if (conditionalSelectModule(clientIdentifier, clientParameters, clientParameters.getClonedUpdateParameters()))
		{
			if (fSelectedModuleEngine.updateParametersDirectly(clientParameters.getClonedUpdateParameters()))
				return (fSelectedModuleEngine.getVariantIdentifiers());
		}
		log.error("Error in getVariantIdentifiers method");
		return (new String[0]);
	}

	public String getSelectedVariantIdentifier(String clientIdentifier)
	{
		// Obtain the last value from clientParameters
		SHClientParametersInterface clientParameters = getClientParameters(clientIdentifier);
		if (clientParameters == null)
			return (SHConstants.STRING_PARAMETER_ERROR); // client might have gone away
		return (clientParameters.getVariantIdentifier());
	}

	public boolean processSync(String clientIdentifier, SHMessageEnvelopeInterface messageEnvelope)
	{
		fMessageQueue.addMessageToQueue(messageEnvelope);
		return (true);
	}

	public boolean pause(String clientIdentifier, boolean moduleHandlesPauseResume)
	{
		if (moduleHandlesPauseResume)
		{
			if (fMessageQueue.pause(clientIdentifier))
			{
				fSelectedModuleEngine.cancel();
				return (true);
			}
			return (false);
		} else
			return (fSelectedModuleEngine.pause());
	}

	public boolean resume(String clientIdentifier, boolean moduleHandlesPauseResume)
	{
		if (moduleHandlesPauseResume)
		{
			if (fMessageQueue.resume(clientIdentifier))
			{
				return (fSelectedModuleEngine.cancel());
			}
			return (false);
		} else
			return (fSelectedModuleEngine.resume());
	}

	public boolean stop(String clientIdentifier)
	{
		return (fSelectedModuleEngine.stop());
	}

	public boolean cancel(String clientIdentifier)
	{
		boolean status1 = false;
		status1 = fMessageQueue.cancel(clientIdentifier);
		boolean status2 = fSelectedModuleEngine.cancel();
		return (status1 && status2);
	}

	public float getVolume(String clientIdentifier)
	{
		// Obtain the last value from clientParameters
		SHClientParametersInterface clientParameters = getClientParameters(clientIdentifier);
		if (clientParameters == null)
			return (SHConstants.FLOAT_PARAMETER_ERROR); // client might have gone away
		return (clientParameters.getVolume());
	}

	public float getRate(String clientIdentifier)
	{
		// Obtain the last value from clientParameters
		SHClientParametersInterface clientParameters = getClientParameters(clientIdentifier);
		if (clientParameters == null)
			return (SHConstants.FLOAT_PARAMETER_ERROR); // client might have gone away
		return (clientParameters.getRate());
	}

	public float getPitch(String clientIdentifier)
	{
		// Obtain the last value from clientParameters
		SHClientParametersInterface clientParameters = getClientParameters(clientIdentifier);
		if (clientParameters == null)
			return (SHConstants.FLOAT_PARAMETER_ERROR); // client might have gone away
		return (clientParameters.getPitch());
	}

	public boolean updateParametersDirectly(String clientIdentifier, SHUpdateParametersInterface updateParameters)
	{
		if (isEngineIdentifiedWithClient(clientIdentifier) == false) // Cannot update directly
		{
			// To take effect quickly, it is assumed that the client issued a cancel command before calling this method
			// Update clientParameters provided they are not marked as disabled
			SHClientParametersInterface clientParameters = getClientParameters(clientIdentifier);
			if (clientParameters == null)
				return (false); // client might have gone away
			String engineIdentifier = updateParameters.getEngineIdentifier();
			if (engineIdentifier.equals(SHConstants.STRING_PARAMETER_DISABLED) == false)
				clientParameters.setEngineIdentifier(engineIdentifier);
			String languageCode = updateParameters.getLanguageCode();
			if (languageCode.equals(SHConstants.STRING_PARAMETER_DISABLED) == false)
				clientParameters.setLanguageCode(languageCode);
			String voiceIdentifier = updateParameters.getVoiceIdentifier();
			if (voiceIdentifier.equals(SHConstants.STRING_PARAMETER_DISABLED) == false)
				clientParameters.setVoiceIdentifier(voiceIdentifier);
			String variantIdentifier = updateParameters.getVariantIdentifier();
			if (variantIdentifier.equals(SHConstants.STRING_PARAMETER_DISABLED) == false)
				clientParameters.setVariantIdentifier(variantIdentifier);
			float volume = updateParameters.getVolume();
			if (volume != SHConstants.FLOAT_PARAMETER_DISABLED)
				clientParameters.setVolume(volume);
			float rate = updateParameters.getRate();
			if (rate != SHConstants.FLOAT_PARAMETER_DISABLED)
				clientParameters.setRate(rate);
			float pitch = updateParameters.getPitch();
			if (pitch != SHConstants.FLOAT_PARAMETER_DISABLED)
				clientParameters.setPitch(pitch);
			return (true); // Would be updated when client connected to engine
		}
		if (fSelectedModuleEngine.updateParametersDirectly(updateParameters))
			return (true);
		return (false);
	}

	private class EventsRouter implements SHEventsListenerInterface
	{

		@Override
		public void messageBegin(String clientIdentifier, int messageIdentifier)
		{
			try
			{
				fEventListenersMap.get(clientIdentifier).messageBegin(clientIdentifier, messageIdentifier);
			} catch (Exception ex)
			{
				log.error("Error in messageBegin method\n:" + ex);
			}
		}

		@Override
		public void messageEnd(String clientIdentifier, int messageIdentifier)
		{
			try
			{
				fEventListenersMap.get(clientIdentifier).messageEnd(clientIdentifier, messageIdentifier);
			} catch (Exception ex)
			{
				log.error("Error in messageEnd method\n:" + ex);
			}
		}

		@Override
		public void messagePause(String clientIdentifier, int messageIdentifier)
		{
			try
			{
				fEventListenersMap.get(clientIdentifier).messagePause(clientIdentifier, messageIdentifier);
			} catch (Exception ex)
			{
				log.error("Error in messagePause method\n:" + ex);
			}
		}

		@Override
		public void messageResume(String clientIdentifier, int messageIdentifier)
		{
			try
			{
				fEventListenersMap.get(clientIdentifier).messageResume(clientIdentifier, messageIdentifier);
			} catch (Exception ex)
			{
				log.error("Error in messageResume method\n:" + ex);
			}
		}

		@Override
		public void messageCancel(String clientIdentifier, int messageIdentifier)
		{
			try
			{
				SHEventsListenerInterface eventsListener = fEventListenersMap.get(clientIdentifier);
				if (eventsListener == null) // This can happen when a client has gone away
					return;
				eventsListener.messageCancel(clientIdentifier, messageIdentifier);
			} catch (Exception ex)
			{
				log.error("Error in messageCancel method\n:" + ex);
			}
		}

		@Override
		public void messageAudioAvailable(String clientIdentifier, int messageIdentifier)
		{
			// Irrelevant generated at the server's client level	
		}

	}

	private class AudioDataRouter implements SHAudioDataHandlerInterface
	{

		public boolean processAudio(byte[] samples)
		{
			// Not applicable
			return (false);
		}

		public boolean processAudio(SHMessageEnvelopeInterface messageEnvelope)
		{
			try
			{
				SHAudioDataHandlerInterface audioDataHandler = fAudioDataHandlersMap.get(messageEnvelope
						.getClientIdentifier());
				if (audioDataHandler == null) // Client has gone away
					return (false);
				return (audioDataHandler.processAudio(messageEnvelope));
			} catch (Exception ex)
			{
				log.error("Error in processAudio method\n:" + ex);
				return (false);
			}

		}

	}

}
