/*
* Copyright (c) Isaac Porat 2011 - 2013 all rights reserved.
* This file is released under the GNU General Public License, version 2, or (at your option) any later version.
* See the file SpeechHub_License.txt for more details.
*/

package sh.server;

import java.util.Timer;
import java.util.TimerTask;

import sh.common.SHConfig;
import sh.connections.SHConstants;
import sh.connections.SHServerShutdownInterface;

import core.utilities.StackTrace;

import log.common.CascadingLoggerInterface;

public class SHServerAutoShutdown
{
	private CascadingLoggerInterface log;
	private static SHServerAutoShutdown fServerAutoShutdown;
	private final static int JAVA_STANDARD_INSTALLATION_TIMEOUT = 60000; // One minute in ms
	private final static int JAVA_STANDARD_INSTALLATION_SERVER_SHUTDOWN_ITERATIONS = 5; // Five minutes - multiply of timeout
	private final static int JAVA_PORTABLE_INSTALLATION_TIMEOUT = 2500; // Two and a half seconds in ms
	private final static int JAVA_PORTABLE_INSTALLATION_SERVER_SHUTDOWN_ITERATIONS = 2; // Five seconds - multiply of timeout
	// Java standard installation by default	
	private int fTimeoutMillis = JAVA_STANDARD_INSTALLATION_TIMEOUT;
	private int fServerShutdownIterations = JAVA_STANDARD_INSTALLATION_SERVER_SHUTDOWN_ITERATIONS;
	private SHServerShutdownInterface fServerShutdown;
	private int fClientsNumber = 0;
	private Timer fServerShutdownTimer;
	private ServerShutdownTask fServerShutdownTask;
	private int fServerShutdownDelayRemaining = 0;
	private boolean fAutoShutdownEnabled = true;

	private SHServerAutoShutdown(CascadingLoggerInterface logger)
	{
		log = logger.clone(SHServerAutoShutdown.class);
		/*
		* Java portable installation delay is much shorter to ensure that when
		* portable media is pulled out the server is shutdown
		* otherwise the server crashes and cannot be restarted unless the process is killed
		*/
		String value = System.getenv("SH_JAVA_INSTALLATION");
		if ((value != null) && (value.equals("portable")))
		{
			fTimeoutMillis = JAVA_PORTABLE_INSTALLATION_TIMEOUT;
			fServerShutdownIterations = JAVA_PORTABLE_INSTALLATION_SERVER_SHUTDOWN_ITERATIONS;
		}
		value = SHConfig.getServerProperty(SHConstants.PROP_SERVER_AUTO_SHUTDOWN);
		fAutoShutdownEnabled = ((value != null) && (value.equalsIgnoreCase("true")));
		if (fAutoShutdownEnabled)
			fServerShutdownTimer = new Timer();
	}

	public synchronized static SHServerAutoShutdown getInstance(CascadingLoggerInterface logger)
	{
		if (fServerAutoShutdown == null)
			fServerAutoShutdown = new SHServerAutoShutdown(logger);
		return (fServerAutoShutdown);
	}

	public void setServerShutdown(SHServerShutdownInterface serverShutdown)
	{
		fServerShutdown = serverShutdown;
	}

	public void startTimer()
	{
		fServerShutdownTask = new ServerShutdownTask();
		fServerShutdownTimer.schedule(fServerShutdownTask, fTimeoutMillis);
	}

	public void setClientsNumber(String clientIdentifier, String clientIdentifierCaption, int clientsNumber)
	{
		if (clientsNumber > fClientsNumber)
			log.notify("clientIdentifier: " + clientIdentifier + " connected"); // clientIdentifierCaption is not known at this stage
		else if (clientsNumber < fClientsNumber)
			log.notify("clientIdentifier: " + clientIdentifier + " clientIdentifierCaption: " + clientIdentifierCaption
					+ " disconnected");
		else
			log.error("Error in setClientsNumber method: no change in clients connections number");
		fClientsNumber = clientsNumber;
	}

	private void shutdown()
	{
		log.notify("No clients connected to server, automatic shutdown");
		if (fServerShutdown != null)
			fServerShutdown.shutdown();
		else
			System.exit(0);
	}

	private void scheduleTimer()
	{
		if (fServerShutdownTask != null)
			fServerShutdownTask.cancel();
		fServerShutdownTask = new ServerShutdownTask();
		fServerShutdownTimer.schedule(fServerShutdownTask, fTimeoutMillis);
	}

	private class ServerShutdownTask extends TimerTask
	{

		@Override
		public void run()
		{
			try
			{
				/*
				 * The following arrangement ensures that new TimerTask is only created at the end of timeout
				 * but clients number  can be notified at any time
				 * Effective timeout not less than n and not greater then n + 1
				 */
				if (fClientsNumber > 0)
				{
					fServerShutdownDelayRemaining = 0;
					scheduleTimer();
					return;
				}
				if (fServerShutdownDelayRemaining < fServerShutdownIterations)
				{
					fServerShutdownDelayRemaining++;
					scheduleTimer();
					return;
				}
				shutdown();
			} catch (Exception ex)
			{
				log.error("Error in run method:\n" + StackTrace.stackToString(ex));
				if (fServerShutdownTask != null)
					fServerShutdownTask.cancel();
				fServerShutdownTask = null;
			}
		}

	}

}
