/*
* Copyright (c) Isaac Porat 2011 - 2013 all rights reserved.
* This file is released under the GNU General Public License, version 2, or (at your option) any later version.
* See the file SpeechHub_License.txt for more details.
*/

package sh.server;

import core.utilities.StackTrace;
import log.common.CascadingLoggerInterface;
import sh.common.SHClientParameters;
import sh.common.SHConfig;
import sh.connections.SHClientParametersInterface;
import sh.connections.SHConstants;
import sh.connections.SHModuleInterface;
import sh.connections.SHModuleTimestampInterface;
import sh.connections.SHServerClientInterface;
import sh.connections.SHAudioDataHandlerInterface;
import sh.connections.SHEventsListenerInterface;
import sh.module.SHModuleTimestamp;
import sh.module.server.SHModuleServerClientWrapper;

/*
 * This class implements server client API
 * It is associated with each client connection and 
 * it is used by the supported protocol specific handlers or might be used on its own
 * to provide server client functionality 
 * @Author Isaac Porat 
 */
public class SHServerClient implements SHServerClientInterface
{
	private CascadingLoggerInterface log;
	private SHModuleInterface fModuleClient;
	private SHEventsListenerInterface fEventsListener;
	private SHAudioDataHandlerInterface fAudioDataHandler;
	private SHMultiplexer fMultiplexer;
	private SHModuleTimestampInterface fModuleTimestamp;
	private String fClientIdentifier;
	private SHClientParametersInterface fClientParameters;
	private String fEngineIdentifierDefault = "";

	public SHServerClient(CascadingLoggerInterface logger)
	{
		log = logger.clone(SHServerClient.class);
		fModuleTimestamp = new SHModuleTimestamp();
		fMultiplexer = SHMultiplexer.getInstance(log);
		fEngineIdentifierDefault = SHConfig.getServerProperty(SHConstants.PROP_SERVER_ENGINE_IDENTIFIER_CLIENT_DEFAULT);
	}

	@Override
	public boolean initialize(String clientIdentifier)
	{
		try
		{
			fClientIdentifier = clientIdentifier;
			fClientParameters = new SHClientParameters(fClientIdentifier);
			fMultiplexer.setClientParameters(fClientIdentifier, fClientParameters);
			String settingsBaseDirectory = SHConfig.getSettingsBaseDirectory();
			fModuleClient = new SHModuleServerClientWrapper(fClientIdentifier, fMultiplexer, fModuleTimestamp,
					settingsBaseDirectory, log);
			fModuleClient.setProperty(SHConstants.PROP_MODULE_ENGINE_IDENTIFIER, fEngineIdentifierDefault);
			// Select default engine
			selectEngine(fEngineIdentifierDefault);
			return (true);
		} catch (Exception ex)
		{
			log.error("Error in initialize method:\n" + StackTrace.stackToString(ex));
			return (false);
		}
	}

	@Override
	public boolean close()
	{
		try
		{
			fMultiplexer.releaseClient(fClientIdentifier);
			return (true);
		} catch (Exception ex)
		{
			log.error("Error in close method:\n" + StackTrace.stackToString(ex));
			return (false);
		}
	}

	@Override
	public void setClientIdentifierCaption(String clientIdentifierCaption)
	{
		if (fClientParameters != null)
			fClientParameters.setClientIdentifierCaption(clientIdentifierCaption);
	}

	@Override
	public boolean setPriority(int priorityLevel, boolean selfCancel)
	{
		return (fModuleClient.setPriority(priorityLevel, selfCancel));
	}

	@Override
	public String[] getEngineIdentifiers()
	{
		return (fModuleClient.getEngineIdentifiers());
	}

	@Override
	public boolean selectEngine(String engineIdentifier)
	{
		try
		{
			fClientParameters.setEngineIdentifier(engineIdentifier);
			fClientParameters.setModuleTimestamp(fModuleTimestamp);
			fClientParameters.setEventsListener(fEventsListener);
			fClientParameters.setAudioDataHandler(fAudioDataHandler);
			fClientParameters.setVoiceIdentifier(SHConstants.ENGINE_VOICE_IDENTIFIER_DEFAULT);
			fClientParameters.setVariantIdentifier(SHConstants.ENGINE_VARIANT_IDENTIFIER_DEFAULT);
			boolean status = fMultiplexer.selectModule(fClientIdentifier, engineIdentifier, fModuleTimestamp,
					fEventsListener, fAudioDataHandler);
			if (status)
			{
				if (fModuleClient != null) // Client module does not exist on initialization 
					fModuleClient.setProperty(SHConstants.PROP_MODULE_ENGINE_IDENTIFIER, engineIdentifier);
			}
			if (status)
				return (fMultiplexer.selectEngine(fClientIdentifier, engineIdentifier));
			return (false);
		} catch (Exception ex)
		{
			log.error("Error in selectEngine method\n" + StackTrace.stackToString(ex));
			return (false);
		}
	}

	@Override
	public String getSelectedEngineIdentifier()
	{
		return (fClientParameters.getEngineIdentifier());
	}

	@Override
	public boolean selectLanguage(String languageCode)
	{
		fClientParameters.setLanguageCode(languageCode);
		return (true);
	}

	@Override
	public String getSelectedLanguageCode()
	{
		return (fClientParameters.getLanguageCode());
	}

	@Override
	public String[] getVoiceIdentifiers()
	{
		return (fModuleClient.getVoiceIdentifiers());
	}

	@Override
	public boolean selectVoice(String voiceIdentifier)
	{
		fClientParameters.setVoiceIdentifier(voiceIdentifier);
		return (true);
	}

	@Override
	public String getSelectedVoiceIdentifier()
	{
		return (fClientParameters.getVoiceIdentifier());
	}

	@Override
	public String[] getVariantIdentifiers()
	{
		return (fModuleClient.getVariantIdentifiers());
	}

	@Override
	public boolean selectVariant(String variantIdentifier)
	{
		fClientParameters.setVariantIdentifier(variantIdentifier);
		return (true);
	}

	@Override
	public String getSelectedVariantIdentifier()
	{
		return (fClientParameters.getVariantIdentifier());
	}

	@Override
	public boolean speakAsync(int messageIdentifier, String message)
	{
		return (fModuleClient.speak(messageIdentifier, message, false));
	}

	@Override
	public boolean speakAsyncChar(int messageIdentifier, String message)
	{
		return (fModuleClient.speakChar(messageIdentifier, message, false));
	}

	@Override
	public boolean speakAsyncKey(int messageIdentifier, String message)
	{
		return (fModuleClient.speakKey(messageIdentifier, message, false));
	}

	@Override
	public boolean pause()
	{
		return (fModuleClient.pause());
	}

	@Override
	public boolean resume()
	{
		return (fModuleClient.resume());
	}

	@Override
	public boolean stop()
	{
		return (fModuleClient.stop());
	}

	@Override
	public boolean cancel()
	{
		return (fModuleClient.cancel());
	}

	@Override
	public boolean setVolume(float volume)
	{
		fClientParameters.setVolume(volume);
		return (true);
	}

	@Override
	public float getVolume()
	{
		return (fClientParameters.getVolume());
	}

	@Override
	public boolean setRate(float rate)
	{
		fClientParameters.setRate(rate);
		return (true);
	}

	@Override
	public float getRate()
	{
		return (fClientParameters.getRate());
	}

	@Override
	public boolean setPitch(float pitch)
	{
		fClientParameters.setPitch(pitch);
		return (true);
	}

	@Override
	public float getPitch()
	{
		return (fClientParameters.getPitch());
	}

	@Override
	public boolean setProperty(String key, String value)
	{
		/*
		 * Note clientParameters setProperty must not be called here
		 * It is only called by the multiplexer
		 * in method modulePropertiesToClientParameters 
		 */
		return (fModuleClient.setProperty(key, value));
	}

	@Override
	public boolean setEventsListener(SHEventsListenerInterface eventsListener)
	{
		fEventsListener = eventsListener;
		return (true);
	}

	@Override
	public boolean setAudioDataHandler(SHAudioDataHandlerInterface audioDataHandler)
	{
		fAudioDataHandler = audioDataHandler;
		return (true);
	}

}
