/*
* Copyright (c) Isaac Porat 2011 - 2013 all rights reserved.
* This file is released under the GNU General Public License, version 2, or (at your option) any later version.
* See the file SpeechHub_License.txt for more details.
*/

package sh.server.essip;

import log.common.CascadingLoggerInterface;
import sh.comm.SHServerCommMessageData;
import sh.connections.SHConstants;
import sh.connections.SHServerClientInterface;
import sh.connections.SHEventsListenerInterface;
import sh.server.SHServerClient;
import sh.settings.SHEngineParameters;

/*
 * Extended SSIP commands to SHServerClientHandler API map conversion
 * This class is associated with each client
 * 
 * @Author Isaac Porat 
 */
public class SHExtendedSSIPCommandMapper
{
	private CascadingLoggerInterface log;
	private SHEventsListenerInterface fEventsListener;
	private SHServerClientInterface fClient;
	private SHServerCommMessageData fServerCommMessageData;

	public SHExtendedSSIPCommandMapper(SHEventsListenerInterface eventsListener, CascadingLoggerInterface logger)
	{
		log = logger.clone(SHExtendedSSIPCommandMapper.class);
		fEventsListener = eventsListener;
		fClient = new SHServerClient(logger);
		fClient.setEventsListener(eventsListener);
	}

	public void initialize(String clientIdentifier)
	{
		fClient.initialize(clientIdentifier);
	}

	public void close()
	{
		fClient.close();
	}

	public void setClientIdentifierCaption(String clientIdentifierCaption)
	{
		fClient.setClientIdentifierCaption(clientIdentifierCaption);
	}

	public boolean setPriority(String priority)
	{
		if (priority == null)
			return (false);
		/* 
		* Map SSIP  to SH API
		* Note that this mapping is an approximation of the SSIP priorities but there are  some differences 
		*/
		int priorityLevel = 2;
		boolean selfCancel = false;
		if (priority.equalsIgnoreCase("important"))
		{
			priorityLevel = 5;
			selfCancel = false;
		} else if (priority.equalsIgnoreCase("message"))
		{
			priorityLevel = 4;
			selfCancel = false;
		} else if (priority.equalsIgnoreCase("text"))
		{
			priorityLevel = 3;
			selfCancel = true;
		} else if (priority.equalsIgnoreCase("progress"))
		{
			priorityLevel = 2;
			selfCancel = false;
		} else if (priority.equalsIgnoreCase("notification"))
		{
			priorityLevel = 1;
			selfCancel = true;
		}
		return (fClient.setPriority(priorityLevel, selfCancel));
	}

	public String[] listOutputModules()
	{
		return (fClient.getEngineIdentifiers());
	}

	public boolean setOutputModule(String engineIdentifier)
	{
		return (fClient.selectEngine(engineIdentifier));
	}

	public boolean setLanguage(String languageCode)
	{
		/*
		* For now returns always true even if language directories for a particular language code is not found
		/* This is not really an error as some clients like NVDA do their own replacements
		* of punctuations etc
		* Language selection for a list of voices in all languages is always done on the client side
		*/
		return (fClient.selectLanguage(languageCode));
	}

	public String[] listSynthesisVoices()
	{
		return (fClient.getVoiceIdentifiers());
	}

	public boolean setSynthesisVoice(String voiceIdentifier)
	{
		return (fClient.selectVoice(voiceIdentifier));
	}

	public String[] listVoices()
	{
		return (fClient.getVariantIdentifiers());
	}

	public boolean setVoice(String voiceIdentifier)
	{
		// Note: variantIdentifier = voiceIdentifier in SSIP
		return (fClient.selectVariant(voiceIdentifier));
	}

	public void speak(int messageIdentifier, String message)
	{
		if (fServerCommMessageData != null)
			fServerCommMessageData.indicateSpeak(messageIdentifier, message, false);
		fClient.speakAsync(messageIdentifier, message);
	}

	public void speakChar(int messageIdentifier, String message)
	{
		if (fServerCommMessageData != null)
			fServerCommMessageData.indicateSpeak(messageIdentifier, message, false);
		fClient.speakAsyncChar(messageIdentifier, message);
	}

	public void speakKey(int messageIdentifier, String message)
	{
		if (fServerCommMessageData != null)
			fServerCommMessageData.indicateSpeak(messageIdentifier, message, false);
		fClient.speakAsyncKey(messageIdentifier, message);
	}

	public void pause()
	{
		fClient.pause();
	}

	public void resume()
	{
		fClient.resume();
	}

	public void stop()
	{
		fClient.stop();
	}

	public void cancel()
	{
		if (fServerCommMessageData != null)
			fServerCommMessageData.cancel();
		fClient.cancel();
	}

	public boolean setVolume(float volume)
	{
		/*
		 * Convert SSIP range +-100 to module 0 - 100
		 */
		volume = (float) ((volume + 100.0) / 2.0);
		return (fClient.setVolume(volume));
	}

	public boolean setRate(float rate)
	{
		return (fClient.setRate(rate));
	}

	public boolean setPitch(float pitch)
	{
		return (fClient.setPitch(pitch));
	}

	public boolean setPunctuation(String level)
	{
		/*
		 * Supported levels none, some, most, all
		 * Note that SSIP and hence SD does not support most
		 * but SH supports most
		 */
		return (fClient.setProperty(SHConstants.PROP_MODULE_PUNCTUATION_LEVEL, level));
	}

	public boolean setCaptionLetterRecognition(String mode)
	{
		/*
		 * SSIP and hence Speech Dispatcher does not support pitch level indication
		 * SpeechHub supports this with a value starting with "pitchLevel=  
		 */
		if (mode.startsWith(SHConstants.MARKER_MODULE_CHAR_CAPITAL_INDICATOR_PITCH_LEVEL))
		{
			mode = mode.replaceAll("^" + SHConstants.MARKER_MODULE_CHAR_CAPITAL_INDICATOR_PITCH_LEVEL, "").trim();
			return (fClient.setProperty(SHConstants.PROP_MODULE_CHAR_CAPITAL_INDICATOR_PITCH_LEVEL, mode));
		}
		/*
		 * SSIP and hence Speech Dispatcher do not support cap pitch indicator
		 * SpeechHub supports cap pitch indicator
		 * SpeechHub does not support yet icon cap indicator
		 * Therefore icon is replaced for now by pitch indicator
		 * none and spell (with the word 'cap' before the capital char  
		 */
		if (mode.equalsIgnoreCase(SHConstants.MODULE_CHAR_CAPITAL_INDICATOR_ICON))
			mode = SHConstants.MODULE_CHAR_CAPITAL_INDICATOR_PITCH;
		return (fClient.setProperty(SHConstants.PROP_MODULE_CHAR_CAPITAL_INDICATOR, mode));
	}

	public boolean setSpelling(String mode)
	{
		mode = mode.equalsIgnoreCase("ON") ? "true" : "false";
		return (fClient.setProperty(SHConstants.PROP_MODULE_SPELL_MODE, mode));
	}

	public boolean setSSML(String mode)
	{
		mode = mode.equalsIgnoreCase("ON") ? "true" : "false";
		return (fClient.setProperty(SHConstants.PROP_MODULE_SSML_MODE, mode));
	}

	public boolean setRateBoost(String mode)
	{
		mode = mode.equalsIgnoreCase("ON") ? "true" : "false";
		return (fClient.setProperty(SHConstants.PROP_ENGINE_USE_SONIC, mode));
	}

	public boolean setSilenceBeforeMessage(String silenceMilliseconds)
	{
		return (fClient.setProperty(SHConstants.PROP_MODULE_SILENCE_BEFORE_MESSAGE, silenceMilliseconds));
	}

	public String getEngineFactoryProperty(String engineIdentifier, String key)
	{
		if ((engineIdentifier == null) || (key == null))
			return (SHConstants.STATUS_NONE);
		String value = SHEngineParameters.getInstance(log).getEngineFactoryProperty(engineIdentifier, key);
		return (value != null ? value : SHConstants.STATUS_NONE);
	}

	public boolean setMessageData(String mode)
	{
		if (mode.equalsIgnoreCase("ON"))
			fServerCommMessageData = new SHServerCommMessageData(fEventsListener, log);
		else
			fServerCommMessageData = null;
		return (fClient.setAudioDataHandler(fServerCommMessageData));
	}

	public String getMessageData(String type, String identifier)
	{
		if ((type == null) || (identifier == null))
			return (SHConstants.STATUS_NONE);
		String messageData = fServerCommMessageData.getMessageData(type, identifier);
		return (messageData != null ? messageData : SHConstants.STATUS_NONE);
	}

}
