/*
* Copyright (c) Isaac Porat 2011 - 2013 all rights reserved.
* This file is released under the GNU General Public License, version 2, or (at your option) any later version.
* See the file SpeechHub_License.txt for more details.
*/

package sh.server.essip;

import hrpc.common.HRHandlerInterface;
import hrpc.common.HRMessageHandlerInterface;

import java.util.ArrayList;
import java.util.List;
import java.util.Timer;
import java.util.TimerTask;

import log.common.BasicLogger;
import log.common.CascadingLoggerInterface;
import sh.common.SHIdentifiers;
import sh.connections.SHEventsListenerInterface;

/*
 * Extended SSIP protocol handler
 * This class is constructed for each connection and thus identifies a client
 * SSIP supports Speech Dispatcher in Linux
 * This handler includes command extensions to SSIP used by SpeechHub
 * These command extensions use message codes not defined at the time by SD in its msg.h header file
 * Message OK 228 229 including a block 263 - 299 seems to be unused by original SSIP
 * Events including 706 and above not used by original SSIP
		* 229 used by GET ENGINE_FACTORY_PROPERTY
		* 228 used by SET SELF RATE_BOOST_MODE 
		* 270 used by SET SELF SILENCE_BEFORE_MESSAGe
		* 705 used by AUDIO_AVAILABLE event
		* 271 used by SET SELF MESSAGE_DATA_MODE
		* 272 used by GET MESSAGE_DATA
		*  
		* The CHAR and KEY commands were modified
		* in that they return a message identifier otherwise there were of limited use
		*   
 * @author Isaac Porat
 * @author Bill Cox
  */
public class SHExtendedSSIPHandler implements HRHandlerInterface
{
	private CascadingLoggerInterface log;
	// Main debug mode specified in configuration file
	private final static boolean DEBUG_MODE_EVENTS_INTERNAL = false;
	// Error codes - '3' ssip code last two digits defined by SpeechHub  
	private final static String ERROR_SERVER_NOT_SUPPORTED = "300";
	private final static String ERROR_SERVER_PROCESSING_REQUEST = "301";
	private final static String ERROR_CLIENT_RESUME = "401";
	private final static String MESSAGE_LINE_SEPARATOR = "\r\n";
	private boolean fDebugMode = false;
	private HRMessageHandlerInterface fMessageHandler;
	private SHExtendedSSIPCommandMapper fCommandMapper;
	private StringBuilder fBlockBuffer;
	private boolean fBlockExpected = false;
	private int fMessageCounter = 0;
	private SHEventsListenerInterface fEventsHandler;
	private List<List<String>> fEventGroups = new ArrayList<List<String>>();
	private boolean fWaitingForResponse = false;
	private Timer fSendEventsTimer;
	private boolean fMessagePaused = false;
	@SuppressWarnings("unused")
	private String fClientUsername = "unidentified";
	private String fClientIdentifierCaption = "unidentified";
	// internal client identifier allows different instances of the same client to interact with a unique client identifier
	private String fClientIdentifierInternal;
	/* 
	*Client identifier unique across  all clients 
	*Returned to the client with events
	*/
	private String fClientIdentifierUnique;
	@SuppressWarnings("unused")
	private String fClientModule = "unidentified";
	private boolean fEventBeginEnabled = false;
	private boolean fEventEndEnabled = false;
	private boolean fEventPauseEnabled = false;
	private boolean fEventResumeEnabled = false;
	private boolean fEventCancelEnabled = false;
	private boolean fEventAudioAvailableEnabled = false;
	private int fLastSpeakMessageQueuedIdentifier = 0;
	private int fLastMessageEndEventIdentifier = 0;
	private int fLastMessageCancelEventIdentifier = 0;

	private void delay()
	{
		/*
		// Delay to debug communication
		try
		{
			Thread.sleep(10);
		} catch (InterruptedException e)
		{
		}
		*/
	}

	public SHExtendedSSIPHandler(boolean debugMode, CascadingLoggerInterface logger)
	{
		fDebugMode = debugMode;
		log = logger.clone(SHExtendedSSIPHandler.class);
		fEventsHandler = new EventsHandler();
		fCommandMapper = new SHExtendedSSIPCommandMapper(fEventsHandler, log);
		fBlockBuffer = new StringBuilder("");
		fSendEventsTimer = new Timer();
	}

	public void setMessageHandler(HRMessageHandlerInterface messageHandler)
	{
		fMessageHandler = messageHandler;
	}

	public void close()
	{
		fCommandMapper.close();
	}

	public boolean processMessage(String command)
	{
		if (command == null)
		{
			log.error("SSIP line  received by server null");
			sendResponse(ERROR_SERVER_NOT_SUPPORTED + " SSIP command received by server null");
			return (true);
		}
		if (fDebugMode)
			BasicLogger.notifyPrintln("Receive " + (System.currentTimeMillis() % 1000) + ": " + command);
		fWaitingForResponse = true;
		if (fBlockExpected)
		{
			fBlockExpected = processBlockAppend(command);
			if (fBlockExpected == false)
			{
				processBlockEnd();
				return (true);
			}
			return (true);
		}
		String[] parts = command.trim().split(" ");
		if (parts.length == 0)
		{
			log.error("Error in process message, no command found, array empty");
			return (true);
		}
		if (parts[0].equalsIgnoreCase("SET"))
		{
			processSet(command, parts);
			return (true);
		}
		if (parts[0].equalsIgnoreCase("GET"))
		{
			processGet(parts);
			return (true);
		}
		// clientIdentifierInternal might have been initialized already, if not initialize client
		if (fClientIdentifierInternal == null)
			initializeClient();
		if (parts[0].equalsIgnoreCase("SPEAK"))
		{
			processSpeak(parts);
			return (true);
		}
		if (parts[0].equalsIgnoreCase("CHAR"))
		{
			processChar(parts);
			return (true);
		}
		if (parts[0].equalsIgnoreCase("KEY"))
		{
			processKey(parts);
			return (true);
		}
		if (parts[0].equalsIgnoreCase("LIST"))
		{
			processList(parts);
			return (true);
		}
		if (parts[0].equalsIgnoreCase("PAUSE"))
		{
			processPause(parts);
			return (true);
		}
		if (parts[0].equalsIgnoreCase("RESUME"))
		{
			processResume(parts);
			return (true);
		}
		if (parts[0].equalsIgnoreCase("STOP"))
		{
			processStop(parts);
			return (true);
		}
		if (parts[0].equalsIgnoreCase("CANCEL"))
		{
			processCancel(parts);
			return (true);
		}
		if (parts[0].equalsIgnoreCase("HISTORY"))
		{
			processHistory(parts);
			return (true);
		}
		if (parts[0].equalsIgnoreCase("QUIT"))
		{
			processQuit(parts);
			return (false);
		}
		if (command.equals("."))
		{
			log.error("Error detected, single dot sent by client without block indication detected.\nError ignored and no response sent to server");
			return (true);
		}
		sendResponse(ERROR_SERVER_NOT_SUPPORTED + " command type " + command + " not supported by server");
		return (true);
	}

	private boolean processBlockAppend(String line)
	{
		if (line == null)
			return (true);
		if (line.equals("."))
			return (false);
		/*
		 * A period at the start of a line needs to be removed
		 * So the way for a user to specify a period is with double periods 
		 */
		if (line.startsWith("."))
		{
			// Efficiency is not important here as this situation is rarely encountered
			line = line.replaceFirst("\\.", "");
		}
		fBlockBuffer.append(line + "\n");
		return (true);
	}

	private void processBlockEnd()
	{
		/*
		 * It is assumed for now that this method is only used for speak	 
		* Otherwise here one needs to test for last command
		*/
		// Note do not use trim in the next line to preserve original message only remove the '\n' from the end of the message
		String message = fBlockBuffer.toString().replaceAll("\n$", "");
		fBlockBuffer.delete(0, fBlockBuffer.length());
		fMessageCounter++;
		List<String> messages = new ArrayList<String>();
		messages.add("225-" + Integer.toString(fMessageCounter));
		messages.add("225 OK  MESSAGE QUEUED");
		sendResponse(messages);
		fLastSpeakMessageQueuedIdentifier = fMessageCounter;
		if (DEBUG_MODE_EVENTS_INTERNAL)
			BasicLogger.notifyPrintln("SpeakMessageQueued: " + message + " clientIdentifierInternal: "
					+ fClientIdentifierCaption + " " + fClientIdentifierInternal + " messageIdentifier: "
					+ fLastSpeakMessageQueuedIdentifier);
		fCommandMapper.speak(fMessageCounter, message);
	}

	private void initializeClient()
	{
		try
		{
			// Create an internal client identifier
			if (fClientIdentifierUnique == null)
				fClientIdentifierUnique = Integer.toString(SHIdentifiers.getNewClientInternalMarker());
			fClientIdentifierInternal = "CI-" + fClientIdentifierUnique;
			fCommandMapper.initialize(fClientIdentifierInternal);
		} catch (Exception ex)
		{
			log.error("Error in initializeClient method:\n" + ex);
		}
	}

	private void processSet(String command, String[] parts)
	{
		String type = parts[2];
		if (type.equalsIgnoreCase("CLIENT_NAME"))
		{
			String name = parts[3];
			String[] nameParts = name.split(":");
			fClientUsername = nameParts[0];
			fClientIdentifierCaption = nameParts[1];
			fClientModule = nameParts[2];
			// clientIdentifierInternal might have been initialized already, if not initialize client
			if (fClientIdentifierInternal == null)
				initializeClient();
			fCommandMapper.setClientIdentifierCaption(fClientIdentifierCaption);
			sendResponse("208 OK CLIENT NAME SET");
			return;
		}
		if (fClientIdentifierInternal == null)
			initializeClient();
		if (type.equalsIgnoreCase("PRIORITY"))
		{
			sendResponse("202 OK PRIORITY SET");
			fCommandMapper.setPriority(parts[3]);
			return;
		} else if (type.equalsIgnoreCase("LANGUAGE"))
		{
			sendResponse("201 OK LANGUAGE SET");
			fCommandMapper.setLanguage(parts[3]);
			return;
		} else if (type.equalsIgnoreCase("NOTIFICATION"))
		{
			String event = parts[3];
			boolean state = false;
			if (parts[4].equalsIgnoreCase("ON"))
				state = true;
			if (event.equalsIgnoreCase("ALL"))
			{
				fEventBeginEnabled = state;
				fEventEndEnabled = state;
				fEventPauseEnabled = state;
				fEventResumeEnabled = state;
				fEventCancelEnabled = state;
				fEventAudioAvailableEnabled = true;
			} else if (event.equalsIgnoreCase("BEGIN"))
				fEventBeginEnabled = state;
			else if (event.equalsIgnoreCase("END"))
				fEventEndEnabled = state;
			else if (event.equalsIgnoreCase("PAUSE"))
				fEventPauseEnabled = state;
			else if (event.equalsIgnoreCase("RESUME"))
				fEventResumeEnabled = state;
			else if (event.equalsIgnoreCase("CANCEL"))
				fEventCancelEnabled = state;
			else if (event.equalsIgnoreCase("AUDIO_AVAILABLE"))
				fEventAudioAvailableEnabled = state;
			sendResponse("220 OK NOTIFICATION SET");
			return;
		} else if (type.equalsIgnoreCase("OUTPUT_MODULE"))
		{
			if (fCommandMapper.setOutputModule(parts[3]))
				sendResponse("216 OK OUTPUT MODULE SET");
			else
				sendResponse(ERROR_SERVER_PROCESSING_REQUEST + " Cannot select requested engine");
			return;
		} else if (type.equalsIgnoreCase("VOICE"))
		{
			/*
			 * This command which suppose to set default voices Male1, Child1 etc is not implemented
			 */
			if (fCommandMapper.setVoice(parts[3]))
				sendResponse("209 OK VOICE SET");
			else
				sendResponse(ERROR_SERVER_PROCESSING_REQUEST + " Cannot select requested voice");
			return;
		} else if (type.equalsIgnoreCase("SYNTHESIS_VOICE"))
		{
			String voice = command.trim().split(" ", 4)[3];
			if (fCommandMapper.setSynthesisVoice(voice))
				sendResponse("209 OK SYNTHESIS VOICE SET");
			else
				sendResponse(ERROR_SERVER_PROCESSING_REQUEST + " Cannot select requested synthesis voice");
			return;
		} else if (type.equalsIgnoreCase("VOLUME"))
		{
			if (fCommandMapper.setVolume(Float.valueOf(parts[3]).floatValue()))
				sendResponse("218 OK VOLUME SET");
			else
				sendResponse(ERROR_SERVER_PROCESSING_REQUEST + " Cannot set requested volume");
			return;
		} else if (type.equalsIgnoreCase("RATE"))
		{
			if (fCommandMapper.setRate(Float.valueOf(parts[3]).floatValue()))
				sendResponse("203 OK RATE SET");
			else
				sendResponse(ERROR_SERVER_PROCESSING_REQUEST + " Cannot set requested rate");
			return;
		} else if (type.equalsIgnoreCase("PITCH"))
		{
			if (fCommandMapper.setPitch(Float.valueOf(parts[3]).floatValue()))
				sendResponse("204 OK PITCH SET");
			else
				sendResponse(ERROR_SERVER_PROCESSING_REQUEST + " Cannot set pitch");
			return;
		} else if (type.equalsIgnoreCase("PUNCTUATION"))
		{
			if (fCommandMapper.setPunctuation(parts[3]))
				sendResponse("205 OK PUNCTUATION SET");
			else
				sendResponse(ERROR_SERVER_PROCESSING_REQUEST + " Cannot set pitch");
			return;
		} else if (type.equalsIgnoreCase("CAP_LET_RECOGN"))
		{
			if (fCommandMapper.setCaptionLetterRecognition(parts[3]))
				sendResponse("206 OK CAP LET RECOGNITION SET");
			else
				sendResponse(ERROR_SERVER_PROCESSING_REQUEST + " Cannot set pitch");
			return;
		} else if (type.equalsIgnoreCase("SPELLING"))
		{
			if (fCommandMapper.setSpelling(parts[3]))
				sendResponse("207 OK SPELLING SET");
			else
				sendResponse(ERROR_SERVER_PROCESSING_REQUEST + " Cannot set pitch");
			return;
		} else if (type.equalsIgnoreCase("SSML_MODE"))
		{
			if (fCommandMapper.setSSML(parts[3]))
				sendResponse("219 OK SSML MODE SET");
			else
				sendResponse(ERROR_SERVER_PROCESSING_REQUEST + " Cannot set SSML");
			return;
			// SpeechHub SSIP extensions
			// See note at the class comments on message codes used
		} else if (type.equalsIgnoreCase("RATE_BOOST_MODE"))
		{
			if (fCommandMapper.setRateBoost(parts[3]))
				sendResponse("228 OK RATE BOOST SET");
			else
				sendResponse(ERROR_SERVER_PROCESSING_REQUEST + " Cannot set rate booster");
			return;
		} else if (type.equalsIgnoreCase("SILENCE_BEFORE_MESSAGE"))
		{
			if (fCommandMapper.setSilenceBeforeMessage(parts[3]))
				sendResponse("270 OK SILENCE BEFORE MESSAGE SET");
			else
				sendResponse(ERROR_SERVER_PROCESSING_REQUEST + " Cannot set silence before message");
			return;
		} else if (type.equalsIgnoreCase("MESSAGE_DATA_MODE"))
		{
			if (fCommandMapper.setMessageData(parts[3]))
				sendResponse("271 OK MESSAGE DATA SET");
			else
				sendResponse(ERROR_SERVER_PROCESSING_REQUEST + " Cannot set message data");
			return;
		}
		sendResponse(ERROR_SERVER_NOT_SUPPORTED + " command type " + type + " not supported by server");
		return;
	}

	private void processGet(String[] parts)
	{
		// SpeechHub SSIP extension
		// See note at the class comments on message codes used
		if (parts[1].equalsIgnoreCase("ENGINE_FACTORY_PROPERTY"))
		{
			sendResponse("229-" + fCommandMapper.getEngineFactoryProperty(parts[2], parts[3]));
			sendResponse("229 OK ENGINE FACTORY PROPERTY SENT");
			return;
		}
		if (parts[1].equalsIgnoreCase("MESSAGE_DATA"))
		{
			List<String> messages = new ArrayList<String>();
			messages.add("272-" + fCommandMapper.getMessageData(parts[2], parts[3]));
			messages.add("272 OK MESSAGE DATA SENT");
			sendResponse(messages);
			// For efficiency, send events not fired because of collision with message data 
			if (fEventGroups.size() > 0)
				sendEventsResponse();
			return;
		}
		sendResponse(ERROR_SERVER_NOT_SUPPORTED + " GET command type " + parts[1] + " not supported by server");
	}

	private void processList(String[] parts)
	{
		String type = parts[1];
		if (type.equalsIgnoreCase("OUTPUT_MODULES"))
		{
			String[] engineIdentifiers = fCommandMapper.listOutputModules();
			if (engineIdentifiers == null)
			{
				sendResponse(ERROR_SERVER_PROCESSING_REQUEST + " Cannot list requested engine identifiers");
				return;
			}
			List<String> messages = new ArrayList<String>();
			for (int n = 0; n < engineIdentifiers.length; n++)
			{
				messages.add("250-" + engineIdentifiers[n]);
			}
			messages.add("250 OK OUTPUT MODULE LIST SENT");
			sendResponse(messages);
			return;
		} else if (type.equalsIgnoreCase("VOICES"))
		{
			String[] voiceIdentifiers = fCommandMapper.listVoices();
			if (voiceIdentifiers == null)
			{
				sendResponse(ERROR_SERVER_PROCESSING_REQUEST + " Cannot list requested voice identifiers");
				return;
			}
			List<String> messages = new ArrayList<String>();
			for (int n = 0; n < voiceIdentifiers.length; n++)
			{
				messages.add("249-" + voiceIdentifiers[n]);
			}
			messages.add("249 OK VOICE LIST SENT");
			sendResponse(messages);
			return;
		} else if (type.equalsIgnoreCase("SYNTHESIS_VOICES"))
		{
			String[] voiceIdentifiers = fCommandMapper.listSynthesisVoices();
			if (voiceIdentifiers == null)
			{
				sendResponse(ERROR_SERVER_PROCESSING_REQUEST + " Cannot list requested voice identifiers");
				return;
			}
			List<String> messages = new ArrayList<String>();
			for (int n = 0; n < voiceIdentifiers.length; n++)
			{
				messages.add("249-" + voiceIdentifiers[n]);
			}
			messages.add("249 OK SYNTHESIS VOICE LIST SENT");
			sendResponse(messages);
			return;
		}
	}

	private void processPause(String[] parts)
	{
		fMessagePaused = true;
		sendResponse("211 OK PAUSED");
		fCommandMapper.pause();
	}

	private void processResume(String[] parts)
	{
		if (fMessagePaused)
		{
			fMessagePaused = false;
			sendResponse("212 OK RESUMED");
			fCommandMapper.resume();
		} else
			sendResponse(ERROR_CLIENT_RESUME + " Cannot resume, not in paused state");
	}

	private void processStop(String[] parts)
	{
		sendResponse("200 OK STOPPED");
		fCommandMapper.stop();
	}

	private void processCancel(String[] parts)
	{
		/*
		 * Note, initialy a response was send to the client before the command was processed
		 * However it was found that with slow engines such as SAPI5,
		 * it caused missing start of speech occasionally
		 * Using a stopwatch the processing delay is usually 0, sometimes 1 and rarely 10 ms
		 * which is acceptable; this way a server gets a response when the cancel command is completed
		 */
		fCommandMapper.cancel();
		sendResponse("213 OK CANCELED");
		fEventsHandler.messageCancel(fClientIdentifierInternal, fLastSpeakMessageQueuedIdentifier);
	}

	private void processHistory(String[] parts)
	{
		if (parts[1].equals("GET"))
		{
			if (parts[2].equals("CLIENT_ID"))
			{
				sendResponse("245-" + fClientIdentifierUnique);
				sendResponse("245 OK CLIENT ID SENT");
				return;
			}
			sendResponse(ERROR_SERVER_NOT_SUPPORTED + " history SET command type " + parts[2]
					+ " not supported by server");
			return;
		}
		sendResponse(ERROR_SERVER_NOT_SUPPORTED + " history command type " + parts[1] + " not supported by server");
	}

	private void processQuit(String[] parts)
	{
		sendResponse("231 OK CONNECTION CLOSED");
	}

	private void processSpeak(String[] parts)
	{
		fBlockExpected = true;
		sendResponse("230 OK RECEIVING DATA");
		return;
	}

	/**
		* Note that this command was enhanced from the original SSIP
		 * in that it returns a message identifier otherwise it was of limited use
		 */
	private void processChar(String[] parts)
	{
		fMessageCounter++;
		List<String> messages = new ArrayList<String>();
		messages.add("225-" + Integer.toString(fMessageCounter));
		messages.add("225 OK CHAR QUEUED");
		sendResponse(messages);
		fLastSpeakMessageQueuedIdentifier = fMessageCounter;
		if (DEBUG_MODE_EVENTS_INTERNAL)
			BasicLogger.notifyPrintln("CharMessageQueued: " + parts[1] + " clientIdentifierInternal: "
					+ fClientIdentifierCaption + " " + fClientIdentifierInternal + " messageIdentifier: "
					+ fLastSpeakMessageQueuedIdentifier);
		fCommandMapper.speakChar(fMessageCounter, parts[1]);
		return;
	}

	/**
		* Note that this command was enhanced from the original SSIP
		 * in that it returns a message identifier otherwise it was of limited use
		 */
	private void processKey(String[] parts)
	{
		fMessageCounter++;
		List<String> messages = new ArrayList<String>();
		messages.add("225-" + Integer.toString(fMessageCounter));
		messages.add("225 OK KEY QUEUED");
		sendResponse(messages);
		fLastSpeakMessageQueuedIdentifier = fMessageCounter;
		if (DEBUG_MODE_EVENTS_INTERNAL)
			BasicLogger.notifyPrintln("KeyMessageQueued: " + parts[1] + " clientIdentifierInternal: "
					+ fClientIdentifierCaption + " " + fClientIdentifierInternal + " messageIdentifier: "
					+ fLastSpeakMessageQueuedIdentifier);
		fCommandMapper.speakKey(fMessageCounter, parts[1]);
		return;
	}

	public void sendResponse(String message)
	{
		synchronized (this)
		{
			if (fDebugMode)
				BasicLogger.notifyPrintln("Response " + (System.currentTimeMillis() % 1000) + ": " + message);
			delay();
			fMessageHandler.dispatchMessage(message);
			fWaitingForResponse = false;
		}
	}

	public void sendResponse(List<String> messages)
	{
		synchronized (this)
		{
			String text = "";
			for (int n = 0; n < messages.size(); n++)
			{
				if (n == 0)
					text = messages.get(n);
				else
					text += MESSAGE_LINE_SEPARATOR + messages.get(n);
			}
			if (fDebugMode)
				BasicLogger.notifyPrintln("Response " + (System.currentTimeMillis() % 1000) + ": " + text);
			delay();
			fMessageHandler.dispatchMessage(text);
			fWaitingForResponse = false;
		}
	}

	private synchronized void sendEventsResponse()
	{
		for (int n = 0; n < fEventGroups.size(); n++)
		{
			String text = "";
			List<String> group = fEventGroups.get(n);
			for (int i = 0; i < group.size(); i++)
			{
				if (i == 0)
					text = group.get(i);
				else
					text += MESSAGE_LINE_SEPARATOR + group.get(i);
			}
			delay();
			fMessageHandler.dispatchMessage(text);
			if (fDebugMode)
				BasicLogger.notifyPrintln("Response " + (System.currentTimeMillis() % 1000) + ": " + text);
		}
		fEventGroups.clear();
	}

	private class EventsHandler implements SHEventsListenerInterface
	{

		@Override
		public synchronized void messageBegin(String clientIdentifier, int messageIdentifier)
		{
			// Check if message cancel event fired already
			if (messageIdentifier <= fLastMessageCancelEventIdentifier)
				return;
			if (DEBUG_MODE_EVENTS_INTERNAL)
				BasicLogger.notifyPrintln("MessageBeginEvent: ClientIdentifierInternal: " + fClientIdentifierInternal
						+ " messageIdentifier: " + messageIdentifier);
			if (fEventBeginEnabled == false)
				return;
			List<String> messages = new ArrayList<String>();
			messages.add("701-" + Integer.toString(messageIdentifier));
			messages.add("701-" + fClientIdentifierUnique);
			messages.add("701 BEGIN");
			fEventGroups.add(messages);
			if (fWaitingForResponse)
				fSendEventsTimer.schedule(new SendEventsTask(), 50);
			else
				sendEventsResponse();
		}

		@Override
		public synchronized void messageEnd(String clientIdentifier, int messageIdentifier)
		{
			// Check if message cancel event fired already
			if (messageIdentifier <= fLastMessageCancelEventIdentifier)
				return;
			fLastMessageEndEventIdentifier = messageIdentifier;
			if (DEBUG_MODE_EVENTS_INTERNAL)
				BasicLogger.notifyPrintln("MessageEndEvent: ClientIdentifierInternal: " + fClientIdentifierInternal
						+ " messageIdentifier: " + messageIdentifier);
			if (fEventEndEnabled == false)
				return;
			List<String> messages = new ArrayList<String>();
			messages.add("702-" + Integer.toString(messageIdentifier));
			messages.add("702-" + fClientIdentifierUnique);
			// Here needs to handle message index if required
			messages.add("702 END");
			fEventGroups.add(messages);
			if (fWaitingForResponse)
				fSendEventsTimer.schedule(new SendEventsTask(), 50);
			else
				sendEventsResponse();
		}

		@Override
		public synchronized void messageCancel(String clientIdentifier, int messageIdentifier)
		{
			// Check if messageCancel event already fired		
			if (messageIdentifier <= fLastMessageCancelEventIdentifier)
				return;
			//Check if messageEnd event fired already
			if (messageIdentifier <= fLastMessageEndEventIdentifier)
				return;
			fLastMessageCancelEventIdentifier = messageIdentifier;
			if (DEBUG_MODE_EVENTS_INTERNAL)
				BasicLogger.notifyPrintln("MessageCancelEvent: ClientIdentifierInternal: " + fClientIdentifierInternal
						+ " messageIdentifier: " + messageIdentifier);
			if (fEventCancelEnabled == false)
				return;
			List<String> messages = new ArrayList<String>();
			messages.add("703-" + Integer.toString(messageIdentifier));
			messages.add("703-" + fClientIdentifierUnique);
			messages.add("703 CANCELED");
			fEventGroups.add(messages);
			if (fWaitingForResponse)
				fSendEventsTimer.schedule(new SendEventsTask(), 50);
			else
				sendEventsResponse();
		}

		@Override
		public synchronized void messagePause(String clientIdentifier, int messageIdentifier)
		{
			if (DEBUG_MODE_EVENTS_INTERNAL)
				BasicLogger.notifyPrintln("MessagePauseEvent: ClientIdentifierInternal: " + fClientIdentifierInternal
						+ " messageIdentifier: " + messageIdentifier);
			if (fEventPauseEnabled == false)
				return;
			List<String> messages = new ArrayList<String>();
			messages.add("704-" + Integer.toString(messageIdentifier));
			messages.add("704-" + fClientIdentifierUnique);
			messages.add("704 PAUSED");
			fEventGroups.add(messages);
			if (fWaitingForResponse)
				fSendEventsTimer.schedule(new SendEventsTask(), 50);
			else
				sendEventsResponse();
		}

		@Override
		public synchronized void messageResume(String clientIdentifier, int messageIdentifier)
		{
			if (DEBUG_MODE_EVENTS_INTERNAL)
				BasicLogger.notifyPrintln("MessageResumeEvent: ClientIdentifierInternal: " + fClientIdentifierInternal
						+ " messageIdentifier: " + messageIdentifier);
			if (fEventResumeEnabled == false)
				return;
			List<String> messages = new ArrayList<String>();
			messages.add("705-" + Integer.toString(messageIdentifier));
			messages.add("705-" + fClientIdentifierUnique);
			messages.add("705 RESUMED");
			fEventGroups.add(messages);
			if (fWaitingForResponse)
				fSendEventsTimer.schedule(new SendEventsTask(), 50);
			else
				sendEventsResponse();
		}

		@Override
		public synchronized void messageAudioAvailable(String clientIdentifier, int messageIdentifier)
		{
			if (DEBUG_MODE_EVENTS_INTERNAL)
				BasicLogger.notifyPrintln("MessageAudioAvailableEvent: ClientIdentifierInternal: "
						+ fClientIdentifierInternal + " messageIdentifier: " + messageIdentifier);
			if (fEventAudioAvailableEnabled == false)
				return;
			List<String> messages = new ArrayList<String>();
			messages.add("706-" + Integer.toString(messageIdentifier));
			messages.add("706-" + fClientIdentifierUnique);
			messages.add("706 Audio available");
			fEventGroups.add(messages);
			if (fWaitingForResponse)
				fSendEventsTimer.schedule(new SendEventsTask(), 50);
			else
				sendEventsResponse();
		}

	}

	private class SendEventsTask extends TimerTask
	{

		@Override
		public void run()
		{
			if (fEventGroups.size() == 0)
				return;
			if (fWaitingForResponse)
			{
				fSendEventsTimer.schedule(new SendEventsTask(), 50);
				log.notify("Warnning: server message Event collision with response detected, timer fired again, this delayed event already by 50ms");
			} else
			{
				sendEventsResponse();
				log.notify("Warnning: server message Event collision with response detected, this delayed event by 50ms");
			}
		}
	}

}
