/*
* Copyright (c) Isaac Porat 2011 - 2013 all rights reserved.
* This file is released under the GNU General Public License, version 2, or (at your option) any later version.
* See the file SpeechHub_License.txt for more details.
*/

package sh.server.independent;

import java.io.File;
import java.io.FileOutputStream;
import java.io.PrintStream;

import core.utilities.StackTrace;

import log.common.BasicLogger;
import log.common.CascadingLoggerInterface;
import log.common.SimpleLogger;
import sh.common.SHConfig;
import sh.connections.SHConstants;
import sh.independent.SHConfigIndependent;
import sh.independent.SHIndependentModulesController;
import sh.server.SHServerLauncher;
import sh.services.SHTTSModulesControllerServiceInterface;

/**
 * Launches  the independent version of the server 
 * (without TC - OSGi support)
 * @author Isaac Porat
 */
public class SHServerIndependent
{
	private CascadingLoggerInterface log;
	private static final String LOG_FILE = "log.txt";
	private final static String LOG_NEW_INSTANCE_FILE = "log_new_instance.txt";
	private static final String LOCK_FILE = "instance.lock";

	@SuppressWarnings("unused")
	private static SHServerLauncher fServerLauncher;

	private static void usage(String format, Object... args)
	{
		BasicLogger.notifyPrintf(format, args);
		BasicLogger.notifyPrintln("Usage: Java classpath SHServerIndependent [parameters]... \n"
				+ "Parameters can be:\n"
				+ "-trt transport type:  Select the  communication transport type with one of the following values:\n"
				+ SHConstants.SERVER_TRANSPORT_MULTI_SOCKET_INET
				+ " - connect to a running already socket inet server which accepts multiple client connections,\n"
				+ "this is the default transport type if no parameters are specified\n"
				+ "the socket port is  obtained from the server configuration file\n"
				+ SHConstants.SERVER_TRANSPORT_SINGLE_SOCKET_INET
				+ " - start a single client connection   socket inet server,\n"
				+ " the socket port must be defined as additional argument.\n"
				+ SHConstants.SERVER_TRANSPORT_MULTI_SOCKET_UNIX
				+ " - start a single client connection   socket unix server,\n"
				+ "the socket file path is  obtained from the server configuration file\n"
				+ SHConstants.SERVER_TRANSPORT_STDIO + " - start the server and connect to it using stdout and stdin\n"
				+ "-sip socket inet port - the socket inet port to use applicable only to "
				+ SHConstants.SERVER_TRANSPORT_SINGLE_SOCKET_INET + " transport type\n");
		System.exit(1);
	}

	/**
	Starting the server from the command line
	 */
	public static void main(String[] args)
	{
		/*
		* Log levels 
		* Error - 1
		* Warning - 2
		* Info - 3
		* Debug - 4
		*/
		int logLevel = 2;
		CascadingLoggerInterface log = new SimpleLogger(SHServerIndependent.class, logLevel);

		String transportType = SHConstants.SERVER_TRANSPORT_MULTI_SOCKET_INET;
		int singleClientServerSocketInetPort = 0;

		int i = 0;
		String arg;
		while (i < args.length && args[i].startsWith("-"))
		{
			arg = args[i++];
			if (arg.equals("-trt"))
			{
				if (i < args.length)
					transportType = args[i++];
				else
					usage("Expected transport type after -trt parameter\n");
			} else if (arg.equals("-sip"))
			{
				if (i < args.length)
				{
					try
					{
						singleClientServerSocketInetPort = Integer.parseInt(args[i++]);
					} catch (NumberFormatException ex)
					{
						usage("Socket inet port %s is not a valid integer\n", args[--i]);
					}
				} else
					usage("Expected socket inet port after -sip parameter\n");
			} else if (arg.equals("-lfo"))
			{
				// Ignore (LogFileOff) for compatibility with the tcf version
				// The same is done here using the transportType parameter  
			} else if (arg.equals("-sio"))
			{
				// Ignore (SingleInstanceOff) for compatibility with the tcf version
				// The same is done here using the transportType parameter  
			} else
				usage("Unrecognized parameter %s\n", arg);
		}
		if (i < args.length)
			usage("Unrecognized parameter %s\n", args[i]);
		// Check for recognized transport types 
		if (((transportType.equals(SHConstants.SERVER_TRANSPORT_MULTI_SOCKET_INET))
				|| (transportType.equals(SHConstants.SERVER_TRANSPORT_SINGLE_SOCKET_INET))
				|| (transportType.equals(SHConstants.SERVER_TRANSPORT_MULTI_SOCKET_UNIX)) || (transportType
				.equals(SHConstants.SERVER_TRANSPORT_STDIO))) == false)
			usage("Unrecognized %s transport type\n", transportType);
		// Check that a port is not specified unless it is SingleClientSocketInet transport type
		if ((transportType.equals(SHConstants.SERVER_TRANSPORT_SINGLE_SOCKET_INET) == false)
				&& (singleClientServerSocketInetPort != 0))
			usage("Port number is not relevant and should not be specified for %s transport type\n", transportType);
		// Check that a port number for SingleClientSocketInet transport type is specified
		if ((transportType.equals(SHConstants.SERVER_TRANSPORT_SINGLE_SOCKET_INET))
				&& (singleClientServerSocketInetPort == 0))
			usage("Port number must be specified for SingleClientSocketInet transport type\n");

		try
		{
			// Configure locations - settings and  engines
			SHConfigIndependent.configureLocations(log);

			String systemBaseDirectoryRW = SHConfig.getSystemBaseDirectory();
			SHServerIndependent serverIndependent = new SHServerIndependent(log);

			// Only check instance and redirect stdout and stderr   if in MultiSocket transport type
			if ((transportType.equals(SHConstants.SERVER_TRANSPORT_MULTI_SOCKET_INET))
					|| (transportType.equals(SHConstants.SERVER_TRANSPORT_MULTI_SOCKET_UNIX)))
			{
				/*
				 * Abort if another instance is running and redirect STDOUT and STDERR to log file		
				 */
				if (serverIndependent.checkInstanceRedirectLog(systemBaseDirectoryRW) == false)
				{
					/* 
					* Note this method returns false
					* when another instance is found or any other error 
					*/
					log.error("Aborting another instance exists");
					return;
				}
			}
			log.notify("Independent server current directory:\n" + (new File(".").getAbsolutePath()));
			serverIndependent.startServer(transportType, singleClientServerSocketInetPort);
		} catch (Exception ex)
		{
			log.error("Error in main method:\n" + StackTrace.stackToString(ex));
		}
	}

	public SHServerIndependent(CascadingLoggerInterface logger)
	{
		log = logger.clone(SHServerIndependent.class);
	}

	private boolean checkInstanceRedirectLog(String systemBaseDirectoryRW)
	{
		try
		{
			File systemBaseDirectoryRWFile = new File(systemBaseDirectoryRW);
			if (systemBaseDirectoryRWFile.exists() == false)
				systemBaseDirectoryRWFile.mkdirs();
			/* 
			* Check if this is the only instance running and if not
			* write command line arguments to the args file
			* for use by the running instance of the profile
			* Note that redirection of STDERR is done
			* to a new instance log file to ensure that for already existing instance 
			* the redirection is not hijacked by this new process
			*/
			String logNewInstanceFilePath = systemBaseDirectoryRW + "/" + LOG_NEW_INSTANCE_FILE;
			File logNewInstanceFile = new File(logNewInstanceFilePath);
			if (logNewInstanceFile.exists())
				logNewInstanceFile.delete();
			FileOutputStream fileOutputStream = new FileOutputStream(logNewInstanceFilePath);
			PrintStream printStream = new PrintStream(fileOutputStream);
			System.setErr(printStream);
			String lockPath = systemBaseDirectoryRW + "/" + LOCK_FILE;
			if (core.utilities.InstanceHandler.lockInstance(lockPath) == false)
				return (false);
			String logFilePath = systemBaseDirectoryRW + "/" + LOG_FILE;
			File logFile = new File(logFilePath);
			if (logFile.exists())
				logFile.delete();
			FileOutputStream logFileOutputStream = new FileOutputStream(logFilePath);
			PrintStream logPrintStream = new PrintStream(logFileOutputStream);
			System.setOut(logPrintStream);
			System.setErr(logPrintStream);
			fileOutputStream.close();
			// Remove empty new instance log file if exists
			if (logNewInstanceFile.exists())
				logNewInstanceFile.delete();
			return (true);
		} catch (Exception ex)
		{
			log.error("Error in checkInstanceRedirectLog method:\n" + ex);
			return (false);
		}
	}

	public void startServer(String transportType, int singleClientServerSocketInetPort)
	{
		// Print banner.
		log.notify("SpeechHub Server Session Log");
		log.notify("=================\n");
		// Controller with server TTS module stack
		boolean serverMode = true;
		SHTTSModulesControllerServiceInterface modulesController = new SHIndependentModulesController(serverMode,
				log);
		try
		{
			fServerLauncher = new SHServerLauncher(transportType, singleClientServerSocketInetPort, modulesController,
					log);
			// Notify start of the specific version
			if (transportType.equals(SHConstants.SERVER_TRANSPORT_MULTI_SOCKET_INET))
				log.notify("SpeechHub server (" + SHConstants.SERVER_TRANSPORT_MULTI_SOCKET_INET + ") started on port "
						+ SHConfig.getServerProperty(SHConstants.PROP_SERVER_SOCKET_INET_PORT));
			else if (transportType.equals(SHConstants.SERVER_TRANSPORT_MULTI_SOCKET_UNIX))
				log.notify("SpeechHub server (" + SHConstants.SERVER_TRANSPORT_MULTI_SOCKET_UNIX
						+ ") started on socket file "
						+ SHConfig.getServerProperty(SHConstants.PROP_SERVER_SOCKET_UNIX_PATH));
			else if (transportType.equals(SHConstants.SERVER_TRANSPORT_SINGLE_SOCKET_INET))
				log.notify("SpeechHub server (" + SHConstants.SERVER_TRANSPORT_SINGLE_SOCKET_INET
						+ ") started on port " + singleClientServerSocketInetPort);
			else if (transportType.equals(SHConstants.SERVER_TRANSPORT_STDIO))
				log.notify("SpeechHub server (" + SHConstants.SERVER_TRANSPORT_STDIO + ") started");
			log.notify("SpeechHub server (independent version) ready");
		} catch (Exception ex)
		{
			log.error("Error in launching SpeechHub server:\n" + StackTrace.stackToString(ex));
		}
	}

}
