/*
* Copyright (c) Isaac Porat 2011 - 2013 all rights reserved.
* This file is released under the GNU General Public License, version 2, or (at your option) any later version.
* See the file SpeechHub_License.txt for more details.
*/

package sh.settings;

import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import log.common.CascadingLoggerInterface;
import sh.common.SHConfig;
import sh.connections.SHConstants;
import core.utilities.StackTrace;
import core.utilities.VersionComparator;

/**
 * Lists engine identifiers
 * Handles common user set and engine 'factory set'  properties
 * @author Isaac Porat
 */
public class SHEngineParameters
{
	private static SHEngineParameters fEngineParameters = null;
	private CascadingLoggerInterface log;
	private Map<String, String> fEngineMap;
	private SHEngineCommonProperties fEngineCommonProperties;
	private Map<String, SHEngineFactoryProperties> fEngineFactoryPropertiesMap;
	private Map<String, SHEngineVoicesFactoryProperties> fEngineVoicesFactoryPropertiesMap;
	private Set<String> fEngineDisabledSet;

	private SHEngineParameters(CascadingLoggerInterface logger)
	{
		log = logger.clone(SHEngineParameters.class);
		fEngineMap = new HashMap<String, String>();
		fEngineFactoryPropertiesMap = new HashMap<String, SHEngineFactoryProperties>();
		fEngineVoicesFactoryPropertiesMap = new HashMap<String, SHEngineVoicesFactoryProperties>();
		fEngineDisabledSet = new HashSet<String>();
	}

	public synchronized static SHEngineParameters getInstance(CascadingLoggerInterface logger)
	{
		if (fEngineParameters == null)
			fEngineParameters = new SHEngineParameters(logger);
		return (fEngineParameters);
	}

	public String getEngineDirectory(String engineIdentifier)
	{
		if (fEngineMap.isEmpty())
			loadEngineMap();
		String engineVersion = fEngineMap.get(engineIdentifier);
		if (engineVersion == null)
			return (null);
		return (engineIdentifier + "-" + engineVersion);
	}

	public String[] getAvailableEngineIdentifiers()
	{
		if (fEngineMap.isEmpty())
			loadEngineMap();
		String[] engineArray = new String[fEngineMap.size()];
		int index = 0;
		for (Iterator<Map.Entry<String, String>> iterator = fEngineMap.entrySet().iterator(); iterator.hasNext();)
		{
			Map.Entry<String, String> entry = iterator.next();
			engineArray[index] = entry.getKey();
			index++;
		}
		return (engineArray);
	}

	private void loadEngineMap()
	{
		try
		{
			String[] engineArray = getAvailableEngineIdentifiersDirectories();
			for (int n = 0; n < engineArray.length; n++)
			{
				String[] parts = engineArray[n].split("-");
				String engineIdentifier = parts[0];
				String engineVersion = parts[1];
				if (fEngineDisabledSet.contains(engineIdentifier)) // Engine disabled
					continue;
				// If map does not contain engine put new engine 
				if (fEngineMap.containsKey(engineIdentifier) == false)
				{
					fEngineMap.put(engineIdentifier, engineVersion);
					continue;
				}
				// Always replace with latest version
				String mapEngineVersion = fEngineMap.get(engineIdentifier);
				if (mapEngineVersion == null)
				{
					fEngineMap.put(engineIdentifier, engineVersion);
					continue;
				}
				if (VersionComparator.isSecondVersionGreater(mapEngineVersion, engineVersion))
					fEngineMap.put(engineIdentifier, engineVersion);
			}
		} catch (Exception ex)
		{
			log.error("Error in loadMap method:\n" + StackTrace.stackToString(ex));
		}
	}

	private String[] getAvailableEngineIdentifiersDirectories()
	{
		String enginesPath = SHConfig.getEnginesBaseDirectory() + "/engine";
		//String enginesPath = "C:/TC/sh/engine";
		File enginesFile = new File(enginesPath);
		if (enginesFile.exists() == false)
		{
			log.error("Error in getEngineIdentifiers method:\n" + enginesFile.getAbsolutePath()
					+ "\nEngines directory does not exist");
			return (new String[0]);
		}
		try
		{
			List<String> fileList = new ArrayList<String>();
			String[] files = enginesFile.list();
			for (int n = 0; n < files.length; n++)
			{
				String enginePath = enginesPath + "/" + files[n];
				// Ignore files
				if (new File(enginePath).isDirectory() == false)
					continue;
				// Ignore dorectories starting with '_' indicating disabled engine 
				if (files[n].startsWith("_"))
					continue;
				// Ignore engines without version number
				if (files[n].contains("-") == false)
					continue;
				fileList.add(files[n]);
			}
			return (fileList.toArray(new String[0]));
		} catch (Exception ex)
		{
			log.error("Error in getAvailableEngineIdentifiers method:\n" + ex);
			return (new String[0]);
		}
	}

	public String[] getProtocolEngineIdentifiers(String requiredEngineProtocol, String requiredEngineProtocolVersion)
	{
		String enginesPath = SHConfig.getEnginesBaseDirectory() + "/engine";
		File enginesFile = new File(enginesPath);
		if (enginesFile.exists() == false)
		{
			log.error("Error in getEngineIdentifiers method:\n" + enginesFile.getAbsolutePath()
					+ "\nEngines directory does not exist");
			return (new String[0]);
		}
		try
		{
			List<String> fileList = new ArrayList<String>();
			String[] engineIdentifiers = getAvailableEngineIdentifiers();
			for (int n = 0; n < engineIdentifiers.length; n++)
			{
				String engineDirectory = getEngineDirectory(engineIdentifiers[n]);
				if (engineDirectory == null)
					continue;
				// Test for protocol compatibility
				String engineProtocol = getEngineFactoryProperty(engineIdentifiers[n], SHConstants.PROP_ENGINE_PROTOCOL);
				String engineProtocolVersion = getEngineFactoryProperty(engineIdentifiers[n],
						SHConstants.PROP_ENGINE_PROTOCOL_VERSION);
				if (engineProtocol.equals(requiredEngineProtocol) == false)
					continue;
				if (engineProtocolVersion.equals(requiredEngineProtocolVersion) == false)
					continue;
				fileList.add(engineIdentifiers[n]);
			}
			return (fileList.toArray(new String[0]));
		} catch (Exception ex)
		{
			log.error("Error in getProtocolEngineIdentifiers method:\n" + ex);
			return (new String[0]);
		}
	}

	public boolean areEnginesSupportingProtocolAvailable(String requiredEngineProtocol,
			String requiredEngineProtocolVersion)
	{
		String[] availableEngineIdentifiers = getProtocolEngineIdentifiers(requiredEngineProtocol,
				requiredEngineProtocolVersion);
		return (availableEngineIdentifiers.length > 0);
	}

	public String getEngineCommonProperty(String key)
	{
		if (fEngineCommonProperties == null)
			fEngineCommonProperties = new SHEngineCommonProperties(log);
		return (fEngineCommonProperties.getProperty(key));
	}

	public String getEngineFactoryProperty(String engineIdentifier, String key)
	{
		try
		{
			SHEngineFactoryProperties engineProperties = fEngineFactoryPropertiesMap.get(engineIdentifier);
			if (engineProperties == null)
			{
				engineProperties = new SHEngineFactoryProperties(log);
				fEngineFactoryPropertiesMap.put(engineIdentifier, engineProperties);
			}
			return (engineProperties.getProperty(engineIdentifier, key));
		} catch (Exception ex)
		{
			log.error("Error in getEngineProperty method:\n" + StackTrace.stackToString(ex));
			return (null);
		}
	}

	public String getEngineVoicesFactoryProperty(String engineIdentifier, String key)
	{
		try
		{
			SHEngineVoicesFactoryProperties voicesProperties = fEngineVoicesFactoryPropertiesMap.get(engineIdentifier);
			if (voicesProperties == null)
			{
				voicesProperties = new SHEngineVoicesFactoryProperties(log);
				fEngineVoicesFactoryPropertiesMap.put(engineIdentifier, voicesProperties);
			}
			return (voicesProperties.getProperty(engineIdentifier, key));
		} catch (Exception ex)
		{
			log.error("Error in getEngineVoicesProperty method:\n" + StackTrace.stackToString(ex));
			return (null);
		}
	}

	public boolean isEngineAvailable(String engineIdentifier)
	{
		if (fEngineMap.isEmpty())
			loadEngineMap();
		return (fEngineMap.containsKey(engineIdentifier));
	}

	public String getAvailableDefaultEngineIdentifier()
	{
		if (fEngineMap.isEmpty())
			loadEngineMap();
		// Use espeak if available
		if (fEngineMap.containsKey(SHConstants.ENGINE_IDENTIFIER_ESPEAK))
			return (SHConstants.ENGINE_IDENTIFIER_ESPEAK);
		// Choose the first engine available
		String[] engineIdentifiers = getAvailableEngineIdentifiers();
		return (engineIdentifiers[0]);
	}

	public String getAvailableDefaultVoiceIdentifier()
	{
		// At present 'default' voice is returned which should be interpreted correctly to suitable default in every implemented engine
		return (SHConstants.ENGINE_VOICE_IDENTIFIER_DEFAULT);
	}

	public void disableEngine(String engineIdentifier)
	{
		fEngineDisabledSet.add(engineIdentifier);
		clearParametersCache();
	}

	/**
	 * Parameters cache needs to be cleard whenever
	 * an engine is added removed or disabled while SH is running
	 * The cache is started automatically when SH starts or after cache clear
	 * This method is called automatically whenever an engine disabled but not in other situations 
	 */
	public void clearParametersCache()
	{
		fEngineMap.clear();
		fEngineFactoryPropertiesMap.clear();
	}

}
