/*
* Copyright (c) Isaac Porat 2011 - 2013 all rights reserved.
* This file is released under the GNU General Public License, version 2, or (at your option) any later version.
* See the file SpeechHub_License.txt for more details.
*/

package sh.settings;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * Handles properties list where the key is regexp
 * Handles the use of '"' for termination of both key and value
 * Place '\' (backslash) before quotes '\" and '\='  to prevent special processing  in keys and values'
 * Similarily use '\\' in front of backslash to prevent special processing  
 * 
 * @author Isaac Porat
 *
 */
public class SHEnhancedPropertiesList
{
	private final static String TEMP_ESCAPE_BACKSLASH = "qQbackslashQq";
	private final static String TEMP_ESCAPE_EQUAL = "qQequalQq";
	private final static String TEMP_ESCAPE_QUOTES = "qQquotesQq";

	private List<String> keyList = new ArrayList<String>();
	private Map<String, String> propertiesMap = new HashMap<String, String>();

	public SHEnhancedPropertiesList()
	{
	}

	public SHEnhancedPropertiesList(String filePath) throws IOException
	{
		String string = core.utilities.DirFileUtilities.fileToString(filePath, "UTF-8");
		String[] lines = string.split("[\n]+");
		for (int n = 0; n < lines.length; n++)
		{
			lines[n] = lines[n].trim();
			if (lines[n].equals(""))
				continue;
			if (lines[n].startsWith("#"))
				continue;
			// Replace temporarily special characters 
			lines[n] = lines[n].replaceAll("\\\\\\\\", TEMP_ESCAPE_BACKSLASH);
			lines[n] = lines[n].replaceAll("\\\\=", TEMP_ESCAPE_EQUAL);
			lines[n] = lines[n].replaceAll("\\\\\"", TEMP_ESCAPE_QUOTES);
			if (lines[n].matches(".+=.+"))
			{
				String[] parts = lines[n].split("=");
				parts[0] = parts[0].trim();
				if (parts[0].equals(""))
					continue;
				parts[1] = parts[1].trim();
				if (parts[1].equals(""))
					continue;
				// Remove '"' if present at the start or end of a part
				parts[0] = parts[0].replaceAll("^\"", "");
				parts[0] = parts[0].replaceAll("\"$", "");
				parts[1] = parts[1].replaceAll("^\"", "");
				parts[1] = parts[1].replaceAll("\"$", "");
				if (parts[0].contains("qQ"))
				{
					parts[0] = parts[0].replaceAll(TEMP_ESCAPE_EQUAL, "\\\\=");
					parts[0] = parts[0].replaceAll(TEMP_ESCAPE_QUOTES, "\\\\\"");
					parts[0] = parts[0].replaceAll(TEMP_ESCAPE_BACKSLASH, "\\\\\\\\");
				}
				if (parts[1].contains("qQ"))
				{
					parts[1] = parts[1].replaceAll(TEMP_ESCAPE_EQUAL, "\\\\=");
					parts[1] = parts[1].replaceAll(TEMP_ESCAPE_QUOTES, "\\\\\"");
					parts[1] = parts[1].replaceAll(TEMP_ESCAPE_BACKSLASH, "\\\\\\\\");
				}
				put(parts[0], parts[1]);
			}
		}
	}

	public void clear()
	{
		propertiesMap.clear();
		keyList.clear();
	}

	public void put(String key, String value)
	{
		propertiesMap.put(key, value);
		if (!keyList.contains(key))
			keyList.add(key);
	}

	public String get(String key)
	{
		return (propertiesMap.get(key));
	}

	public boolean containsKey(String key)
	{
		return (propertiesMap.containsKey(key));
	}

	public String[] toKeysArray()
	{
		String[] keyArray =
		{};
		return (keyList.toArray(keyArray));
	}

}
