/*
* Copyright (c) Isaac Porat 2011 - 2013 all rights reserved.
* This file is released under the GNU General Public License, version 2, or (at your option) any later version.
* See the file SpeechHub_License.txt for more details.
*/

package sh.settings;

import java.io.File;
import java.util.ArrayList;
import java.util.List;

import log.common.CascadingLoggerInterface;

import sh.common.SHConfig;
import sh.connections.SHConstants.PunctuationLevel;

/**
 * Locations for settings files which can be modified by the user
 * Engine executable, data and properties locations are handled by other classes
 * Notably SHEngineFactoryProperties handles 'factory' set engine properties
 * @author Isaac Porat
 */
public class SHPropertiesFileLocations
{
	private CascadingLoggerInterface log;
	private String fCommonReplacementBaseDirectory;
	private String fEngineReplacementBaseDirectory;
	private String fCommonConfigurationBaseDirectory;

	public SHPropertiesFileLocations(CascadingLoggerInterface logger)
	{
		log = logger.clone(SHPropertiesFileLocations.class);
		String settingsBaseDirectory = SHConfig.getSettingsBaseDirectory();
		fCommonReplacementBaseDirectory = settingsBaseDirectory + "/common/replacement";
		fEngineReplacementBaseDirectory = settingsBaseDirectory + "/engine";
		fCommonConfigurationBaseDirectory = settingsBaseDirectory + "/common/configuration";
	}

	public boolean isLanguageReplacementDirectoryAvailable(String languageCode)
	{
		if (getReplacementLanguagePropertiesDirectory(fCommonReplacementBaseDirectory, languageCode) == null)
		{
			log.notify("Warning, language code '"
					+ languageCode
					+ "' replacements directory not found.\n Language specific replacements such as punctuation, char and key where applicable delegated to engines\n");
			return (false);
		}
		return (true);
	}

	public String[] getPunctuationPropertiesFilesPath(String languageCode, PunctuationLevel punctuation)
	{
		List<String> filePathList = new ArrayList<String>();
		String baseDirectory = fCommonReplacementBaseDirectory;
		getPunctuationPropertiesFilesPath(filePathList, baseDirectory, languageCode, punctuation);
		return (filePathList.toArray(new String[0]));
	}

	private void getPunctuationPropertiesFilesPath(List<String> filePathList, String baseDirectory,
			String languageCode, PunctuationLevel punctuation)
	{
		String languageDirectory = getReplacementLanguagePropertiesDirectory(baseDirectory, languageCode);
		if (languageDirectory == null)
		{
			if (languageCode == null)
				languageCode = "null";
			// Warning moved to isLanguageReplacementDirectoryAvailable method  
			//			log.notify("Warning in getPunctuationPropertiesFilesPath method: language directory for language code: "
			//+ languageCode + " not found, properties skipped!");
			return;
		}
		String filePath = null;
		if (punctuation.equals(PunctuationLevel.SOME))
		{
			filePath = languageDirectory + "/punctuation-some.txt";
			if (new File(filePath).exists())
				filePathList.add(filePath);
		} else if (punctuation.equals(PunctuationLevel.MOST))
		{
			filePath = languageDirectory + "/punctuation-most.txt";
			if (new File(filePath).exists())
				filePathList.add(filePath);
		} else if (punctuation.equals(PunctuationLevel.ALL))
		{
			filePath = languageDirectory + "/punctuation-all.txt";
			if (new File(filePath).exists())
				filePathList.add(filePath);
		}
	}

	public String[] getStandardPropertiesFilesPath(String engineIdentifier, String languageCode)
	{
		List<String> filePathList = new ArrayList<String>();
		String baseDirectory = fCommonReplacementBaseDirectory;
		getStandardPropertiesFilesPath(filePathList, baseDirectory, languageCode, true);
		baseDirectory = fEngineReplacementBaseDirectory + "/" + engineIdentifier + "/replacement";
		getStandardPropertiesFilesPathAllDefault(filePathList, baseDirectory, languageCode, false); // do not report error replacement per engine optional
		return (filePathList.toArray(new String[0]));
	}

	private void getStandardPropertiesFilesPath(List<String> filePathList, String baseDirectory, String languageCode,
			boolean reportError)
	{
		String languageDirectory = getReplacementLanguagePropertiesDirectory(baseDirectory, languageCode);
		if (languageDirectory == null)
		{
			if (reportError)
			{
				if (languageCode == null)
					languageCode = "null";
				// Warning moved to isLanguageReplacementDirectoryAvailable method  
				//				log.notify("Warning in getStandardPropertiesFilesPath method: language directory for language code: "
				//+ languageCode + " not found, properties skipped!");
			}
			return;
		}
		// General replacement	
		String filePath = languageDirectory + "/general.txt";
		if (new File(filePath).exists())
			filePathList.add(filePath);
		return;
	}

	private void getStandardPropertiesFilesPathAllDefault(List<String> filePathList, String baseDirectory,
			String languageCode, boolean reportError)
	{
		String languageDirectory = getReplacementLanguagePropertiesDirectoryAllDefault(baseDirectory, languageCode);
		if (languageDirectory == null)
		{
			if (reportError)
			{
				if (languageCode == null)
					languageCode = "null";
				// Warning moved to isLanguageReplacementDirectoryAvailable method  
				//log.notify("Warning in getStandardPropertiesFilesPathAllDefault method: language directory for language code: "
				//+ languageCode + " not found, properties skipped!");
			}
			return;
		}
		// General replacement	
		String filePath = languageDirectory + "/general.txt";
		if (new File(filePath).exists())
			filePathList.add(filePath);
		return;
	}

	public String[] getCharPropertiesFilesPath(String engineIdentifier, String languageCode)
	{
		List<String> filePathList = new ArrayList<String>();
		String baseDirectory = fCommonReplacementBaseDirectory;
		getCharPropertiesFilesPath(filePathList, baseDirectory, languageCode);
		baseDirectory = fEngineReplacementBaseDirectory + "/" + engineIdentifier + "/replacement";
		getCharPropertiesFilesPathAllDefault(filePathList, baseDirectory, languageCode, false); // do not report error replacement per engine optional
		return (filePathList.toArray(new String[0]));
	}

	private void getCharPropertiesFilesPath(List<String> filePathList, String baseDirectory, String languageCode)
	{
		String languageDirectory = getReplacementLanguagePropertiesDirectory(baseDirectory, languageCode);
		if (languageDirectory == null)
		{
			if (languageCode == null)
				languageCode = "null";
			// Warning moved to isLanguageReplacementDirectoryAvailable method  
			//			log.notify("Warning in getCharPropertiesFilesPath method: language directory for language code "
			//+ languageCode + " not found, properties skipped!");
			return;
		}
		String filePath = null;
		filePath = languageDirectory + "/char.txt";
		if (new File(filePath).exists())
			filePathList.add(filePath);
		return;
	}

	private void getCharPropertiesFilesPathAllDefault(List<String> filePathList, String baseDirectory,
			String languageCode, boolean reportError)
	{
		String languageDirectory = getReplacementLanguagePropertiesDirectoryAllDefault(baseDirectory, languageCode);
		if (languageDirectory == null)
		{
			if (reportError)
			{
				if (languageCode == null)
					languageCode = "null";
				log.notify("Error in getCharPropertiesFilesPathAllDefault method: language directory for language code: "
						+ languageCode + " not found, properties skipped!");
			}
			return;
		}
		// Char replacement	
		String filePath = languageDirectory + "/char.txt";
		if (new File(filePath).exists())
			filePathList.add(filePath);
		return;
	}

	public String[] getKeyPropertiesFilesPath(String engineIdentifier, String languageCode)
	{
		List<String> filePathList = new ArrayList<String>();
		String baseDirectory = fCommonReplacementBaseDirectory;
		getKeyPropertiesFilesPath(filePathList, baseDirectory, languageCode);
		baseDirectory = fEngineReplacementBaseDirectory + "/" + engineIdentifier + "/replacement";
		getKeyPropertiesFilesPathAllDefault(filePathList, baseDirectory, languageCode, false); // do not report error replacement per engine optional
		return (filePathList.toArray(new String[0]));
	}

	private void getKeyPropertiesFilesPath(List<String> filePathList, String baseDirectory, String languageCode)
	{
		String languageDirectory = getReplacementLanguagePropertiesDirectory(baseDirectory, languageCode);
		if (languageDirectory == null)
		{
			if (languageCode == null)
				languageCode = "null";
			// Warning moved to isLanguageReplacementDirectoryAvailable method  
			//log.notify("Warning in getKeyPropertiesFilesPath method: language directory not found for language code: "
			//+ languageCode + " properties skipped!");
			return;
		}
		String filePath = null;
		filePath = languageDirectory + "/key.txt";
		if (new File(filePath).exists())
			filePathList.add(filePath);
		return;
	}

	private void getKeyPropertiesFilesPathAllDefault(List<String> filePathList, String baseDirectory,
			String languageCode, boolean reportError)
	{
		String languageDirectory = getReplacementLanguagePropertiesDirectoryAllDefault(baseDirectory, languageCode);
		if (languageDirectory == null)
		{
			if (reportError)
			{
				if (languageCode == null)
					languageCode = "null";
				// Warning moved to isLanguageReplacementDirectoryAvailable method  
				//log.notify("Warning in getKeyPropertiesFilesPathAllDefault method: language directory not found for language code: "
				//+ languageCode + " properties skipped!");
			}
			return;
		}
		// Key replacement	
		String filePath = languageDirectory + "/key.txt";
		if (new File(filePath).exists())
			filePathList.add(filePath);
		return;
	}

	private String getReplacementLanguagePropertiesDirectory(String baseDirectory, String languageCode)
	{
		String languageDirectory = baseDirectory + "/" + languageCode;
		if (new File(languageDirectory).exists() == false)
		{
			// If includes sub code check if primary code directory exists   
			if (languageDirectory.contains("-") == false)
				return (null);
			// Try to find primary language code		
			languageDirectory = languageDirectory.replaceAll("-[a-zA-Z]+$", "");
			if (new File(languageDirectory).exists() == false)
				return (null);
		}
		return (languageDirectory);
	}

	/*
	 * This method substitutes 'all' language directory
	 * if requested language code not foune
	 */
	private String getReplacementLanguagePropertiesDirectoryAllDefault(String baseDirectory, String languageCode)
	{
		String languageDirectory = baseDirectory + "/" + languageCode;
		if (new File(languageDirectory).exists() == false)
		{
			// If includes sub code check if primary code directory exists
			if (!languageDirectory.matches("-[a-zA-Z]+$"))
			{
				// Try 'all'
				languageDirectory = baseDirectory + "/all";
				if (new File(languageDirectory).exists() == false)
					return (null);
				return (languageDirectory);
			}
			// Try to find primary language code		
			languageDirectory = languageDirectory.replaceAll("-[a-zA-Z]+$", "");
			if (new File(languageDirectory).exists() == false)
			{
				// Try 'all'
				languageDirectory = baseDirectory + "/en";
				if (new File(languageDirectory).exists() == false)
					return (null);
				return (languageDirectory);
			}
		}
		return (languageDirectory);
	}

	public String[] getConfigurationModuleCommonPropertiesFilesPath()
	{
		List<String> filePathList = new ArrayList<String>();
		String baseDirectory = fCommonConfigurationBaseDirectory;
		String fileName = "module-config.txt";
		getPropertiesFilesPath(filePathList, baseDirectory, fileName);
		return (filePathList.toArray(new String[0]));
	}

	public String[] getConfigurationEngineCommonPropertiesFilesPath()
	{
		List<String> filePathList = new ArrayList<String>();
		String baseDirectory = fCommonConfigurationBaseDirectory;
		String fileName = "engine-config.txt";
		getPropertiesFilesPath(filePathList, baseDirectory, fileName);
		return (filePathList.toArray(new String[0]));
	}

	public String getConfigurationServerPropertiesFilesPath()
	{
		String baseDirectory = fCommonConfigurationBaseDirectory;
		String fileName = "server-config.txt";
		return (baseDirectory + "/" + fileName);
	}

	private void getPropertiesFilesPath(List<String> filePathList, String baseDirectory, String fileName)
	{
		String filePath = baseDirectory + "/" + fileName;
		if (new File(filePath).exists())
			filePathList.add(filePath);
		return;
	}

}
