/*
* Copyright (c) Isaac Porat 2011 - 2013 all rights reserved.
* This file is released under the GNU General Public License, version 2, or (at your option) any later version.
* See the file SpeechHub_License.txt for more details.
*/

package sh.settings;

import java.io.File;
import java.io.IOException;

import log.common.CascadingLoggerInterface;
import sh.common.SHConfig;
import core.utilities.DirFileUtilities;

/**
 * Copy read only to read write settings directories
 * @author Isaac Porat
 */
public class SHSettingsDirectoriesHandler
{
	private static SHSettingsDirectoriesHandler fHandler = null;
	private CascadingLoggerInterface log;
	private SHEngineParameters fEngineParameters;

	private SHSettingsDirectoriesHandler(CascadingLoggerInterface logger)
	{
		log = logger.clone(SHSettingsDirectoriesHandler.class);
		fEngineParameters = SHEngineParameters.getInstance(logger);
	}

	public synchronized static SHSettingsDirectoriesHandler getInstance(CascadingLoggerInterface logger)
	{
		if (fHandler == null)
			fHandler = new SHSettingsDirectoriesHandler(logger);
		return (fHandler);
	}

	public synchronized boolean verifySettingsDirectories(String settingsDirReadPath, String settingsDirReadWritePath)
	{
		if (settingsDirReadPath == null)
		{
			log.error("Error in verifySettingsDirectories method: source directory null");
			return (false);
		}
		File settingsDirReadFile = new File(settingsDirReadPath);
		if (settingsDirReadFile.exists() == false)
		{
			log.error("Error in verifySettingsDirectories method:\n source directory " + settingsDirReadPath
					+ " not found");
			return (false);
		}
		if (copyDirectory(settingsDirReadFile, new File(settingsDirReadWritePath)) == false)
		{
			log.error("Error in verifySettingsDirectories method: can't copy to:\n" + settingsDirReadWritePath);
			return (false);
		}
		/*	 
		* Copy engine replacement directory for all available engines
		* to settings read write directory
		*/
		String[] engineIdentifiers = SHEngineParameters.getInstance(log).getAvailableEngineIdentifiers();
		for (int n = 0; n < engineIdentifiers.length; n++)
		{
			String replacementDirReadPath = SHConfig.getEnginesBaseDirectory() + "/engine/"
					+ fEngineParameters.getEngineDirectory(engineIdentifiers[n]) + "/sh/replacement";
			File replacementDirReadFile = new File(replacementDirReadPath);
			if (replacementDirReadFile.exists()) // Replacement directory is optional for an engine
			{
				String replacementDirReadWritePath = settingsDirReadWritePath + "/engine/" + engineIdentifiers[n]
						+ "/replacement";
				File replacementDirReadWriteFile = new File(replacementDirReadWritePath);
				if (copyDirectory(replacementDirReadFile, replacementDirReadWriteFile) == false)
					log.error("Error in verifySettingsDirectories method:\n could not copy " + engineIdentifiers[n]
							+ " replacement directory to settings read write directory");
			}
		}
		return (true);
	}

	private boolean copyDirectory(File sourceFile, File targetFile)
	{
		try
		{
			if (sourceFile.isDirectory())
			{
				if (targetFile.exists() == false)
					targetFile.mkdirs();
				String files[] = sourceFile.list();
				for (String file : files)
				{
					File sourceChildFile = new File(sourceFile, file);
					File targetChildFile = new File(targetFile, file);
					copyDirectory(sourceChildFile, targetChildFile);
				}
			} else
			{
				//if file, then  conditional copy
				conditionalCopyFile(sourceFile, targetFile);
			}
			return (true);
		} catch (Exception ex)
		{
			log.error("Error in copyDirectory method:\n" + ex);
			;
			return (false);
		}
	}

	private void conditionalCopyFile(File sourceFile, File targetFile) throws IOException
	{
		/*
		 * Update policy at this stage is simple; if file does not exist copy it
		 * At a later stage more sophistacated policy based for example on version number could be applied
		 */
		if (targetFile.exists() == false)
			DirFileUtilities.copyFile(sourceFile, targetFile);
	}

}
