/*
* Copyright (c) Isaac Porat 2011 - 2013 all rights reserved.
* This file is released under the GNU General Public License, version 2, or (at your option) any later version.
* See the file SpeechHub_License.txt for more details.
*/

package sh.tts;

import core.utilities.StackTrace;
import sh.common.SHConfig;
import sh.common.SHCustomeHandlersGlobal;
import sh.connections.SHClientEngineVoiceSelectorHandlerInterface;
import sh.connections.SHAudioDataHandlerInterface;
import sh.connections.SHEventsListenerInterface;
import sh.connections.SHTTSInterface;
import sh.services.SHTTSModulesControllerServiceInterface;
import log.common.CascadingLoggerInterface;

/**
 * TTS incorporating and able to switch between modules
 * This implementation uses the independent version
 * There is no implementation using TC
 *   
 * @author Isaac Porat
 *
 */
public class SHTTS implements SHTTSInterface, SHClientEngineVoiceSelectorHandlerInterface
{
	private static SHTTS fTTSClient = null;
	private CascadingLoggerInterface log;
	private SHTTSModulesControllerServiceInterface fModulesController;
	private SHTTSInterface fTTS;

	private SHTTS(String clientIdentifier, CascadingLoggerInterface logger)
	{
		log = logger.clone(SHTTS.class);
		fModulesController = SHConfig.getTTSModulesControllerService();
		SHCustomeHandlersGlobal.getInstance().setClientEngineVoiceSelectorHandler(this);
	}

	// Only one SHTTSObject can operate in a virtual machine 
	public synchronized static SHTTS getInstance(String clientIdentifier, CascadingLoggerInterface logger)
	{
		if (fTTSClient == null)
			fTTSClient = new SHTTS(clientIdentifier, logger);
		return (fTTSClient);
	}

	@Override
	public boolean initialize()
	{
		return (fTTS.initialize());
	}

	@Override
	public boolean close()
	{
		return (fTTS.close());
	}

	@Override
	public String[] getEngineIdentifiers()
	{
		try
		{
			return (fModulesController.getTTSEngineIdentifiers());
		} catch (Exception ex)
		{
			log.error("Error in getEngineIdentifiers method:\n" + StackTrace.stackToString(ex));
			return (null);
		}
	}

	@Override
	public boolean selectEngine(String engineIdentifier)
	{
		try
		{
			if (fModulesController.selectTTSModule(engineIdentifier))
			{
				fTTS = (SHTTSInterface) fModulesController.getSelectedTTSModule();
				if (fTTS != null)
					return (fTTS.selectEngine(engineIdentifier));
				return (false);
			}
			return (false);
		} catch (Exception ex)
		{
			log.error("Error in selectEngine method:\n" + StackTrace.stackToString(ex));
			return (false);
		}
	}

	@Override
	public String getSelectedEngineIdentifier()
	{
		return (fTTS.getSelectedEngineIdentifier());
	}

	@Override
	public boolean selectLanguage(String languageCode)
	{
		return (fTTS.selectLanguage(languageCode));
	}

	@Override
	public String getSelectedLanguageCode()
	{
		return (fTTS.getSelectedLanguageCode());
	}

	@Override
	public String[] getVoiceIdentifiers()
	{
		return (fTTS.getVoiceIdentifiers());
	}

	@Override
	public boolean selectVoice(String voiceIdentifier)
	{
		return (fTTS.selectVoice(voiceIdentifier));
	}

	@Override
	public String getSelectedVoiceIdentifier()
	{
		return (fTTS.getSelectedVoiceIdentifier());
	}

	@Override
	public String[] getVariantIdentifiers()
	{
		return (fTTS.getVariantIdentifiers());
	}

	@Override
	public boolean selectVariant(String variantIdentifier)
	{
		return (fTTS.selectVariant(variantIdentifier));
	}

	@Override
	public String getSelectedVariantIdentifier()
	{
		return (fTTS.getSelectedVariantIdentifier());
	}

	@Override
	public boolean speak(int messageIdentifier, String message, boolean speakSync)
	{
		return (fTTS.speak(messageIdentifier, message, speakSync));
	}

	@Override
	public boolean speakChar(int messageIdentifier, String message, boolean speakSync)
	{
		return (fTTS.speakChar(messageIdentifier, message, speakSync));
	}

	@Override
	public boolean speakKey(int messageIdentifier, String message, boolean speakSync)
	{
		return (fTTS.speakKey(messageIdentifier, message, speakSync));
	}

	@Override
	public boolean speakSpell(int messageIdentifier, String message, boolean speakSync)
	{
		return (fTTS.speakSpell(messageIdentifier, message, speakSync));
	}

	@Override
	public boolean pause()
	{
		return (fTTS.pause());
	}

	@Override
	public boolean resume()
	{
		return (fTTS.resume());
	}

	@Override
	public boolean stop()
	{
		return (fTTS.stop());
	}

	@Override
	public boolean cancel()
	{
		return (fTTS.cancel());
	}

	@Override
	public boolean setVolume(float volume)
	{
		return (fTTS.setVolume(volume));
	}

	@Override
	public float getVolume()
	{
		return (fTTS.getVolume());
	}

	@Override
	public boolean setRate(float rate)
	{
		return (fTTS.setRate(rate));
	}

	@Override
	public float getRate()
	{
		return (fTTS.getRate());
	}

	@Override
	public boolean setRateBoostMode(boolean mode)
	{
		return (fTTS.setRateBoostMode(mode));
	}

	@Override
	public boolean setPitch(float pitch)
	{
		return (fTTS.setPitch(pitch));
	}

	@Override
	public float getPitch()
	{
		return (fTTS.getPitch());
	}

	@Override
	public boolean setPunctuationLevel(String level)
	{
		return (fTTS.setPunctuationLevel(level));
	}

	@Override
	public boolean setCharCapitalIndicator(String charCapitalIndicator, String arg)
	{
		return (fTTS.setCharCapitalIndicator(charCapitalIndicator, arg));
	}

	@Override
	public boolean setSilenceBeforeMessage(int silenceMilliseconds)
	{
		return (fTTS.setSilenceBeforeMessage(silenceMilliseconds));
	}

	@Override
	public boolean setProperty(String key, String value)
	{
		return (fTTS.setProperty(key, value));
	}

	@Override
	public String getProperty(String key)
	{
		return (fTTS.getProperty(key));
	}

	@Override
	public boolean setEventsListener(SHEventsListenerInterface eventsListener)
	{
		return (fTTS.setEventsListener(eventsListener));
	}

	@Override
	public boolean setAudioDataHandler(SHAudioDataHandlerInterface audioDataHandler)
	{
		return (fTTS.setAudioDataHandler(audioDataHandler));
	}

	/*
	 *This method is intended to be used by SHEngineWatchdog to select alternative available engine 
	 *and voice when the watchdog failed to restart a crashed engine 
	 */
	@Override
	public boolean selectClientEngineVoice(String clientIdentifier, String engineIdentifier, String voiceIdentifier)
	{
		// Errors logged by methods
		if (selectEngine(engineIdentifier))
			return (selectVoice(voiceIdentifier));
		return (false);
	}

}
