/*
* Copyright (c) Isaac Porat 2011 - 2013 all rights reserved.
* This file is released under the GNU General Public License, version 2, or (at your option) any later version.
* See the file SpeechHub_License.txt for more details.
*/

package sh.tts;

import log.common.CascadingLoggerInterface;
import sh.audio.SHAudioDeliveryParallelInterface;
import sh.audio.javasound.SHWindowsTrimThreadAudioDeliveryParallelJavasound;
import sh.connections.SHAudioDataHandlerInterface;
import sh.connections.SHCustomeHandlersInterface;
import sh.connections.SHEngineInterface;
import sh.connections.SHEventsListenerInterface;
import sh.connections.SHMessageEnvelopeInterface;
import sh.connections.SHModuleInterface;
import sh.connections.SHModuleTimestampInterface;
import sh.connections.SHTTSInterface;
import sh.connections.SHUpdateParametersInterface;
import core.utilities.OS;
import core.utilities.StackTrace;

/**
 * TTS audio player javasound adaptor
 * Provides audio javasound player using message envelope returned from TTS
 *  
 * @author Isaac Porat
 *
 */
public class SHTTSExternalAudioPlayerParallelJavasoundAdaptor implements SHTTSInterface, SHModuleInterface,
		SHEngineInterface
{
	private CascadingLoggerInterface log;
	private SHTTSInterface fTTS;
	private SHAudioDeliveryParallelInterface fAudioDelivery;
	private SHAudioDataHandlerInterface fAudioDataHandler;
	private int fMessageIdentifierCounter = 0;
	private boolean fCatchupModeMessageDelivered = false;

	public SHTTSExternalAudioPlayerParallelJavasoundAdaptor(SHTTSInterface tts, CascadingLoggerInterface logger)
	{
		log = logger.clone(SHTTSExternalAudioPlayerParallelJavasoundAdaptor.class);
		fTTS = tts;
		loadAudioDelivery();
		fAudioDataHandler = new AudioDataHandler();
		fTTS.setAudioDataHandler(fAudioDataHandler);
	}

	private void loadAudioDelivery()
	{
		try
		{
			if (fAudioDelivery != null)
				fAudioDelivery.close();
			if (OS.isWindows())
				fAudioDelivery = new SHWindowsTrimThreadAudioDeliveryParallelJavasound(log);
			else
			{
				/*
				* Assume Linux
				* Replace the Windows class with your own Linux class
				* A good start would be to copy the existing Windows class and modify as necessary
				* Note old Linux specific code left in the Windows class for future Linuxx implementation
				*/
				fAudioDelivery = new SHWindowsTrimThreadAudioDeliveryParallelJavasound(log);
			}
		} catch (Exception ex)
		{
			log.error("Error in loadAudioLine method:\n" + StackTrace.stackToString(ex));
		}
	}

	@Override
	public boolean configureTiming(SHModuleTimestampInterface moduleTimestamp, boolean masterCancel)
	{
		// Not applicable
		return (false);
	}

	@Override
	public boolean initialize()
	{
		return (fTTS.initialize());
	}

	@Override
	public boolean close()
	{
		return (fTTS.close());
	}

	@Override
	public boolean setPriority(int priorityLevel, boolean selfCancel)
	{
		// Not applicable
		return (false);
	}

	@Override
	public String[] getEngineIdentifiers()
	{
		try
		{
			return (fTTS.getEngineIdentifiers());
		} catch (Exception ex)
		{
			log.error("Error in getEngineIdentifiers method:\n" + StackTrace.stackToString(ex));
			return (null);
		}
	}

	@Override
	public boolean selectEngine(String engineIdentifier)
	{
		try
		{
			fAudioDelivery.disableCatchupMode(); // Reset to default conversion mode
			return (fTTS.selectEngine(engineIdentifier));
		} catch (Exception ex)
		{
			log.error("Error in selectEngine method:\n" + StackTrace.stackToString(ex));
			return (false);
		}
	}

	@Override
	public String getSelectedEngineIdentifier()
	{
		return (fTTS.getSelectedEngineIdentifier());
	}

	@Override
	public boolean selectLanguage(String languageCode)
	{
		return (fTTS.selectLanguage(languageCode));
	}

	@Override
	public String getSelectedLanguageCode()
	{
		return (fTTS.getSelectedLanguageCode());
	}

	@Override
	public String[] getVoiceIdentifiers()
	{
		return (fTTS.getVoiceIdentifiers());
	}

	@Override
	public boolean selectVoice(String voiceIdentifier)
	{
		fAudioDelivery.disableCatchupMode(); // Reset to default conversion mode
		return (fTTS.selectVoice(voiceIdentifier));
	}

	@Override
	public String getSelectedVoiceIdentifier()
	{
		return (fTTS.getSelectedVoiceIdentifier());
	}

	@Override
	public String[] getVariantIdentifiers()
	{
		return (fTTS.getVariantIdentifiers());
	}

	@Override
	public boolean selectVariant(String variantIdentifier)
	{
		return (fTTS.selectVariant(variantIdentifier));
	}

	@Override
	public String getSelectedVariantIdentifier()
	{
		return (fTTS.getSelectedVariantIdentifier());
	}

	@Override
	public boolean speak(SHMessageEnvelopeInterface messageEnvelope)
	{
		// Not applicable
		return (false);
	}

	private String conditionallyNotifyCatchupMode(String message)
	{
		if (fAudioDelivery == null)
			return (message);
		if (fAudioDelivery.getCatchupMode())
		{
			if (message.length() == 1) // Postponed  message
				return (message);
			if (fCatchupModeMessageDelivered == false)
			{
				// Replace current message for maximum effect
				message = "SpeechHub message: Slow synthesizer response detected.  ";
				fCatchupModeMessageDelivered = true;
			}
		} else
			fCatchupModeMessageDelivered = false;
		return (message);
	}

	@Override
	public boolean speak(int messageIdentifier, String message, boolean speakSync)
	{
		message = conditionallyNotifyCatchupMode(message);
		// If message identifier 0 generate message identifier automatically
		if (messageIdentifier == 0)
			messageIdentifier = fMessageIdentifierCounter++;
		return (fTTS.speak(messageIdentifier, message, speakSync));
	}

	@Override
	public boolean speakChar(int messageIdentifier, String message, boolean speakSync)
	{
		// If message identifier 0 generate message identifier automatically
		if (messageIdentifier == 0)
			messageIdentifier = fMessageIdentifierCounter++;
		return (fTTS.speakChar(messageIdentifier, message, speakSync));
	}

	@Override
	public boolean speakKey(int messageIdentifier, String message, boolean speakSync)
	{
		// If message identifier 0 generate message identifier automatically
		if (messageIdentifier == 0)
			messageIdentifier = fMessageIdentifierCounter++;
		return (fTTS.speakKey(messageIdentifier, message, speakSync));
	}

	@Override
	public boolean speakSpell(int messageIdentifier, String message, boolean speakSync)
	{
		// If message identifier 0 generate message identifier automatically
		if (messageIdentifier == 0)
			messageIdentifier = fMessageIdentifierCounter++;
		return (fTTS.speakSpell(messageIdentifier, message, speakSync));
	}

	@Override
	public boolean pause()
	{
		return (fTTS.pause());
	}

	@Override
	public boolean resume()
	{
		return (fTTS.resume());
	}

	@Override
	public boolean stop()
	{
		return (fTTS.stop());
	}

	@Override
	public boolean cancel()
	{
		try
		{
			boolean status = fTTS.cancel();
			fAudioDelivery.cancel();
			return (status);
		} catch (Exception ex)
		{
			log.error("Error in cancel method:\n" + ex);
			return (false);
		}
	}

	@Override
	public boolean setVolume(float volume)
	{
		return (fTTS.setVolume(volume));
	}

	@Override
	public float getVolume()
	{
		return (fTTS.getVolume());
	}

	@Override
	public boolean setRate(float rate)
	{
		fAudioDelivery.disableCatchupMode(); // Reset to default conversion mode
		return (fTTS.setRate(rate));
	}

	@Override
	public float getRate()
	{
		return (fTTS.getRate());
	}

	@Override
	public boolean setRateBoostMode(boolean mode)
	{
		return (fTTS.setRateBoostMode(mode));
	}

	@Override
	public boolean setPitch(float pitch)
	{
		return (fTTS.setPitch(pitch));
	}

	@Override
	public float getPitch()
	{
		return (fTTS.getPitch());
	}

	@Override
	public boolean setPunctuationLevel(String level)
	{
		return (fTTS.setPunctuationLevel(level));
	}

	@Override
	public boolean setCharCapitalIndicator(String charCapitalIndicator, String arg)
	{
		return (fTTS.setCharCapitalIndicator(charCapitalIndicator, arg));
	}

	@Override
	public boolean setSilenceBeforeMessage(int silenceMilliseconds)
	{
		return (fTTS.setSilenceBeforeMessage(silenceMilliseconds));
	}

	@Override
	public boolean setProperty(String key, String value)
	{
		return (fTTS.setProperty(key, value));
	}

	@Override
	public String getProperty(String key)
	{
		return (fTTS.getProperty(key));
	}

	@Override
	public boolean setEventsListener(SHEventsListenerInterface eventsListener)
	{
		return (fTTS.setEventsListener(eventsListener));
	}

	@Override
	public boolean setAudioDataHandler(SHAudioDataHandlerInterface audioDataHandler)
	{
		// Disabled handler is declared and applied in this class
		return (false);
	}

	@Override
	public boolean setCustomeHandlers(SHCustomeHandlersInterface customeHandlers)
	{
		// Not applicable
		return (false);
	}

	@Override
	public boolean updateParametersDirectly(SHUpdateParametersInterface updateParameters)
	{
		// Not applicable
		return (false);
	}

	@Override
	public boolean setUpdateIdentifier(int updateIdentifier)
	{
		// Not applicable
		return (false);
	}

	@Override
	public boolean isUpdateAccepted(int updateIdentifier)
	{
		// Not applicable
		return (false);
	}

	private class AudioDataHandler implements SHAudioDataHandlerInterface
	{

		@Override
		public boolean processAudio(byte[] samples)
		{
			// Unused
			return (false);
		}

		@Override
		public boolean processAudio(SHMessageEnvelopeInterface messageEnvelope)
		{
			// Called method trap and displays exceptions
			return (fAudioDelivery.playAudioData(messageEnvelope));
		}

	}

}
