/*
* Copyright (c) Isaac Porat 2011 - 2013 all rights reserved.
* This file is released under the GNU General Public License, version 2, or (at your option) any later version.
* See the file SpeechHub_License.txt for more details.
*/

package sh.tts;

import log.common.CascadingLoggerInterface;
import sh.common.SHClientParameters;
import sh.connections.SHClientParametersInterface;
import sh.connections.SHConstants;
import sh.connections.SHCustomeHandlersInterface;
import sh.connections.SHEngineInterface;
import sh.connections.SHMessageEnvelopeInterface;
import sh.connections.SHModuleInterface;
import sh.connections.SHModuleTimestampInterface;
import sh.connections.SHAudioDataHandlerInterface;
import sh.connections.SHEventsListenerInterface;
import sh.connections.SHTTSInterface;
import sh.connections.SHUpdateParametersInterface;
import sh.module.SHModuleTimestamp;
import sh.module.tts.SHModuleTTSWrapper;

/**
 * TTS to module adaptor
 * Implements complete stack module wrapping a single connector
 * 
 * @author Isaac Porat
 */
public class SHTTSModuleAdaptor implements SHTTSInterface, SHModuleInterface, SHEngineInterface
{
	private CascadingLoggerInterface log;
	private SHModuleInterface fModule;
	private SHClientParametersInterface fClientParameters;

	public SHTTSModuleAdaptor(SHEngineInterface engine, String clientIdentifier, CascadingLoggerInterface logger)
	{
		log = logger.clone(SHTTSModuleAdaptor.class);
		fClientParameters = new SHClientParameters(clientIdentifier);
		fClientParameters.setModuleTimestamp(new SHModuleTimestamp());
		fModule = new SHModuleTTSWrapper(engine, fClientParameters, logger);
	}

	@Override
	public boolean initialize()
	{
		try
		{
			return (fModule.initialize());
		} catch (Exception ex)
		{
			log.error("Error in initialize method:\n" + ex);
			return (false);
		}
	}

	@Override
	public boolean close()
	{
		try
		{
			return (fModule.close());
		} catch (Exception ex)
		{
			log.error("Error in close method:\n" + ex);
			return (false);
		}
	}

	@Override
	public String[] getEngineIdentifiers()
	{
		try
		{
			return (fModule.getEngineIdentifiers());
		} catch (Exception ex)
		{
			log.error("Error in getEngineIdentifiers method:\n" + ex);
			return (null);
		}
	}

	@Override
	public boolean selectEngine(String engineIdentifier)
	{
		try
		{
			fClientParameters.setEngineIdentifier(engineIdentifier);
			return (fModule.updateParametersDirectly(fClientParameters.getClonedUpdateParameters()));
		} catch (Exception ex)
		{
			log.error("Error in selectEngine Priority method:\n" + ex);
			return (false);
		}
	}

	@Override
	public String getSelectedEngineIdentifier()
	{
		try
		{
			return (fModule.getSelectedEngineIdentifier());
		} catch (Exception ex)
		{
			log.error("Error in getSelectedEngineIdentifier Priority method:\n" + ex);
			return (null);
		}
	}

	@Override
	public String getSelectedLanguageCode()
	{
		try
		{
			return (fModule.getSelectedLanguageCode());
		} catch (Exception ex)
		{
			log.error("Error in getLanguage method:\n" + ex);
			return (null);
		}
	}

	@Override
	public boolean selectLanguage(String languageCode)
	{
		try
		{
			fClientParameters.setLanguageCode(languageCode);
			return (true);
		} catch (Exception ex)
		{
			log.error("Error in setLanguage method:\n" + ex);
			return (false);
		}
	}

	@Override
	public String[] getVoiceIdentifiers()
	{
		try
		{
			return (fModule.getVoiceIdentifiers());
		} catch (Exception ex)
		{
			log.error("Error in getVoiceIdentifiers method:\n" + ex);
			return (null);
		}
	}

	@Override
	public boolean selectVoice(String voiceIdentifier)
	{
		try
		{
			fClientParameters.setVoiceIdentifier(voiceIdentifier);
			return (true);
		} catch (Exception ex)
		{
			log.error("Error in selectVoice method:\n" + ex);
			return (false);
		}
	}

	@Override
	public String getSelectedVoiceIdentifier()
	{
		try
		{
			return (fModule.getSelectedVoiceIdentifier());
		} catch (Exception ex)
		{
			log.error("Error in getSelectedVoiceIdentifier method:\n" + ex);
			return (null);
		}
	}

	@Override
	public String[] getVariantIdentifiers()
	{
		try
		{
			return (fModule.getVariantIdentifiers());
		} catch (Exception ex)
		{
			log.error("Error in getVariantIdentifiers method:\n" + ex);
			return (null);
		}
	}

	@Override
	public boolean selectVariant(String variantIdentifier)
	{
		try
		{
			fClientParameters.setVariantIdentifier(variantIdentifier);
			return (true);
		} catch (Exception ex)
		{
			log.error("Error in selectVariant method:\n" + ex);
			return (false);
		}
	}

	@Override
	public String getSelectedVariantIdentifier()
	{
		try
		{
			return (fModule.getSelectedVariantIdentifier());
		} catch (Exception ex)
		{
			log.error("Error in getSelectedVariantIdentifier method:\n" + ex);
			return (null);
		}
	}

	@Override
	public boolean speak(int messageIdentifier, String message, boolean speakSync)
	{
		try
		{
			return (fModule.speak(messageIdentifier, message, speakSync));
		} catch (Exception ex)
		{
			log.error("Error in speak method:\n" + ex);
			return (false);
		}
	}

	@Override
	public boolean speakChar(int messageIdentifier, String message, boolean speakSync)
	{
		try
		{
			return (fModule.speakChar(messageIdentifier, message, speakSync));
		} catch (Exception ex)
		{
			log.error("Error in speakChar method:\n" + ex);
			return (false);
		}
	}

	@Override
	public boolean speakKey(int messageIdentifier, String message, boolean speakSync)
	{
		try
		{
			return (fModule.speakKey(messageIdentifier, message, speakSync));
		} catch (Exception ex)
		{
			log.error("Error in speakKey method:\n" + ex);
			return (false);
		}
	}

	@Override
	public boolean speakSpell(int messageIdentifier, String message, boolean speakSync)
	{
		try
		{
			if (setProperty(SHConstants.PROP_MODULE_SPELL_MODE, "true") == false)
				return (false);
			return (fModule.speak(messageIdentifier, message, speakSync));
		} catch (Exception ex)
		{
			log.error("Error in speakSpell method:\n" + ex);
			return (false);
		} finally
		{
			setProperty(SHConstants.PROP_MODULE_SPELL_MODE, "false");
		}
	}

	@Override
	public boolean pause()
	{
		try
		{
			return (fModule.pause());
		} catch (Exception ex)
		{
			log.error("Error in pause method:\n" + ex);
			return (false);
		}
	}

	@Override
	public boolean resume()
	{
		try
		{
			return (fModule.resume());
		} catch (Exception ex)
		{
			log.error("Error in resume method:\n" + ex);
			return (false);
		}
	}

	@Override
	public boolean stop()
	{
		// Not applicable
		return (false);
	}

	@Override
	public boolean cancel()
	{
		try
		{
			return (fModule.cancel());
		} catch (Exception ex)
		{
			log.error("Error in cancel method:\n" + ex);
			return (false);
		}
	}

	@Override
	public boolean setVolume(float volume)
	{
		try
		{
			fClientParameters.setVolume(volume);
			return (true);
		} catch (Exception ex)
		{
			log.error("Error in setVolume method:\n" + ex);
			return (false);
		}
	}

	@Override
	public float getVolume()
	{
		try
		{
			return (fModule.getVolume());
		} catch (Exception ex)
		{
			log.error("Error in getVolume method:\n" + ex);
			return (SHConstants.FLOAT_PARAMETER_ERROR);
		}
	}

	@Override
	public boolean setRate(float rate)
	{
		try
		{
			fClientParameters.setRate(rate);
			return (true);
		} catch (Exception ex)
		{
			log.error("Error in setRate method:\n" + ex);
			return (false);
		}
	}

	@Override
	public float getRate()
	{
		try
		{
			return (fModule.getRate());
		} catch (Exception ex)
		{
			log.error("Error in getRate method:\n" + ex);
			return (SHConstants.FLOAT_PARAMETER_ERROR);
		}
	}

	@Override
	public boolean setRateBoostMode(boolean mode)
	{
		try
		{
			return (fModule.setProperty(SHConstants.PROP_ENGINE_USE_SONIC, mode ? "true" : "false"));
		} catch (Exception ex)
		{
			log.error("Error in setRateBoost method:\n" + ex);
			return (false);
		}
	}

	@Override
	public boolean setPitch(float pitch)
	{
		try
		{
			fClientParameters.setPitch(pitch);
			return (true);
		} catch (Exception ex)
		{
			log.error("Error in setPitch method:\n" + ex);
			return (false);
		}
	}

	@Override
	public float getPitch()
	{
		try
		{
			return (fModule.getPitch());
		} catch (Exception ex)
		{
			log.error("Error in getPitch method:\n" + ex);
			return (SHConstants.FLOAT_PARAMETER_ERROR);
		}
	}

	@Override
	public boolean setPunctuationLevel(String level)
	{
		try
		{
			return (fModule.setProperty(SHConstants.PROP_MODULE_PUNCTUATION_LEVEL, level));
		} catch (Exception ex)
		{
			log.error("Error in setPunctuationLevel method:\n" + ex);
			return (false);
		}
	}

	@Override
	public boolean setCharCapitalIndicator(String charCapitalIndicator, String arg)
	{
		try
		{
			if (charCapitalIndicator == null)
				return (false);
			if (fModule.setProperty(SHConstants.PROP_MODULE_CHAR_CAPITAL_INDICATOR, charCapitalIndicator))
			{
				// Only argument for pitch level is supported
				if ((charCapitalIndicator.equals(SHConstants.MODULE_CHAR_CAPITAL_INDICATOR_PITCH)) && (arg != null))
					return (fModule.setProperty(SHConstants.PROP_MODULE_CHAR_CAPITAL_INDICATOR_PITCH_LEVEL, arg));
				return (true); // For all other properties with no arg requirement
			}
			return (false);
		} catch (Exception ex)
		{
			log.error("Error in setCapitalIndicator method:\n" + ex);
			return (false);
		}
	}

	@Override
	public boolean setSilenceBeforeMessage(int silenceMilliseconds)
	{
		try
		{
			return (fModule.setProperty(SHConstants.PROP_MODULE_SILENCE_BEFORE_MESSAGE,
					Integer.toString(silenceMilliseconds)));
		} catch (Exception ex)
		{
			log.error("Error in setSilenceBeforeMessage method:\n" + ex);
			return (false);
		}
	}

	@Override
	public boolean setProperty(String key, String value)
	{
		try
		{
			return (fModule.setProperty(key, value));
		} catch (Exception ex)
		{
			log.error("Error in setProperty method:\n" + ex);
			return (false);
		}
	}

	@Override
	public String getProperty(String key)
	{
		try
		{
			return (fModule.getProperty(key));
		} catch (Exception ex)
		{
			log.error("Error in getProperty method:\n" + ex);
			return (null);
		}
	}

	@Override
	public boolean setEventsListener(SHEventsListenerInterface eventsListener)
	{
		try
		{
			return (fModule.setEventsListener(eventsListener));
		} catch (Exception ex)
		{
			log.error("Error in setEventsListener method:\n" + ex);
			return (false);
		}
	}

	@Override
	public boolean setAudioDataHandler(SHAudioDataHandlerInterface audioDataHandler)
	{
		try
		{
			return (fModule.setAudioDataHandler(audioDataHandler));
		} catch (Exception ex)
		{
			log.error("Error in setAudioDataHandler method:\n" + ex);
			return (false);
		}
	}

	/*
	 * Compatibility with SHModuleInterface only required to ensure
	 * storage in tcf version module controllers
	*/

	@Override
	public boolean configureTiming(SHModuleTimestampInterface moduleTimestamp, boolean masterCancel)
	{
		// Not applicable
		return (false);
	}

	@Override
	public boolean setPriority(int priorityLevel, boolean selfCancel)
	{
		// Not applicable
		return (false);
	}

	@Override
	public boolean speak(SHMessageEnvelopeInterface messageEnvelope)
	{
		// Not applicable
		return (false);
	}

	@Override
	public boolean setCustomeHandlers(SHCustomeHandlersInterface customeHandlers)
	{
		// Not applicable
		return (false);
	}

	@Override
	public boolean updateParametersDirectly(SHUpdateParametersInterface updateParameters)
	{
		// Not applicable
		return (false);
	}

	@Override
	public boolean setUpdateIdentifier(int updateIdentifier)
	{
		// Not applicable
		return (false);
	}

	@Override
	public boolean isUpdateAccepted(int updateIdentifier)
	{
		// Not applicable
		return (false);
	}

}
