/*
* Copyright (c) Isaac Porat 2011 - 2013 all rights reserved.
* This file is released under the GNU General Public License, version 2, or (at your option) any later version.
* See the file SpeechHub_License.txt for more details.
*/

package sh.tts.independent;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.nio.MappedByteBuffer;
import java.nio.channels.FileChannel;
import java.nio.charset.Charset;

import core.utilities.StackTrace;

import log.common.BasicLogger;
import log.common.CascadingLoggerInterface;
import log.common.SimpleLogger;
import sh.audio.SHWaveListener;
import sh.common.SHConfig;
import sh.connections.SHConstants;
import sh.connections.SHTTSInterface;
import sh.independent.SHConfigIndependent;
import sh.independent.SHIndependentModulesController;
import sh.services.SHTTSModulesControllerServiceInterface;
import sh.tts.SHTTS;

/**
 * Independent (no server) multi full stack modules TTS with CLI user interface 
 * @author Bill Cox
 * 
 */
public class SHTTSCli
{
	private final static boolean SERVER_MODE = false;
	private final static String CLIENT_IDENTIFIER = "DefaultClient";

	private static void usage(String format, Object... args)
	{
		BasicLogger.notifyPrintf(format, args);
		BasicLogger.notifyPrintln("Usage: sh-say [parameter]... [text]\n" + "Parameters can be:\n"
				+ "-d currendDirectory: Tell speech-hub the current directory\n"
				+ "           Without this, -w and -f flags must use absolute paths.\n"
				+ "-s speed:  Set the speed of speech.  -100 is slowest, 100 is fastest.\n"
				+ "-p pitch:  Set pitch.  -100 is lowest, 100 is highest.\n" + "-f file:   Speak the text in file.\n"
				+ "-e engine: Switch to the engine, one of: espeak, ibmtts\n"
				+ "-f file:   Read input text from the file.\n" + "-v voice:  Switch to the named voice\n"
				+ "-l:        List supported voices\n"
				+ "-w file:   Write speech output to a file rather than speak it\n"
				+ "--:        Indicate end of options so the next parameter will be taken as text\n");
		System.exit(1);
	}

	public static String readFile(String path, String currentDirectory)
	{
		try
		{
			if (currentDirectory != null && !path.startsWith("/"))
			{
				path = currentDirectory + "/" + path;
			}
			FileInputStream stream = new FileInputStream(new File(path));
			try
			{
				FileChannel fc = stream.getChannel();
				MappedByteBuffer bb = fc.map(FileChannel.MapMode.READ_ONLY, 0, fc.size());
				/* Instead of using default, pass in a decoder. */
				return Charset.defaultCharset().decode(bb).toString();
			} finally
			{
				stream.close();
			}
		} catch (IOException e)
		{
			usage("Unable to read file %s\n", path);
		}
		return null; // Dummy return
	}

	public static float stringToFloat(String s)
	{
		try
		{
			float f = Float.valueOf(s.trim()).floatValue();
			return f;
		} catch (NumberFormatException nfe)
		{
			usage("Expected floating point number after flag");
		}
		return 0.0f; // Dummy return;
	}

	public static void main(String[] args)
	{
		float speed = 1.0f;
		float pitch = 1.0f;
		String voiceName = null;
		int xArg = 0;
		char c;
		boolean listVoices = false;
		String outputFile = null;
		String currentDirectory = null;

		String text = "Hello, world!";
		String engineName = SHConstants.ENGINE_IDENTIFIER_ESPEAK;

		while (xArg < args.length && args[xArg].charAt(0) == '-')
		{
			c = args[xArg].charAt(1);
			if (c == '-')
			{
				xArg++;
				break;
			}
			switch (c)
			{
			case 'd':
				if (++xArg == args.length)
				{
					usage("Expected current directory after -d parameter\n");
				}
				currentDirectory = args[xArg];
				break;
			case 'f':
				if (++xArg == args.length)
				{
					usage("Expected file name after -f parameter\n");
				}
				text = readFile(args[xArg], currentDirectory);
				break;
			case 'l':
				listVoices = true;
				break;
			case 'p':
				if (++xArg == args.length)
				{
					usage("Expected pitch after -p parameter\n");
				}
				pitch = stringToFloat(args[xArg]);
				break;
			case 's':
				if (++xArg == args.length)
				{
					usage("Expected speed after -s parameter\n");
				}
				speed = stringToFloat(args[xArg]);
				break;
			case 'e':
				if (++xArg == args.length)
				{
					usage("Expected engine name after -e parameter\n");
				}
				engineName = args[xArg];
				break;
			case 'v':
				if (++xArg == args.length)
				{
					usage("Expected voice name after -v parameter\n");
				}
				voiceName = args[xArg];
				break;
			case 'w':
				if (++xArg == args.length)
				{
					usage("Expected wav file name after -w parameter\n");
				}
				outputFile = args[xArg];
				break;
			default:
				usage("Unknown option %c\n", c);
			}
			xArg++;
		}
		if (xArg + 1 == args.length)
		{
			text = args[xArg];
		} else if (xArg != args.length)
		{
			usage("Only one non-option parameter allowed.");
		}
		int logLevel = 2;
		CascadingLoggerInterface log = new SimpleLogger(SHTTSCli.class, logLevel);

		SHTTSModulesControllerServiceInterface modulesController;
		SHTTSInterface tts = null;
		try
		{
			// Configure locations - settings and  engines
			SHConfigIndependent.configureLocations(log);

			modulesController = new SHIndependentModulesController(SERVER_MODE, log);
			// Register the modulesController globally
			SHConfig.setTTSModulesControllerService(modulesController);
			tts = SHTTS.getInstance(CLIENT_IDENTIFIER, log);
			tts.selectEngine(engineName);
		} catch (Exception ex)
		{
			log.error("Error in main method:\n" + StackTrace.stackToString(ex));
			return;
		}

		if (listVoices)
		{
			String voices[] = tts.getVoiceIdentifiers();
			for (int i = 0; i < voices.length; i++)
			{
				BasicLogger.notifyPrintln(voices[i]);
			}
		} else
		{
			tts.setRate(speed);
			tts.setPitch(pitch);
			if (voiceName != null)
			{
				tts.selectVoice(voiceName);
			}
			SHWaveListener listener = null;
			if (outputFile != null)
			{
				if (currentDirectory != null && !outputFile.startsWith("/"))
				{
					outputFile = currentDirectory + "/" + outputFile;
				}
				int sampleRate = Integer.parseInt(tts.getProperty(SHConstants.PROP_ENGINE_SAMPLE_RATE));
				listener = new SHWaveListener(outputFile, sampleRate, log);
				tts.setAudioDataHandler(listener);
			}
			tts.speak(0, text, true);
			if (outputFile != null)
			{
				listener.save();
			}
		}
		tts.close();
		System.exit(0);
	}

}
