/***************************** License **********************************

    Copyright(c)2005, Communicative Machines
    All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are 
met:

- Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright 
  notice, this list of conditions and the following disclaimer in the 
  documentation and/or other materials provided with the distribution.

- Neither the name of the copyright holders nor the names of its 
  contributors may be used to endorse or promote products derived from 
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER 
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

************************************************************************/
// Communicator.java


package com.cmlabs.air;

import java.util.*;
import org.w3c.dom.*;

public class Communicator extends Thread implements MessageReceiver, MessageAcceptor, MessageSender {

  class Maintainer extends Thread {
    Communicator parent = null;
    int timeout = 5000;
    boolean shouldContinue = true;
    boolean isRunning = false;
    public Maintainer(Communicator parent) {
      this.parent = parent;
    }
    public void run() {
      isRunning = true;
      long lastRun = System.currentTimeMillis();
      while (shouldContinue) {
	Utils.pause(50);
	if ((shouldContinue) && (System.currentTimeMillis() - lastRun > timeout)) {
	  parent.doMaintenance();
	  lastRun = System.currentTimeMillis();
	}
      }
      isRunning = false;
    }
    public boolean terminate() {
      shouldContinue = false;
      long start = System.currentTimeMillis();
      while ((isRunning) && (System.currentTimeMillis() - start < 100))
	Utils.pause(20);
      return (!isRunning);
    }
  }

  boolean debug = false;
  boolean shouldListen = true;
  TCPLocation cnsLocation = null;
  NetworkConnection cnsCon = null;
  IncomingSocketConnector incomingListener = null;
  public String compName = "";
  String configuration = "";
  String description = "";
  double priority = 2;
//  Hashtable cnsCache = new Hashtable();
//  Hashtable sendConnections = new Hashtable();
//  Hashtable receiveConnections = new Hashtable();
  ObjectDictionary cnsCache = new ObjectDictionary();
  ObjectDictionary sendConnections = new ObjectDictionary();
  ObjectDictionary receiveConnections = new ObjectDictionary();
  TextCollection lostSendConnections = new TextCollection();
  ObjectDictionary lostReceiveConnections = new ObjectDictionary();
  ObjectDictionary parameters = new ObjectDictionary();
  TextDictionary globalInfo = new TextDictionary();

  MessageAcceptor messageAcceptor = null;
  TextDictionary statusDict = new TextDictionary();
  Time createdTime = new Time();
  public PerfStat perfStat = new PerfStat();

  Maintainer maintainer = null;

  // test
  public ObjectDictionary getParameters() { return parameters; }

  public Communicator(String name, String cnsHost, int cnsPort) {
    cnsLocation = new TCPLocation(cnsHost, cnsPort, "CNS");
    cnsCache.put("CNS", cnsLocation);
    compName = name;
    perfStat.name = name;
    statusDict.put("Started", createdTime.printString());
  }

  public void destroy() {
    shutdown();
  }

  public boolean init() {
    if (!connectToServer())
	    return false;

    maintainer = new Maintainer(this);
    maintainer.start();

    configuration = getConfiguration("");
    parseConfiguration(configuration);
    return true;
  }

  public boolean initNoListen() {
    shouldListen = false;
    if (!connectToServer())
      return false;

    maintainer = new Maintainer(this);
    maintainer.start();

    configuration = getConfiguration("");
    parseConfiguration(configuration);
    return true;
  }

  synchronized public boolean shutdown() {
    if (cnsCon != null)
      cnsCon.destroy();
    cnsCon = null;
    if (incomingListener != null)
      incomingListener.stopAll();
    incomingListener = null;
    disconnectNetworkConnections();
    return true;
  }

  public void run() {

  }

  public String getConfiguration(String name) {

    Message send;

    if (globalInfo.getCount() == 0) {
      send = new Message(compName, "Config", "GET_ENTRY");
      send.content = "Global";
      Message recvglobal = sendReceiveMessage("Config", send);
      if ( (recvglobal != null) && (recvglobal.type.equals("ENTRY")) &&
	  (recvglobal.object != null)) {
	globalInfo = (TextDictionary) recvglobal.object;
      }
    }

    // if name is empty, use compName
    String entry = name;
    if (entry.length() == 0)
      entry = compName;

    send = new Message(compName, "Config", "GET_ENTRY", entry, "");
    Message recv = sendReceiveMessage("Config", send);

    if ( (recv != null) && (recv.type.equals("ENTRY")) ) {
      return recv.getContent();
    }

    return "";
  }

  public String getComponentName() {
    return compName;
  }

  public boolean parseConfiguration(String confXML) {
    // In configXML we find our own configuration <xxx name="xxx"...><parameter name="">...
    XMLElement node = Utils.xmlGetRootNodeFromXML(confXML);
    if (node == null)
      return false;

    if (node.hasAttribute("priority"))
      priority = node.getDoubleAttribute("priority");
    setPlugPriority(priority);

    XMLElement xmlNode;
    Parameter parameter;

    for (Enumeration e=node.enumerateChildren(); e.hasMoreElements(); ) {
      xmlNode = (XMLElement) e.nextElement();
      if (xmlNode != null) {
	if (xmlNode.getName().equalsIgnoreCase("description")) {
	  description = xmlNode.getContent();
	}
	else if (xmlNode.getName().equalsIgnoreCase("parameter")) {
	  parameter = new Parameter(xmlNode);
	  if (parameter.name.length() > 0) {
	    parameters.put(parameter.name, parameter);
	  }
	}
      }
    }
    return true;
  }


  public boolean disconnectNetworkConnections() {

    NetworkConnection con;
    Enumeration e;

    for (int n=0; n<sendConnections.getCount(); n++) {
      con = (NetworkConnection) sendConnections.get(n);
      con.destroy();
    }
    sendConnections.removeAll();

    for (int n=0; n<receiveConnections.getCount(); n++) {
      con = (NetworkConnection) receiveConnections.get(n);
      con.destroy();
    }
    receiveConnections.removeAll();
    return true;
  }

  // connects to cns
  synchronized public boolean connectToServer() {

    int PSYPORTTRIES = 3;

      if (incomingListener != null) return true;

      int startPort = cnsLocation.port;
      if (!isConnected()) {
	synchronized (this) {
	  cnsCon = null;
	  while (cnsCon == null) {
	    try {
	      cnsCon = new NetworkConnection(cnsLocation.addr, cnsLocation.port,
					     new NetMessageProtocol());
	      cnsCon.name = "CNS";
	    }
	    catch (Exception e) {
	      cnsCon = null;
	    }

	    if (cnsCon == null) {
	      cnsLocation.port++;
	      if (cnsLocation.port - startPort > PSYPORTTRIES) {
		// System.out.println("JavaAIRPlug: Could not connect to the Server at all... giving up...");
		return false;
	      }
	      else {
		//System.out.println(
		//    "JavaAIRPlug: Could not connect to the Server, retrying " +
		//    (cnsLocation.port - startPort) + " of 10...");
	      }
	    }
	  }
	}
      }

      int port = bind();

      if (port == TCPLocation.NO_LISTEN) {
      }
      else if (port <= 0) {
	System.out.println("AIR: Could not bind to the Server on '" + cnsLocation.toString() + "...");
	return false;
      }

      if (!openReceivingConnection(this, "CNS"))
	return false;

      if (shouldListen) {
	incomingListener = new IncomingSocketConnector(port,
	    new NetMessageProtocol(), this);
	// if the bind port returned doesn't work, get another. Try this 5 times, then fail.
	int bindAttempts = 0;
	while (!incomingListener.init()) {
	  System.out.println("AIR: Could not reserve local port " + port + ", retry binding to the Server...");
	  // Maybe we are an applet?
	  port = bind();
	  bindAttempts++;
	  if (port <= 0) {
	    System.out.println("AIR: Could not bind (" + bindAttempts +
			       ". attempt) to the Server on '" + cnsLocation.toString() + "...");
	    return false;
	  }
	  incomingListener = new IncomingSocketConnector(port,
	      new NetMessageProtocol(), this);
	  if (bindAttempts > 50) {
	    incomingListener = null;
	    return false;
	  }
	}
	incomingListener.start();
      }

      return true;
  }

  public int bind() {

    int port = 0;
    String bindstr = "BIND";
    if (!shouldListen)
      bindstr = "BIND_NOLISTEN";

    Message msg = sendReceiveMessage("CNS", new Message(compName, "CNS", bindstr)); // tcploaction obj - language xml - or not content
    if (msg == null) {
      //System.out.println("PlugConnection: connect: Connection to host: " + cnsHost + " and port: " + cnsPort + " failed. The problem was: connectAcceptMessage == null");
      System.out.println("Error code JAP-2");
      return 0;
    }
    if (! ("BIND_ACCEPT".equals(msg.type))) {
      //System.out.println("PlugConnection: connect: Connection to host: " + cnsHost + " and port: " + cnsPort + " failed. The problem was: !(\"CONNECT_ACCEPT\".equals(connectAcceptMessage.type)");
      //System.out.println("PlugConnection: connect: the message received was: " + connectAcceptMessage.toXml());
      System.out.println("Error code JAP-3 {" + msg.type + "}");
      return 0;
    }
    port = Utils.str2int(msg.content);
    return port;
  }

  public boolean cnsRebind() {
    if ((incomingListener == null) || (incomingListener.port <= 0)) {
      incomingListener = null;
      return connectToServer();
    }

    int port = incomingListener.port;
    return cnsRebindOnPort(port);
  }

  public boolean cnsRebindOnPort(int newport) {

    Message msg;
    if (!shouldListen) {
      msg = new Message(compName, "CNS", "BIND_NOLISTEN");
    }
    else {
      msg = new Message(compName, "CNS", "BIND");
      msg.object = new TCPLocation("", newport, compName);
    }
    msg = sendReceiveMessage("CNS", msg);

    if (msg == null) {
      //System.out.println("PlugConnection: connect: Connection to host: " + cnsHost + " and port: " + cnsPort + " failed. The problem was: connectAcceptMessage == null");
      System.out.println("Error code JAP-2");
      return false;
    }
    if (! ("BIND_ACCEPT".equals(msg.type))) {
      //System.out.println("PlugConnection: connect: Connection to host: " + cnsHost + " and port: " + cnsPort + " failed. The problem was: !(\"CONNECT_ACCEPT\".equals(connectAcceptMessage.type)");
      //System.out.println("PlugConnection: connect: the message received was: " + connectAcceptMessage.toXml());
      System.out.println("Error code JAP-3 {" + msg.type + "}");
      return false;
    }
    int port = Utils.str2int(msg.content);
    return (port == newport);
  }

  synchronized public boolean isConnected() {
//    return ((cnsCon != null) && (cnsCon.isConnected()) && (cnsCon.isAlive()));
    return ((cnsCon != null) && (cnsCon.isConnected()));
  }

  public boolean ping(String name) {

    // long beginTime = System.currentTimeMillis();
    Message msg = sendReceiveMessage(name, new Message(compName, name, "PING"));
    // long endTime = System.currentTimeMillis();
    if (msg == null) {
      //System.out.println("PlugConnection: Could not ping CNS...");
      // System.out.println("Error code JAP-5");
      return false;
    }
    if (! ("PING_SUCCESS".equals(msg.type))) {
      //System.out.println("PlugConnection: CNS did not accept ping, got back: " + connectAcceptMessage.type);
      // System.out.println("Error code JAP-6");
      return false;
    }

    return true;
  }

  public NetworkConnection findExistingConnection(String name) {

    if (name.equalsIgnoreCase("CNS")) {
      synchronized (this) {
	if (checkPrimaryConnection())
	  return cnsCon;
	else
	  return null;
      }
    }

    NetworkConnection con = null;
    if (sendConnections.containsKey(name)) {
      con = (NetworkConnection) sendConnections.get(name);
      if (!con.isConnected()) {
	System.out.println("Found dead connection to " + name);
	sendConnections.remove(con);
	// return findExistingConnection(name);
	return null;
      }
    }
    return con;
  }

  public NetworkConnection openConnection(String name) {

    TCPLocation loc;
    if (name.equalsIgnoreCase("CNS")) {
      loc = cnsLocation;
    }
    else {
      loc = resolve(name);
    }

    if ((loc == null) || (!loc.isValid()))
      return null;

    NetworkConnection con = null;
    try {
    //  con = new NetworkConnection(loc.addr, loc.port, new NetMessageProtocol());
      con = new NetworkConnection(loc, new NetMessageProtocol());
      con.name = name;
    }
    catch (Exception e) {
      // loc is no longer valid...
      cnsCache.remove(name);
      if ("CNS".equalsIgnoreCase(name))
	return null;
      loc = resolve(name);
      if ((loc == null) || (!loc.isValid()))
	return null;
      try {
      //  con = new NetworkConnection(loc.addr, loc.port, new NetMessageProtocol());
        con = new NetworkConnection(loc, new NetMessageProtocol());
	con.name = name;
      }
      catch (Exception ee) {
	cnsCache.remove(name);
	return null;
      }
      // System.out.println("JavaAIRPlug: Could not connect to " + name + " on '" +
	//		 cnsLocation.toString() + "...");
      //e.printStackTrace();
      //System.out.println("Error code JAP-1");
      return null;
    }
    return con;

  }

  public boolean openReceivingConnection(MessageReceiver receiver, String name) {

    NetworkConnection con = openConnection(name);
    if (con == null)
      return false;
    if (!con.initializeAsReceiver(receiver, compName)) {
      con.destroy();
      return false;
    }
    receiveConnections.put(name, con);
    con.name = name+"*";
    return true;
  }

  public TCPLocation resolve(String name) {

    TCPLocation loc = null;
    if (cnsCache.containsKey(name)) {
      loc = (TCPLocation) cnsCache.get(name);
      if ((loc == null) || (loc.isValid()))
	return loc;
    }

    Message msg = sendReceiveMessage("CNS", new Message(compName, "CNS", "RESOLVE_REQUEST", name, "English"));
    if (msg == null) {
      //System.out.println("PlugConnection: Could not ping CNS...");
      // System.out.println("Error code JAP-5");
      return null;
    }
    if (!("RESOLVE_SUCCESS".equals(msg.type))) {
      //System.out.println("PlugConnection: CNS did not accept ping, got back: " + connectAcceptMessage.type);
      // System.out.println("Error code JAP-6");
      return null;
    }
    if (!("xml".equalsIgnoreCase(msg.language))) {
      //System.out.println("PlugConnection: CNS did not accept ping, got back: " + connectAcceptMessage.type);
      // System.out.println("Error code JAP-6");
      return null;
    }

    if (msg.object != null) {
      loc = (TCPLocation) msg.object;
    }
    else if (msg.xmlContentNode != null)
      loc = new TCPLocation(msg.xmlContentNode);
    else
      loc = new TCPLocation(msg.content);
    if (!loc.isValid())
      return null;

    loc.addr = getNonPrivateAddress(loc.addr);

    cnsCache.put(name, loc);
    return loc;
  }

  synchronized public Message sendReceiveMsgTo(String to, Message message) {
    return sendReceiveMessage(to, message);
  }

  synchronized public Message sendReceiveMessage(String to, Message message) {

    NetworkConnection con = findExistingConnection(to);

    if (con == null) {
      if (to.equalsIgnoreCase("CNS")) {
	// We have tried, giving up...
	return null;
      }

      con = openConnection(to);
      if (con == null) {
	if (isConnected()) {
	  message.to = to;
	  con = cnsCon;
	}
	else
	  return null;
      }
      else
	sendConnections.put(to, con);
    }

    long beginTime = System.currentTimeMillis();
    Message msg = con.sendReceiveMessage(message);
    long endTime = System.currentTimeMillis();
    if (msg == null) {
      //System.out.println("PlugConnection: Could not ping CNS...");
      // System.out.println("Error code JAP-5");
      return null;
    }
    else if (msg.type.equalsIgnoreCase("FORWARD_FAILED")) {
      // System.out.println("Message");
      return null;
    }
    return msg;
  }

  public boolean sendMsgTo(String to, Message message) {
    return sendMessage(to, message);
  }

  public boolean sendMessage(String to, Message message) {

    Message msg = sendReceiveMessage(to, message);

    if (msg == null)
      return false;

    if (!msg.type.equalsIgnoreCase("RECEIVE_ACCEPT"))
      return false;

    return true;
  }

  public Message receiveMessage(Message message) {

    if (message == null)
      return null;

    if (message.isresponse.length() > 0) {
      System.out.println("Warning: " + this.compName + " received unsolicited reply message:\n-> " + message.toString() + "\n");
      return null;
    }

    message.receivedTime = new Time();

    String type = message.type;
    String msgID = message.id;

    Message outMsg = null;

    // System.out.println("Received msg with type: " + type);

    if (type.equalsIgnoreCase("ping")) {
      outMsg = new Message(compName, message.from, "PING_SUCCESS");
      outMsg.isresponse = msgID;
      return outMsg;
    }
    else if (type.equalsIgnoreCase("status")) {
      outMsg = new Message(compName, message.from, "STATUS_SUCCESS", "", "");
      outMsg.object = statusDict;
      outMsg.isresponse = msgID;
      return outMsg;
    }
    else if (type.equalsIgnoreCase("status")) {
      outMsg = new Message(compName, message.from, "STATUS_SUCCESS", "", "");
      outMsg.object = statusDict;
      outMsg.isresponse = msgID;
      return outMsg;
    }
    else if (type.equals("PRIORITY_SET")) {
       if (setPlugPriority(Utils.str2int(message.getContent())))
	  return new Message(compName, message.from, "PRIORITY_SET_SUCCESS");
       else
	  return new Message(compName, message.from, "PRIORITY_SET_FAILED");
    }
    else if (type.equals("PRIORITY_GET")) {
       return new Message(compName, message.from, "PRIORITY_GET_SUCCESS", "" + getPlugPriority(), "");
    }
    else if (type.equals("PARAM_SET")) {
       TextCollection oc = Utils.splitString(message.getContent(),"=");
       String parametername = oc.get(0).trim();
       if (!hasParameter(parametername))
	  return new Message(compName, message.from, "PARAM_NOT_EXIST", parametername, "");
       else {
	  String parametervalue = oc.get(1).trim();
	  if (setParameterString(parametername, parametervalue))
	     return new Message(compName, message.from, "PARAM_SET_SUCCESS");
	  else
	     return new Message(compName, message.from, "PARAM_SET_FAILED");
       }
    }
    else if (type.equals("PARAM_ADD_ITEM")) {
       TextCollection oc = Utils.splitString(message.getContent(),"=");
       String parametername = oc.get(0).trim();
       String parametervalue = oc.get(1).trim();
       if (!hasParameter(parametername))
	  return new Message(compName, message.from, "PARAM_NOT_EXIST", parametername, "");
       else {
	  if (addParameterItem(parametername, parametervalue))
	     return new Message(compName, message.from, "PARAM_ADD_ITEM_SUCCESS");
	  else
	     return new Message(compName, message.from, "PARAM_ADD_ITEM_FAILED");
       }
    }
    else if (type.equals("PARAM_REMOVE_ITEM")) {
       TextCollection oc = Utils.splitString(message.getContent(),"=");
       String parametername = oc.get(0).trim();
       String parametervalue = oc.get(1).trim();
       if (!hasParameter(parametername))
	  return new Message(compName, message.from, "PARAM_NOT_EXIST", parametername, "");
       else {
	  if (removeParameterItem(parametername, parametervalue))
	     return new Message(compName, message.from, "PARAM_REMOVE_ITEM_SUCCESS");
	  else
	     return new Message(compName, message.from, "PARAM_REMOVE_ITEM_FAILED");
       }
    }
    else if (type.equals("PARAM_GET_ALL_SPECS")) {
      return new Message(compName, message.from, "PARAM_GET_ALL_SPECS_SUCCESS", parameters);
    }
    else if (type.equals("PARAM_GET_SPEC")) {
       String parametername = message.getContent().trim();
       Parameter parameter = getParameter(parametername);
       if (!hasParameter(parametername))
	  return new Message(compName, message.from, "PARAM_NOT_EXIST", parametername, "");
       else {
	  return new Message(compName, message.from, "PARAM_GET_SPEC_SUCCESS", parameter);
       }
    }
    else if (type.equals("PARAM_GET")) {
       String parametername = message.getContent().trim();
       if (!hasParameter(parametername))
	  return new Message(compName, message.from, "PARAM_NOT_EXIST", parametername, "");
       else {
	  return new Message(compName, message.from, "PARAM_GET_SUCCESS", getParameterString(parametername), "");
       }
    }
    else if (type.equals("PARAM_INCREMENT")) {
      TextCollection oc = Utils.splitString(message.getContent(),"=");
       String parametername = oc.get(0).trim();
       if (!hasParameter(parametername))
	  return new Message(compName, message.from, "PARAM_NOT_EXIST", parametername, "");
       int steps = Utils.str2int(oc.get(1).trim());
       if (!increaseParameter(parametername, steps))
	  return new Message(compName, message.from, "PARAM_INCREMENT_FAILED", parametername, "");
       else {
	  return new Message(compName, message.from, "PARAM_INCREMENT_SUCCESS", getParameterString(parametername), "");
       }
    }
    else if (type.equals("PARAM_DECREMENT")) {
      TextCollection oc = Utils.splitString(message.getContent(),"=");
       String parametername = oc.get(0).trim();
       if (!hasParameter(parametername))
	  return new Message(compName, message.from, "PARAM_NOT_EXIST", parametername, "");
	int steps = Utils.str2int(oc.get(1).trim());
       if (!decreaseParameter(parametername, steps))
	  return new Message(compName, message.from, "PARAM_DECREMENT_FAILED", parametername, "");
       else {
	  return new Message(compName, message.from, "PARAM_DECREMENT_SUCCESS", getParameterString(parametername), "");
       }
    }
    else if (type.equals("PARAM_RESET")) {
       String parametername = message.getContent().trim();
       if (!hasParameter(parametername))
	  return new Message(compName, message.from, "PARAM_NOT_EXIST", parametername, "");
       if (!resetParameter(parametername))
	  return new Message(compName, message.from, "PARAM_RESET_FAILED", parametername, "");
       else {
	  return new Message(compName, message.from, "PARAM_RESET_SUCCESS", getParameterString(parametername), "");
       }
    }
    else if (type.equals("GET_CURRENT_PERFSTAT")) {
      return new Message(compName, message.from, "GET_CURRENT_PERFSTAT_SUCCESS", perfStat);
    }
//    else if (type.equalsIgnoreCase("Psyclone.Module.Report.PhaseChange")) {
//      return processPhaseMessage(message);
//    }
    else if (type.equalsIgnoreCase("AIR.Instruct.Wakeup")) {
      Message msg = MessageProcessor.getTriggerMessage(message);
      if (msg != null) {
	// System.out.println("JavaReader: received wakeup message " + entry + "  with " + msgs.getCount() +" retrieved messages and a phase spec of " + phasespec.getXML().length() + " bytes: \n" + msg);
	statusDict.put("LastMsgWoken", msg.postedTime.toString() + " | " + msg.type + " | " + msg.from + " | " + message.from);
	long recv = Utils.str2long(statusDict.get("MsgReceived"));
	if (recv < 0)
		recv = 1;
	else
		recv++;
	statusDict.put("MsgReceived", ""+recv);
        perfStat.msgWoken++;
        perfStat.lastMessageTrigger = msg.shallowClone();
      }
    }

    if (messageAcceptor != null)
      outMsg = messageAcceptor.acceptMessage(message);
    else
      outMsg = this.acceptMessage(message);

    if (outMsg != null) {
      outMsg.isresponse = msgID;
      if (message.inReplyTo != null)
	outMsg.inReplyTo = message.inReplyTo;
      else
	outMsg.inReplyTo = new Reference(message);
    }
    return outMsg;
  }

//  public Message processPhaseMessage(Message message) {
//    return null;
//  }

  public Message acceptMessage(Message message) {
    return null;
  }

  boolean setPlugPriority(double pri) {
    priority = pri;
    setCompPriority(90 - (int)(priority*10.0));
    return true;
  }

  double getPlugPriority() {
    return priority;
  }


  /************* UTILITY FUNCTIONS ***************************/

  public boolean setCompPriority(int priority) {
    this.priority = priority;
    int nativePriority = (int)(priority/10);
    if (nativePriority < 1)
      nativePriority = 1;
    else if (nativePriority > 10)
      nativePriority = 10;
    try {
      setPriority(nativePriority);
    }
    catch(Exception e) {
      return false;
    }
    return true;
  }

  public int getCompPriority() {
    int nativePriority = getPriority() * 10;
    if (nativePriority < 0)
      nativePriority = 0;
    else if (nativePriority > 100)
      nativePriority = 100;
    return nativePriority;
  }


  public String getNonPrivateAddress(String addr) {
      if ((isIPAddressPrivate(addr)) && (!isIPAddressPrivate(cnsLocation.addr))) return cnsLocation.addr;
      return addr;
  }

  public boolean isIPAddressPrivate(String address) {
      if (address.startsWith("127."))
        return true;
      else if (address.equalsIgnoreCase("localhost"))
        return true;
      else if (address.startsWith("10."))
        return true;
      else if (address.startsWith("192.168."))
        return true;
      else
        return false;
  }

  void doMaintenance() {
    checkPrimaryConnection();
    checkReceivingConnections();
    checkOpenConnections();
    tryReconnectingLostConnections();
  }


  boolean checkReceivingConnections() {

    NetworkConnection con;
    MessageReceiver rec;
    String to;
    for (int n=0; n<receiveConnections.getCount(); n++) {
      to = (String)receiveConnections.getKey(n);
      con = (NetworkConnection) receiveConnections.get(n);
      if (con != null) {
	if (con.isConnected()) {
	  // Good...
	}
	else {
	  rec = con.messageReceiver;
	  con.destroy();
	  receiveConnections.remove(to);
	  if (!openReceivingConnection(rec, to)) {
	    if (debug) System.out.println("AIR: Lost connection to " + obscureInternals(to));
	    lostReceiveConnections.put(to, rec);
	  }
	  else
	    if (debug) System.out.println("AIR: Connection to " + obscureInternals(to) + " reconnected.");
	  // We messed up the collection, start from scratch...
	  return checkReceivingConnections();
	}
      }
    }
    return true;
  }

  boolean checkOpenConnections() {
    //System.out.println("in checkOpenConnections");
    NetworkConnection con;
    String to;
    for (int n=0; n<sendConnections.getCount(); n++) {
      to = (String)sendConnections.getKey(n);
      con = (NetworkConnection) sendConnections.get(n);
      if (con != null) {
	if (con.isConnected()) {
	  // Good...
	}
	else {
	  con.destroy();
	  sendConnections.remove(to);
	  if (!lostSendConnections.contains(to))
	    lostSendConnections.add(to);
	  if (!ping(to))
	    if (debug) System.out.println("AIR: Disconnected from " + obscureInternals(to) + ".");
	  else {
	    if (debug) System.out.println("AIR: Reconnected to " + obscureInternals(to) + ".");
	  }
	  // We messed up the collection, start from scratch...
	  return checkOpenConnections();
	}
      }
    }
    return true;
  }

  boolean tryReconnectingLostConnections() {
    int n;
    String to;
    for (n=0; n<lostSendConnections.getCount(); n++) {
      to = lostSendConnections.get(n);
      if (ping(to)) {
	if (debug) System.out.println("AIR: Reconnected to " + obscureInternals(to) + ".");
	lostSendConnections.remove(n);
	n--;
      }
    }

    MessageReceiver rec;
    for (n=0; n<lostReceiveConnections.getCount(); n++) {
      to = (String) lostReceiveConnections.getKey(n);
      rec = (MessageReceiver) lostReceiveConnections.get(n);
      if ( (to.length() > 0) && (rec != null) )
	if (openReceivingConnection(rec, to)) {
	  if (debug) System.out.println("AIR: Reconnected to " + obscureInternals(to) + ".");
	  lostReceiveConnections.remove(n);
	  n--;
	}
    }
    return true;

  }

  boolean checkPrimaryConnection() {

    boolean justDisconnected = false;

    synchronized (this) {
      if (cnsCon != null) {
	if (cnsCon.isConnected())
	  return true;
	else {
	  cnsCon.destroy();
	  cnsCon = null;
	  justDisconnected = true;
	}
      }

      try {
	cnsCon = new NetworkConnection(cnsLocation.addr, cnsLocation.port,
				       new NetMessageProtocol());
	cnsCon.name = "CNS";
      }
      catch (Exception e) {
	//  System.out.println("JavaAIRPlug: Could not connect to the Server on '" +
	//                     cnsLocation.addr + "' port " + cnsLocation.port +
	//                     "...");
	cnsCon = null;
	if (justDisconnected) {
	  // This means that we should report the disconnect
	  primaryDisconnectNotify();
	}
	return false;
      }
    }

    if (cnsRebind()) {
      primaryReconnectNotify();
      return true;
    }
    else {
      if (cnsCon != null)
	cnsCon.destroy();
      cnsCon = null;
      // This means that we should report the disconnect
      if (justDisconnected) {
	// This means that we should report the disconnect
	primaryDisconnectNotify();
      }
      return false;
    }
  }

  public boolean primaryReconnectNotify() {
    return true;
  }

  public boolean primaryDisconnectNotify() {
    return true;
  }


  public boolean addParameter(String paramname, String xml) {
    Parameter parameter = new Parameter(xml);
    if ((parameter.name.length() > 0) && (parameter.name.equals(paramname)))
      return parameters.put(parameter.name, parameter);
    else
      return false;
  }

  public boolean hasParameter(String name) {
    return parameters.containsKey(name);
  }

  public String getParameterString(String module, String parameter) {
    return getParameter(module, parameter);
  }

  public int getParameterInteger(String module, String parameter) {
    String str = getParameter(module, parameter);
    if (str.startsWith("PARAM_ERROR"))
      return Parameter.PARAM_ERROR;
    else
      return Utils.str2int(str);
  }

  public double getParameterDouble(String module, String parameter) {
    String str = getParameter(module, parameter);
    if (str.startsWith("PARAM_ERROR"))
      return Parameter.PARAM_ERROR;
    else
      return Utils.str2double(str);
  }

  public boolean parameterSetString(String module, String parameter, String value, String notifyComponent) {
    return parameterSet(module, parameter, value, notifyComponent);
  }

  public boolean parameterSetInteger(String module, String parameter, int value, String notifyComponent) {
    return parameterSet(module, parameter, ""+value, notifyComponent);
  }

  public boolean parameterSetDouble(String module, String parameter, double value, String notifyComponent) {
    return parameterSet(module, parameter, ""+value, notifyComponent);
  }


  public Parameter getParameter(String name) {
    return (Parameter) parameters.get(name);
  }

  public ObjectDictionary getParameterSpecs(String module) {
    Message msg = sendReceiveMessage(module,
				   new Message(compName, module, "PARAM_GET_ALL_SPECS"));
    if ( (msg != null) && (msg.type.equals("PARAM_GET_ALL_SPECS_SUCCESS"))) {
      ObjectDictionary content = (ObjectDictionary) msg.object;
      return content;
    }
    else {
      return null;
    }
  }

  public Parameter getParameterSpec(String module, String parameter) {
    Message msg = sendReceiveMessage(module,
				   new Message(compName, module, "PARAM_GET_SPEC", parameter, ""));
    if ( (msg != null) && (msg.type.equals("PARAM_GET_SPEC_SUCCESS"))) {
      Parameter content = (Parameter) msg.object;
      return content;
    }
    else {
      return null;
    }
  }

  public String getParameter(String module, String parameter) {
    Message msg = sendReceiveMessage(module,
				   new Message(compName, module, "PARAM_GET", parameter, ""));
    if ( (msg != null) && (msg.type.equals("PARAM_GET_SUCCESS"))) {
      String content = msg.getContent();
      return content;
    }
    else if (msg != null) {
      return "PARAM_ERROR: " + msg.type;
    }
    else {
      return "PARAM_ERROR: Unable to connect to parameter";
    }
  }

  public boolean parameterSet(String module, String parameter, String value, String notifyComponent) {
    Message notifyMsg = new Message(compName, "",
				    "AIR.Param.Set",
				    "Module: " + module + "  Parameter: " + parameter +
				    "  Value: " + value, "");
    sendMessage(notifyComponent, notifyMsg);
    Message msg = sendReceiveMessage(module,
				   new Message(compName, module, "PARAM_SET",
					       parameter + "=" + value, ""));
    if ( (msg != null) && (msg.type.equals("PARAM_SET_SUCCESS"))) {
      value = msg.getContent();
      // AIR.Param.  Tweak.Up / Tweak.Down / Set / Reset				content: N
      // AIR.Param.  Tweak.Up / Tweak.Down / Set / Reset  .Success		content: Resulting value
      notifyMsg = new Message(compName, "",
			      "AIR.Param.Set.Complete",
			      "Module: " + module + "  Parameter: " + parameter +
			      "  Value: " + value, "");
      notifyMsg.addReferenceToHistory(msg, compName);
      sendMessage(notifyComponent, notifyMsg);
      return true;
    }
    else {
      if (msg == null) {
	value = "No Response";
      }
      else {
	value = msg.type;
      }
      notifyMsg = new Message(compName, "",
			      "AIR.Param.Set.Failed",
			      "Module: " + module + "  Parameter: " + parameter +
			      "  Reason: " + value, "");
      sendMessage(notifyComponent, notifyMsg);
      return false;
    }
  }

  public boolean parameterAddItem(String module, String parameter, String value, String notifyComponent) {
    Message notifyMsg = new Message(compName, "",
				    "AIR.Param.AddItem",
				    "Module: " + module + "  Parameter: " + parameter +
				    "  Value: " + value, "");
    sendMessage(notifyComponent, notifyMsg);
    Message msg = sendReceiveMessage(module,
				   new Message(compName, module, "PARAM_ADD_ITEM",
					       parameter + "=" + value, ""));
    if ( (msg != null) && (msg.type.equals("PARAM_ADD_ITEM_SUCCESS"))) {
      value = msg.getContent();
      // AIR.Param.  Tweak.Up / Tweak.Down / Set / Reset				content: N
      // AIR.Param.  Tweak.Up / Tweak.Down / Set / Reset  .Success		content: Resulting value
      notifyMsg = new Message(compName, "",
			      "AIR.Param.AddItem.Complete",
			      "Module: " + module + "  Parameter: " + parameter +
			      "  Value: " + value, "");
      notifyMsg.addReferenceToHistory(msg, compName);
      sendMessage(notifyComponent, notifyMsg);
      return true;
    }
    else {
      if (msg == null) {
	value = "No Response";
      }
      else {
	value = msg.type;
      }
      notifyMsg = new Message(compName, "",
			      "AIR.Param.AddItem.Failed",
			      "Module: " + module + "  Parameter: " + parameter +
			      "  Reason: " + value, "");
      sendMessage(notifyComponent, notifyMsg);
      return false;
    }
  }

  public boolean parameterRemoveItem(String module, String parameter, String value, String notifyComponent) {
    Message notifyMsg = new Message(compName, "",
				    "AIR.Param.RemoveItem",
				    "Module: " + module + "  Parameter: " + parameter +
				    "  Value: " + value, "");
    sendMessage(notifyComponent, notifyMsg);
    Message msg = sendReceiveMessage(module,
				   new Message(compName, module, "PARAM_REMOVE_ITEM",
					       parameter + "=" + value, ""));
    if ( (msg != null) && (msg.type.equals("PARAM_REMOVE_ITEM_SUCCESS"))) {
      value = msg.getContent();
      // AIR.Param.  Tweak.Up / Tweak.Down / Set / Reset				content: N
      // AIR.Param.  Tweak.Up / Tweak.Down / Set / Reset  .Success		content: Resulting value
      notifyMsg = new Message(compName, "",
			      "AIR.Param.RemoveItem.Complete",
			      "Module: " + module + "  Parameter: " + parameter +
			      "  Value: " + value, "");
      notifyMsg.addReferenceToHistory(msg, compName);
      sendMessage(notifyComponent, notifyMsg);
      return true;
    }
    else {
      if (msg == null) {
	value = "No Response";
      }
      else {
	value = msg.type;
      }
      notifyMsg = new Message(compName, "",
			      "AIR.Param.RemoveItem.Failed",
			      "Module: " + module + "  Parameter: " + parameter +
			      "  Reason: " + value, "");
      sendMessage(notifyComponent, notifyMsg);
      return false;
    }
  }

  public boolean parameterIncrease(String module, String parameter, int steps, String notifyComponent) {
    String value = ""+steps;
    Message notifyMsg = new Message(compName, "", "AIR.Param.Tweak.Up",
				    "Module: " + module + "  Parameter: " + parameter +
				    "  Value: " + value, "");
    sendMessage(notifyComponent, notifyMsg);
    Message msg = sendReceiveMessage(module,
			       new Message(compName, module, "PARAM_INCREMENT",
					   parameter + "=" + value, ""));
    if ( (msg != null) && (msg.type.equals("PARAM_INCREMENT_SUCCESS"))) {
      value = msg.getContent();
      notifyMsg = new Message(compName, "",
			      "AIR.Param.Tweak.Up.Complete",
			      "Module: " + module + "  Parameter: " + parameter +
			      "  Value: " + value, "");
      notifyMsg.addReferenceToHistory(msg, compName);
      sendMessage(notifyComponent, notifyMsg);
      return true;
    }
    else {
      if (msg == null) {
	value = "No Response";
      }
      else {
	value = msg.type;
      }
      notifyMsg = new Message(compName, "",
			      "AIR.Param.Tweak.Up.Failed",
			      "Module: " + module + "  Parameter: " + parameter +
			      "  Reason: " + value, "");
      sendMessage(notifyComponent, notifyMsg);
      return false;
    }
  }

  public boolean parameterDecrease(String module, String parameter, int steps, String notifyComponent) {
    String value = ""+steps;
    Message notifyMsg = new Message(compName, "", "AIR.Param.Tweak.Down",
				"Module: " + module + "  Parameter: " + parameter +
				"  Value: " + value, "");
    sendMessage(notifyComponent, notifyMsg);
    Message msg = sendReceiveMessage(module,
			       new Message(compName, module, "PARAM_DECREMENT",
					   parameter + "=" + value, ""));
    if ( (msg != null) && (msg.type.equals("PARAM_DECREMENT_SUCCESS"))) {
      value = msg.getContent();
      notifyMsg = new Message(compName, "",
			      "AIR.Param.Tweak.Down.Complete",
			      "Module: " + module + "  Parameter: " + parameter +
			      "  Value: " + value, "");
      notifyMsg.addReferenceToHistory(msg, compName);
      sendMessage(notifyComponent, notifyMsg);
      return true;
    }
    else {
      if (msg == null) {
	value = "No Response";
      }
      else {
	value = msg.type;
      }
      notifyMsg = new Message(compName, "",
			      "AIR.Param.Tweak.Down.Failed",
			      "Module: " + module + "  Parameter: " + parameter +
			      "  Reason: " + value, "");
      sendMessage(notifyComponent, notifyMsg);
      return false;
    }
  }

  public boolean parameterReset(String module, String parameter, String notifyComponent) {
    String value;
    Message notifyMsg = new Message(compName, "", "AIR.Param.Reset",
				"Module: " + module + "  Parameter: " + parameter, "");
    sendMessage(notifyComponent, notifyMsg);
    Message msg = sendReceiveMessage(module,
			       new Message(compName, module, "PARAM_RESET", parameter,
					   ""));
    if ( (msg != null) && (msg.type.equals("PARAM_RESET_SUCCESS"))) {
      value = msg.getContent();
      notifyMsg = new Message(compName, "",
			      "AIR.Param.Reset.Complete",
			      "Module: " + module + "  Parameter: " + parameter +
			      "  Value: " + value, "");
      notifyMsg.addReferenceToHistory(msg, compName);
      sendMessage(notifyComponent, notifyMsg);
      return true;
    }
    else {
      if (msg == null) {
	value = "No Response";
      }
      else {
	value = msg.type;
      }
      notifyMsg = new Message(compName, "",
			      "AIR.Param.Reset.Failed",
			      "Module: " + module + "  Parameter: " + parameter +
			      "  Reason: " + value, "");
      sendMessage(notifyComponent, notifyMsg);
      return false;
    }
  }

  // if the 'to' string is CNS, return Server instead
  // Other names that should be hidden in output messgages
  // should be added here
  public String obscureInternals(String to) {
      if (to.equalsIgnoreCase("CNS")) return "Server";
      return to;
  }

  public String getParameterString(String param) {
    Parameter parameter = getParameter(param);
    if (parameter == null)
      return "";
    return parameter.getAsString();
  }

  public int getParameterInteger(String param) {
    Parameter parameter = getParameter(param);
    if (parameter == null)
      return 0;
    return parameter.getInteger();
  }

  public double getParameterDouble(String param) {
    Parameter parameter = getParameter(param);
    if (parameter == null)
      return 0;
    return parameter.getDouble();
  }

  public boolean setParameterString(String param, String value) {
    Parameter parameter = getParameter(param);
    if (parameter == null)
      return false;
    return parameter.set(value);
  }


 public boolean addParameterItem(String param, String value) {
    Parameter parameter = getParameter(param);
    if (parameter == null)
      return false;
    return parameter.addItem(value);
  }

 public boolean removeParameterItem(String param, String value) {
    Parameter parameter = getParameter(param);
    if (parameter == null)
      return false;
    return parameter.removeItem(value);
  }

 public boolean setParameterInteger(String param, int value) {
    Parameter parameter = getParameter(param);
    if (parameter == null)
      return false;
    return parameter.setInteger(value);
  }

  public boolean setParameterDouble(String param, double value) {
    Parameter parameter = getParameter(param);
    if (parameter == null)
      return false;
    return parameter.setDouble(value);
  }

  public boolean resetParameter(String param) {
    Parameter parameter = getParameter(param);
    if (parameter == null)
      return false;
    return parameter.reset();
  }

  public boolean increaseParameter(String param, int steps) {
    Parameter parameter = getParameter(param);
    if (parameter == null)
      return false;
    return parameter.increment(steps);
  }

  public boolean decreaseParameter(String param, int steps) {
    Parameter parameter = getParameter(param);
    if (parameter == null)
      return false;
    return parameter.decrement(steps);
  }

  public TextCollection getParameterItems(String param) {
    Parameter parameter = getParameter(param);
    if (parameter == null)
      return new TextCollection();
    return parameter.values;
  }

  public boolean setStatus(String entry, String text) {
    return statusDict.put(entry, text);
  }

  public void setDebug(boolean d) {
      debug = d;
  }
}


