/***************************** License **********************************

    Copyright(c)2005, Communicative Machines
    All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are 
met:

- Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright 
  notice, this list of conditions and the following disclaimer in the 
  documentation and/or other materials provided with the distribution.

- Neither the name of the copyright holders nor the names of its 
  contributors may be used to endorse or promote products derived from 
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER 
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

************************************************************************/
// ConnectionProfile.java


package com.cmlabs.air;

public class ConnectionProfile extends BaseObject {

  String name;
  long speedUploadBps;
  long speedDownloadBps;
  long avgUploadBps;
  long avgDownloadBps;
  long maxAvgUploadBps;
  long maxAvgDownloadBps;
  double totalUpload;
  double totalDownload;
  ObjectDictionary uploadHistory = new ObjectDictionary();
  ObjectDictionary downloadHistory = new ObjectDictionary();
  int maxConnectionHistory = 0;
  boolean isDirectlyConnected = false;

  public ConnectionProfile(String name) {
    this.name = name;
  }

  public ConnectionProfile(XMLElement node) {
    fromXML(node);
  }

  public boolean fromXML(XMLElement node) {
    if (node == null)
      return false;

    if (!node.getName().equalsIgnoreCase("connectionprofile"))
      return false;

    name = node.getStringAttribute("name");
    speedUploadBps = Utils.str2long(node.getStringAttribute("speeduploadbps"));
    speedDownloadBps = Utils.str2long(node.getStringAttribute("speeddownloadbps"));
    avgUploadBps = Utils.str2long(node.getStringAttribute("avguploadbps"));
    avgDownloadBps = Utils.str2long(node.getStringAttribute("avgdownloadbps"));
    maxAvgUploadBps = Utils.str2long(node.getStringAttribute("maxavguploadbps"));
    maxAvgDownloadBps = Utils.str2long(node.getStringAttribute("maxavgdownloadbps"));
    totalUpload = Utils.str2double(node.getStringAttribute("totalupload"));
    totalDownload = Utils.str2double(node.getStringAttribute("totaldownload"));

    return true;
  }

  public String toXML() {
    if (isDirectlyConnected)
      return "<connectionprofile name=\"" + Utils.xmlStringEncode(name) + "\" speeduploadbps=\"" + speedUploadBps + "\" speeddownloadbps=\"" + speedDownloadBps + "\" avguploadbps=\"" + avgUploadBps + "\" avgdownloadbps=\"" + avgDownloadBps + "\" maxavguploadbps=\"" + maxAvgUploadBps + "\" maxavgdownloadbps=\"" + maxAvgDownloadBps + "\"  totalupload=\"" + totalUpload + "\" totaldownload=\"" + totalDownload + "\" maxhistory=\"" + maxConnectionHistory + "\" direct=\"yes\" />";
    else
      return "<connectionprofile name=\"" + Utils.xmlStringEncode(name) + "\" speeduploadbps=\"" + speedUploadBps + "\" speeddownloadbps=\"" + speedDownloadBps + "\" avguploadbps=\"" + avgUploadBps + "\" avgdownloadbps=\"" + avgDownloadBps + "\" maxavguploadbps=\"" + maxAvgUploadBps + "\" maxavgdownloadbps=\"" + maxAvgDownloadBps + "\"  totalupload=\"" + totalUpload + "\" totaldownload=\"" + totalDownload + "\" maxhistory=\"" + maxConnectionHistory + "\" direct=\"no\" />";
  }

  public String print() {
    return "speed up " + speedUploadBps + " down " + speedDownloadBps + "bps avg up " + avgUploadBps + " down " + avgDownloadBps + "bps max up " + maxAvgUploadBps + " down " + maxAvgDownloadBps + "bps  total up " + totalUpload + " down " + totalDownload + "b ";
  }

  public boolean setSpeed(long upbps, long downbps) {
    speedUploadBps = upbps;
    speedDownloadBps = downbps;
    return true;
  }

  public boolean addDataStat(long upbytes, long downbytes, Time time) {
    if (upbytes > 0) {
      totalUpload += upbytes;
      uploadHistory.put(time, "" + upbytes);
      while (uploadHistory.getCount() > maxConnectionHistory) {
	// printf("@@@ %s\n", (char*) uploadHistory.print()); fflush(stdout);
	uploadHistory.remove(0);
      }
    }
    if (downbytes > 0) {
      totalDownload += downbytes;
      downloadHistory.put(time, "" + downbytes);
      while (downloadHistory.getCount() > maxConnectionHistory) {
	// printf("### %s\n", (char*) downloadHistory.print()); fflush(stdout);
	downloadHistory.remove(0);
      }
    }
    return true;
  }

  public long getSpeedUpload() {
    recalc();
    return speedUploadBps;
  }

  public long getSpeedDownload() {
    recalc();
    return speedDownloadBps;
  }

  public long getAvgUpload() {
    recalc();
    return avgUploadBps;
  }

  public long getAvgDownload() {
    recalc();
    return avgDownloadBps;
  }

  public long getMaxAvgUpload() {
    recalc();
    return maxAvgUploadBps;
  }

  public long getMaxAvgDownload() {
    recalc();
    return maxAvgDownloadBps;
  }

  public double getTotalUpload() {
    recalc();
    return totalUpload;
  }

  public double getTotalDownload() {
    recalc();
    return totalDownload;
  }

  String getSpeedString() {
    return Utils.bytifyRates(speedUploadBps, speedDownloadBps);
  }

  String getRateString() {
    recalc();
    return Utils.bytifyRates(avgUploadBps, avgDownloadBps);
  }

  String getMaxRateString() {
    recalc();
    return Utils.bytifyRates(maxAvgUploadBps, maxAvgDownloadBps);
  }

  String getTotalString() {
    return Utils.bytifySizes(totalUpload, totalDownload);
  }

  public Time getLastActivity() {
    Time up = (Time) uploadHistory.getLastKeyObject();
    Time down = (Time) downloadHistory.getLastKeyObject();
    if ( (up == null) && (down == null)) {
      return null;
    }
    else if (up == null)
      return down;
    else if (down == null)
      return up;
    else {
      if (up.difference(down) > 0)
	return up;
      else
	return down;
    }
  }

  public boolean recalc() {

    Time now = new Time();
    Time t;
    int n;
    double totalup = 0;
    double totaldown = 0;

    if (uploadHistory.getCount() > 0) {
      t = (Time) uploadHistory.getKeyObject(0);
      double upduration = now.difference(t) / 1000000.0;
      if (upduration > 0) {
	for (n = 0; n < uploadHistory.getCount(); n++)
	  totalup += Utils.str2long( (String) uploadHistory.get(n));
	avgUploadBps = (long) (totalup / upduration);
	if (avgUploadBps > maxAvgUploadBps)
	  maxAvgUploadBps = avgUploadBps;
      }
    }
    if (downloadHistory.getCount() > 0) {
      t = (Time) downloadHistory.get(0);
      double downduration = now.difference(t) / 1000000.0;
      if (downduration > 0) {
	for (n = 0; n < downloadHistory.getCount(); n++)
	  totaldown += Utils.str2long( (String) downloadHistory.get(n));
	avgDownloadBps = (long) (totaldown / downduration);
	if (avgDownloadBps > maxAvgDownloadBps)
	  maxAvgDownloadBps = avgDownloadBps;
      }
    }
    return true;
  }

}

