/***************************** License **********************************

    Copyright(c)2005, Communicative Machines
    All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are 
met:

- Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright 
  notice, this list of conditions and the following disclaimer in the 
  documentation and/or other materials provided with the distribution.

- Neither the name of the copyright holders nor the names of its 
  contributors may be used to endorse or promote products derived from 
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER 
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

************************************************************************/
// Messenger.java


package com.cmlabs.air;

import java.util.*;

public class Messenger extends BaseObject {

  public MessageSender messageSender = null;
  public String name;
  public String wakeupMessageType;
  public String phaseChangeType;

  public ObjectQueue inputMessages = new ObjectQueue();
  public ObjectCollection outputMessages = new ObjectCollection();
  public ObjectDictionary streams = new ObjectDictionary();
  public Message currentMessage = null;
  public PhaseSpec currentPhaseSpec = null;

  public int totalTriggerCount = 0;
  public boolean isStillRunning = true;
  public boolean shouldContinue = true;

  public Time startCrankTime = null;
  public Time startProcessTime = null;
  public int verbose = 0;
  public double priority = 0;

  PerfStat perfStat = null;

  public Messenger(String name, MessageSender sender, String wakeupMessageType, String phaseChangeType) {
    messageSender = sender;
    this.name = name;
    this.wakeupMessageType = wakeupMessageType;
    this.phaseChangeType = phaseChangeType;
  }

  public String getName() {
    return name;
  }

  public TextCollection getDestinations() {
    TextCollection destinations = new TextCollection();

    if (messageSender == null)
      return destinations;

    PhaseSpec spec = getPhaseSpec();
    if (spec == null)
      return destinations;

    PostSpec post;
    for (int n=0; n<spec.posts.getCount(); n++) {
      if ( (post = (PostSpec) spec.posts.get(n)) != null) {
        destinations.add(post.to);
      }
    }
    return destinations;
  }

  public boolean isWakeupMessage(Message msg) {
    if ( (msg == null) || (msg.object == null))
      return false;
    return msg.type.equalsIgnoreCase(wakeupMessageType);
  }

  public int getInputMessageCount() {
    return inputMessages.getCount();
  }

  public int getOutputMessageCount() {
    return outputMessages.getCount();
  }

  public int getTriggerCount() {
    return totalTriggerCount;
  }

  public String getTriggerAlias() {
    PhaseSpec spec = getPhaseSpec();
    if (spec == null)
      return "";
    else
      return spec.lastTriggerName;
  }

  public Message waitForNewMessage(int timeout) {
    if (!shouldContinue)
      return null;

    Message msg = (Message) inputMessages.waitForNewEntry(timeout);

    if (msg == null)
      return null;

    PhaseSpec newSpec;
    if (isWakeupMessage(msg)) {
      ObjectDictionary dict = (ObjectDictionary) msg.object;
      if (dict == null) {
        newSpec = null;
      }
      else {
        newSpec = (PhaseSpec) dict.get("PhaseSpec");
      }

      if (!checkNewPhaseSpec(newSpec, msg.postedTime)) {
        // We should ignore this message, it is too old
        return null;
      }

      currentMessage = msg;
      msg = getTriggerMessage();
    }
    else {
      // Remove current message, but do not set this new one...
      if (currentMessage != null) {
        currentMessage = null;
      }
    }

    startProcessTime = new Time();
    return msg;

/*    currentMessage = msg;
    if (currentMessage == null) {
      ObjectDictionary dict = (ObjectDictionary) currentMessage.object;
      if (dict == null) {
        newSpec = null;
      }
      else {
        newSpec = (PhaseSpec) dict.get("PhaseSpec");
      }
      checkNewPhaseSpec(newSpec, currentMessage.postedTime);
    }
*/


//    msg = getTriggerMessage();
/*
    if (msg == null) {
      this.messageSender - >
          log(1, "Received error wakeup message from '%s' type '%s'...\n",
              (char * ) currentMessage - > getFrom(),
              (char * ) currentMessage - > getType());
    }
    else {
      if (verbose > 2)
        messageSender - >
            log(1, "Triggered by message from '%s' type '%s'...\n",
                (char * ) msg - > getFrom(), (char * ) msg - > getType());
      else
        messageSender - >
            log(3, "Triggered by message from '%s' type '%s'...\n",
                (char * ) msg - > getFrom(), (char * ) msg - > getType());
    }
 */
//    startProcessTime = new Time();
//    return getTriggerMessage();
  }

  boolean processPhaseSpec(PhaseSpec newSpec, Time t) {
    return checkNewPhaseSpec(newSpec, t);
  }


  boolean checkNewPhaseSpec(PhaseSpec newSpec, Time msgtime) {

    if ( (newSpec != null) && (currentPhaseSpec != null) ) {
      if (newSpec.equals(currentPhaseSpec)) {
        // Process message with current spec
        currentPhaseSpec = newSpec;
        // System.out.println("Got PhaseSpec (ready)");
        return true;
      }
      else {
        // Is the new phase spec older than our current one
        if (newSpec.isOlderThan(currentPhaseSpec)) {
          // ignore this message all together
          //System.out.println("Got PhaseSpec (out of date)");
          return false;
        }
        else {
          // new spec is newer, use this
          currentPhaseSpec = newSpec;
          //System.out.println("Got PhaseSpec (new)");
          clearOldMessagesFromQueue(currentPhaseSpec.activationTime);
          return true;
        }
      }
    }
    else if (newSpec != null) {
      // This is a new spec and we haven't got an old one
      currentPhaseSpec = newSpec;
      //System.out.println("Got PhaseSpec (brand new)");
      clearOldMessagesFromQueue(currentPhaseSpec.activationTime);
      return true;
    }
    else {
      // newSpec is NULL set this as our current spec...
      currentPhaseSpec = null;
      //System.out.println("Got PhaseSpec (out of context)");
      clearOldMessagesFromQueue(msgtime);
      return true;
    }
  }














  public Message getWakeupMessage() {
    if ((!shouldContinue) || (currentMessage == null))
      return null;

    if (!isWakeupMessage(currentMessage)) {
      return null;
    }
    return currentMessage;
}


  public Message getTriggerMessage() {
    if ((!shouldContinue) || (currentMessage == null))
      return null;

    if (!isWakeupMessage(currentMessage)) {
      return currentMessage;
    }

    ObjectDictionary dict = (ObjectDictionary) currentMessage.object;
    if (dict == null) {
      return null;
    }

    Message hmsg;
    if ( (hmsg = (Message) dict.get("TriggerMessage")) == null ) {
      return null;
    }
    return hmsg;
}

  public int getRetrievedMessageCount() {
    ObjectCollection retrievedMessages = getAllRetrievedMessages();
    if (retrievedMessages == null)
      return 0;
    else
      return retrievedMessages.getCount();
}

  public ObjectCollection getAllRetrievedMessages() {
    if ((!shouldContinue) || (currentMessage == null) || (!isWakeupMessage(currentMessage)))
      return null;

    ObjectDictionary dict = (ObjectDictionary) currentMessage.object;
    if (dict == null) {
      return null;
    }

    ObjectCollection retrievedMessages = (ObjectCollection) dict.get("RetrievedMessages");
    if (retrievedMessages == null) {
      return null;
    }
    return retrievedMessages;
  }

  public Message getRetrievedMessage(int pos) {
    ObjectCollection retrievedMessages = getAllRetrievedMessages();
    if (retrievedMessages != null) {
      return (Message) retrievedMessages.get(pos);
    }
    else
      return null;
  }

  public PhaseSpec getPhaseSpec() {
    if (!shouldContinue) // || (currentMessage == null) || (!isWakeupMessage(currentMessage)))
      return null;

    return currentPhaseSpec;

/*    ObjectDictionary dict = (ObjectDictionary) currentMessage.object;
    if (dict == null) {
      return null;
    }

    PhaseSpec spec = (PhaseSpec) dict.get("PhaseSpec");
    if (spec == null) {
      return null;
    }
    return spec;*/
}

  public ObjectCollection retrieveMessages(String from, String retrieveSpecXML) {
    ObjectCollection col = null;

    Message msg;
    if (retrieveSpecXML.toLowerCase().startsWith("<retrieves"))
      msg = new Message(name, from, "AIR.Retrieves", retrieveSpecXML, "xml");
    else
      msg = new Message(name, from, "AIR.Retrieve", retrieveSpecXML, "xml");
    Message rmsg = messageSender.sendReceiveMsgTo(from, msg);
    if ( (rmsg != null) && (rmsg.type.toLowerCase().indexOf(".success") >= 0) ) {
      col = (ObjectCollection) rmsg.object;
    }
/*
    if (col == null) {
      this.messageSender.log(1, "retrieve from '%s' failed...\n", (char*) retrieveFrom);
    }
    else {
      if (verbose > 3)
        this.messageSender.log(1, "retrieve from '%s' returned %d messages...\n", (char*) retrieveFrom, col.getCount());
      else
        this.messageSender.log(4, "retrieve from '%s' returned %d messages...\n", (char*) retrieveFrom, col.getCount());
    }
 */
    return col;
  }

  public ObjectCollection retrieveMessages(String from, RetrieveSpec retrieveSpec) {
    ObjectCollection col = null;
    Message msg = new Message(name, from, "AIR.Retrieve");
    msg.object = retrieveSpec;
    Message rmsg = messageSender.sendReceiveMsgTo(from, msg);
    if ( (rmsg != null) && (rmsg.type.equalsIgnoreCase("AIR.Retrieve.Success")) ) {
      col = (ObjectCollection) rmsg.object;
    }
/*
    if (col == null) {
      this.messageSender.log(1, "retrieve from '%s' failed...\n", (char*) retrieveFrom);
    }
    else {
      if (verbose > 3)
        this.messageSender.log(1, "retrieve from '%s' returned %d messages...\n", (char*) retrieveFrom, col.getCount());
      else
        this.messageSender.log(4, "retrieve from '%s' returned %d messages...\n", (char*) retrieveFrom, col.getCount());
    }
 */
    return col;
  }

  public ObjectCollection retrieveMessages(String from, ObjectCollection retrieveSpecs) {
    ObjectCollection col = null;
    Message msg = new Message(name, from, "AIR.Retrieves");
    msg.object = retrieveSpecs;
    Message rmsg = messageSender.sendReceiveMsgTo(from, msg);
    if ( (rmsg != null) && (rmsg.type.equalsIgnoreCase("AIR.Retrieves.Success")) ) {
      col = (ObjectCollection) rmsg.object;
    }
/*
    if (col == null) {
      this.messageSender.log(1, "retrieve from '%s' failed...\n", (char*) retrieveFrom);
    }
    else {
      if (verbose > 3)
        this.messageSender.log(1, "retrieve from '%s' returned %d messages...\n", (char*) retrieveFrom, col.getCount());
      else
        this.messageSender.log(4, "retrieve from '%s' returned %d messages...\n", (char*) retrieveFrom, col.getCount());
    }
 */
    return col;
  }



  public TextCollection getStreamNames() {
    return streams.getAllKeys();
  }

  public MediaConnection getStreamConnection(String name) {
    return (MediaConnection) streams.get(name);
  }


  public boolean addOutputMessage(Message msg) {
    return outputMessages.add(msg);
  }

  public boolean addOutputMessage(Message msg, String destination) {
    PhaseSpec spec = getPhaseSpec();
    if (spec == null) {
      msg.type = destination;
      return outputMessages.add(msg);
    }

    PostSpec post;
    for (int n=0; n<spec.posts.getCount(); n++) {
      if ( (post = (PostSpec) spec.posts.get(n)) != null) {
        if (post.alias.equalsIgnoreCase(destination)) {
          msg.type = post.type;
          return outputMessages.add(msg);
        }
      }
    }

    msg.type = destination;
    return outputMessages.add(msg);
  }

  public boolean sendOutputMessages() {
    if (messageSender == null)
      return false;


    PhaseSpec spec = getPhaseSpec();
    if (spec == null)
      return false;


    Message triggerMsg = getTriggerMessage();
    // if (triggerMsg == null)
    //    return false;

    Message msg;
    Message msgCopy;
    ObjectCollection destinations = new ObjectCollection();
    PostSpec post;
    String str = "";
    String type = "", posttype = "", phaseChange = "";
    int n;

    while ( (msg = (Message) outputMessages.getFirst()) != null) {
        //System.out.println("JAP: got into outputMessage while loop");
      destinations.removeAll();
      if (msg.to.length() > 0) {
        if (msg.priority == 0)
          msg.priority = priority;
        msg.from = getName();
        msg.postedTime = new Time();
        // Send message
//        str = JString::format("message from '%s' to '%s' type '%s'", (char*) msg.getFrom(), (char*) msg.getTo(), (char*) msg.getType());
        if (!messageSender.sendMsgTo(msg.to, msg)) {
//          messageSender.log(1, "Posting of %s failed.\n", (char*) str);
        }
        else {
          if (perfStat != null) {
            perfStat.msgPosted++;
            perfStat.lastMessagePosted = msg.shallowClone();
          }
/*          if (verbose > 1)
            messageSender.log(1, "Posted %s...\n", (char*) str);
          else
            messageSender.log(2, "Posted %s...\n", (char*) str);*/
        }
      }
      else {
        type = msg.type;
        for (n=0; n<spec.posts.getCount(); n++) {
          if ( (post = (PostSpec) spec.posts.get(n)) != null) {
            if ( (type.length() > 0) &&
                 (post.type.length() > 0) &&
                 (!Utils.isPartOf(type, post.type)) && (!type.equalsIgnoreCase(post.alias)) ) {
              // Ignore...
            }
            else {
              destinations.add(post);
              if ( (post.phasechange.length() > 0) && (phaseChange.length() == 0) )
                phaseChange = post.to;
              if (type.equalsIgnoreCase(post.alias))
                msg.type = post.type;
                //System.out.println("JAP:phaseChange = " + phaseChange);
            }
          }
        }
        if (destinations.getCount() == 0) {
          for (n=0; n<spec.posts.getCount(); n++) {
            if ( (post = (PostSpec) spec.posts.get(n)) != null) {
              posttype = post.type;
              if ( (type.length() > 0) && (posttype.toLowerCase().indexOf(type.toLowerCase()) >= 0) ) {
                destinations.add(post);
                if ( (post.phasechange.length() > 0) && (phaseChange.length() == 0) )
                  phaseChange = post.to;
                msg.type = posttype;
                if (msg.priority == 0)
                  msg.priority = post.priority;
                break;
              }
            }
          }
        }
        if (destinations.getCount() == 0) {
        }
        else {
          if (phaseChange.length() > 0)
            sendPhaseChange(phaseChange);
          for (int k=0; k < destinations.getCount(); k++) {
            if ( (post = (PostSpec) destinations.get(k)) != null ) {
              msgCopy = msg;
              // if (shouldChangePhase) {
              // msgCopy.setPhaseChangeMsg(true);
              // shouldChangePhase = false;
              // }
              // If from is hardcoded, overwrite
              if (post.from.length() > 0)
                msgCopy.from = post.from;
              else
                msgCopy.from = getName();
              if (msgCopy.type.length() == 0)
                msgCopy.type = post.type;
              if (msgCopy.priority == 0)
                msgCopy.priority = post.priority;
                // If content hardcoded, overwrite
              if (post.content.length() > 0) {
                msgCopy.content = post.content;
                msgCopy.language = post.language;
              }
              msgCopy.to = post.to;
              msgCopy.cc = post.cc;
              msgCopy.postedTime = new Time();

              if (msgCopy.inReplyTo == null) {
                if ( (triggerMsg != null) && (post.useInReplyTo) && (triggerMsg.inReplyTo != null))
                  msgCopy.inReplyTo = triggerMsg.inReplyTo;
                else if ((post.useInReplyTo) && (triggerMsg != null) && (triggerMsg.inReplyTo == null))
                  msgCopy.inReplyTo = new Reference(triggerMsg);
              }

              // Send message
              // str = JString::format("message from '%s' to '%s' type '%s'", (char*) msg.getFrom(), (char*) msgCopy.getTo(), (char*) msgCopy.getType());
              if (!messageSender.sendMsgTo(msgCopy.to, msgCopy)) {
                // messageSender.log(1, "Posting of %s failed.\n", (char*) str);
              }
              else {
                if (perfStat != null) {
                  perfStat.msgPosted++;
                  perfStat.lastMessagePosted = msgCopy.shallowClone();
                }
              /*  if (verbose > 1)
                  messageSender.log(1, "Posted %s...\n", (char*) str);
                else
                  messageSender.log(2, "Posted %s...\n", (char*) str);*/
              }
            }
          }
        }
      }
      outputMessages.remove(0);
    }
    return true;
  }

  public boolean sendPhaseChange() {
    return sendPhaseChange("");
  }

  public boolean sendPhaseChange(String dest) {
    if (messageSender == null)
      return false;

    PhaseSpec spec = getPhaseSpec();
    if (spec == null)
      return false;

    int n;
    PostSpec post = null;
    if (dest.length() > 0) {
      for (n=0; n<spec.posts.getCount(); n++) {
        if ( (post = (PostSpec) spec.posts.get(n)) != null) {
          if (post.to.equalsIgnoreCase(dest))
            break;
        }
      }
    }
    else {
      for (n=0; n<spec.posts.getCount(); n++) {
        if ( (post = (PostSpec) spec.posts.get(n)) != null) {
          if (post.phasechange.equalsIgnoreCase("yes"))
            break;
        }
      }

      if (post == null)
        post = (PostSpec) spec.posts.getFirst();
    }

    if (post == null)
      return false;

    String to = post.to;

    Message msg = new Message(getName(), to, phaseChangeType);
    msg.postedTime = new Time();

//    String str = JString::format("message from '%s' to '%s' type '%s'", (char*) msg->getFrom(), (char*) to, (char*) msg->getType());
    if (!messageSender.sendMsgTo(to, msg)) {
      return false;
    }
    else {
//      if (verbose > 1)
  //      messageSender->log(1, "Posted %s...\n", (char*) str);
    //  else
      //  messageSender->log(2, "Posted %s...\n", (char*) str);
    }
//    if (perfStat != null)
//      perfStat->msgPosted++;

//	printf("[%s] PHASECHANGE End!\n", (char*) getName());
    return true;
}


  boolean clearOldMessagesFromQueue(Time stamp) {
    if (stamp == null)
      return false;
    String id;
    ObjectDictionary dict;
    PhaseSpec spec;
    Message hmsg = null;
    if (inputMessages.getCount() > 0) {
      synchronized(inputMessages) {
        for (int n=0; n<inputMessages.getCount(); n++) {
          hmsg = (Message) inputMessages.viewEntry(n);
          if (hmsg != null) {
            dict = (ObjectDictionary) hmsg.object;
            if (dict != null) {
              spec = (PhaseSpec) dict.get("PhaseSpec");
              if (spec != null) {
                if ( (!spec.equals(currentPhaseSpec)) && (spec.activationTime != null) &&
                    (spec.activationTime.difference(stamp) < 0)) {
                  inputMessages.retrieveEntry(n);
                  n--;
                }
              }
            }
          }
        }
      }
    }
    return true;
  }


  // Module Interface
  public boolean addInputMessage(Message msg) {
    synchronized(inputMessages) {
      inputMessages.add(msg);
    }
    return true;
  }

  public boolean clearStreams() {
    streams.removeAll();
//    MediaConnection media;
//    while ( (media = (MediaConnection) streams.get(0)) != null) {
//      media.stopContinuousBackgroundReceive();
//      streams.remove(0);
//    }
    return true;
  }

  public boolean addStream(String name, MediaConnection media) {
    return streams.put(name, media);
  }


  public boolean postMessageToAllDestinations(Message msg) {

    if (msg.priority == 0)
      msg.priority = priority;
    Message hmsg;
    TextCollection destinations = getDestinations();
    int count = destinations.getCount();
    for (int n=0; n<count; n++) {
      hmsg = msg;
      hmsg.to = destinations.get(n);
      messageSender.sendMsgTo(destinations.get(n), hmsg);
      if (perfStat != null) {
        perfStat.msgPosted++;
        perfStat.lastMessagePosted = hmsg.shallowClone();
      }
    }
    return true;
  }
}



