/***************************** License **********************************

    Copyright(c)2005, Communicative Machines
    All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are 
met:

- Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright 
  notice, this list of conditions and the following disclaimer in the 
  documentation and/or other materials provided with the distribution.

- Neither the name of the copyright holders nor the names of its 
  contributors may be used to endorse or promote products derived from 
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER 
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

************************************************************************/
// NetworkConnection.java


package com.cmlabs.air;

import java.net.Socket;
import java.io.IOException;

public class NetworkConnection extends Thread{

    Socket socket;
    NetProtocol netProtocol;
    MessageReceiver messageReceiver = null;
    IncomingSocketConnector isc;
    Time lastRead = new Time();
    Time lastWrite = new Time();
    long keepAliveTimeout = 60000;

    String host = "";
    String name = "";
    int port;

    int type;

    public static int OUTGOING = 0;
    public static int RECEIVER = 1;
    public static int INCOMING_MULTI_USE = 2;

    public long timeout = 60000; // 1 minute
    public long timeLastUsed;

    public boolean shouldBeRunning = true;

   // Outgoing connections don't run in their own thread
    public NetworkConnection(String host, int port, NetProtocol netProtocol) throws Exception {
	this.netProtocol = netProtocol;
	this.host = host;
	this.port = port;
	this.socket = new Socket(host, port);
	this.socket.setSoTimeout(10);
	// this.socket.setKeepAlive(false);
	this.socket.setSoLinger(true, 0); //so it goes away when you close it
	this.messageReceiver = null;
	this.type = OUTGOING;
	this.start();
    }

    // Outgoing connections don't run in their own thread
     public NetworkConnection(TCPLocation loc, NetProtocol netProtocol) throws Exception {
         this.netProtocol = netProtocol;
         TextCollection hosts = loc.getAllAddresses();
         this.port = loc.port;

         Exception lastException = null;
         for (int n=0; n<hosts.getCount(); n++) {
           host = hosts.get(n);
           if (host.length() > 0) {
             lastException = null;
             try {
               this.socket = new Socket(host, port);
             }
             catch(Exception e) {
               lastException = e;
               continue;
             }
             break;
           }
         }
         if (lastException != null) {
           throw lastException;
         }

         this.socket.setSoTimeout(10);
         // this.socket.setKeepAlive(false);
         this.socket.setSoLinger(true, 0); //so it goes away when you close it
         this.messageReceiver = null;
         this.type = OUTGOING;
         this.start();
     }

    // incoming connection, created by serverSocket - runs in its own thread
    public NetworkConnection(Socket socket, NetProtocol netProtocol, MessageReceiver messageReceiver, IncomingSocketConnector isc) throws Exception {
	this.netProtocol = netProtocol;
	this.socket = socket;
	this.socket.setSoTimeout(10);
	// this.socket.setKeepAlive(false);
	this.socket.setSoLinger(true, 0); //so it goes away when you close it
	this.messageReceiver = messageReceiver;
	this.isc = isc;
	this.type = INCOMING_MULTI_USE;
	timeLastUsed = System.currentTimeMillis();
	this.start();
    }

    public void destroy() {
	shouldBeRunning = false;
	if ( (type == RECEIVER) || (type == INCOMING_MULTI_USE) ) {
	  Utils.pause(150);
	}
	// once this is all done, close the socket
	  try {
	    if (socket != null)
	      socket.close();
	  }
	  catch (IOException ioe) {
	  }
    }

    public synchronized boolean initializeAsReceiver(MessageReceiver messageReceiver, String myname) {
      if (type == RECEIVER)
	return true;
      if (type == INCOMING_MULTI_USE)
	return false;
      if (!netProtocol.initializeAsReceiver(socket, myname))
	return false;
      this.messageReceiver = messageReceiver;
      type = RECEIVER;
      return true;
    }

    public synchronized boolean sendMessage(Message message) {
      if (type != OUTGOING)
	return false;
      if (!netProtocol.sendMessage(socket, message))
	return false;
      lastWrite = new Time();
      Message msg = netProtocol.receiveMessage(socket, 10000);
      if ((msg == null) || (!msg.type.equalsIgnoreCase("RECEIVE_ACCEPT")))
	return false;

      lastRead = new Time();
      return true;
    }

    public synchronized boolean sendReplyMessage(Message message) {
      if (!netProtocol.sendMessage(socket, message))
	return false;

      lastWrite = new Time();
      return true;
    }

    public Message sendReceiveMessage(Message message) {
      return sendReceiveMessage(message, 10000);
    }

    public synchronized Message sendReceiveMessage(Message message, int timeout) {
      if (type != OUTGOING)
	return new Message("", "", "SEND_FAILED");
      if (!netProtocol.sendMessage(socket, message))
	return new Message("", "", "SEND_FAILED");
      lastWrite = new Time();
      Message msg = netProtocol.receiveMessage(socket, timeout);
     // if (netProtocol.getLastError().length() > 0) {
     //   System.out.println("["+name+"] SEND: " + netProtocol.getLastError());
     // }

     if (msg == null)
	return new Message("", "", "RECEIVE_FAILED");

      lastRead = new Time();
      return msg;
    }

    public synchronized Message receiveMessage(int timeout) {
	Message msg = netProtocol.receiveMessage(socket, timeout);
      // if (netProtocol.getLastError().length() > 0) {
      //   System.out.println("["+name+"] RECV: " + netProtocol.getLastError());
      // }
	if (!netProtocol.isSocketConnected(socket)) {
	  socket = null;
	  return null;
	}
	//if (!isConnected())
	  //socket = null;
	if (msg != null) {
	  lastRead = new Time();
	  if (msg.type.equalsIgnoreCase("KEEPALIVE")) {
	    msg = new Message("", "", "KEEPALIVESUCCESS");
	    sendMessage(msg);
	    msg = null;
	  }
	}
//	      System.out.println("Con Receiving Message type [" + msg.type + "]");
	return msg;
    }

    public void run() {

      Message outMsg = null;
      Message incomingMessage = null;

      while (shouldBeRunning) {
	if (type == OUTGOING) {
	  if (msSinceLastActivity() > keepAliveTimeout) {
	    sendKeepAliveMessage();
	  }
	  Utils.pause(20);
	}
	else {
	  incomingMessage = receiveMessage(100);
	  timeLastUsed = System.currentTimeMillis();
	  if (incomingMessage != null) {
	/*    if (type == RECEIVER)
	      System.out.println("Received Callback Message: " + incomingMessage.toString());
	    else
	      System.out.println("Received Normal Message: " + incomingMessage.toString());*/
	 /*   if (type == RECEIVER)
	      System.out.print("c");
	    else
	      System.out.print(".");*/
	    outMsg = messageReceiver.receiveMessage(incomingMessage);
	    if (outMsg == null) {
		sendReplyMessage(new Message(incomingMessage.to, incomingMessage.from, "RECEIVE_ACCEPT"));
	    }
	    else {
	      sendReplyMessage(outMsg);
	    }
	  }
	  else {
	    if (!isConnected())
	      shouldBeRunning = false;
	  }

	}

      }

	// once this is all done, close the socket
	if ((type != OUTGOING) && (socket != null)) {
	  try {
	    socket.close();
	  }
	  catch (IOException ioe) {
	    System.out.println("Error code NC-2");
	  } // should log
	}
    }

    public boolean isClosed() {
      if (socket == null)
	return true;
      else
	return false;
    }

    public boolean isConnected() {
      if (socket == null)
	return false;
      if (type == OUTGOING) {
	this.receiveMessage(10);
      }
      if (!netProtocol.isSocketConnected(socket))
	return false;
//      if (isClosed())
//	return false;
//      if (!socket.isConnected())
//	return false;
      if (socket.getInetAddress() == null)
	return false;
/*      if (socket.isClosed())
	return false;
      if (!socket.isConnected())
	return false;*/
      return true;
    }

    long msSinceLastRead() {
      Time now = new Time();
      return (now.fullMS - lastRead.fullMS);
    }

    long msSinceLastWrite() {
      Time now = new Time();
      return(now.fullMS - lastWrite.fullMS);
    }

    long msSinceLastActivity() {
      long a = msSinceLastRead();
      long b = msSinceLastWrite();
      return (a > b ? a : b);
    }

    boolean setKeepAliveTimeout(long timeout) {
      keepAliveTimeout = timeout;
      return true;
    }

    long getKeepAliveTimeout() {
      return keepAliveTimeout;
    }

    boolean useKeepAlive(boolean doKeepAlive) {
      if ( (doKeepAlive) && (keepAliveTimeout <= 0))
	keepAliveTimeout = 60000;
      else if (!doKeepAlive)
	keepAliveTimeout = -1;
      return true;
    }

    boolean sendKeepAliveMessage() {
      if (keepAliveTimeout <= 0)
	      return true;

      if (!isConnected())
	return false;

      Message msg = new Message("", "", "KEEPALIVE");
      msg = sendReceiveMessage(msg, 5000);

      if (msg == null)
	return false;
      else if (!msg.type.equalsIgnoreCase("KEEPALIVESUCCESS"))
	return false;
      else
	return true;
    }

}

