/***************************** License **********************************

    Copyright(c)2005, Communicative Machines
    All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are 
met:

- Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright 
  notice, this list of conditions and the following disclaimer in the 
  documentation and/or other materials provided with the distribution.

- Neither the name of the copyright holders nor the names of its 
  contributors may be used to endorse or promote products derived from 
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER 
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

************************************************************************/
// ObjectQueue.java

package com.cmlabs.air;

import java.util.*;

public class ObjectQueue extends BaseObject {

  ObjectDictionary dict = new ObjectDictionary();

  public ObjectQueue() {
  }

  public boolean isEmpty() {
    synchronized(dict) {
      return (dict.getCount() == 0);
    }
  }

  public int getCount() {
    synchronized(dict) {
      return dict.getCount();
    }
  }

  boolean removeAll() {
    synchronized(dict) {
      return dict.removeAll();
    }
  }

  ObjectCollection retrieveAll() {
    synchronized(dict) {
      if (dict.getCount() == 0) {
        return null;
      }
      ObjectCollection col = dict.getAllObjects();
      dict.removeAll();
      return col;
    }
  }

  String add(Object obj) {
    synchronized(dict) {
      String id = Utils.generateID("queue");
      dict.put(id, obj);
      dict.notify();
      return id;
    }
  }

  String add(String id, Object obj) {
    synchronized(dict) {
      dict.put(id, obj);
      dict.notify();
      return id;
    }
  }

  String waitForNewEntryID(long ms) {
    synchronized(dict) {
      if (dict.getCount() > 0)
        return dict.getFirstKey().toString();

      if (ms > 0) {
        try {
          dict.wait(ms);
        } catch(Exception e) {
          return "";
        }
        return dict.getFirstKey().toString();
      }
    }
    return "";
  }

  Object waitForNewEntry(long ms) {
    Object obj = null;
    synchronized(dict) {
      if (dict.getCount() > 0) {
        obj = dict.getFirst();
        dict.remove(0);
        return obj;
      }

      else if (ms > 0) {
        try {
          dict.wait(ms);
        }
        catch (Exception e) {
          return null;
        }
        if (dict.getCount() > 0) {
          obj = dict.getFirst();
          dict.remove(0);
          return obj;
        }
      }
    }
    return obj;
  }

  synchronized boolean waitForNewEntryToAppear(long ms) {
    synchronized(dict) {
      if (dict.getCount() > 0)
        return true;

      else if (ms > 0) {
        try {
          dict.wait(ms);
        }
        catch (Exception e) {
          return false;
        }
        return (dict.getCount() > 0);
      }
    }
    return false;
  }

  synchronized Object waitForNewEntry(String id, long ms) {
    Object obj = null;
    synchronized(dict) {
      if ( (obj = dict.get(id)) != null) {
        dict.remove(id);
        return obj;
      }

      else if (ms > 0) {
        long t_end = System.currentTimeMillis();
        long t = t_end - System.currentTimeMillis();
        while (t >= 0) {
          try {
            dict.wait(t);
          }
          catch (Exception e) {
            return null;
          }

          if ( (obj = dict.get(id)) != null) {
            dict.remove(id);
            return obj;
          }
          t = t_end - System.currentTimeMillis();
        }
      }
    }
    return null;
  }


  Object retrieveEntry(String id) {
    synchronized(dict) {
      Object obj = dict.get(id);
      dict.remove(id);
      return obj;
    }
  }

  Object retrieveEntry(int pos) {
    synchronized(dict) {
      Object obj = dict.get(pos);
      dict.remove(pos);
      return obj;
    }
  }

  Object viewEntry(String id) {
    synchronized(dict) {
      return dict.get(id);
    }
  }

  Object viewEntry(int pos) {
    synchronized (dict) {
      return dict.get(pos);
    }
  }

}




