/***************************** License **********************************

    Copyright(c)2005, Communicative Machines
    All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are 
met:

- Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright 
  notice, this list of conditions and the following disclaimer in the 
  documentation and/or other materials provided with the distribution.

- Neither the name of the copyright holders nor the names of its 
  contributors may be used to endorse or promote products derived from 
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER 
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

************************************************************************/
// Parameter.java


package com.cmlabs.air;

import java.util.*;
import org.w3c.dom.*;

public class Parameter
    extends BaseObject {

  static int NOVALUE = -87654321;
  static int PARAM_ERROR = -99999999;

  public String name = "";
  public String value = "";
  public String defaultValue = "";
  public String valueType = "";
  public String parameterType = "";
  public double step = 0;

  boolean locked = false;
  double minValue = 0, maxValue = 0;
  int current = 0;
  int def = 0;
  TextCollection values = new TextCollection();

  public Parameter(String parametername, String type, TextCollection collValues) {
    name = parametername;
    valueType = type;
    values.removeAll();
    values = collValues;
    maxValue = minValue = NOVALUE;
    def = current = 0;
    step = 1;
  }

  public Parameter(String parametername, String type, String val, String defaultVal) {
    name = parametername;
    valueType = type;
    value = val;
    defaultValue = defaultVal;
    maxValue = minValue = NOVALUE;
    def = current = 0;
    step = 1;
  }

  public Parameter(String parametername, String type) {
    name = parametername;
    valueType = type;
    maxValue = minValue = NOVALUE;
    step = 1;
  }

  public Parameter(String xml) {
    valueType = "String";
    def = current = 0;
    maxValue = minValue = NOVALUE;
    step = 1;
    fromXML(xml);
  }

  public Parameter(XMLElement node) {
    valueType = "String";
    def = current = 0;
    maxValue = minValue = NOVALUE;
    step = 1;
    fromXML(node);
  }

// ***********************************************************************************
// XML Functions
// ***********************************************************************************

  public String toXML() {
    // <parameter type="erasedFromMax" value="15" units="integer" />
    String minmax = "";
    if (minValue != NOVALUE)
      minmax = " min=\"" + minValue + "\"";
    if (maxValue != NOVALUE)
      minmax += " max=\"" + maxValue + "\"";

    String xml;
    if (values.getCount() > 0) {
      xml = "<parameter name=\"" + Utils.xmlStringEncode(getName()) +
	  "\" type=\"" + Utils.xmlStringEncode(getType()) +
	  "\" value=\"" + Utils.xmlStringEncode(getString()) +
	  "\" default=\"" + Utils.xmlStringEncode(defaultValue) +
	  "\" step=\"" + step + "\"" + minmax + ">\n" + Utils.xmlIndent(values.toXML()) + "</parameter>";
    }
    else {
      xml = "<parameter name=\"" + Utils.xmlStringEncode(getName()) +
	  "\" type=\"" + Utils.xmlStringEncode(getType()) +
	  "\" value=\"" + Utils.xmlStringEncode(getString()) +
	  "\" default=\"" + Utils.xmlStringEncode(defaultValue) +
	  "\" step=\"" + step + "\"" + minmax + " />";
    }
    return xml;
  }

  public boolean fromXML(String xml) {
    return fromXML(Utils.xmlGetRootNodeFromXML(xml));
  }

  public boolean fromXML(XMLElement node) {
    if (node == null)
      return false;

    name = node.getStringAttribute("name");
    setType(node.getStringAttribute("type"));
    value = node.getStringAttribute("value");
    if (node.hasAttribute("default"))
      defaultValue = node.getStringAttribute("default");
    else
      defaultValue = value;
    if (node.hasAttribute("step"))
      step = node.getDoubleAttribute("step");
    if (node.hasAttribute("min"))
      minValue = node.getDoubleAttribute("min");
    if (node.hasAttribute("max"))
      maxValue = node.getDoubleAttribute("max");

    if (value.length() == 0)
      value = defaultValue;
    else if (defaultValue.length() == 0)
      defaultValue = value;

    XMLElement subNode = node.getChild("collection");
    if (subNode != null)
      values.fromXML(subNode);
    else
      values.removeAll();

    return true;
  }

// ***********************************************************************************
// Setters
// ***********************************************************************************

  public boolean set(String val) {

    if (valueType.equalsIgnoreCase("Integer")) {
      return setInteger(Utils.str2int(val));
    }
    else if (valueType.equalsIgnoreCase("Double")) {
      return setDouble(Utils.str2double(val));
    }
    else {
      return setString(val);
    }
  }

  public boolean setType(String type) {
    valueType = type;
    return true;
  }

  public boolean setMinMax(double min, double max) {
    minValue = min;
    maxValue = max;
    return true;
  }

  public boolean addItem(String val) {
    if (values.getCount() == 0)
      return false;
    int pos = values.getPos(val);
    if (pos >= 0)
      return true;
    else
      return values.add(val);
  }

  public boolean removeItem(String val) {
    if (values.getCount() < 2)
      return false;
    int pos = values.getPos(val);
    if (pos < 0)
      return true;
    else {
      if ((current == pos) && (current > 0)) {
	current--;
      }
      return values.remove(pos);
    }
  }

  public boolean setString(String val) {

    if (valueType.length() == 0)
      valueType = "String";

    if (!valueType.equalsIgnoreCase("String"))
      return false;

    // If we have no collection of values
    if (values.getCount() == 0) {
      value = val;
    }
    else {
      if (!values.contains(val))
	return false;
      current = values.getPos(val);
    }
    return true;
  }

  public boolean setInteger(int val) {

    if (valueType.length() == 0)
      valueType = "Integer";

    if (!valueType.equalsIgnoreCase("Integer"))
      return false;

    // If we have no collection of values
    if (values.getCount() == 0) {
      if ( (minValue != NOVALUE) && (maxValue != NOVALUE) &&
	  ( (val < minValue) || (val > maxValue)))
	return false;
      value = ""+val;
    }
    else {
      if (!values.contains(""+val))
	return false;
      current = values.getPos(""+val);
    }
    return true;
  }

  public boolean setDouble(double val) {

    if (valueType.length() == 0)
      valueType = "Double";

    if (!valueType.equalsIgnoreCase("Double"))
      return false;

    // If we have no collection of values
    if (values.getCount() == 0) {
      if ( (minValue != NOVALUE) && (maxValue != NOVALUE) &&
	  ( (val < minValue) || (val > maxValue)))
	return false;
      value = ""+val;
    }
    else {
      if (!values.contains(""+val))
	return false;
      current = values.getPos(""+val);
    }
    return true;
  }

// ***********************************************************************************
// Getters
// ***********************************************************************************

  public String getType() {
    return valueType;
  }

  public String getName() {
    return name;
  }

  public int getInteger() {

    if (values.getCount() == 0)
      return Utils.str2int(value);
    else {
      if (current < 0)
	return 0;
      if (!valueType.equals("Integer"))
	return 0;
      return Utils.str2int(values.get(current));
    }
  }

  public double getDouble() {
    if (values.getCount() == 0)
      return Utils.str2double(value);
    else {
      if (current < 0)
	return 0;
      if (!valueType.equals("Double"))
	return 0;
      return Utils.str2double(values.get(current));
    }
  }

  public String getString() {
    if (values.getCount() == 0)
      return value;
    else {
      if (current < 0)
	return "";
      return values.get(current);
    }
  }

  public String getAsString() {

    if (values.getCount() == 0)
      return value;
    else {
      if (current < 0)
	return "";
      return values.get(current);
    }
  }

  public boolean reset() {
    current = def;
    value = defaultValue;
    return true;
  }

  public boolean increment(int steps) {

    if (valueType.equalsIgnoreCase("Integer")) {
      return setInteger(Utils.str2int(value) + (int) (step * steps));
    }
    if (valueType.equalsIgnoreCase("Double")) {
      return setDouble(Utils.str2double(value) + (step * steps));
    }
    if (values.getCount() == 0)
      return true;

    if (current < 0)
      return false;

    if ( (current + steps) < (values.getCount() - 1)) {
      current += steps;
      return true;
    }
    else
      return false;
  }

  public boolean decrement(int steps) {
    if (valueType.equalsIgnoreCase("Integer")) {
      return setInteger(Utils.str2int(value) - (int) (step * steps));
    }
    if (valueType.equalsIgnoreCase("Double")) {
      return setDouble(Utils.str2double(value) - (step * steps));
    }
    if (values.getCount() == 0)
      return true;

    if (current < 0)return false;

    if (current - steps >= 0) {
      current -= steps;
      return true;
    }
    else
      return false;
  }

}
