/***************************** License **********************************

    Copyright(c)2005, Communicative Machines
    All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are 
met:

- Redistributions of source code must retain the above copyright notice,
  this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright 
  notice, this list of conditions and the following disclaimer in the 
  documentation and/or other materials provided with the distribution.

- Neither the name of the copyright holders nor the names of its 
  contributors may be used to endorse or promote products derived from 
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER 
OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

************************************************************************/
// PhaseSpec.java


package com.cmlabs.air;

import java.util.*;

public class PhaseSpec extends BaseObject {

  public String name = "";
  public String pid = "";
  public String moduleName = "";
  public String context = "";
  public boolean allowSelfTriggering = false;
  public String lastTriggerName = "";
  public Time activationTime = null;

  //! Contains TriggerSpecs
  public ObjectCollection triggers = new ObjectCollection();

  //! Contains RetrieveSpecs
  public ObjectCollection retrieves = new ObjectCollection();

  //! Contains CrankSpecs
  public ObjectCollection cranks = new ObjectCollection();

  //! Contains PostSpecs
  public ObjectCollection posts = new ObjectCollection();

  //! Contains StreamAccesses
  public ObjectCollection streams = new ObjectCollection();

  public PhaseSpec() {
    pid = Utils.generateID();
  }

  public PhaseSpec(String xml) {
    pid = Utils.generateID();
    fromXML(Utils.xmlGetRootNodeFromXML(xml));
  }

  public PhaseSpec(XMLElement node) {
    pid = Utils.generateID();
    fromXML(node);
  }

  public boolean equals(PhaseSpec spec) {
    if (spec == null)
      return false;
    return (spec.pid.equalsIgnoreCase(pid));
//    return ( (!spec.moduleName.equals(moduleName)) || (!spec.context.equals(context)) );
  }

  public boolean fromXML(String xml) {
    return fromXML(Utils.xmlGetRootNodeFromXML(xml));
  }

  public boolean fromXML(XMLElement node) {

    if (node == null)
      return false;

    String nodename = node.getName();
    if ( (nodename == null) || (!nodename.equalsIgnoreCase("phase")))
      return false;

    triggers.removeAll();
    retrieves.removeAll();
    cranks.removeAll();
    posts.removeAll();

    if (node.hasAttribute("id"))
      name = node.getStringAttribute("id");
    else
      name = node.getStringAttribute("name");
    if (node.hasAttribute("pid")) {
      if (node.getStringAttribute("pid").length() > 0)
        pid = node.getStringAttribute("pid");
    }
    String str = node.getStringAttribute("module");
    if (str.length() > 0)
      moduleName = str;
    str = node.getStringAttribute("context");
    if (str.length() > 0)
      context = str;

    XMLElement xmlNode = node.getChild("allowselftriggering");
    if ((xmlNode != null) && (xmlNode.getContent().equalsIgnoreCase("yes"))) {
      allowSelfTriggering = true;
    }
    if ((xmlNode != null) && (xmlNode.getContent().equalsIgnoreCase("no"))) {
      allowSelfTriggering = false;
    }

    lastTriggerName = node.getStringAttribute("lasttrigger");

    TriggerSpec triggerSpec;
    RetrieveSpec retrieveSpec;
    CrankSpec crankSpec;
    PostSpec postSpec;
    StreamAccess accessSpec;

    String key;
    XMLElement child;
    XMLElement sub;
    for (Enumeration e = node.enumerateChildren(); e.hasMoreElements(); ) {
      child = (XMLElement) e.nextElement();
      if (child != null) {
        key = child.getName();
        if (key.equalsIgnoreCase("triggers")) {
          for (Enumeration ee = child.enumerateChildren(); ee.hasMoreElements(); ) {
            sub = (XMLElement) ee.nextElement();
            if (sub != null) {
              key = sub.getName();
              if (key.equalsIgnoreCase("trigger")) {
                triggerSpec = new TriggerSpec(sub,
                                              child.getStringAttribute("from"));
                triggers.add(triggerSpec);
              }
            }
          }
        }

        else if (key.equalsIgnoreCase("retrieves")) {
          for (Enumeration ee = child.enumerateChildren(); ee.hasMoreElements(); ) {
            sub = (XMLElement) ee.nextElement();
            if (sub != null) {
              key = sub.getName();
              if (key.equalsIgnoreCase("retrieve")) {
                retrieveSpec = new RetrieveSpec(sub,
                                              child.getStringAttribute("from"));
              retrieves.add(retrieveSpec);
              }
            }
          }
        }

        else if (key.equalsIgnoreCase("cranks")) {
          for (Enumeration ee = child.enumerateChildren(); ee.hasMoreElements(); ) {
            sub = (XMLElement) ee.nextElement();
            if (sub != null) {
              key = sub.getName();
              if (key.equalsIgnoreCase("crank")) {
                crankSpec = new CrankSpec(sub);
                cranks.add(crankSpec);
              }
            }
          }
        }

        else if (key.equalsIgnoreCase("posts")) {
          for (Enumeration ee = child.enumerateChildren(); ee.hasMoreElements(); ) {
            sub = (XMLElement) ee.nextElement();
            if (sub != null) {
              key = sub.getName();
              if (key.equalsIgnoreCase("post")) {
                postSpec = new PostSpec(sub, child.getStringAttribute("from"));
                      posts.add(postSpec);
              }
            }
          }
        }

        else if (key.equalsIgnoreCase("streams")) {
          for (Enumeration ee = child.enumerateChildren(); ee.hasMoreElements(); ) {
            sub = (XMLElement) ee.nextElement();
            if (sub != null) {
              key = sub.getName();
              if (key.toLowerCase().indexOf("stream") >= 0) {
                accessSpec = new StreamAccess(sub);
                streams.add(accessSpec);
              }
            }
          }
        }

        else if (key.equalsIgnoreCase("activationtime")) {
          activationTime = new Time();
          if (!activationTime.fromXML(child))
            activationTime = null;
        }

      }
    }

    return true;
  }

  public String toXML() {

    String xml = "", subXML = "";
    int n;

    if (allowSelfTriggering)
      xml += "<allowselftriggering>Yes</allowselftriggering>\n";

    TriggerSpec triggerSpec;
    if (triggers.getCount() > 0) {
      subXML = "";
      for (n=0; n<triggers.getCount(); n++) {
        triggerSpec = (TriggerSpec) triggers.get(n);
        if (triggerSpec != null)
          subXML += triggerSpec.toXML() + "\n";
      }
      xml += "<triggers>\n" + Utils.xmlIndent(subXML) + "</triggers>\n";
    }

    RetrieveSpec retrieveSpec;
    if (retrieves.getCount() > 0) {
      subXML = "";
      for (n = 0; n < retrieves.getCount(); n++) {
        retrieveSpec = (RetrieveSpec) retrieves.get(n);
        if (retrieveSpec != null)
          subXML += retrieveSpec.toXML() + "\n";
      }
      xml += "<retrieves>\n" + Utils.xmlIndent(subXML) + "</retrieves>\n";
    }

    CrankSpec crankSpec;
    if (cranks.getCount() > 0) {
      subXML = "";
      for (n=0; n<cranks.getCount(); n++) {
        crankSpec = (CrankSpec) cranks.get(n);
        if (crankSpec != null)
          subXML += crankSpec.toXML() + "\n";
      }
      xml += "<cranks>\n" + Utils.xmlIndent(subXML) + "</cranks>\n";
    }

    PostSpec postSpec;
    if (posts.getCount() > 0) {
      subXML = "";
      for (n=0; n<posts.getCount(); n++) {
        postSpec = (PostSpec) posts.get(n);
        if (postSpec != null)
          subXML += postSpec.toXML() + "\n";
      }
      xml += "<posts>\n" + Utils.xmlIndent(subXML) + "</posts>\n";
    }

    StreamAccess accessSpec;
    if (streams.getCount() > 0) {
      subXML = "";
      for (n=0; n<streams.getCount(); n++) {
        accessSpec = (StreamAccess) streams.get(n);
        if (accessSpec != null)
          subXML += accessSpec.toXML() + "\n";
      }
      xml += "<streams>\n" + Utils.xmlIndent(subXML) + "</streams>\n";
    }

    if (activationTime != null)
      xml += activationTime.toXML("activationtime") + "\n";
    else
      xml += Time.toXMLInvalid("activationtime") + "\n";

    return "<phase name=\"" + Utils.xmlStringEncode(name) + "\" module=\"" + Utils.xmlStringEncode(moduleName) + "\" context=\"" + Utils.xmlStringEncode(context) + "\" lasttrigger=\"" + Utils.xmlStringEncode(lastTriggerName) + "\" pid=\"" + Utils.xmlStringEncode(pid) + "\">\n" + Utils.xmlIndent(xml) + "</phase>";
  }

  public String print() {
    return "phase name=\"" + Utils.xmlStringEncode(name) + "\" module=\"" + Utils.xmlStringEncode(moduleName) + "\" context=\"" + Utils.xmlStringEncode(context) + "\"\n";
  }

  public String toHTML() {
    return print();
  }

  public boolean isNewerThan(PhaseSpec spec) {
    if (spec == null)
      return true;
    return (activationTime.difference(spec.activationTime) > 0);
  }

  public boolean isOlderThan(PhaseSpec spec) {
    if (spec == null)
      return false;
    return (activationTime.difference(spec.activationTime) < 0);
  }

}
